! $Id: AccessStmts.f90 2811 2019-03-23 23:58:52Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AccessStmts module.


!*******************************************************************************
!!
!> Procedures and definitions for parsing /access-stmt/'s and /access-spec/'s, 
!! including some simple constants to represent the various options for an 
!! access specification.

MODULE AccessStmts
  
  USE CompilerKinds
  USE GenericSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAccessStmt
  PUBLIC :: WriteAccess
  PUBLIC :: GenericSpec           ! From GenericSpecs.
  
  !-----------------------------------------------------------------------------
  
  !> Access has not been specified.
  INTEGER, PARAMETER, PUBLIC :: iaccNone = 0
  !> Access is specified to be PRIVATE.
  INTEGER, PARAMETER, PUBLIC :: iaccPrivate = 1
  !> Access is specified to be PUBLIC.
  INTEGER, PARAMETER, PUBLIC :: iaccPublic = 2
  
  !> Stores the results of parsing an /access-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: AccessStmt
    !> The iacc* value that corresponds to the statement.
    INTEGER :: access
    !> List of identifiers provided in the /access-stmt/.
    !!
    !! Always allocated, but may be zero length.
    TYPE(GenericSpec), ALLOCATABLE :: access_id_list(:)
  CONTAINS
    PROCEDURE :: Dump => access_Dump
    PROCEDURE :: Query => access_Query
    PROCEDURE :: Modify => access_Modify
    PROCEDURE, NOPASS :: What => access_What
    PROCEDURE :: Advertise => access_Advertise
  END TYPE AccessStmt
  
  !-----------------------------------------------------------------------------
  
  !> The keywords that can comprise an access statement or specifier.
  !!
  !! Indices must be consistent with the iacc* constants (bar iaccNone).
  CHARACTER(*,scck), PARAMETER, PUBLIC :: AccessKeywords(2)  &
      = [scck_'PRIVATE', scck_'PUBLIC ']
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an access-stmt (R524).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    access_stmt       The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An access statement is:
  !!
  !! /access-spec/  [ [ :: ] /access-id-list/ ]
  
  SUBROUTINE ParseAccessStmt(tlist, access_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(AccessStmt), INTENT(OUT) :: access_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /access-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R524'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error code for procedure calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: it             ! Start index in tlist for the list.
    INTEGER :: item_count     ! Number of items in the list.
    INTEGER :: item_start     ! Start index in tlist for a particular item.
    INTEGER :: item_finish    ! Finish index in tlist for a particular item.
    INTEGER :: item           ! Item index
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /access-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The opening keyword
    
    CALL ExpectKeyword(  &
        tlist, 1,  &
        AccessKeywords,  &
        '/access-spec/',  &
        access_stmt%access,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The keyword on its own is ok.
    IF (SIZE(tlist) == 1) THEN
      ALLOCATE(access_stmt%access_id_list(0))
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Optional double colon.
    
    IF (Match(tlist, 2, ittDeclare, scck_'::')) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! The /access-id-list/
    
    item_count = ListItemCount(tlist(it:))
    ALLOCATE(access_stmt%access_id_list(item_count))
    item_finish = 0
    DO item = 1, item_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_access_id(  &
          tlist(item_start:item_finish),  &
          access_stmt%access_id_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseAccessStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /access-id/ (R525).  This is just a 
  !! wrapper for the /generic-spec/ parsing code in GenericSpecs::Parse.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    access_id         The results of parsing the /access-id/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_access_id(tlist, access_id, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(GenericSpec), INTENT(OUT) :: access_id
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /access-id/.
    CHARACTER(*), PARAMETER :: comp = 'R525'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /access-id/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ParseGenericSpec(tlist, access_id, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_access_id
  
  
  !*****************************************************************************
  !!
  !> Implementation of AccessStmt%Dump - write a debugging representation of 
  !! the results of parsing an access statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE access_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AccessStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Item list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/access-spec/ = ')", ADVANCE='NO') pfx
    CALL WriteAccess(unit, pnode%access)
    WRITE (unit, "()")
      
    WRITE (unit, "(A,'/access-id-list/')") pfx
    DO i = 1, SIZE(pnode%access_id_list)
      WRITE (unit, "(A,I0,': ')", ADVANCE='NO') pfx // '  ', i
      CALL WriteGenericSpec(unit, pnode%access_id_list(i))
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE access_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AccessStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE access_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AccessStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%access_id_list)
      CALL pnode%access_id_list(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE access_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AccessStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE access_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AccessStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%access_id_list)
      CALL pnode%access_id_list(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE access_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AccessStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'access-stmt'.
  
  FUNCTION access_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'access-stmt'
    
  END FUNCTION access_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AccessStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE access_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AccessStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! decl index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%access_id_list)))
    DO i = 1, SIZE(pnode%access_id_list)
      IF (pnode%access == iaccPublic) THEN
        ident_list(i) = IdentifierNature(  &
            pnode%access_id_list(i),  &
            [NatureDefPublic],  &
            [.Locationof. pnode] )
      ELSE
        ident_list(i) = IdentifierNature(  &
            pnode%access_id_list(i),  &
            [NatureDefPrivate],  &
            [.Locationof. pnode] )
      END IF
    END DO
    
  END SUBROUTINE access_Advertise
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of an access specifier to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the designator to.
  !!
  !! @param[in]     access            The access specifier to write out.
  
  SUBROUTINE WriteAccess(log_unit, access)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    INTEGER, INTENT(IN) :: access
    
    !***************************************************************************    
    
    SELECT CASE (access)
    CASE (iaccNone)
      WRITE (log_unit, "(A)", ADVANCE='NO') '(access not specified)'
    CASE (iaccPrivate, iaccPublic)
      WRITE (log_unit, "(A)", ADVANCE='NO') AccessKeywords(access)
    CASE DEFAULT
      STOP 'Internal error in AccessStmts%%WriteAccess, &
          &undefined access spec.'
    END SELECT
    
  END SUBROUTINE WriteAccess
  
END MODULE AccessStmts
