! $Id: ArrayConstructors.f90 2800 2019-03-22 18:13:05Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ArrayConstructors module.


!*******************************************************************************
!!
!> Procedures and types for parsing /array-constructors/.

MODULE ArrayConstructors
  
  USE DeclarationTypeSpecs
  USE Expressions
  USE ParseNodes
  USE ImpliedDos
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseArrayConstructor
  PUBLIC :: WriteArrayConstructor
  PUBLIC :: TypeSpec
  
  !-----------------------------------------------------------------------------
  
  !> The parent type for values in an array constructor.
  TYPE, EXTENDS(TermParseNode), ABSTRACT, PUBLIC :: ACValue
  CONTAINS
    PROCEDURE, NOPASS :: What => acv_What
  END TYPE ACValue
  
  !> An element in a list of ACValue's.
  TYPE, PUBLIC :: ACValueList
    CLASS(ACValue), ALLOCATABLE :: item
  CONTAINS
    PROCEDURE :: Write => acvl_Write
    PROCEDURE :: Query => acvl_Query
    PROCEDURE :: Modify => acvl_Modify
  END TYPE ACValueList
  
  !> An ACValue that is an expression.
  TYPE, PUBLIC, EXTENDS(ACValue) :: ACValueExpr
    !> The expression.
    CLASS(Expression), ALLOCATABLE :: ex
  CONTAINS
    PROCEDURE :: Write => acve_Write
    PROCEDURE :: Query => acve_Query
    PROCEDURE :: Modify => acve_Modify
    PROCEDURE :: GetLocation => acve_GetLocation
  END TYPE ACValueExpr
  
  !> Represents an /ac-implied-do-control/.
  TYPE, EXTENDS(ImpliedDoControl), PUBLIC :: ACImpliedDoControl
  CONTAINS
    PROCEDURE, NOPASS :: What => acidcontrol_What
  END TYPE
  
  !> An ACValue that is an /ac-implied-do/.
  TYPE, PUBLIC, EXTENDS(ACValue) :: ACValueImpliedDo
    !> The list of values.
    !!
    !! Size of at least one.
    TYPE(AcValueList), ALLOCATABLE :: ac_value_list(:)
    !> Implied do control.
    TYPE(ACImpliedDoControl) :: control
  CONTAINS
    PROCEDURE :: Write => acvid_Write
    PROCEDURE :: Query => acvid_Query
    PROCEDURE :: Modify => acvid_Modify
    PROCEDURE :: GetLocation => acvid_GetLocation
  END TYPE ACValueImpliedDo
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /array-constructor/ (R468).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /array-constructor/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the end of the 
  !! /array-constructor/ in @a tlist.
  !!
  !! @param[out]    type_spec         The type spec in the array constructor, 
  !! if present, not allocated if no type spec was present.  The type spec 
  !! is checked for `:` and `*` as /type-param-value/s.
  !!
  !! @param[out]    ac_value_list     The sequence of value objects in the 
  !! constructor.  Always allocated, but may be zero length (in which case 
  !! type_spec will be allocated).
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The aggregate type that represents an array constructor is currently 
  !! an extension of expression.  Array constructors only ever appear as 
  !! primaries in expressions (unlike structure constructors - which can 
  !! also appear as data statement constants), so we've not gone to the 
  !! bother of creating a separate parse node.
  
  ! An array of structures with array components could potentially 
  ! result in a recursive call - hence the attribute.
  RECURSIVE SUBROUTINE ParseArrayConstructor( tlist, start, finish,  &
        type_spec, ac_value_list, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(TypeSpec), INTENT(OUT), ALLOCATABLE :: type_spec
    TYPE(ACValueList), INTENT(OUT), ALLOCATABLE :: ac_value_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /array-constructor/.
    CHARACTER(*), PARAMETER :: comp = 'R468'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Option return from Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: declare        ! Index of the declare token.
    INTEGER :: ispec          ! Type parameter specifier index.
    
    ! Error list for procedure calls 
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_start     ! Index of the starting token of the value list.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /array-constructor/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! We don't care which parenthesis option (value of opt) is in use.
    CALL Expect( &
        tlist,  &
        start,  &
        [ittDelimiter, ittDelimiter],  &
        [scck_'[ ', scck_'(/'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The ac-spec.  This starts with an optional /type-spec/ ::
    
    declare = FindBare(tlist, ittDeclare, scck_'::', start + 1)
    IF (declare /= 0) THEN
      ALLOCATE(type_spec)
      CALL ParseTypeSpec(tlist(start+1:declare-1), type_spec, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      list_start = declare + 1
      
      ! Check that the /type-param-value/s are not invalid in an 
      ! array constructor (they cannot be : or *).
      !
      ! A little debatable whether we want to do this here or in the 
      ! semantic pass.
      DO ispec = 1, SIZE(type_spec%params)
        IF (type_spec%params(ispec)%is_colon) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C402 on R401 via ' // comp,  &
              LOCATION=type_spec%params(ispec)%GetLocation(),  &
              MSG='A colon may not be used for a /type-param-value/ in &
                &the /type-param-spec/ of an /array-constructor/.' )
        ELSE IF (type_spec%params(ispec)%is_star) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C488 on R454 via ' // comp,  &
              LOCATION=type_spec%params(ispec)%GetLocation(),  &
              MSG='An asterisk may not be used in a /type-param-value/ in &
                &the /type-param-spec/ of an /array-constructor/.' )
        END IF
      END DO
      
      ! That can be it.
      IF (list_start == closing) THEN
        ALLOCATE(ac_value_list(0))
        CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
        RETURN
      END IF
    ELSE
      list_start = start + 1  ! After the opening parenthesis.
    END IF
    
    !---------------------------------------------------------------------------
    ! The /ac-value-list/
    
    CALL parse_ac_value_list(  &
        tlist,  &
        list_start, closing - 1,  &
        ac_value_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE ParseArrayConstructor
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /ac-value-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /ac-value-list/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the end of the 
  !! /ac-value-list/ in @a tlist.
  !!
  !! @param[out]    ac_value_list     The sequence of value objects.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! /ac-value-list/ could have (ultimately) as one of its values another 
  !! /ac-value-list/ - hence RECURSIVE.
  
  RECURSIVE SUBROUTINE parse_ac_value_list( tlist, start, finish,  &
      ac_value_list, err_list )
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(AcValueList), INTENT(OUT), ALLOCATABLE :: ac_value_list(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the list.
    INTEGER :: item_start     ! Index of the starting token for an item.
    INTEGER :: item_finish    ! Index of the finishing token for an item.
    INTEGER :: item           ! Item index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(start:finish))
    ALLOCATE(ac_value_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      
      CALL parse_ac_value(  &
          tlist,  &
          item_start, item_finish,  &
          ac_value_list(item)%item,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE parse_ac_value_list
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /ac-value/ (R472).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /ac-value/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the end of the 
  !! /ac-value/ in @a tlist.
  !!
  !! @param[out]    ac_value          The value object.  Note that this is
  !! polymorphic - it may be a value that is designated by an expression
  !! or by an /ac-implied-do/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An /ac-value/ could be (ultimately) another /ac-value/ - hence 
  !! RECURSIVE.
  
  RECURSIVE SUBROUTINE parse_ac_value( tlist, start, finish, ac_value,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(ACValue), INTENT(OUT), ALLOCATABLE :: ac_value
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /ac-value/.
    CHARACTER(*), PARAMETER :: comp = 'R472'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of closing parenthesis.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: assign         ! Index of the assign (=) token.
    
    ! Allocatable objects used to simply construction extensions of ACValue.
    ! After being built these are MOVE_ALLOC'd into the ac_value argument.
    CLASS(ACValueImpliedDo), ALLOCATABLE :: implied_do
    CLASS(ACValueExpr), ALLOCATABLE :: expr
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /ac-value/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Check for ac-implied-do.
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      assign = FindBare(tlist(start+1:closing-1), ittAssign, scck_'=')
      IF (assign /= 0) THEN
        ! We appear to have an /ac-implied-do/ as the item.
        ALLOCATE(implied_do)
        
        CALL parse_ac_implied_do_content(  &
            tlist,  &
            start + 1, closing - 1,  &
            implied_do,  &
            sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        
        CALL MOVE_ALLOC(implied_do, ac_value)
        
        CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
        RETURN
      END IF
    END IF
    
    !---------------------------------------------------------------------------
    ! The ac-value is an expression
    
    ALLOCATE(expr)
    CALL ParseExpr(tlist, start, finish, expr%ex, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    CALL MOVE_ALLOC(expr, ac_value)
    
  END SUBROUTINE parse_ac_value
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the content (the part within the 
  !! outer parenthesis) of an /ac-implied-do/ (R472).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the content of the 
  !! /ac-implied-do/ in @a tlist.  This should be the index of the token 
  !! that is immediately after the opening parenthesis.
  !!
  !! @param[in]     finish            Index of the end of the content of the 
  !! /ac-implied-do/ in @a tlist.  This should be the index of the token that 
  !! is immediately before the closing parenthesis.
  !!
  !! @param[out]    implied_do        Results of parsing the /ac-implied-do/.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! An array constructor could say "I do" multiple times, hence RECURSIVE.
  RECURSIVE SUBROUTINE parse_ac_implied_do_content( tlist, start, finish,  &
      implied_do, err_list )
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(AcValueImpliedDo), INTENT(OUT) :: implied_do
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /ac-implied-do/.
    CHARACTER(*), PARAMETER :: comp = 'R473'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count   ! Number of value items in an /ac-implied-do/.
    INTEGER :: item_start   ! Starting index of an item in an /ac-implied-do/.
    INTEGER :: item_finish  ! Finishing index of an item in an /ac-implied-do/.
    INTEGER :: item         ! Item index.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    CALL ParseImpliedDo(  &
        tlist,  &
        start, finish,  &
        'ac-value',  &
        '/ac-implied-do/',  &
        comp,  &
        list_count,  &
        implied_do%control,  &
        err_list )
    IF (Failure(err_list)) RETURN
    
    ALLOCATE(implied_do%ac_value_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(  &
          tlist,  &
          start, finish,  &
          item_start,  &
          item_finish )
      CALL parse_ac_value(  &
          tlist,  &
          item_start, item_finish,  &
          implied_do%ac_value_list(item)%item,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END DO
    
  END SUBROUTINE parse_ac_implied_do_content
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of an array constructor.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the array constructor to.
  !!
  !! @param[in]     type_spec         The type spec of the array constructor 
  !! to write out, if one was present or not allocated if not.
  !!
  !! @param[in]     values            The /ac-value-list/.
  
  RECURSIVE SUBROUTINE WriteArrayConstructor(log_unit, type_spec, values)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(TypeSpec), INTENT(IN), ALLOCATABLE :: type_spec
    TYPE(ACValueList), INTENT(IN) :: values(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Array constructor value index.
    
    !***************************************************************************
    
    WRITE (log_unit, "('[')", ADVANCE='NO')
    IF (ALLOCATED(type_spec)) THEN
      CALL WriteTypeSpec(log_unit, type_spec)
      WRITE (log_unit, "(' :: ')", ADVANCE='NO')
    END IF
    IF (SIZE(values) > 0) THEN
      i = 1
      DO
        CALL values(i)%item%Write(log_unit)
        IF (i == SIZE(values)) EXIT
        WRITE (log_unit, "(', ')", ADVANCE='NO')
        i = i + 1
      END DO
    END IF
    WRITE (log_unit, "(']')", ADVANCE='NO')
    
  END SUBROUTINE WriteArrayConstructor
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValue%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'ac-value'
  !!
  !! @todo We should specialize this for the extensions of ACValue.
  
  FUNCTION acv_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'ac-value'
    
  END FUNCTION acv_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in AcValueExpr
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValueExpr%Write - write a debugging representation 
  !! of an array constructor value to a logical unit.
  !!
  !! @param[in]     pnode             The expression to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  SUBROUTINE acve_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueExpr), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%ex%Write(unit)
    
  END SUBROUTINE acve_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValueExpr%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an array value can contain an array value via nested 
  ! array constructors.
  RECURSIVE SUBROUTINE acve_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueExpr), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The contained expression.
    CALL pnode%ex%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE acve_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValueExpr%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an array value can contain an array value via nested 
  ! array constructors.
  RECURSIVE SUBROUTINE acve_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueExpr), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The contained expression.
    CALL pnode%ex%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE acve_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValueExpr%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION acve_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(AcValueExpr), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%ex
    
  END FUNCTION acve_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in ACValueImpliedDo
  
  
  !*****************************************************************************
  !!
  !> Implementation of AcValueImpliedDo%Write - write a debugging representation 
  !! of an array constructor value to a logical unit.
  !!
  !! @param[in]     pnode             The expression to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  SUBROUTINE acvid_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueImpliedDo), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! value index.
    
    !***************************************************************************
    
    WRITE (unit, "('(ac-implied-do: ')", ADVANCE='NO')
    
    DO i = 1, SIZE(pnode%ac_value_list)
      CALL pnode%ac_value_list(i)%Write(unit)
    END DO
    
    CALL pnode%control%Write(unit)
    
    WRITE (unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE acvid_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ACValueImpliedDo%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an array value can contain an array value via nested 
  ! array constructors.
  RECURSIVE SUBROUTINE acvid_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ACValueImpliedDo), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%ac_value_list)
      CALL pnode%ac_value_list(i)%Query(visitor)
    END DO
    
    CALL pnode%control%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE acvid_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ACValueImpliedDo%Modify- traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an array value can contain an array value via nested 
  ! array constructors.
  RECURSIVE SUBROUTINE acvid_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ACValueImpliedDo), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%ac_value_list)
      CALL pnode%ac_value_list(i)%Modify(visitor)
    END DO
    
    CALL pnode%control%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE acvid_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ACValueImpliedDo%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION acvid_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ACValueImpliedDo), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%ac_value_list(1)%item
    
  END FUNCTION acvid_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in AcValueList.
  
  
  !*****************************************************************************
  !!
  !> Forward a write operation onto the item.
  !!
  !! @param[in]     pnode             The parse node element to write.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  ! Array constructor values can contain other acv's, hence RECURSIVE.
  RECURSIVE SUBROUTINE acvl_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueList), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%item%Write(unit)
    
  END SUBROUTINE acvl_Write
  
  
  !*****************************************************************************
  !!
  !> Forward a query operation onto the item.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Array constructor values can contain other acv's, hence RECURSIVE.
  RECURSIVE SUBROUTINE acvl_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueList), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Query(visitor)
    
  END SUBROUTINE acvl_Query
  
  
  !*****************************************************************************
  !!
  !> Forward a modify operation onto the item.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Array constructor values can contain other acv's, hence RECURSIVE.
  RECURSIVE SUBROUTINE acvl_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AcValueList), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Modify(visitor)
    
  END SUBROUTINE acvl_Modify
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ACImpliedDoControl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of ACImpliedDoControl%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION acidcontrol_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'ac-implied-do-control'
    
  END FUNCTION acidcontrol_What
  
END MODULE ArrayConstructors
