! $Id: AssignmentStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AssignmentStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing assignment statements.

MODULE AssignmentStmts
  
  USE Designators
  USE Expressions
  USE ParseNodes
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAssignmentStmt
  
  !> The results of parsing an assignment statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: AssignmentStmt
    !> The variable on the left hand side of the assignment statement.
    CLASS(Variable), ALLOCATABLE :: var
    !> The expression on the right hand side of the assignment statement.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Dump => assignment_Dump
    PROCEDURE :: Query => assignment_Query
    PROCEDURE :: Modify => assignment_Modify
    PROCEDURE, NOPASS :: What => assignment_What
    PROCEDURE :: Advertise => assignment_Advertise
  END TYPE AssignmentStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parsing a sequence of tokens that is an /assignment-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    var               The variable being assigned to.  In 
  !! F2008 this can also be the result of a function that returns a POINTER., 
  !! hence this is parsed as a generic expression.
  !!
  !! @param[out]    expr              The expression being assigned.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAssignmentStmt(tlist, assign_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(AssignmentStmt), INTENT(OUT) :: assign_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_List(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /assignment-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R732'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: assign         ! Index of the assign token.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /assignment-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Find the location of the '='
    assign = FindBare(tlist, ittAssign, scck_'=')
    IF (assign == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='This /assignment-stmt/ appears to be missing a "=".' )
      RETURN
    ELSE IF (assign == 1) THEN
      ! Trap this so we don't get a bizarre "expected expression" error 
      ! from the expression parsing code.
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
!          LOCATION=tlist(1)%location,  &
          MSG='Expected /variable/.' )
    END IF
    
    ! Parse the variable bit.
    CALL ParseVariable(tlist, 1, assign - 1, assign_stmt%var, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Parse the expression bit.
    CALL ParseExpr(  &
        tlist,  &
        assign + 1, SIZE(tlist),  &
        assign_stmt%expr,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseAssignmentStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssignmentStmt%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE assignment_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssignmentStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/variable/ = ')", ADVANCE='NO') pfx
    CALL WriteVariable(unit, pnode%var)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'/expr/ = ')", ADVANCE='NO') pfx
    CALL WriteExpr(unit, pnode%expr)
    WRITE (unit, "()")
    
  END SUBROUTINE assignment_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssignmentStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE assignment_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssignmentStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%var%Query(visitor)
    
    CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE assignment_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssignmentStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE assignment_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssignmentStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%var%Modify(visitor)
    
    CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE assignment_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssignmentStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'assignment-stmt'.
  
  FUNCTION assignment_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'assignment-stmt'
    
  END FUNCTION assignment_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssignmentStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE assignment_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssignmentStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
  END SUBROUTINE assignment_Advertise
  
END MODULE AssignmentStmts
