! $Id: AssociateConstructs.f90 2200 2016-06-07 23:02:23Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AssociateConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /associate-construct/'s.

MODULE AssociateConstructs
  
  USE AssociateStmts
  USE ExecutionParts
  USE Identifiers
  USE Labels
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAssociateConstruct
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an associate construct.
  TYPE, EXTENDS(BlockStorage), PUBLIC :: AssociateConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The associations defined in the opening associate statement.
    TYPE(Association), ALLOCATABLE :: associations(:)
    !> The label on the end associate statement, if present, not allocated 
    !! otherwise.
    TYPE(Label), ALLOCATABLE :: end_label
  CONTAINS
    PROCEDURE :: Dump => associate_Dump
    PROCEDURE :: Query => associate_Query
    PROCEDURE :: Modify => associate_Modify
    PROCEDURE, NOPASS :: What => associate_What
    PROCEDURE :: Advertise => associate_Advertise
  END TYPE AssociateConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for an /associate-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    assoc_construct   The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAssociateConstruct(st, assoc_construct, err_list)
    
    USE AssociateConstructStNodes
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(AssociateConstructStNode), INTENT(INOUT) :: st
    TYPE(AssociateConstruct), INTENT(OUT) :: assoc_construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening ASSOCIATE statement.
    
    CALL ParseAssociateStmt(  &
        st%first%tlist,  &
        assoc_construct%associations,  &
        assoc_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Executable children.
    
    CALL ParseBlock(st%exec, assoc_construct%children, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! The closing end associate statement.
    
    CALL ParseEndAssociateStmt(  &
        st%last%tlist,  &
        assoc_construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    ! Save the statement label of the last statement, if any, in case 
    ! it is referenced by other statements (which is valid if those 
    ! statements are within the executable children of the construct).
    IF (ALLOCATED(st%last%statement_label)) THEN
      assoc_construct%end_label = st%last%statement_label
    END IF
    
  END SUBROUTINE ParseAssociateConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssociateConstruct%Dump - write a debugging 
  !! representation of the results of parsing an associate statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE associate_Dump(pnode, unit, pfx)
    
    USE Expressions
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssociateConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Association index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'associations = ')") pfx
      
    DO i = 1, SIZE(pnode%associations)
      WRITE ( unit, "(A,'/associate-name/ = ',A,', /selector/ = ')",  &
          ADVANCE='NO' ) pfx // '  ', .NameOf. pnode%associations(i)%name
      CALL WriteExpr(unit, pnode%associations(i)%selector)
      WRITE (unit, "()")
    END DO
      
    ! The construct name
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    WRITE (unit, "(A,'child executable statements:')") pfx
    CALL DumpBlock(pnode%children, unit, pfx // '  ')
    
  END SUBROUTINE associate_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssociateConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'associate-construct'.
  
  FUNCTION associate_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'associate-construct'
    
  END FUNCTION associate_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssociateConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE associate_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AssociateConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Associate index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%associations)
      CALL pnode%associations(i)%selector%Query(visitor)
    END DO
    
    CALL QueryBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE associate_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssociateConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE associate_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AssociateConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Associate index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%associations)
      CALL pnode%associations(i)%selector%Modify(visitor)
    END DO
    
    CALL ModifyBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE associate_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AssociateConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE associate_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssociateConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE associate_Advertise
  
END MODULE AssociateConstructs
