! $Id: AssociateStmts.f90 2819 2019-03-24 04:09:58Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the AssociateStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /associate-stmt/'s and 
!! /end-associate-stmt/'s.

MODULE AssociateStmts
  
  USE Expressions
  USE Identifiers
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseAssociateStmt
  PUBLIC :: ParseEndAssociateStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an individual /association/.
  TYPE, PUBLIC :: Association
    !> The associate name.
    TYPE(Identifier) :: name
    !> The selector.
    CLASS(Expression), ALLOCATABLE :: selector
  END TYPE Association
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /associate-stmt/ (R803).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /associate-stmt/.
  !!
  !! @param[out]    associations      The associations defined by this 
  !! associate statement.
  !!
  !! @param[out]    construct_name    The /construct-name/ in the 
  !! /associate-stmt/ if present, or not allocated if not.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseAssociateStmt(tlist, associations, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Association), INTENT(OUT), ALLOCATABLE :: associations(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /associate-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R803'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the association list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /associate-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! ASSOCIATE ...
    
    CALL ExpectKeyword(tlist, it, scck_'ASSOCIATE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ASSOCIATE(...)
    
    CALL Expect(tlist, it + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The /association-list/
    
    list_count = ListItemCount(tlist(it+2:closing-1))
    ALLOCATE(associations(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it + 2, closing - 1, item_start, item_finish)
      CALL parse_association(  &
          tlist,  &
          item_start,  &
          item_finish,  &
          associations(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Carry on in event of error.
    END DO
    
    !---------------------------------------------------------------------------
    ! Thats it!
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseAssociateStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an association (R804).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! parent statement.
  !!
  !! @param[in]     start             Start index of the association in 
  !! @a tlist.
  !!
  !! @param[in]     finish            End index of the association in @a tlist.
  !!
  !! @param[out]    assoc             The results of parsing the association.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_association(tlist, start, finish, assoc, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Association), INTENT(OUT) :: assoc
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /association/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R804'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error code for child calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /association/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! name => 
    
    CALL ExpectName(tlist, start, '/associate-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    assoc%name = IdentifierIfortBug(tlist(start))
    
    CALL Expect(tlist, start + 1, ittAssociate, scck_'=>', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The selector.
    
    CALL ParseExpr(tlist, start + 2, finish, assoc%selector, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_association
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-associate-stmt/ (R806).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-associate-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /associate-stmt/ if one was present, not allocated if a construct 
  !! name was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndAssociateStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-associate-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R806'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-associate-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDASSOCIATE or END ASSOCIATE ...
    
    CALL Expect(tlist, 1, scck_'END', scck_'ASSOCIATE', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/associate-stmt/',  &
        '/end-associate-stmt/',  &
        'C805 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndAssociateStmt
  
END MODULE AssociateStmts
