! $Id: CallStmts.f90 2826 2019-03-24 09:20:59Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the CallStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing procedure references.
!!
!! A procedure reference is either a /function-reference/ or a 
!! /subroutine-reference/.

MODULE CallStmts
  
  USE CompilerKinds
  USE Designators
  USE Expressions
  USE Identifiers
  USE Labels
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCallStmt
  PUBLIC :: ParseFunctionReference
  PUBLIC :: WriteProcedureReference
  
  PUBLIC :: PartRef       ! From Designators.
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing an /actual-arg-spec/.
  TYPE, PUBLIC :: ActualArgSpec
    !> Optional keyword.
    !!
    !! Not allocated if the keyword is not present.
    TYPE(Identifier), ALLOCATABLE :: keyword
    !> The /label/ of the /alt-return-spec/ if the /actual-arg/ is such a spec 
    !! (* /label/) or not allocated if the /actual-arg/ is something else.
    TYPE(Label), ALLOCATABLE :: alt_return_label
    !> The /actual-arg/ is not a /atl-return-spec/.
    !!
    !! Not allocated if the /actual-arg/ is an /alt-return-spec/.
    CLASS(Expression), ALLOCATABLE :: expr
  END TYPE ActualArgSpec
  
  !> The results of parsing a /call-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CallStmt
    !> The /part-ref/'s that make up the /procedure-designator/ of 
    !! the /call-stmt/, if any (may be zero sized).
    TYPE(PartRef), ALLOCATABLE :: parts(:)
    !> The procedure name, procedure pointer component name or binding name.
    TYPE(Identifier) :: name
    !> The actual arguments in the call-stmt.
    !!
    !! Always allocated after a successful parse, may be zero sized.
    TYPE(ActualArgSpec), ALLOCATABLE :: args(:)
  CONTAINS
    PROCEDURE :: Dump => call_Dump
    PROCEDURE :: Query => call_Query
    PROCEDURE :: Modify => call_Modify
    PROCEDURE :: Advertise => call_Advertise
    PROCEDURE, NOPASS :: What => call_What
  END TYPE CallStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /call-stmt/ (R1220)
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    call_stmt         The results of parsing the call 
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCallStmt(tlist, call_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(CallStmt), INTENT(OUT) :: call_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /call-stmt/
    CHARACTER(*), PARAMETER :: comp = 'R1220'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /call-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! CALL ...
    
    CALL ExpectKeyword(tlist, 1, scck_'CALL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Parse the procedure-designator that follows the CALL keyword.
    
    CALL parse_procedure_designator(  &
        tlist,  &
        2, SIZE(tlist),  &
        call_stmt%parts,  &
        call_stmt%name,  &
        it,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That can be it.
    IF (it > SIZE(tlist)) THEN
      ALLOCATE(call_stmt%args(0))
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ( /actual-arg-spec-list/ )
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing > it + 1) THEN
      CALL parse_actual_arg_spec_list(  &
          tlist,  &
          it + 1, closing - 1,  &
          call_stmt%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseCallStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /function-reference/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /function-reference/.
  !!
  !! @param[out]    finish            Index of the last token in @a tlist 
  !! of the /function-reference/
  !!
  !! @param[out]    parts             The /part-ref/'s that make up the 
  !! /procedure-designator/ of the /function-reference/, if any (may be zero 
  !! sized).
  !!
  !! @param[out]    name              The procedure name, procedure 
  !! pointer component name or binding name.
  !!
  !! @param[out]    args              The actual arguments of the 
  !! /call-stmt/.  Always allocated, may be size zero.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Function references are currently stored in FunctionRefExpr objects, 
  !! which are parse nodes.  Whether there needs to be another ParseNode 
  !! below a FunctionRefExpr needs to be considered (see the model for 
  !! StructureConstructorExpr).  Otherwise some of the error checking 
  !! code below needs to be moved into FunctionRefExpr%Check.
  
  ! A function reference can have as an argument another function 
  ! reference - hence the RECURSIVE attribute.
  RECURSIVE SUBROUTINE ParseFunctionReference( tlist, start, finish,  &
      parts, name, args, err_list )
    
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PartRef), INTENT(OUT), ALLOCATABLE :: parts(:)
    TYPE(Identifier), INTENT(OUT) :: name
    TYPE(ActualArgSpec), INTENT(OUT), ALLOCATABLE :: args(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /function-reference/.
    CHARACTER(*), PARAMETER :: comp = 'R1219'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /function-reference/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Parse the procedure-designator at the start of the reference.
    
    CALL parse_procedure_designator(  &
        tlist,  &
        start, finish,  &
        parts,  &
        name,  &
        it,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! ( /actual-arg-spec-list/ )
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing > it + 1) THEN
      CALL parse_actual_arg_spec_list(  &
          tlist,  &
          it + 1, closing - 1,  &
          args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ALLOCATE(args(0))
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it.
    
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE ParseFunctionReference
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /procedure-designator/ (R1221).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /procedure-designator/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the parent syntax rule.
  !!
  !! @param[out]    parts             The /part-ref/'s that make up the 
  !! /procedure-designator/, if any (always allocated, may be zero sized).
  !!
  !! @param[out]    name              The procedure name, procedure 
  !! pointer component name or binding name.
  !!
  !! @param[out]    after             Index of the token in @a tlist that 
  !! immediately follows the /procedure-designator/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_procedure_designator( tlist, start, finish, parts, name,  &
       after, err_list )
    
    USE Errors
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PartRef), INTENT(OUT), ALLOCATABLE :: parts(:)
    TYPE(Identifier), INTENT(OUT) :: name
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /procedure-designator/.
    CHARACTER(*), PARAMETER :: comp = 'R1221'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: percent        ! Index of the next percent character.
    INTEGER :: num_parts      ! Number of parts before the last %.
    INTEGER :: ip             ! Part index.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Parse all the /part-ref/'s (if any) up until the procedure name.
    
    percent = start - 1
    num_parts = 0
    DO
      percent = FindBare(tlist, ittPercent, scck_'%', percent + 1)
      IF ((percent == 0) .OR. (percent > finish)) EXIT
      
      num_parts = num_parts + 1
    END DO
    
    ALLOCATE(parts(num_parts))
    after = start
    DO ip = 1, num_parts
      percent = FindBare(tlist, ittPercent, scck_'%', after)
      CALL ParsePartRef(tlist, after, percent - 1, parts(ip), sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      after = percent + 1
    END DO
    ! it here is the index of the token after the last %.
    
    !---------------------------------------------------------------------------
    
    IF (num_parts == 0) THEN
      CALL ExpectName(tlist, after, '/procedure-name/', err)
    ELSE
      CALL ExpectName( tlist, after,  &
          '/binding-name/ or /procedure-component-name/', err )
    END IF
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    name = IdentifierIfortBug(tlist(after))
    
    after = after + 1
    
  END SUBROUTINE parse_procedure_designator
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /actual-arg-spec-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /actual-arg-spec-list/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /actual-arg-spec-list/.
  !!
  !! @param[out]    args              Results of parsing the argument list.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This shares much in common with the parsing of the component spec list 
  !! of a structure constructor.
  !!
  !! Because this is invoked as part of function reference processing, 
  !! and because the arguments to a function reference can invoke 
  !! functions that themselves have arguments, the parsing needs to 
  !! be RECURSIVE.
  
  RECURSIVE SUBROUTINE parse_actual_arg_spec_list( tlist, start, finish,  &
      args, err_list )
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ActualArgSpec), INTENT(OUT), ALLOCATABLE :: args(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the argument list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Flag that indicates that at least one KEYWORD= style component 
    ! specification has been seen.
    LOGICAL :: seen_keyword
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    seen_keyword = .FALSE.
    list_count = ListItemCount(tlist(start:finish))
    ALLOCATE(args(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_actual_arg_spec(  &
          tlist,  &
          item_start, item_finish,  &
          args(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
      
      IF (seen_keyword) THEN
        IF (.NOT. ALLOCATED(args(item)%keyword)) THEN
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT='C1232 on R1222',  &
              LOCATION=QueryLocation(tlist(item_start)),  &
              MSG='The keyword= may be omitted from a /actual-arg-spec/ &
                &only if the keyword= has been omitted from each &
                &preceding /actual-arg-spec/.' )
        END IF
      ELSE IF (ALLOCATED(args(item)%keyword)) THEN
        seen_keyword = .TRUE.
      END IF
    END DO
    
  END SUBROUTINE parse_actual_arg_spec_list
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /actual-arg-spec/ (R1222).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /actual-arg-spec/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /actual-arg-spec/.
  !!
  !! @param[out]    args              Results of parsing the argument.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This has some things in common with the parsing of a 
  !! /component-data-source/ of a /structure-constructor/, except that 
  !! actual arguments can be alternate return specifiers.
  !!
  !! Because this is invoked as part of function reference processing, 
  !! and because the arguments to a function reference can invoke 
  !! functions that themselves have arguments, the parsing needs to 
  !! be RECURSIVE.
  
  RECURSIVE SUBROUTINE parse_actual_arg_spec( tlist, start, finish, arg,  &
      err_list )
    
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ActualArgSpec), INTENT(OUT) :: arg
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /actual-arg-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R1222'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: assign         ! Index of = if present.
    TYPE(Error) :: err        ! Error object for expect calls.
    INTEGER :: it             ! Index of the current token.
    
    ! Error list for chlid parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    assign = FindBare(tlist, ittAssign, scck_'=', start)
    
    IF ((assign /= 0) .AND. (assign <= finish)) THEN
      CALL Expect(tlist, start, '/keyword/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      arg%keyword = IdentifierIfortBug(tlist(start))
      
      CALL ExpectEnd(tlist(:assign-1), start + 1, comp, err_list)
      it = assign + 1
    ELSE
      it = start
    END IF
    ! it here the index of the start of the /actual-arg/.
    
    !---------------------------------------------------------------------------
    ! Could be * /label/ or an expression.
    
    IF (Match(tlist, it, ittOperator, scck_'*')) THEN
      ! Looks like a alt-return-specifier.
      ALLOCATE(arg%alt_return_label)
      CALL parse_alt_return_spec(  &
          tlist,  &
          it, finish,  &
          arg%alt_return_label,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Looks like an expression.
      CALL ParseExpr(tlist, it, finish, arg%expr, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
  END SUBROUTINE parse_actual_arg_spec
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /alt-return-spec/ (R1224).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /alt-return-spec/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /alt-return-spec/.
  !!
  !! @param[out]    alt_return_label  Results of parsing the label in the 
  !! /alt-return-spec/.  Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! No need for RECURSIVE here - an /alt-return-spec/ cannot be 
  !! a function/expresion/etc.
  
  SUBROUTINE parse_alt_return_spec( tlist, start, finish,  &
      alt_return_label, err_list )
    
    USE Errors
    USE ErrorCodes
    USE ErrorLevels
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Label), INTENT(OUT) :: alt_return_label
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /alt-return-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R1224'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /alt-return-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start, ittOperator, scck_'*', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL ParseLabel(  &
        tlist,  &
        start + 1,  &
        finish,  &
        alt_return_label,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Give a warning.
    
    ! @todo Move to Check.
    CALL Add( err_list,  &
        CODE=errObsolescentAlternateReturn,  &
        LEVEL=errLevelWarning,  &
        COMPONENT=comp,  &
        LOCATION=QueryLocation(tlist(start)),  &
        MSG='An /alt-return-spec/ is an obsolescent feature.' )
    
  END SUBROUTINE parse_alt_return_spec
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of a procedure reference (from a call 
  !! statement or function reference) to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     parts             The /part-ref/'s that make up the 
  !! /procedure-designator/.
  !!
  !! @param[in]     name              The procedure name, binding name or 
  !! procedure pointer component name.
  !!
  !! @parma[in]     args              The actual argument spec list.
  
  SUBROUTINE WriteProcedureReference(log_unit, parts, name, args)
    
    USE CompilerKinds
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(PartRef), INTENT(IN) :: parts(:)
    TYPE(Identifier), INTENT(IN) :: name
    TYPE(ActualArgSpec), INTENT(IN) :: args(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Part-ref and argument index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(parts)
      CALL WritePartRef(log_unit, parts(i))
      WRITE (log_unit, "(' % ')", ADVANCE='NO')
    END DO
    WRITE (log_unit, "(A)", ADVANCE='NO') .NameOf. name
      
    WRITE (log_unit, "(' ( ')", ADVANCE='NO')
    DO i = 1, SIZE(args)
      WRITE (log_unit, "(4X)", ADVANCE='NO')
      IF (ALLOCATED(args(i)%keyword))  &
          WRITE (log_unit, "(A,' = ')", ADVANCE='NO') .NameOf. (args(i)%keyword)
      IF (ALLOCATED(args(i)%alt_return_label)) THEN
        WRITE (log_unit, "('* ', I0)", ADVANCE='NO')  &
            .ValueOf. args(i)%alt_return_label
      END IF
      IF (ALLOCATED(args(i)%expr))  &
          CALL WriteExpr(log_unit, args(i)%expr)
      IF (i /= SIZE(args)) WRITE (log_unit, "(', ')", ADVANCE='NO')
    END DO
    WRITE (log_unit, "(' )')", ADVANCE='NO')
      
  END SUBROUTINE WriteProcedureReference
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%Dump - write a debugging representation of 
  !! the results of parsing a call statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE call_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CallStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Argument and part index.
    
    !***************************************************************************
    
    ! This duplicates somewhat the code in 
    ! ParsingTestUtils::WriteProcedureReference, but we do this separately 
    ! here so we can have a friendlier output format.
    WRITE (unit, "(A,'/procedure-designator/ = ')", ADVANCE='NO') pfx
    DO i = 1, SIZE(pnode%parts)
      CALL WritePartRef(unit, pnode%parts(i))
      WRITE (unit, "(' % ')", ADVANCE='NO')
    END DO
    WRITE (unit, "(A)") .NameOf. pnode%name
      
    WRITE (unit, "(A,'/actual-arg-spec-list/')") pfx
    DO i = 1, SIZE(pnode%args)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      IF (ALLOCATED(pnode%args(i)%keyword))  &
          WRITE (unit, "(A,' = ')", ADVANCE='NO')  &
            .NameOf. (pnode%args(i)%keyword)
      IF (ALLOCATED(pnode%args(i)%alt_return_label)) THEN
        WRITE (unit, "('* ', I0)", ADVANCE='NO')  &
            .ValueOf. pnode%args(i)%alt_return_label
      END IF
      IF (ALLOCATED(pnode%args(i)%expr))  &
          CALL WriteExpr(unit, pnode%args(i)%expr)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE call_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE call_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CallStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Part and arg reference.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%args)
      IF (ALLOCATED(pnode%args(i)%expr)) THEN
        CALL pnode%args(i)%expr%Query(visitor)
      END IF
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE call_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE call_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CallStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Part and arg reference.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%parts)
      CALL pnode%parts(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%args)
      IF (ALLOCATED(pnode%args(i)%expr)) THEN
        CALL pnode%args(i)%expr%Modify(visitor)
      END IF
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE call_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE call_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CallStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ! If we have CALL abc%xyz, then we don't consider xyz a subroutine 
    ! identifier, but we do consider abc a data object.
    IF (SIZE(pnode%parts) == 0) THEN
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureRefSubroutine],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    ELSE
      ALLOCATE(ident_list(1))
      ident_list(1) = IdentifierNature(  &
          !=====================================================================
          ! ifort 19.0 bug - see ticket #48.
          !NAME=.NameOf. pnode%parts(1)%part_name,  &
          !=====================================================================
          NAME=name_bug(pnode%parts(1)%part_name),  &
          !=====================================================================
          NATURES=[NatureRefDataObject],  &
          !=====================================================================
          ! ifort 19.0 bug - see ticket #48.
          !LOCATIONS=[.LocationOf. pnode%parts(1)%part_name] )
          !=====================================================================
          LOCATIONS=[location_bug(pnode%parts(1)%part_name)] )
          !=====================================================================
    END IF
    
  END SUBROUTINE call_Advertise
  
  
  !=============================================================================
  ! ifort 19.0 bug - see ticket #48.
  !=============================================================================
  function name_bug(ident)
    type(identifier), intent(in) :: ident
    character(:), allocatable :: name_bug
    
    name_bug = .NameOf. ident
  end function name_bug
  !=============================================================================
  
  
  !=============================================================================
  ! ifort 19.0 bug - see ticket #48.
  !=============================================================================
  function location_bug(ident)
    use SourceLocations
    
    type(identifier), intent(in) :: ident
    type(sourcelocation), allocatable :: location_bug
    
    location_bug = .LocationOf. ident
  end function location_bug
  !=============================================================================
  
  
  !*****************************************************************************
  !!
  !> Implementation of CallStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'call-stmt'.
  
  FUNCTION call_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'call-stmt'
    
  END FUNCTION call_What
  
END MODULE CallStmts
