! $Id: CodimensionStmts.f90 2161 2016-06-01 23:18:33Z ian $
!> @file
!! Defines the CodimensionStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /codimension-stmts/.

MODULE CodimensionStmts
  
  USE CompilerKinds
  USE DimensionStmts
  USE Identifiers
  USE ParseNodes
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseCodimensionStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an /codimension-decl/.
  TYPE, EXTENDS(DimensionDecl), PUBLIC :: CodimensionDecl
  CONTAINS
    PROCEDURE, NOPASS :: What => codimdecl_What
    PROCEDURE :: Advertise => codimdecl_Advertise
  END TYPE CodimensionDecl
  
  !> Stores the results of parsing a /codimension-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: CodimensionStmt
    !> The items in the of /codimension-decl-list/.
    TYPE(CodimensionDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => codim_Dump
    PROCEDURE :: Query => codim_Query
    PROCEDURE :: Modify => codim_Modify
    PROCEDURE, NOPASS :: What => codim_What
  END TYPE CodimensionStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /codimension-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    codim_stmt        The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCodimensionStmt(tlist, codim_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(CodimensionStmt), INTENT(OUT) :: codim_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /codimension-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R531'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err      ! Error object for Expect calls.
    INTEGER :: opt          ! Expect option.
    INTEGER :: it           ! Current token
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /codimension-decl-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /codimension-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'CODIMENSION', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! option :: before name list.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  & 
        '"::" or /coarray-name/',  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! /allocatable-decl-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(codim_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_codimension_decl(  &
          tlist,  &
          item_start, item_finish,  &
          codim_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParseCodimensionStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /codimension-decl/ (R532).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Start index of the declaration in 
  !! @a tlist.
  !!
  !! @param[in]     finish            End index of the declaration in @a tlist.
  !!
  !! @param[out]    decl              The result of parsing the 
  !! /codimension-decl/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_codimension_decl(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(CodimensionDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /codimension-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R527'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err          ! Error object for Expect calls.
    INTEGER :: closing          ! Index of closing parenthesis.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect(tlist, start, '/coarray-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    decl%name = IdentifierIfortBug(tlist(start))
    
    !---------------------------------------------------------------------------
    ! lbracket coarray-spec rbracket
    
    CALL Expect(tlist, start + 1, ittDelimiter, scck_'[', err)    
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start + 1, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
        
    CALL ParseCoarraySpec(  &
        tlist(start+2:closing-1),  &
        decl%array_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's all folks.
    CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
    
  END SUBROUTINE parse_codimension_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of CodimensionStmt%Dump - write a debugging representation 
  !! of the results of parsing a codimension statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE codim_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CodimensionStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A,'/coarray-name/ = ',A,' [')",  ADVANCE='NO') pfx,  &
          .NameOf. pnode%decls(i)%name
      CALL WriteDimSpec(unit, pnode%decls(i)%array_spec)
      WRITE (unit, "(']')")
    END DO
    
  END SUBROUTINE codim_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of CodimensionStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE codim_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CodimensionStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE codim_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CodimensionStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE codim_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CodimensionStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Declaration list index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE codim_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CodimensionStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'codimension-stmt'.
  
  FUNCTION codim_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'codimension-stmt'
    
  END FUNCTION codim_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in CodimensioNDecl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of CodimensionDecl%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'codimension-decl'.
  
  FUNCTION codimdecl_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'codimension-decl'
    
  END FUNCTION codimdecl_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IfConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE codimdecl_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CodimensionDecl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureDefCodimension], &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE codimdecl_Advertise
  
END MODULE CodimensionStmts
