! $Id: ContinueStmts.f90 2161 2016-06-01 23:18:33Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ContinueStmts module.


!*******************************************************************************
!!
!> Procedure for parsing /continue-stmt/'s.

MODULE ContinueStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseContinueStmt
  
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ContinueStmt
  CONTAINS
    PROCEDURE :: Dump => continue_Dump
    PROCEDURE, NOPASS :: What => continue_What
    PROCEDURE :: Query => continue_Query
    PROCEDURE :: Modify => continue_Modify
  END TYPE ContinueStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /continue-stmt/ (R854).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /continue-stmt/ consists solely of the keyword "CONTINUE".
  
  SUBROUTINE ParseContinueStmt(tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /continue-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R854'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /continue-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! CONTINUE ...
    
    CALL ExpectKeyword(tlist, 1, scck_'CONTINUE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    ! Nothing else!
    CALL ExpectEOS(tlist, 2, comp, err_list)
    
  END SUBROUTINE ParseContinueStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContinueStmt%Dump - write a debugging 
  !! representation of the ParseNode to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE continue_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ContinueStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    ! Nothing to do.
    
  END SUBROUTINE continue_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContinueStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE continue_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ContinueStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE continue_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContinueStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE continue_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ContinueStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE continue_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ContinueStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'continue-stmt'.
  
  FUNCTION continue_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'continue-stmt'
    
  END FUNCTION continue_What
  
END MODULE ContinueStmts
