! $Id: DataStmts.f90 2800 2019-03-22 18:13:05Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DataStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /data-stmts/.

MODULE DataStmts
  
  USE CompilerKinds
  USE ComplexLiteralConstants
  USE Designators
  USE Expressions
  USE ImpliedDos
  USE ParseNodes
  USE StructureConstructors
  USE SubscriptLists
  USE Tokens
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDataStmt
  
  !-----------------------------------------------------------------------------
  
  !> Parent type for storing the result of parsing a /data-stmt-object/.
  TYPE, EXTENDS(TermParseNode), PUBLIC, ABSTRACT :: DataStmtObject
  CONTAINS
    PROCEDURE, NOPASS :: What=> dso_What
  END TYPE DataStmtObject
  
  !> An element in a sequence of /data-stmt-object/'s.
  TYPE, PUBLIC :: DataStmtObjectList
    CLASS(DataStmtObject), ALLOCATABLE :: item
  END TYPE DataStmtObjectList
  
  !> Parent type for storing a /data-i-do-object/.
  TYPE, EXTENDS(TermParseNode), PUBLIC, ABSTRACT :: DataIDoObject
  CONTAINS
    PROCEDURE, NOPASS :: What => dido_What
  END TYPE DataIDoObject
  
  !> An element in a sequence of /data-i-do-object/'s.
  TYPE, PUBLIC :: DataIDoObjectList
    CLASS(DataIDoObject), ALLOCATABLE :: item
  END TYPE DataIDoObjectList
  
  !> Parent type for a /data-stmt-constant/.
  TYPE, PUBLIC, ABSTRACT :: DataStmtConstant
  CONTAINS
    PROCEDURE(dsc_Query), DEFERRED :: Query
    PROCEDURE(dsc_Modify), DEFERRED :: Modify
  END TYPE DataStmtConstant
  
  ! Interfaces for bindings in DataStmtConstant
  ABSTRACT INTERFACE
    !> Forward the Query operation.
    SUBROUTINE dsc_Query(constant, visitor)
      IMPORT :: DataStmtConstant
      IMPORT :: ParseNodeQueryVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(DataStmtConstant), INTENT(IN) :: constant
      CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE dsc_Query
    
    !> Forward the Modify operation.
    SUBROUTINE dsc_Modify(constant, visitor)
      IMPORT :: DataStmtConstant
      IMPORT :: ParseNodeModifyVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(DataStmtConstant), INTENT(INOUT) :: constant
      CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE dsc_Modify
  END INTERFACE
  
  !> Parent type for a /data-stmt-repeat/.
  TYPE, PUBLIC, ABSTRACT :: DataStmtRepeat
  CONTAINS
    PROCEDURE(dsr_Query), DEFERRED :: Query
    PROCEDURE(dsr_Modify), DEFERRED :: Modify
  END TYPE DataStmtRepeat
  
  ! Interfaces for bindings in DataStmtRepeat.
  ABSTRACT INTERFACE
    !> Forward the Query operation.
    SUBROUTINE dsr_Query(repeat, visitor)
      IMPORT :: DataStmtRepeat
      IMPORT :: ParseNodeQueryVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(DataStmtRepeat), INTENT(IN) :: repeat
      CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE dsr_Query
    
    !> Forward the Modify operation.
    SUBROUTINE dsr_Modify(repeat, visitor)
      IMPORT :: DataStmtRepeat
      IMPORT :: ParseNodeModifyVisitor
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(DataStmtRepeat), INTENT(INOUT) :: repeat
      CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    END SUBROUTINE dsr_Modify
  END INTERFACE
  
  !> A /data-stmt-value/.
  TYPE, PUBLIC :: DataStmtValue
    !> The repeat value, not allocatated if not present.
    CLASS(DataStmtRepeat), ALLOCATABLE :: repeat
    !> The constant.
    !!
    !! Always allocated.
    CLASS(DataStmtConstant), ALLOCATABLE :: constant
  END TYPE DataStmtValue
  
  !> Stores the results of parsing a /data-stmt-set/.
  TYPE, PUBLIC :: DataStmtSet
    !> The object list
    TYPE(DataStmtObjectList), ALLOCATABLE :: objects(:)
    !> List of values.
    TYPE(DataStmtValue), ALLOCATABLE :: values(:)
  END TYPE DataStmtSet
  
  !> Stores the results of parsing a /data-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: DataStmt
    !> The list of /data-stmt-set/'s.
    TYPE(DataStmtSet), ALLOCATABLE :: data_set_list(:)
  CONTAINS
    PROCEDURE :: Dump => data_Dump
    PROCEDURE :: Query => data_Query
    PROCEDURE :: Modify => data_Modify
    PROCEDURE, NOPASS :: What => data_What
  END TYPE DataStmt
  
  !-----------------------------------------------------------------------------
  ! Extensions of DataStmtObject.
  !
  ! What is implemented by DataStmtObject.  Extensions must implement 
  ! Write, Query, Modify and GetLocation.
  
  !> A /data-stmt-object/ that is a /variable/.
  TYPE, PUBLIC, EXTENDS(DataStmtObject) :: DataStmtObjectVariable
    !> The variable nominated by a designator.  Constraints on the variable 
    !! mean that it cannot be a function reference - this is subject to 
    !! an interpretation F08/0077.
    !!
    !! @todo Resolve.
    TYPE(Designator) :: desig
  CONTAINS
    PROCEDURE :: Write => dsov_Write
    PROCEDURE :: Query => dsov_Query
    PROCEDURE :: Modify => dsov_Modify
    PROCEDURE :: GetLocation => dsov_GetLocation
  END TYPE DataStmtObjectVariable
  
  !> What would be a /data-implied-do-control/, if they had bothered to 
  !! split out that particular bit of syntax.
  TYPE, PUBLIC, EXTENDS(ImpliedDoControl) :: DataImpliedDoControl
  CONTAINS
    PROCEDURE, NOPASS :: What => datacontrol_What
  END TYPE DataImpliedDoControl
  
  !> A /data-implied-do/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: DataImpliedDo
    !> The items in the /data-i-do-object-list/.
    TYPE(DataIDoObjectList), ALLOCATABLE :: items(:)
    !> The control part of the implied do.
    TYPE(DataImpliedDoControl) :: do_control
  CONTAINS
    PROCEDURE :: Write => did_Write
    PROCEDURE :: Query => did_Query
    PROCEDURE :: Modify => did_Modify
    PROCEDURE, NOPASS :: What => did_What
    PROCEDURE :: GetLocation => did_GetLocation
  END TYPE DataImpliedDo
  
  !> A /data-stmt-object/ that is a /data-implied-do/.
  TYPE, PUBLIC, EXTENDS(DataStmtObject) :: DataStmtObjectDataImpliedDo
    TYPE(DataImpliedDo) :: obj
  CONTAINS
    PROCEDURE :: Write => dsodid_Write
    PROCEDURE :: Query => dsodid_Query
    PROCEDURE :: Modify => dsodid_Modify
    PROCEDURE :: GetLocation => dsodid_GetLocation
  END TYPE DataStmtObjectDataImpliedDo
  
  !-----------------------------------------------------------------------------
  ! Extensions of DataIDoObject
  !
  ! What is implemented by DataIDoObject.  Extensions must implement 
  ! Write, Query, Modify and GetLocation.
  
  !> A /data-i-do-object/ that is an /array-element/.
  TYPE, PUBLIC, EXTENDS(DataIDoObject) :: DataIDoObjectDesignator
    ! The designator of the .
    TYPE(Designator) :: desig
  CONTAINS
    PROCEDURE :: Write => didod_Write
    PROCEDURE :: Query => didod_Query
    PROCEDURE :: Modify => didod_Modify
    PROCEDURE :: GetLocation => didod_GetLocation
  END TYPE DataIDoObjectDesignator
  
  !> A /data-i-do-object/ that is an /data-implied-do/.
  TYPE, PUBLIC, EXTENDS(DataIDoObject) :: DataIDoObjectDataImpliedDo
    TYPE(DataImpliedDo) :: obj
  CONTAINS
    PROCEDURE :: Write => didodid_Write
    PROCEDURE :: Query => didodid_Query
    PROCEDURE :: Modify => didodid_Modify
    PROCEDURE :: GetLocation => didodid_GetLocation
  END TYPE DataIDoObjectDataImpliedDo
  
  !-----------------------------------------------------------------------------
  ! Extensions of DataStmtRepeat
  
  !> A /data-stmt-repeat/ that is a designator, which is a named constant.
  TYPE, PUBLIC, EXTENDS(DataStmtRepeat) :: DataStmtRepeatDesig
    TYPE(Designator) :: desig
  CONTAINS
    PROCEDURE :: Query => dsrd_Query
    PROCEDURE :: Modify => dsrd_Modify
  END TYPE DataStmtRepeatDesig
  
  !> A /data-stmt-repeat/ that is an integer literal constant.
  TYPE, PUBLIC, EXTENDS(DataStmtRepeat) :: DataStmtRepeatLiteral
    TYPE(Token) :: tok
  CONTAINS
    PROCEDURE :: Query => dsrl_Query
    PROCEDURE :: Modify => dsrl_Modify
  END TYPE DataStmtRepeatLiteral  
  
  !-----------------------------------------------------------------------------
  ! Extensions of DataStmtConstant
  
  !> A /data-stmt-constant/ that is a designator of some sort.
  TYPE, PUBLIC, EXTENDS(DataStmtConstant) :: DataStmtConstantDesig
    !> The designator.
    TYPE(Designator) :: desig
  CONTAINS
    PROCEDURE :: Query => dscd_Query
    PROCEDURE :: Modify => dscd_Modify
  END TYPE DataStmtConstantDesig
  
  !> A /data-stmt-constant/ that is a /structure-constructor/.
  TYPE, PUBLIC, EXTENDS(DataStmtConstant) :: DataStmtConstantSC
    !> The results of parsing the structure constructor.
    TYPE(StructureConstructor) :: sc
  CONTAINS
    PROCEDURE :: Query => dscsc_Query
    PROCEDURE :: Modify => dscsc_Modify
  END TYPE DataStmtConstantSC
  
  ! The following types store tokens.  This may not be appropriate - as 
  ! per the comments at the top of the PrimaryExpressions module we may 
  ! want to conceptually have abandoned tokens by the end of parsing.
  
  !> A /data-stmt-constant/ that is a simple literal.
  TYPE, PUBLIC, EXTENDS(DataStmtConstant) :: DataStmtConstantSimpleLiteral
    !> The sign token, if present.  
    !!
    !! Not allocated if the sign was not present.  The sign token is never 
    !! present if \a tok is a character or logical literal token.
    TYPE(Token), ALLOCATABLE :: sign_tok
    !> The value part of the simple literal.
    TYPE(Token) :: tok
  CONTAINS
    PROCEDURE :: Query => dscsl_Query
    PROCEDURE :: Modify => dscsl_Modify
  END TYPE DataStmtConstantSimpleLiteral
  
  !> A /data-stmt-constant/ that is a complex literal.
  !!
  !! The contents of this are the same as the contents of the 
  !! ComplexLiteralExpr type.  Perhaps we should factor this out.
  TYPE, PUBLIC, EXTENDS(DataStmtConstant) :: DataStmtConstantComplexLiteral
    !> The results of prasing the constant.
    TYPE(ComplexLiteralConstant) :: constant
  CONTAINS
    PROCEDURE :: Query => dsccl_Query
    PROCEDURE :: Modify => dsccl_Modify
  END TYPE DataStmtConstantComplexLiteral
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /data-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    data_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDataStmt(tlist, data_stmt, err_list)
    
    USE Errors
    USE ErrorCodes
    USE LanguageBindingSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(DataStmt), INTENT(OUT) :: data_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule for a /data-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R534'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Index of the current token.
    INTEGER :: it_end         ! End of the current syntax part.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Temporary for growing the list of sets.
    TYPE(DataStmtSet), ALLOCATABLE :: tmp_sets(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /data-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! DATA...
    
    CALL ExpectKeyword(tlist, 1, scck_'DATA', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! A list of /data-stmt-set/'s.
    
    ! We can't use list processing routines becase the first part of a 
    ! /data-stmt-set/ could itself be a list.
    
    ! There will always be one item in a valid statement.
    ALLOCATE(data_stmt%data_set_list(1))
    it = 2
    DO
      CALL parse_data_stmt_set(  &
          tlist, it,  &
          data_stmt%data_set_list(SIZE(data_stmt%data_set_list)),  &
          it_end,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (it_end > SIZE(tlist)) EXIT
      it = it_end
      
      ! Create space for the next item
      ALLOCATE(tmp_sets(SIZE(data_stmt%data_set_list) + 1))
      tmp_sets(:SIZE(data_stmt%data_set_list)) = data_stmt%data_set_list
      
      CALL MOVE_ALLOC(tmp_sets, data_stmt%data_set_list)
      
      ! Move over the comma
      CALL Expect(tlist, it, ittComma, scck_',', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = it + 1
    END DO
    
  END SUBROUTINE ParseDataStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /data-stmt-set/ (R535).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token that start the 
  !! /data-stmt-set/.
  !!
  !! param[out]     data_set          The results of parsing the 
  !! /data-stmt-set/.
  !!
  !! @param[out]    after             Index of the token following the 
  !! /data-stmt-set/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE parse_data_stmt_set(tlist, start, data_set, after, err_list)
    
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(DataStmtSet), INTENT(OUT) :: data_set
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule for a /data-stmt-set/.
    CHARACTER(*), PARAMETER :: comp = 'R535'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: slash          ! Index of the slash.
    INTEGER :: list_finish    ! Index of the end of one of the lists.
    
    INTEGER :: list_count     ! Number of items in the /data-stmt-object-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! /data-stmt-object-list/, terminated by a /.
    
    ! Find the /.  If we don't find it we process the remainder of the 
    ! statement as the list.
    slash = FindBare(tlist, ittOperator, scck_'/', start)
    IF (slash == 0) THEN
      list_finish = SIZE(tlist)
    ELSE
      list_finish = slash - 1
    END IF
    
    list_count = ListItemCount(tlist(start:list_finish))
    ALLOCATE(data_set%objects(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, list_finish, item_start, item_finish)
      CALL parse_data_stmt_object( tlist, item_start, item_finish,  &
          data_set%objects(item)%item, sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! The terminating slash
    CALL Expect(tlist, list_finish + 1, ittOperator, scck_'/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! This should be the first token of the value list.
    it = list_finish + 2
    
    !---------------------------------------------------------------------------
    ! /data-stmt-value-list/, terminated by a /.
    
    ! Find the /.  If we don't find it we process the remainder of the 
    ! statement as the list.
    slash = FindBare(tlist, ittOperator, scck_'/', it)
    IF (slash == 0) THEN
      list_finish = SIZE(tlist)
    ELSE
      list_finish = slash - 1
    END IF
    
    list_count = ListItemCount(tlist(it:list_finish))
    ALLOCATE(data_set%values(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, list_finish, item_start, item_finish)
      CALL parse_data_stmt_value( tlist, item_start, item_finish,  &
          data_set%values(item), sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! The terminating slash
    CALL Expect(tlist, list_finish + 1, ittOperator, scck_'/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Position after the end of the /data-stmt-set/.
    after = list_finish + 2
    
  END SUBROUTINE parse_data_stmt_set
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-stmt-object/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The start of the /data-stmt-object/ in 
  !! the sequence of tokens.
  !!
  !! @param[in]     finish            The end of the /data-stmt-object/ in 
  !! the sequence of tokens.
  !!
  !! @param[in]     object            An object with the results of parsing 
  !! the /data-stmt-object/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /data-stmt-object/ can be a /variable/ or a /data-implied-do/.
  
  SUBROUTINE parse_data_stmt_object(tlist, start, finish, object, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(DataStmtObject), INTENT(OUT), ALLOCATABLE :: object
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule for a /data-stmt-object/.
    CHARACTER(*), PARAMETER :: comp = 'R536'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of closing delimiter.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: assign         ! Index of =.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    CLASS(DataStmtObjectDataImpliedDo), ALLOCATABLE :: object_implied_do
    CLASS(DataStmtObjectVariable), ALLOCATABLE :: object_variable
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /data-stmt-object/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps an implied do?
    
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Is there a bare '=' in the parenthesised tokens?
      assign = FindBare(tlist(start+1:closing-1), ittAssign, scck_'=')
      IF (assign /= 0) THEN
        ! We appears to have a /data-implied-do/.
        ALLOCATE(object_implied_do)
        CALL parse_data_implied_do_content(tlist, start + 1, closing - 1,  &
            object_implied_do%obj, sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(object_implied_do, object)
        CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
        RETURN
      END IF
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a variable ?
    
    ALLOCATE(object_variable)
    CALL ParseDesignator( tlist(start:finish), object_variable%desig,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    CALL MOVE_ALLOC(object_variable, object)
    
  END SUBROUTINE parse_data_stmt_object
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is the content of /data-implied-do/.
  !!
  !! Recursive because /data-implied-do/'s can have child /data-implied-do/'s.
  !!
  !! @param[in]     tlist             Sequence of tokens.
  !!
  !! @param[in]     start             Index of the token after the opening 
  !! parenthesis.
  !!
  !! @param[in]     finish            Index of the token before the closing 
  !! parenthesis.
  !!
  !! @param[out]    implied_do        The results of parsing the 
  !! /data-implied-do/.
  !!
  !! @param[out]    err_list          List of errors.
  
  RECURSIVE SUBROUTINE parse_data_implied_do_content( tlist, start,  &
      finish, implied_do, err_list )
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DataImpliedDo), INTENT(OUT) :: implied_do
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule of a /data-implied-do/.
    CHARACTER(*), PARAMETER :: comp = 'R537'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the implied do.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
        
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Work out how many items and parse the loop control parts.
    
    CALL ParseImpliedDo( tlist, start, finish, 'data-i-do-object',  &
        '/data-implied-do/', comp, list_count, implied_do%do_control,  &
        sub_err_list )
        
    !---------------------------------------------------------------------------
    ! Parse the items
    
    ALLOCATE(implied_do%items(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL parse_data_i_do_object( tlist, item_start, item_finish,  &
          implied_do%items(item)%item, sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE parse_data_implied_do_content
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-i-do-object/ (R538).
  !!
  !! A /data-i-do-object/ can be an /array-element/, a 
  !! /scalar-structure-component/ or a /data-implied-do/.
  !!
  !! @param[in]     tlits             The sequence of tokens.
  !!
  !! @param[in]     start             Index in @a tlist of the token that is 
  !! the start of the /data-i-do-object/ in @a tlist..
  !!
  !! @param[in]     finish            Index in @a tlist of the token that is 
  !! the end of the /data-i-do-object/.
  !!
  !! @param[out]    object            The results of parsing the 
  !! /data-implied-do-object/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Recursive because child /data-implied-do/ objects will have child 
  !! (grandchild) /data-i-do-object/'s.
  
  RECURSIVE SUBROUTINE parse_data_i_do_object( tlist, start, finish, object,  &
      err_list )
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(DataIDoObject), INTENT(OUT), ALLOCATABLE :: object
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule of a /data-i-do-object/.
    CHARACTER(*), PARAMETER :: comp = 'R538'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of closing parenthesis.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Temporary for construction of a /data-i-do-object/ that is an implied do.
    CLASS(DataIDoObjectDataImpliedDo), ALLOCATABLE :: object_implied_do
    
    ! Temporary for construction of a /data-i-do-object/ that is a designator.
    CLASS(DataIDoObjectDesignator), ALLOCATABLE :: object_desig
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /data-i-do-object/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a nested /data-implied-do/ ?
    
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      ! Looks like a /data-implied-do/.
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ALLOCATE(object_implied_do)
      CALL parse_data_implied_do_content(tlist, start + 1, closing - 1,  &
          object_implied_do%obj, sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(object_implied_do, object)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Some sort of designator.
    
    ALLOCATE(object_desig)
    CALL ParseDesignator( tlist(start:finish), object_desig%desig,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    CALL MOVE_ALLOC(object_desig, object)
    
  END SUBROUTINE parse_data_i_do_object
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-stmt-value/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index of the start of the 
  !! /data-stmt-value/ in the token sequence.
  !!
  !! @param[in]     finish            The index of the finish of the 
  !! /data-stmt-value/ in the token sequence.
  !!
  !! @param[out]    value             The results of parsing the 
  !! /data-stmt/value.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_data_stmt_value(tlist, start, finish, value, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DataStmtValue), INTENT(OUT) :: value
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule of a /data-stmt-value/.
    CHARACTER(*), PARAMETER :: comp = 'R540'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: star     ! Index of the * token.
    INTEGER :: it       ! Current token index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Look for a repeat
    
    star = FindBare(tlist(:finish), ittOperator, scck_'*', start)
    IF (star /= 0) THEN
      CALL parse_data_stmt_repeat(  &
          tlist,  &
          start,  &
          star - 1,  &
          value%repeat,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      it = star + 1
    ELSE
      it = start
    END IF
    
    ! it here is the index of the token that starts the constant
    
    !---------------------------------------------------------------------------
    ! The constant part of the value.
    
    CALL parse_data_stmt_constant(  &
        tlist,  &
        it,  &
        finish,  &
        value%constant,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
  END SUBROUTINE parse_data_stmt_value
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-stmt-repeat/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index of the start of the 
  !! /data-stmt-repeat/ in @a tlist.
  !!
  !! @param[in]     finish            The index of the end of the 
  !! /data-stmt-repeat in @a tlist.
  !!
  !! @param[out]    repeat            Results of parsing the 
  !! /data-stmt-repeat/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /data-stmt-repeat/ is a specific form of designator, and that's 
  !! how we treat it.
  
  SUBROUTINE parse_data_stmt_repeat(tlist, start, finish, repeat, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(DataStmtRepeat), INTENT(OUT), ALLOCATABLE :: repeat
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! The syntax rule of a /data-stmt-repeat/.
    CHARACTER(*), PARAMETER :: comp = 'R541'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option
    TYPE(Error) :: err        ! Expect error object.
    
    ! Temporary for construction of an object to store the results of parsing 
    ! a /data-stmt-repeat/ that is a literal.
    CLASS(DataStmtRepeatLiteral), ALLOCATABLE :: lit_repeat
    
    ! Temporary for construction of an object to store the results of parsing 
    ! a /data-stmt-repeat/ that is a designator.
    CLASS(DataStmtRepeatDesig), ALLOCATABLE :: desig_repeat
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect( tlist, start, [ittIntegerLiteral, ittName],  &
        [scck_'', scck_''], '/integer-literal/ or /name/', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ALLOCATE(lit_repeat)
      lit_repeat%tok = tlist(start)
      CALL MOVE_ALLOC(lit_repeat, repeat)
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      RETURN
    END IF
    
    ALLOCATE(desig_repeat)
    CALL ParseDesignator( tlist(start:finish), desig_repeat%desig,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    CALL MOVE_ALLOC(desig_repeat, repeat)
  
  END SUBROUTINE parse_data_stmt_repeat
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /data-stmt-constant/.
  !!
  !! @param[in]     tlist             The sequence of tokens for the parent 
  !! statement.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /data-stmt-constant/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /data-stmt-constant/ in @a tlist.
  !!
  !! @param[out]    constant          Results of parsing the constant.  Not 
  !! allocated if the constant could not be parsed.  May still be allocated 
  !! even if errors occurred, but some sense of the constant could be made.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_data_stmt_constant(tlist, start, finish, constant, err_list)
    
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE StructureConstructors
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(DataStmtConstant), INTENT(OUT), ALLOCATABLE :: constant
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /data-stmt-constant/.
    CHARACTER(*), PARAMETER :: comp = 'R542'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Options for the constant.  One of these will be allocated and 
    ! defined before being MOVE_ALLOC'd to the constant dummy argument.
    TYPE(DataStmtConstantDesig), ALLOCATABLE :: dsr_desig
    TYPE(DataStmtConstantSC), ALLOCATABLE :: dsr_struct
    TYPE(DataStmtConstantSimpleLiteral), ALLOCATABLE :: dsr_simple
    TYPE(DataStmtConstantComplexLiteral), ALLOCATABLE :: dsr_complex
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    TYPE(Error) :: err        ! Error object for Expect call.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (start > finish) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /data-stmt-constant/.' )
      RETURN
    END IF
    
    ! @todo Rewrite this using a great big Expect for the initial token 
    ! types.
    
    !---------------------------------------------------------------------------
    ! Perhaps a structure constructor?  We can't tell some of these apart from 
    ! designators, so we just look here for the ones that are definitely 
    ! not designators - they have a keyword= type form or no subscripts.
    !
    ! If the user is being a little pathological, and has an accessible type 
    ! named NULL in scope, then the structure constructor for NULL starts 
    ! to look like a null-init.  If the structure constructor requires 
    ! no component specifications, then it looks identically like a null-init.
    
    IF (test_structure_constructor(tlist(start:finish))) THEN
      ALLOCATE(dsr_struct)
      CALL ParseStructureConstructor(  &
          tlist,  &
          start,  &
          finish,  &
          dsr_struct%sc,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      CALL MOVE_ALLOC(dsr_struct, constant)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a literal?
    
    ! Look for unsigned literal (except character - they are treated 
    ! separately because they can start designators).
    IF (ANY( tlist(start)  &
        == [ ittLogicalLiteral, ittIntegerLiteral, ittRealLiteral ] )) THEN
      ALLOCATE(dsr_simple)
      dsr_simple%tok = tlist(start)
      CALL MOVE_ALLOC(dsr_simple, constant)
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      RETURN
    END IF
    
    ! Maybe +/- integer or +/- real?
    IF ( IsToken(tlist(start), ittOperator, scck_'+')  &
        .OR. IsToken(tlist(start), ittOperator, scck_'-') ) THEN
      CALL Expect( tlist,  &
          start + 1,  &
          [ittIntegerLiteral, ittRealLiteral],  &
          ['', ''],  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      ALLOCATE(dsr_simple)
      dsr_simple%sign_tok = tlist(start)
      dsr_simple%tok = tlist(start+1)
      CALL MOVE_ALLOC(dsr_simple, constant)
      CALL ExpectEnd(tlist(:finish), start + 2, comp, err_list)
      RETURN
    END IF
    
    ! Maybe a complex constant.
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ALLOCATE(dsr_complex)
      CALL ParseComplexLiteralConstant(  &
          tlist,  &
          start + 1,  &
          closing - 1,  &
          dsr_complex%constant,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      CALL MOVE_ALLOC(dsr_complex, constant)
      CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
      RETURN
    END IF
    
    ! Maybe a simple character literal?
    IF ( (tlist(start) == ittCharacterLiteral)  &
        .AND. (.NOT. Match(tlist, start + 1, ittDelimiter, scck_'(')) ) THEN
      ALLOCATE(dsr_simple)
      dsr_simple%tok = tlist(start)
      CALL MOVE_ALLOC(dsr_simple, constant)
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! A designator
    
    ALLOCATE(dsr_desig)
    CALL ParseDesignator(tlist(start:finish), dsr_desig%desig, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    CALL MOVE_ALLOC(dsr_desig, constant)
    
  END SUBROUTINE parse_data_stmt_constant
  
  
  !*****************************************************************************
  !!
  !> Tests a sequence of tokens to see whether it could be considered a 
  !! designator or whether it must be a structure constructor.
  !!
  !! Something which looks like a designator could also just be a structure
  !! constructor, but the reverse does not necessarily apply.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns .TRUE. if the sequence is definitely a structure constructor, 
  !! .FALSE. otherwise.
  !!
  !! Similar code in ParseExpressions::test_structure_constructor, but here 
  !! we don't have to be so fussy, as the number of valid options for 
  !! a /data-stmt-constant/ is less - notably - no general function references 
  !! (a reference to the intrinsic NULL is permitted, but we consider that 
  !! a structure constructor for now).  Basically - we are trying to exclude 
  !! something from being a designator.
  
  FUNCTION test_structure_constructor(tlist) RESULT(b)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    LOGICAL :: b
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of the closing delimiter.
    
    !***************************************************************************
    
    ! Structure constructor must start with a name.
    b = Match(tlist, 1, ittName)
    IF (.NOT. b) RETURN
    
    ! Structure constructor cannot have a bare '%'
    b = FindBare(tlist, ittPercent, scck_'%') == 0
    IF (.NOT. b) RETURN
    
    ! Second token of structure constructor must be a (.
    b = Match(tlist, 2, ittDelimiter, scck_'(')
    IF (.NOT. b) RETURN
    
    ! That ( must be matched.
    closing = FindBare(tlist, ittDelimiter, scck_')', 2)
    b = closing /= 0
    IF (.NOT. b) RETURN
    
    ! abc() is not a designator but it could be a structure constructor 
    ! (one for a type that has no components or all default initialised 
    ! components).
    b = closing == 3
    IF (b) RETURN
    
    ! If there is a second set of parentheses - we have a structure 
    ! constructor.
    b = Match(tlist, closing + 1, ittDelimiter, scck_'(')
    IF (b) RETURN
    
    ! Here if only one parenthesis pair.
    
    ! Is there a bare '=' in the parenthesised tokens?
    b = FindBare(tlist(3:closing-1), ittAssign, scck_'=') /= 0
    IF (b) RETURN
    
  END FUNCTION test_structure_constructor
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmt%Dump - write a debugging 
  !! representation of the results of parsing a data statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE data_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: is             ! Set index.
    INTEGER :: i              ! Object index in a set.
    
    !***************************************************************************
    
    DO is = 1, SIZE(pnode%data_set_list)
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      DO i = 1, SIZE(pnode%data_set_list(is)%objects)
        ! We have a binding for DataImpliedDo to call, but not for 
        ! DataStmtObject* extensions in general.
        SELECT TYPE (obj => pnode%data_set_list(is)%objects(i)%item)
        TYPE IS (DataStmtObjectVariable)
          CALL WriteDesignator(unit, obj%desig)
        TYPE IS (DataStmtObjectDataImpliedDo)
          CALL did_Write(obj%obj, unit)
        END SELECT 
      END DO
      
      WRITE (unit, "('/')", ADVANCE='NO')
      DO i = 1, SIZE(pnode%data_set_list(is)%values)
        ASSOCIATE(val => pnode%data_set_list(is)%values(i))
          IF (ALLOCATED(val%repeat)) THEN
            CALL write_data_stmt_repeat(unit, val%repeat)
            WRITE (unit, "(' * ')", ADVANCE='NO')
          END IF
          CALL write_data_stmt_constant(unit, val%constant)
        END ASSOCIATE
      END DO
      WRITE (unit, "('/')")
    END DO
    
  END SUBROUTINE data_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE data_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! Set index.
    INTEGER :: i              ! value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO is = 1, SIZE(pnode%data_set_list)
      ASSOCIATE(set => pnode%data_set_list(is))
        CALL QueryDataStmtObjectList(set%objects, visitor)
        DO i = 1, SIZE(set%values)
          IF (ALLOCATED(set%values(i)%repeat)) THEN
            CALL set%values(i)%repeat%Query(visitor)
          END IF
          CALL set%values(i)%constant%Query(visitor)
        END DO
      END ASSOCIATE
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE data_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE data_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: is             ! Set index.
    INTEGER :: i              ! value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO is = 1, SIZE(pnode%data_set_list)
      ASSOCIATE(set => pnode%data_set_list(is))
        CALL ModifyDataStmtObjectList(set%objects, visitor)
        DO i = 1, SIZE(set%values)
          IF (ALLOCATED(set%values(i)%repeat)) THEN
            CALL set%values(i)%repeat%Modify(visitor)
          END IF
          CALL set%values(i)%constant%Modify(visitor)
        END DO
      END ASSOCIATE
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE data_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'data-stmt'.
  
  FUNCTION data_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'data-stmt'
    
  END FUNCTION data_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataStmtObject and extensions.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObject%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'data-stmt-object'.
  
  FUNCTION dso_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'data-stmt-object'
    
  END FUNCTION dso_What
  
  
  !*****************************************************************************
  !!
  !> Forward a Query operation to all elements in an array of type 
  !! DataStmtObjectList.
  !!
  !! @param[in]    array              
  
  SUBROUTINE QueryDataStmtObjectList(array, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DataStmtObjectList), INTENT(IN) :: array(:)
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! object index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(array)
      CALL array(i)%item%Query(visitor)
    END DO
    
  END SUBROUTINE QueryDataStmtObjectList
  
  
  !*****************************************************************************
  !!
  !> Forward a Modify operation to all elements in an array of type 
  !! DataStmtObjectList.
  
  SUBROUTINE ModifyDataStmtObjectList(array, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DataStmtObjectList), INTENT(INOUT) :: array(:)
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! object index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(array)
      CALL array(i)%item%Modify(visitor)
    END DO
    
  END SUBROUTINE ModifyDataStmtObjectList
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%Write - write a debugging representation 
  !! of an array constructor value to a logical unit.
  !!
  !! @param[in]     pnode             The expression to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  RECURSIVE SUBROUTINE dsov_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectVariable), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%desig%Write(unit)
    
  END SUBROUTINE dsov_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectVariable%Query - perform a query 
  !! operation on the contents of the parse node.
  !!
  !! @param[in]     pnode             The data statement object being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dsov_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectVariable), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%desig%Query(visitor)
    
  END SUBROUTINE dsov_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectVariable%Modify - perform a modify 
  !! operation on the contents of the parse node.
  !!
  !! @param[in,out] pnode             The data statement object being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dsov_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectVariable), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%desig%Modify(visitor)
    
  END SUBROUTINE dsov_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectVariable%GetLocation - get the start 
  !! of the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION dsov_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DataStmtObjectVariable), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%desig
    
  END FUNCTION dsov_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectDataImpliedDo%Write - write a debugging 
  !! representation of an data statement object to a logical unit.
  !!
  !! @param[in]     pnode             The data statement object to write.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  ! Recursive because an implied do can contain another implied do.
  RECURSIVE SUBROUTINE dsodid_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectDataImpliedDo), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%obj%Write(unit)
    
  END SUBROUTINE dsodid_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectDataImpliedDo%Query - perform a query 
  !! operation on the contents of the object.
  !!
  !! @param[in]     pnode             The data statement object being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dsodid_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectDataImpliedDo), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%obj%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dsodid_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectDataImpliedDo%Modify - perform a modify 
  !! operation on the contents of the object.
  !!
  !! @param[in,out] pnode             The data statement object being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dsodid_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtObjectDataImpliedDo), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%obj%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dsodid_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtObjectDataImpliedDo%GetLocation - get the start 
  !! of the source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION dsodid_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DataStmtObjectDataImpliedDo), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%obj
    
  END FUNCTION dsodid_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataImpliedDo.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'data-stmt-object'.
  
  FUNCTION did_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'data-implied-do'
    
  END FUNCTION did_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%Write - write a debugging representation 
  !! of an array constructor value to a logical unit.
  !!
  !! @param[in]     pnode             The expression to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  ! Recursive because an implied do can contain another implied do.
  RECURSIVE SUBROUTINE did_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataImpliedDo), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! item index.
    
    !***************************************************************************
    
    WRITE (unit, "('(data-implied-do: ')", ADVANCE='NO')
    DO i = 1, SIZE(pnode%items)
      SELECT TYPE (obj => pnode%items(i)%item)
      TYPE IS (DataIDoObjectDesignator)
        CALL WriteDesignator(unit, obj%desig)
      TYPE IS (DataIDoObjectDataImpliedDo)
        CALL obj%obj%Write(unit)
      CLASS DEFAULT
        WRITE (unit, "('???')", ADVANCE='NO')
      END SELECT
      WRITE (unit, "(',')", ADVANCE='NO')
    END DO
    
    CALL pnode%do_control%Write(unit)
    
    WRITE (unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE did_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an implied do can contain another implied do.
  RECURSIVE SUBROUTINE did_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataImpliedDo), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%items)
      CALL pnode%items(i)%item%Query(visitor)
    END DO
    
    CALL pnode%do_control%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE did_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%Modify- traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive because an implied do can contain another implied do.
  RECURSIVE SUBROUTINE did_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataImpliedDo), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array value index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%items)
      CALL pnode%items(i)%item%Modify(visitor)
    END DO
    
    CALL pnode%do_control%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE did_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDo%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  ! Recursive because an implied do can contain another implied do.
  RECURSIVE FUNCTION did_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DataImpliedDo), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%items(1)%item
    
  END FUNCTION did_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataIDoObject and extensions.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObject%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'data-construct'.
  
  FUNCTION dido_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'data-implied-do'
    
  END FUNCTION dido_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDesignator%Write - write a debugging 
  !! representation of a /data-i-do-object/ that is a designator to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The /data-i-do-object/ to write.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  SUBROUTINE didod_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDesignator), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%desig%Write(unit)
    
  END SUBROUTINE didod_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDesignator%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE didod_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDesignator), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%desig%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE didod_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDesignator%Modify- traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE didod_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDesignator), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%desig%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE didod_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDesignator%GetLocation - get the start 
  !! of the source that corresponds to the /data-i-do-object/.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION didod_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DataIDoObjectDesignator), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%desig
    
  END FUNCTION didod_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDataImpliedDo%Write - write a debugging 
  !! representation of a /data-i-do-object/ that is a designator to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The /data-i-do-object/ to write.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  
  ! Recursive becase an implied do can contain another implied do.
  RECURSIVE SUBROUTINE didodid_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDataImpliedDo), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%obj%Write(unit)
    
  END SUBROUTINE didodid_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDataImpliedDo%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive becase an implied do can contain another implied do.
  RECURSIVE SUBROUTINE didodid_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDataImpliedDo), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%obj%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE didodid_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDataImpliedDo%Modify- traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  ! Recursive becase an implied do can contain another implied do.
  RECURSIVE SUBROUTINE didodid_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataIDoObjectDataImpliedDo), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%obj%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE didodid_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataIDoObjectDataImpliedDo%GetLocation - get the start 
  !! of the source that corresponds to the /data-i-do-object/.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  ! Recursive becase an implied do can contain another implied do.
  RECURSIVE FUNCTION didodid_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DataIDoObjectDataImpliedDo), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%obj
    
  END FUNCTION didodid_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataStmtRepeat and extensions.
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a data statement repeat to a 
  !! formatted file.
  !!
  !! @param[in]     log_unit          Logical unit for output.
  !!
  !! @param[in]     repeat            The data statement repeat.
  
  SUBROUTINE write_data_stmt_repeat(log_unit, repeat)
    
    USE Designators
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    CLASS(DataStmtRepeat), INTENT(IN) :: repeat
    
    !***************************************************************************
    
    SELECT TYPE (repeat)
    TYPE IS (DataStmtRepeatDesig)
      CALL WriteDesignator(log_unit, repeat%desig)
    TYPE IS (DataStmtRepeatLiteral)
      WRITE (log_unit, "(A)", ADVANCE='NO') QueryValue(repeat%tok)
    CLASS DEFAULT
      STOP 'Unknown DataStmtRepeat'
    END SELECT
    
  END SUBROUTINE write_data_stmt_repeat
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtRepeatDesig%Query - perform a query 
  !! operation on the contents of the repeat.
  !!
  !! @param[in]     constant          The data statement repeat being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dsrd_Query(repeat, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtRepeatDesig), INTENT(IN) :: repeat
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL repeat%desig%Query(visitor)
    
  END SUBROUTINE dsrd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtRepeatDesig%Modify - perform a modify 
  !! operation on the contents of the repeat.
  !!
  !! @param[in,out] constant          The data statement repeat being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dsrd_Modify(repeat, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtRepeatDesig), INTENT(INOUT) :: repeat
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL repeat%desig%Modify(visitor)
    
  END SUBROUTINE dsrd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtRepeatLiteral%Query - perform a query 
  !! operation on the contents of the repeat.
  !!
  !! @param[in]     constant          The data statement repeat being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dsrl_Query(repeat, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtRepeatLiteral), INTENT(IN) :: repeat
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse node to Query.
    
  END SUBROUTINE dsrl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtRepeatLiteral%Modify - perform a modify 
  !! operation on the contents of the repeat.
  !!
  !! @param[in,out] constant          The data statement repeat being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dsrl_Modify(repeat, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtRepeatLiteral), INTENT(INOUT) :: repeat
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse node to Modify.
    
  END SUBROUTINE dsrl_Modify
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataStmtConstant and extensions.
  !
  ! Just bundling everything into the one procedure with a SELECT TYPE 
  ! (as was done for Write) is much shorter than the individual Query 
  ! and Modify bindings, but feels a bit hacky.
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a data statement constant to 
  !! a formatted file.
  !!
  !! @param[in]     log_unit          Logical unit for output.
  !!
  !! @param[in]     const             The data statement constant.
  
  SUBROUTINE write_data_stmt_constant(log_unit, const)
    
    USE Designators
    USE StructureConstructors
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    CLASS(DataStmtConstant), INTENT(IN) :: const
    
    !***************************************************************************
    
    SELECT TYPE (const)
    TYPE IS (DataStmtConstantDesig)
      CALL WriteDesignator(log_unit, const%desig)
      
    TYPE IS (DataStmtConstantSC)
      CALL const%sc%Write(log_unit)
      
    TYPE IS (DataStmtConstantSimpleLiteral)
      IF (ALLOCATED(const%sign_tok)) THEN
        WRITE (log_unit, "(A,' ')", ADVANCE='NO') QueryValue(const%sign_tok)
      END IF
      WRITE (log_unit, "(A)", ADVANCE='NO') QueryValue(const%tok)
      
    TYPE IS (DataStmtConstantComplexLiteral)
      CALL const%constant%Write(log_unit)
      
    CLASS DEFAULT
      STOP 'Unknown DataStmtConstant'
      
    END SELECT 
    
  END SUBROUTINE write_data_stmt_constant
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantDesig%Query - perform a query 
  !! operation on the contents of the constant.
  !!
  !! @param[in]     constant          The data statement constant being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dscd_Query(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantDesig), INTENT(IN) :: constant
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL constant%desig%Query(visitor)
    
  END SUBROUTINE dscd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantDesig%Modify - perform a modify 
  !! operation on the contents of the constant.
  !!
  !! @param[in,out] constant          The data statement constant being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dscd_Modify(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantDesig), INTENT(INOUT) :: constant
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL constant%desig%Modify(visitor)
    
  END SUBROUTINE dscd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantSC%Query - perform a query 
  !! operation on the contents of the constant.
  !!
  !! @param[in]     constant          The data statement constant being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dscsc_Query(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantSC), INTENT(IN) :: constant
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL constant%sc%Query(visitor)
    
  END SUBROUTINE dscsc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantSC%Modify - perform a modify 
  !! operation on the contents of the constant.
  !!
  !! @param[in,out] constant          The data statement constant being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dscsc_Modify(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantSC), INTENT(INOUT) :: constant
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL constant%sc%Modify(visitor)
    
  END SUBROUTINE dscsc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantSimpleLiteral%Query - perform a query 
  !! operation on the contents of the constant.
  !!
  !! @param[in]     constant          The data statement constant being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dscsl_Query(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantSimpleLiteral), INTENT(IN) :: constant
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse nodes to query.
    
  END SUBROUTINE dscsl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantSimpleLiteral%Modify - perform a modify 
  !! operation on the contents of the constant.
  !!
  !! @param[in,out] constant          The data statement constant being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dscsl_Modify(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantSimpleLiteral), INTENT(INOUT) :: constant
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse nodes to modify.
    
  END SUBROUTINE dscsl_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantComplexLiteral%Query - perform a query 
  !! operation on the contents of the constant.
  !!
  !! @param[in]     constant          The data statement constant being 
  !! queried.
  !!
  !! @param[in,out] visitor           The object executing the query 
  !! operation.
  
  SUBROUTINE dsccl_Query(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantComplexLiteral), INTENT(IN) :: constant
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse nodes to query.
    
  END SUBROUTINE dsccl_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataStmtConstantComplexLiteral%Modify - perform a modify 
  !! operation on the contents of the constant.
  !!
  !! @param[in,out] constant          The data statement constant being 
  !! modified.
  !!
  !! @param[in,out] visitor           The object executing the modify 
  !! operation.
  
  SUBROUTINE dsccl_Modify(constant, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmtConstantComplexLiteral), INTENT(INOUT) :: constant
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    ! No parse nodes to modify.
    
  END SUBROUTINE dsccl_Modify
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for DataImpliedDoControl.
  
  
  !*****************************************************************************
  !!
  !> Implementation of DataImpliedDoControl%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION datacontrol_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = '(data-implied-do-control)'
    
  END FUNCTION datacontrol_What
  
END MODULE DataStmts
