! $Id: DerivedTypeDefs.f90 2887 2019-04-06 22:52:29Z ian $
! ff08 source code copyright 2014 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the DerivedTypeDefs module.


!*******************************************************************************
!!
!> Defines the DerivedTypeDef extension of ParseNode to represent 
!! /derived-type-def/'s in specification child construct lists.

MODULE DerivedTypeDefs
  
  USE ComponentDefStmts
  USE DeclarationTypeSpecs
  USE DerivedTypeStmts
  USE Expressions
  USE FinalProcedureStmts
  USE GenericSpecs
  USE Identifiers
  USE ParseNodes
  USE TypeParamDefStmts
  USE TypeBoundProcedureStmts
  USE TypeBoundGenericStmts
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseDerivedTypeDef
  
  PUBLIC :: FindParam
  PUBLIC :: FindParamDef
  
  !-----------------------------------------------------------------------------
  
  !> An element in an array of component definitions.
  TYPE :: comp_def_element
    CLASS(ComponentDefStmt), ALLOCATABLE :: item
  CONTAINS
    PROCEDURE :: Dump => cde_Dump
    PROCEDURE :: Query => cde_Query
    PROCEDURE :: Modify => cde_Modify
  END TYPE comp_def_element
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /derived-type-def/.
  !!
  !! We maintain three separate lists for the bindings becuase the relative 
  !! order is not important.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: DerivedTypeDef
    !> The results of parsing the opening /derived-type-stmt/.
    TYPE(DerivedTypeStmt) :: first
    !> The results of parsing the type parameter definitions.
    TYPE(TypeParamDefStmt), ALLOCATABLE :: params(:)
    !> Flag to indicate that the sequence statement was present.
    LOGICAL :: sequence = .FALSE.
    !> Flag to indicate that a private components statement was present.
    LOGICAL :: private_components = .FALSE.
    !> The component definitions.  These can either be data components or 
    !! procedure components.
    !!
    !! We maintain the different components in the one list to allow 
    !! correctness checking of the definition order of things used in 
    !! constant expressions.
    TYPE(comp_def_element), ALLOCATABLE :: comps(:)
    !> Flag to indicate that a binding private statement was present.
    LOGICAL :: private_bindings = .FALSE.
    
    ! Note that the following components for storing bindings do not 
    ! maintain any record of the order of the bindings.
    
    !> The type bound procedure definitions.
    !!
    !! Always allocated, may be size zero.
    TYPE(TypeBoundProcedureStmt), ALLOCATABLE :: tbps(:)
    !> The type bound generic definitions.
    !!
    !! Always allocated, may be size zero.
    TYPE(TypeBoundGenericStmt), ALLOCATABLE :: gens(:)
    !> The final binding definitions.
    !!
    !! Always allocated, may be size zero.
    TYPE(FinalProcedureStmt), ALLOCATABLE :: finals(:)
  CONTAINS
    PROCEDURE :: Dump => dtdef_Dump
    PROCEDURE :: Query => dtdef_Query
    PROCEDURE :: Modify => dtdef_Modify
    PROCEDURE, NOPASS :: What => dtdef_What
    PROCEDURE :: Who => dtdef_Who
    PROCEDURE :: Advertise => dtdef_Advertise
  END TYPE DerivedTypeDef
  
  !> Get the index of a type parameter.
  INTERFACE FindParam
    MODULE PROCEDURE FindParam_
  END INTERFACE FindParam
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse the statement tree node associated with a /derived-type-def/ 
  !! declaration construct.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    dt_def            The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseDerivedTypeDef(st, dt_def, err_list)
    
    USE ContainsStmts
    USE DerivedTypeDefStNodes
    USE Errors
    USE ErrorCodes
    USE FinalProcedureStmts
    USE Statements
    USE SourceLocations
    USE TypeBoundGenericStmts
    USE TypeBoundProcedureStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DerivedTypeDefStNode), INTENT(INOUT) :: st
    TYPE(DerivedTypeDef), INTENT(OUT) :: dt_def
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: is             ! Type parameter statement index.
    INTEGER :: ip1            ! Type parameter index (in the type param stmt).
    INTEGER :: ip2            ! Type parameter index (in the opening stmt).
    INTEGER :: ic             ! Component or binding index.
    
    ! Temporary for a data component.
    TYPE(DataComponentDefStmt), ALLOCATABLE :: data_comp
    
    ! Temporary for a procedure component.
    TYPE(ProcComponentDefStmt), ALLOCATABLE :: proc_comp
    
    ! Temporary for a single component definition statement.
    CLASS(ComponentDefStmt), ALLOCATABLE :: comp_item
    
    ! Temporary for growing the tbos component.
    TYPE(TypeBoundProcedureStmt), ALLOCATABLE :: tmp_tbps(:)
    
    ! Temporary for growing the gens component.
    TYPE(TypeBoundGenericStmt), ALLOCATABLE :: tmp_gens(:)
    
    ! Temporary for getting final procedure names.
    TYPE(FinalProcedureStmt), ALLOCATABLE :: tmp_finals(:)
    
    ! Number of times a type parameter is defined.
    INTEGER :: tp_def_count
    
    ! Last location that a type parameter was defined.
    TYPE(SourceLocation) :: tp_def_loc
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Opening statement.
    
    CALL ParseDerivedTypeStmt(st%first%tlist, dt_def%first, sub_err_list)
    CALL Add(err_list, sub_err_list)
    ! Abort if the opening statement has been botched.
    IF (Failure(sub_err_list)) RETURN
    
    !-------------------------------------------------------------------------
    ! Type parameters.
    
    ALLOCATE(dt_def%params(SIZE(st%type_params)))
    DO is = 1, SIZE(st%type_params)
      CALL ParseTypeParamDefStmt(  &
          st%type_params(is)%tlist,  &
          dt_def%params(is),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      ! Carry on in the event of failure.
      
      CALL dt_def%params(is)%Init(st%type_params(is))
      
      ! Check that the type parameter appeared in the opening statement.
      def_param_search: DO ip1 = 1, SIZE(dt_def%params(is)%params)
        ASSOCIATE(tp_name => dt_def%params(is)%params(ip1)%name)
          ip2 = FindParam(dt_def, .NameOf. tp_name)
          IF (ip2 == 0) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C437 on R431',  &
                LOCATION=.LocationOf. dt_def%params(is)%params(ip1),  &
                MSG='The type parameter "' // .NameOf. tp_name // '" was not &
                  &specified in the opening /derived-type-stmt/ of the &
                  &/derived-type-def/.' )
          END IF
        END ASSOCIATE
      END DO def_param_search
    END DO
    
    IF (.NOT. Failure(err_list)) THEN
      ! Check that each type parameter in the opening statement has a 
      ! corresponding single type parameter definition.  We only do this if 
      ! all type parameter statements were parsed correctly.
      !
      ! Code in DerivedTypeSemNodes has to do pretty much the same thing, 
      ! so perhaps we are better off deferring this check till then.
      opening_param_search: DO ip2 = 1, SIZE(dt_def%first%params)
        ASSOCIATE(tp_name => dt_def%first%params(ip2))
          tp_def_count = 0
          DO is = 1, SIZE(dt_def%params)
            DO ip1 = 1, SIZE(dt_def%params(is)%params)
              IF ( .NameOf. tp_name  &
                  == .NameOf. dt_def%params(is)%params(ip1)%name ) THEN
                tp_def_count = tp_def_count + 1
                IF (tp_def_count == 2) THEN
                  tp_def_loc = .LocationOf. dt_def%params(is)%params(ip1)%name
                END IF
              END IF
            END DO
          END DO
          IF (tp_def_count == 0) THEN 
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C438 on R431',  &
                LOCATION=.LocationOf. tp_name,  &
                MSG='The type parameter "' // .NameOf. tp_name // '" was &
                  &not specified in a /type-param-def-stmt/ of the &
                  &/derived-type-def/.' )
          ELSE IF (tp_def_count > 1) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C438 on R431',  &
                LOCATION=tp_def_loc,  &
                MSG='The type parameter "' // .NameOf. tp_name // '" was &
                  &specified multiple times in the /type-param-def-stmt/''s &
                  &of the /derived-type-def/.' )
          END IF
        END ASSOCIATE
      END DO opening_param_search
    END IF
    
    !---------------------------------------------------------------------------
    ! Private (for components) or sequence 
    
    IF (ALLOCATED(st%priv_or_seq)) THEN
      SELECT CASE (st%priv_or_seq%ist)
      CASE (istSequence)
        CALL ParseSequenceStmt(st%priv_or_seq%tlist, sub_err_list)
        dt_def%sequence = .TRUE.
      CASE (istPrivateComponents)
        CALL ParsePrivateComponentsStmt(st%priv_or_seq%tlist, sub_err_list)
        dt_def%private_components = .TRUE.
      CASE DEFAULT
        STOP 'Internal error in DerivedTypeDefs%%ParseDerivedTypeDef, &
            &the private or sequence statement is not a private or&
            &sequence statement.'
      END SELECT
      CALL Add(err_list, sub_err_list)
      ! Carry on in the event of failure.
    END IF
    
    !-------------------------------------------------------------------------
    ! Components
    
    ! Allocate to zero here so that if no components get parsed (successfully) 
    ! we still have a zero sized array on exit.
    ALLOCATE(dt_def%comps(0))
    DO ic = 1, SIZE(st%components)
      SELECT CASE (st%components(ic)%ist)
      CASE (istDataComponentDef)
        ! Allocate a temporary for the results of the parse.
        ALLOCATE(data_comp)
        ! Do the parse.
        CALL ParseDataComponentDefStmt(  &
            st%components(ic)%tlist,  &
            data_comp,  &
            sub_err_list )
        ! Put the temporary in the generic object.
        CALL MOVE_ALLOC(data_comp, comp_item)
        
      CASE (istProcComponentDef)
        ! Allocate a temporary for the results of the parse.
        ALLOCATE(proc_comp)
        ! Do the parse.
        CALL ParseProcComponentDefStmt(  &
            st%components(ic)%tlist,  &
            proc_comp,   &
            sub_err_list )
        ! Put the temporary in the generic object.
        CALL MOVE_ALLOC(proc_comp, comp_item)
        
      CASE DEFAULT
        STOP 'Internal error in DerivedTypeDefs%%ParseDerivedTypeDef, &
            &a component part statement was not valid for a component part.'
        
      END SELECT
      
      CALL Add(err_list, sub_err_list)
      ! If things parsed ok, then add the item to the list.
      IF (.NOT. Failure(sub_err_list)) CALL add_comp(dt_def%comps, comp_item)
      ! Carry on in the event of failure.
    END DO
    
    !-------------------------------------------------------------------------
    ! /type-bound-procedure-part/
    
    IF (ALLOCATED(st%tbp)) THEN
      ! Parse the contains statement that separates the bindings from the 
      ! components.
      CALL ParseContainsStmt(st%tbp%contains%tlist, sub_err_list)
      CALL Add(err_list, sub_err_list)
      ! Carry on in the event of error.
      
      ! Was there a /binding-private-stmt/?
      IF (ALLOCATED(st%tbp%binding_private)) THEN
        CALL ParseBindingPrivateStmt( st%tbp%binding_private%tlist,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        dt_def%private_bindings = .TRUE.
      END IF
      
      ! Parse each binding in turn, populate the three lists.
      ALLOCATE(dt_def%tbps(0))
      ALLOCATE(dt_def%gens(0))
      ALLOCATE(dt_def%finals(0))
      DO ic = 1, SIZE(st%tbp%type_bound_procs)
        ! Parse the particular statement in the type bound procedure part.
        !
        ! Any errors reported during parsing are appended to the main 
        ! error list at the end of the SELECT CASE.
        SELECT CASE (st%tbp%type_bound_procs(ic)%ist)
        CASE (istTypeBoundProcedure)      ! Binding for a procedure.
          ! Grow list
          ALLOCATE(tmp_tbps(SIZE(dt_def%tbps) + 1))
          ASSOCIATE(tmp => tmp_tbps(SIZE(tmp_tbps)))
            CALL ParseTypeBoundProcedureStmt(  &
                st%tbp%type_bound_procs(ic)%tlist,  &
                tmp,  &
                sub_err_list )
          END ASSOCIATE
          ! On success, copy previous bindings to the temporary array, then 
          ! move the temporary array across to the list of type bound 
          ! procedures.  On failure we simply discard the temporary array.
          IF (.NOT. Failure(sub_err_list)) THEN
            tmp_tbps(:SIZE(dt_def%tbps)) = dt_def%tbps
            CALL MOVE_ALLOC(tmp_tbps, dt_def%tbps)
          ELSE
            DEALLOCATE(tmp_tbps)
          END IF
          
        CASE (istTypeBoundGeneric)        ! Generic binding.
          ! Grow list
          ALLOCATE(tmp_gens(SIZE(dt_def%gens) + 1))
          ASSOCIATE(tmp => tmp_gens(SIZE(tmp_gens)))
            CALL ParseTypeBoundGenericStmt(  &
                st%tbp%type_bound_procs(ic)%tlist,  &
                tmp,  &
                sub_err_list )
          END ASSOCIATE
          ! On success, copy previous bindings to the temporary array, then 
          ! move the temporary array across to the list of type bound 
          ! procedures.  On failure we simply discard the temporary array.
          IF (.NOT. Failure(sub_err_list)) THEN
            tmp_gens(:SIZE(dt_def%gens)) = dt_def%gens
            CALL MOVE_ALLOC(tmp_gens, dt_def%gens)
          ELSE
            DEALLOCATE(tmp_gens)
          END IF
          
        CASE (istFinalProcedure)          ! Final binding.
          ! Grow list
          ALLOCATE(tmp_finals(SIZE(dt_def%finals) + 1))
          ASSOCIATE(tmp => tmp_finals(SIZE(tmp_finals)))
            CALL ParseFinalProcedureStmt(  &
                st%tbp%type_bound_procs(ic)%tlist,  &
                tmp,  &
                sub_err_list )
          END ASSOCIATE
          ! On success, copy previous finals to the temporary array, then 
          ! move the temporary array across to the list of final procedures.  
          ! On failure we simply discard the temporary array.
          IF (.NOT. Failure(sub_err_list)) THEN
            tmp_finals(:SIZE(dt_def%finals))= dt_def%finals
            CALL MOVE_ALLOC(tmp_finals, dt_def%finals)
          ELSE
            DEALLOCATE(tmp_finals)
          END IF
          
        CASE DEFAULT
          STOP 'Internal error in DerivedTypeDefs%%ParseDerivedTypeDef, &
              &unexpected statement type in type_bound_procs.'
          
        END SELECT
        CALL Add(err_list, sub_err_list)
        ! Carry on in the event of failure.
      END DO
    END IF
    
    !-------------------------------------------------------------------------
    ! /end-type-stmt/
    
    CALL ParseEndTypeStmt(  &
        st%last%tlist,  &
        .NameOf. dt_def%first%name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseDerivedTypeDef
  
  
  !*****************************************************************************
  !!
  !> Find the type parameter index of a type parameter.
  !!
  !! @param[in]     dt_def            The results of parsing the derived 
  !! type definition, at least after the opening statement and the type 
  !! parameter definitions have been parsed.
  !!
  !! @param[in]     tp_name           The name of the type parameter to 
  !! search for.
  !!
  !! @returns The index of the type parameter in the opening statement 
  !! of the derived type definition (i.e. in the type parameter order), 
  !! or zero if no type parameter was found.
  
  FUNCTION FindParam_(dt_def, tp_name) RESULT(idx)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(DerivedTypeDef), INTENT(IN) :: dt_def
    CHARACTER(*), INTENT(IN) :: tp_name
    
    ! Function result
    INTEGER :: idx
    
    !***************************************************************************
    
    idx = FindParam(dt_def%first, tp_name)
    
  END FUNCTION FindParam_
  
  
  !*****************************************************************************
  !!
  !> Find the statement index and parameter index of a type parameter.
  !!
  !! @param[in]     dt_def            The results of parsing the derived 
  !! type definition, at least after the opening statement and the type 
  !! parameter definitions have been parsed.
  !!
  !! @param[in]     tp_name           The name of the type parameter to 
  !! search for.
  !!
  !! @param[out]    is                Type parameter definition statement 
  !! index.  Zero if the parameter name was not found.
  !!
  !! @param[out]    ip                Parameter index within a statement.  
  !! Zero if the parameter name was not found.
  
  SUBROUTINE FindParamDef(dt_def, tp_name, is, ip)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DerivedTypeDef), INTENT(IN) :: dt_def
    CHARACTER(*), INTENT(IN) :: tp_name
    INTEGER, INTENT(OUT) :: is
    INTEGER, INTENT(OUT) :: ip
    
    !***************************************************************************
    
    ! Scan through the type parameter definition statements.
    DO is = 1, SIZE(dt_def%params)
      ! Scan through the names defined by each statement.
      DO ip = 1, SIZE(dt_def%params(is)%params)
        IF (tp_name == .NameOf. dt_def%params(is)%params(ip)%name) RETURN
      END DO
    END DO
    
    ! Indicate failure to find.
    is = 0
    ip = 0
    
  END SUBROUTINE FindParamDef
  
  
  !*****************************************************************************
  !!
  !> Add a new component (by moving) to a list of component definitions.
  !!
  !! @param[in]     comps             The list of component definitions to be 
  !! added to.  Must be allocated on input.
  !!
  !! @param[in]     item              The component definition to add.
  
  SUBROUTINE add_comp(comps, item)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(comp_def_element), INTENT(INOUT), ALLOCATABLE :: comps(:)
    CLASS(ComponentDefStmt), INTENT(INOUT), ALLOCATABLE :: item
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for growing array.
    TYPE(comp_def_element), ALLOCATABLE :: tmp(:)
    
    INTEGER :: i              ! Array index.
    
    !***************************************************************************
    
    ALLOCATE(tmp(LBOUND(comps,1):UBOUND(comps,1)+1))
    DO i = LBOUND(comps, 1), UBOUND(comps, 1)
      CALL MOVE_ALLOC(comps(i)%item, tmp(i)%item)
    END DO
    
    CALL MOVE_ALLOC(item, tmp(SIZE(tmp))%item)
    CALL MOVE_ALLOC(tmp, comps)
    
  END SUBROUTINE add_comp
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%Dump - write a debugging representation 
  !! of the results of parsing a derived type definition to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE dtdef_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeDef), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Type parameter index (opening statement).
    INTEGER :: is             ! Index of the type parameter statement.
    INTEGER :: ip             ! Index of the parameter in the statement.
    INTEGER :: ic             ! Component or binding statement index.
    
    !***************************************************************************
    
    ! The type name.
    WRITE (unit, "(A,'/type-name/ = ',A)") pfx, .NameOf. pnode%first%name
    
    !---------------------------------------------------------------------------
    ! Attributes in the derived type statement.
    
    IF (pnode%first%abstract) WRITE (unit, "(A,'ABSTRACT')") pfx
    IF (ALLOCATED(pnode%first%extends))  &
        WRITE (unit, "(A,'EXTENDS( ',A,' )')") pfx,  &
            .NameOf. pnode%first%extends
    
    IF (pnode%first%access /= 0) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx
      CALL WriteAccess(unit, pnode%first%access)
      WRITE (unit, "()")
    END IF
    
    SELECT CASE (pnode%first%lang)
    CASE (0)  ! Ignore
    CASE (1)      ; WRITE (unit, "(A,'BIND(C)')") pfx
    CASE DEFAULT
      STOP 'Internal error in DerivedTypeDefs%%dtdef_Dump, &
          &unknown language specification.'
    END SELECT
    
    !---------------------------------------------------------------------------
    ! Type parameters.
    
    WRITE (unit, "(A,'type parameters:')") pfx
    DO i = 1, SIZE(pnode%first%params)
      ! Look up the parameter name in the type parameter definition statements.
      CALL FindParamDef(pnode, .NameOf. pnode%first%params(i), is, ip)
      
      WRITE (unit, "(A,I0,': ',A)", ADVANCE='NO') pfx // '  ', i,  &
          .NameOf. pnode%first%params(i)
      SELECT CASE (pnode%params(is)%attr)
      CASE (1)      ; WRITE (unit, "(' (KIND)')")
      CASE (2)      ; WRITE (unit, "(' (LEN)')")
      CASE DEFAULT
        STOP 'Internal error in DerivedTypeDefs%%dtdef_Dump, &
            &unknown type parameter attribute.'
      END SELECT
      
      IF (ALLOCATED(pnode%params(is)%kind_expr)) THEN
        WRITE (unit, "(A,'/kind-selector/ = ')", ADVANCE='NO') pfx // '    '
        CALL WriteExpr(unit, pnode%params(is)%kind_expr)
        WRITE (unit, "()")
      END IF
      IF (ALLOCATED(pnode%params(is)%params(ip)%default)) THEN
        WRITE (unit, "(A,'/scalar-int-constant-expr/ = ')", ADVANCE='NO')  &
            pfx // '    '
        CALL WriteExpr(unit, pnode%params(is)%params(ip)%default)
        WRITE (unit, "()")
      END IF
    END DO
    
    !---------------------------------------------------------------------------
    ! Other attributes.
    
    IF (pnode%sequence) THEN
      WRITE (unit, "(A,'the type is a SEQUENCE type')") pfx
    END IF
    
    IF (pnode%private_components) THEN
      WRITE (unit, "(A,'components are PRIVATE by default')") pfx
    END IF
    
    !---------------------------------------------------------------------------
    ! Components
    
    WRITE (unit, "(A,'components:')") pfx
    DO ic = 1, SIZE(pnode%comps)
      CALL pnode%comps(ic)%Dump(unit, pfx // '  ')
    END DO
    
    !---------------------------------------------------------------------------
    ! Bindings.
    
    IF (pnode%private_bindings) THEN
      WRITE (unit, "(A,'bindings are PRIVATE by default')") pfx
    END IF
    
    ! Specific procedure bindings.
    WRITE (unit, "(A,'specific procedure bindings:')") pfx
    DO ic = 1, SIZE(pnode%tbps)
      CALL pnode%tbps(ic)%Dump(unit, pfx // '  ')
    END DO
    
    ! Generic bindings.
    WRITE (unit, "(A,'generic bindings:')") pfx
    DO ic = 1, SIZE(pnode%gens) 
      CALL pnode%gens(ic)%Dump(unit, pfx // '  ')
    END DO
    
    ! Final bindings
    WRITE (unit, "(A,'final bindings:')") pfx
    DO ic = 1, SIZE(pnode%finals)
      CALL pnode%finals(ic)%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE dtdef_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE dtdef_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DerivedTypeDef), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Parameter, component or binding index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%first%Query(visitor)
    
    DO i = 1, SIZE(pnode%params)
      CALL pnode%params(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%comps)
      CALL pnode%comps(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%tbps)
      CALL pnode%tbps(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%gens)
      CALL pnode%gens(i)%Query(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%finals)
      CALL pnode%finals(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dtdef_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE dtdef_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(DerivedTypeDef), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Parameter, component or binding index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%first%Modify(visitor)
    
    DO i = 1, SIZE(pnode%params)
      CALL pnode%params(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%comps)
      CALL pnode%comps(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%tbps)
      CALL pnode%tbps(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%gens)
      CALL pnode%gens(i)%Modify(visitor)
    END DO
    
    DO i = 1, SIZE(pnode%finals)
      CALL pnode%finals(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE dtdef_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'derived-type-def'.
  
  FUNCTION dtdef_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'derived-type-def'
    
  END FUNCTION dtdef_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%Who - return the identifier text 
  !! for the parse node.
  !!
  !! @param[in]     pnode             The parse node to query.
  !!
  !! @returns The name of the subroutine.
  
  FUNCTION dtdef_Who(pnode) RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(DerivedTypeDef), INTENT(IN) :: pnode
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = .NameOf. pnode%first%name
    
  END FUNCTION dtdef_Who
  
  
  !*****************************************************************************
  !!
  !> Implementation of DerivedTypeDef%Advertise - advertise the identifiers 
  !! referenced or defined by the block.
  !!
  !! @param[in]     pnode             The node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE dtdef_Advertise(pnode, ident_list)
    
    USE AccessStmts
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeDef), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER, ALLOCATABLE :: natures(:)
    TYPE(SourceLocation), ALLOCATABLE :: locations(:)
    
    INTEGER :: ig             ! Type bound generic index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    ! The type name.
    SELECT CASE (pnode%first%access)
    CASE (iaccPrivate)
      natures = [NatureDefType, NatureDefPrivate]
      locations = [.LocationOf. pnode%first%name, .LocationOf. pnode]
    CASE (iaccPublic)
      natures = [NatureDefType, NatureDefPublic]
      locations = [.LocationOf. pnode%first%name, .LocationOf. pnode]
    CASE DEFAULT
      natures = [NatureDefType]
      locations = [.LocationOf. pnode%first%name]
    END SELECT
    
    CALL MergeNatures(  &
        ident_list,  &
        IdentifierNature(  &
          NAME=.NameOf. pnode%first%name,  &
          NATURES=natures,  &
          LOCATIONS=locations ) )
    
    ! Parent type, if present.
    IF (ALLOCATED(pnode%first%extends)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%first%extends,  &
            NATURES=[NatureRefType],  &
            LOCATIONS=[.LocationOf. pnode%first%extends] ) )
    END IF
    
    DO ig = 1, SIZE(pnode%gens)
      ! Advertise type bound generic identifiers that are not names.
      IF (pnode%gens(ig)%gs%keyword /= 0) THEN
        CALL MergeNatures(  &
            ident_list,  &
            IdentifierNature(  &
              SPEC=pnode%gens(ig)%gs,  &
              NATURES=[NatureDefGeneric],  &
              LOCATIONS=[.LocationOf. pnode%gens(ig)%gs] ) )
      END IF
    END DO
    
  END SUBROUTINE dtdef_Advertise
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for comp_def_element.
  
  
  !*****************************************************************************
  !!
  !> Implementation of comp_def_element%Dump - write a debugging 
  !! representation of the contained item to a logical unit.
  !!
  !! @param[in]     pnode             The node element to dump.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  !!
  !! Simply forwards to the Dump binding of the item.
  
  SUBROUTINE cde_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(comp_def_element), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    CALL pnode%item%Dump(unit, pfx)
    
  END SUBROUTINE cde_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of comp_def_element%Query - traverse the contained item 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The node element to query.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  !!
  !! Simply forwards to the Query binding of the item.
  
  SUBROUTINE cde_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(comp_def_element), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Query(visitor)
    
  END SUBROUTINE cde_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of comp_def_element%Modify - traverse the contained item 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The node element to modify.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  !!
  !! Simply forwards to the Modify binding of the item.
  
  SUBROUTINE cde_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(comp_def_element), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL pnode%item%Modify(visitor)
    
  END SUBROUTINE cde_Modify
  
END MODULE DerivedTypeDefs