! $Id: Designators.f90 2815 2019-03-24 02:47:45Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Designators module.


!*******************************************************************************
!!
!> Types and procedures for storing and parsing /designators/ and other 
!! source rules defined in chapters 6.1 through to 6.6.
!!
!! A /designator/ (R601) is:
!! - /object-name/
!! - /array-element/
!! - /array-section/
!! - /coindexed-named-object/
!! - /complex-part-designator/
!! - /structure-component/
!! - /substring/
!!
!! A /substring/ consists of a /parent-string/ and a /substring-range/.  The 
!! /parent-string/ (R609) can then be:
!! - /scalar-variable-name/  (which looks just like /object-name/)
!! - /array-element/ (which is already an option for a /designator/)
!! - /coindexed-named-object/ (as above)
!! - /scalar-structure-component/ (as above, but scalar)
!! - /scalar-constant/ (which is a single token)
!!
!! All the options for /parent-string/ are scalar - but we can't determine that 
!! during basic parsing.
!!
!! Note that array type substring references are possible - but that's 
!! because array-sections can have a substring-range at their end.
!!
!! A complex part designator just looks like a component reference.
!!
!! /type-param-inquiry/'s also are undistinguishable from designators 
!! (see note 6.7), so they end up getting parsed into this as well.
!!
!! The actual parsing is done in ParseExpressions until submodule support 
!! is sufficient.

MODULE Designators
  
  USE CompilerKinds
  USE Identifiers
  USE ParseNodes
  USE SourceLocations
  USE SubscriptLists
  USE SubstringRanges
  USE Tokens
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------    
  
  PUBLIC :: ParseDesignator
  PUBLIC :: ParsePartRef
  PUBLIC :: ParseStructureComponent
  
  PUBLIC :: WriteDesignator
  PUBLIC :: WritePartRef
  
  PUBLIC :: CountParts
  
  !-----------------------------------------------------------------------------    
  
  !> A part-ref (R612).
  TYPE, EXTENDS(TermParseNode), PUBLIC :: PartRef
    !> The name of the part.
    TYPE(Identifier) :: part_name
    
    !> The subscripts (if any) for the part.
    !!
    !! If not allocated then there are no subscripts.
    TYPE(DimSpec), ALLOCATABLE :: subscripts(:)
    
    !> The image select for the part.
    !!
    !! This looks like a restricted form of a section-subscript-list.
    !!
    !! If not allocated, there are none.
    TYPE(DimSpec), ALLOCATABLE :: image_selector(:)
  CONTAINS
    PROCEDURE :: Write => part_Write
    PROCEDURE :: Query => part_Query
    PROCEDURE :: Modify => part_Modify
    PROCEDURE, NOPASS :: What => part_What
    PROCEDURE :: GetLocation => part_GetLocation
  END TYPE PartRef
  
  !> Type that can store all different types of designator.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: Designator
    !> The data-ref or parent-string, if the parent_string is not a literal.
    !!
    !! If the designator starts with a literal then this will not be 
    !! allocated.
    !!
    !! (Either part_refs or literal will be allocated.)
    !!
    !! If it is allocated then its size will be greater than zero.
    TYPE(PartRef), ALLOCATABLE :: part_refs(:)
    
    !> The value of the /parent-string/ if the designator is a literal 
    !! constant.
    !!
    !! Not allocated if designator is not a subscripted literal constant.
    !!
    !! (Either part_refs or literal will be allocated.)
    !!
    !! This cannot be an identifier as we have to store the actual text 
    !! of the literal constant.
    CHARACTER(:,KIND=scck), ALLOCATABLE :: literal
    
    !> The kind of the /parent-string/.
    !!
    !! Only allocated if the designator has a /parent-string/, and only if 
    !! there is a kind parameter.
    CHARACTER(:), ALLOCATABLE :: literal_kind
    
    !> The start in source of the parent string.
    !!
    !! Allocated if literal is allocated.
    TYPE(SourceLocation), ALLOCATABLE :: literal_location
    
    !> Substring reference (must be the last thing in the designator text).
    !!
    !! If not allocated, its not there.
    TYPE(SubstringRange), ALLOCATABLE :: substring
  CONTAINS
    PROCEDURE :: Write => desig_Write
    PROCEDURE :: Query => desig_Query
    PROCEDURE :: Modify => desig_Modify
    PROCEDURE, NOPASS :: What => desig_What
    PROCEDURE :: Advertise => desig_Advertise
    PROCEDURE :: GetLocation => desig_GetLocation
  END TYPE Designator
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /designator/ (R601).
  !!
  !! All options for designators, bar three, involve (or look like) 
  !! /data-refs/ (R611).  Those three exceptions are all substrings.
  !!
  !! The first is a substring that is of a character literal:
  !!   'xyz'(a:b)`
  !!
  !! The second is a substring that is of a array section:
  !!   array(xx)(a:b)
  !!
  !! The third is a substring that is of a coindexed named object.
  !!   coarray[xx](a:b)
  !!
  !! The combination of the second and third is not permitted by the syntax 
  !! rules for /substring/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    desig             Results of parsing the designator - not 
  !! defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  
  RECURSIVE SUBROUTINE ParseDesignator(tlist, desig, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Designator), INTENT(OUT) :: desig
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /designator/.
    CHARACTER(*), PARAMETER :: comp = 'R601'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: finish
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Error object for child Expect calls.
    TYPE(Error) :: err
    
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: pass           ! Scan pass number
    INTEGER :: part           ! part number
    INTEGER :: i              ! Position of the end of the last part
    INTEGER :: substart       ! Substring parenthesis start.
    INTEGER :: part_finish    ! Finish of the current part-ref.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /designator/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! substring of a literal character constant
    
    IF (tlist(1) == ittCharacterLiteral) THEN
      IF (SIZE(tlist) == 1) THEN
        CALL Add( err_list,  &
            COMPONENT='R608',  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='Expected ( /substring-range/ )' )
        RETURN
      END IF
      
      IF (.NOT. IsToken(tlist(2), ittDelimiter, scck_'(')) THEN
        CALL Add( err_list,  &
            COMPONENT='R608',  &
            LOCATION=QueryLocation(tlist(2)),  &
            MSG='Expected ( /substring-range/ )' )
        RETURN
      END IF
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT='R608')
      IF (Failure(err)) RETURN
      
      IF (closing == 3) THEN
        CALL Add( err_list,  &
            COMPONENT='R608',  &
            LOCATION=QueryLocation(tlist(2)),  &
            MSG='Missing /substring-range/.' )
        RETURN
      END IF
      
      ALLOCATE(desig%substring)
      CALL ParseSubstringRange(  &
          desig%substring,  &
          tlist(3:closing-1),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT='R608')
      IF (Failure(sub_err_list)) RETURN
      
      CALL ExpectEnd(tlist, closing + 1, 'R608', err_list)
      
      desig%literal = QueryValue(tlist(1))
      desig%literal_location = QueryLocation(tlist(1))
      IF (HasKind(tlist(1))) desig%literal_kind = QueryKind(tlist(1))
      
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Everything else - which starts off looking like a data-ref (R611)
    
    ! First pass we count, second pass we parse.
    pass_loop: DO pass = 1, 2
      ! i is the position of the end of the previous part, so at the start it 
      ! is one before the start.
      i = 1
      
      ! Current part-ref number and the number of part-refs at the end of the 
      ! first pass.
      part = 0
      
      token_loop: DO WHILE (i <= SIZE(tlist))
        
        finish = FindBare(tlist, ittPercent, scck_'%', i)
        ! Set finish to point at the token at the end of the part-ref.
        IF (finish == 0) THEN
          finish = SIZE(tlist)        ! last part-ref
        ELSE
          finish = finish - 1
        END IF
        
        part = part + 1
        IF (pass == 2) THEN
          ! Substring bits can only come in the last part.
          IF (part == SIZE(desig%part_refs)) THEN
            ! Do we have a array subcript followed by a substring?
            ! Do we have a coindexed named object followed by a substring?
            !
            ! The coindexed named object case is only valid if there is one 
            ! part (i.e. part == 1).  Here we allow multiple parts.
            CALL test_substring(tlist(i:finish), substart)
            IF (substart == 0) THEN
              part_finish = finish
            ELSE
              ! Make substart relative to tlist.
              substart = substart + i - 1
              part_finish = substart - 1
            END IF
          ELSE
            substart = 0
            part_finish = finish
          END IF
          ! A substart of zero here means that there was nothing that looked 
          ! like a substring (or we we are not expecting a substring.  
          ! Regardless, part_finish is the finish of the non-substring tokens 
          ! of the current part.
          
          ! Parse the non-substring tokens of the part.
          CALL ParsePartRef(  &
              tlist,  &
              i,  &
              part_finish,  &
              desig%part_refs(part),  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT='R611')
          IF (Failure(sub_err_list)) RETURN
          
          IF (substart /= 0) THEN
            ! Find the closing parenthesis for the second parenthesis pair
            ! or parenthesis pair following a bracket pair).
            CALL ExpectClosing(tlist, substart, part_finish, err)
            CALL Add(err_list, err, COMPONENT='R608')
            ! Carry on!
            
            ! Parse the substring range.
            ALLOCATE(desig%substring)
            CALL ParseSubstringRange(  &
                desig%substring,  &
                tlist(substart+1:part_finish-1),  &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT='R608')
            
            ! If we missed the closing parenthesis - stop parsing.
            IF (Failure(err)) RETURN
            ! Make sure the final closing parenthesis was the last gasp.
            CALL ExpectEnd(tlist, part_finish + 1, 'R608', sub_err_list)
            ! This is the last iteration of the second pass - so we are done.
          END IF
        END IF
        
        IF (finish == SIZE(tlist)) EXIT
        
        IF (finish == SIZE(tlist) - 1) THEN
          ! Our designator ends with a %
          CALL Add( err_list,  &
              COMPONENT='R611',  &
              LOCATION=QueryLocation(tlist(finish+1)),  &
              MSG='Expected part-name')
          RETURN
        ELSE
          i = finish + 2
        END IF
      END DO token_loop
      
      IF (pass == 1) ALLOCATE(desig%part_refs(part))
      
    END DO pass_loop
    
  END SUBROUTINE ParseDesignator
  
  
  !*****************************************************************************
  !!
  !> Test whether a sequence of tokens looks like an array reference followed 
  !! by a substring, or a coindexed named object followed by a substring.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    after             Start of the substring subscripts, 
  !! which is the token after the array subscripts closing parenthesis or 
  !! the coarray subscript closing bracket.  Zero if there was no such 
  !! substring subscript.
  
  SUBROUTINE test_substring(tlist, after)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(OUT) :: after
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: nest           ! Parenthesis nesting count.
    LOGICAL :: seen_first     ! Have we seen the first '(' or '[' yet?
    
    !***************************************************************************
    
    nest = 0
    seen_first = .FALSE.
    DO after = 1, SIZE(tlist)
      IF (IsToken(tlist(after), ittDelimiter, scck_'(')) THEN
        IF (nest == 0) THEN
          IF (seen_first) RETURN
          seen_first = .TRUE.
        END IF
        nest = nest + 1
      ELSE IF (IsToken(tlist(after), ittDelimiter, scck_'[')) THEN
        nest = nest + 1
        seen_first = .TRUE.
      ELSE IF ( IsToken(tlist(after), ittDelimiter, scck_')')  &
          .OR. IsToken(tlist(after), ittDelimiter, scck_']') ) THEN
        nest = nest - 1
      END IF
    END DO
    
    ! If we make it to here, no start of a second parenthesis group found.
    after = 0
    
  END SUBROUTINE test_substring
  
  
  !*****************************************************************************
  !!
  !> Implementation of Designator%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE desig_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Designator), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteDesignator(unit, pnode)
    
  END SUBROUTINE desig_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of Designator%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE desig_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Designator), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Part reference.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%part_refs)) THEN
      DO ip = 1, SIZE(pnode%part_refs)
        CALL pnode%part_refs(ip)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%substring)) THEN
      CALL pnode%substring%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE desig_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Designator%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE desig_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Designator), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ip             ! Part reference.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%part_refs)) THEN
      DO ip = 1, SIZE(pnode%part_refs)
        CALL pnode%part_refs(ip)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%substring)) THEN
      CALL pnode%substring%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE desig_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Designator%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name represented by parse node.
  
  FUNCTION desig_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'designator'
    
  END FUNCTION desig_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of DesignatorExpr%Advertise - advertise 
  !! the identifiers referenced or defined in the designator.
  !!
  !! @param[in]     pnode             The designator to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  ! Recursive because a designator may contain another designator 
  ! in a subscript or substring range or similar.
  RECURSIVE SUBROUTINE desig_Advertise(pnode, ident_list)
    
    USE CharacterTypes
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Designator), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! ifort 15.0.2 bug workaround - ticket 44.
    character(:), allocatable :: tmp_name
    
    !***************************************************************************
    
    ! This duplicates logic in PrimaryExpressions::advertise_kind, but 
    ! there really isn't a convenient common place for it, and things 
    ! are a little different in terms of how the literal is stored.
    IF (ALLOCATED(pnode%literal_kind)) THEN
      IF (IsLetterChar(pnode%literal_kind(1:1))) THEN
        !=====================================================================
        ! ifort 15.0.2 bug - ice with the following - we need a temp for 
        ! the character variable.
        !
        !CALL MergeNatures(  &
        !    ident_list,  &
        !    IdentifierNature(  &
        !      NAME=pnode%literal_kind,  &
        !      NATURES=[NatureRefDataObject],  &
        !      LOCATIONS=[pnode%literal_location] ) )
        !=====================================================================
        tmp_name = pnode%literal_kind
        ALLOCATE(ident_list(1))
        ident_list(1) = IdentifierNature(  &
            NAME=tmp_name,  &
            NATURES=[NatureRefDataObject],  &
            LOCATIONS=[pnode%literal_location] )
        !=====================================================================
      END IF
    ELSE
      ! part_refs will be allocated.
      ALLOCATE(ident_list(1))
      IF (SIZE(pnode%part_refs) > 1) THEN
        IF (SIZE(pnode%part_refs(1)%subscripts) /= 0) THEN
          ident_list(1) = IdentifierNature(  &
              NAME=.NameOf. pnode%part_refs(1)%part_name,  &
              NATURES=[NatureRefDataObject, NatureRefArray],  &
              LOCATIONS=[ .LocationOf. pnode%part_refs(1)%part_name,  &
                  .LocationOf. pnode%part_refs(1)%part_name ] )
        ELSE
          ident_list(1) = IdentifierNature(  &
              NAME=.NameOf. pnode%part_refs(1)%part_name,  &
              NATURES=[NatureRefDataObject],  &
              LOCATIONS=[.LocationOf. pnode%part_refs(1)%part_name] )
        END IF
      ELSE
        ! Do we have a substring reference.
        IF (ALLOCATED(pnode%substring)) THEN
          ! It is definitely a substring.  Perhaps it is also an array.
          IF (SIZE(pnode%part_refs(1)%subscripts) /= 0) THEN
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefSubstringLike, NatureRefArray],  &
                LOCATIONS=[ .LocationOf. pnode%part_refs(1)%part_name,  &
                    .LocationOf. pnode%part_refs(1)%part_name ] )
          ELSE
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefSubstringLike],  &
                LOCATIONS=[.LocationOf. pnode%part_refs(1)%part_name] )
          END IF
        ELSE IF (SIZE(pnode%part_refs(1)%subscripts) == 1) THEN
          ! It might be a substring, an array, or a function reference.
          IF (IsSubstringRangeLike(pnode%part_refs(1)%subscripts)) THEN
            ! Cannot be a function reference.
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefDataObject, NatureRefSubstringLike],  &
                LOCATIONS=[ .LocationOf. pnode%part_refs(1)%part_name,  &
                    .LocationOf. pnode%part_refs(1)%part_name ] )
          ELSE IF (IsNotFunctionArgList(pnode%part_refs(1)%subscripts)) THEN
            ! Not a function reference or substring - must be an array.
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefDataObject, NatureRefArray],  &
                LOCATIONS=[ .LocationOf. pnode%part_refs(1)%part_name,  &
                    .LocationOf. pnode%part_refs(1)%part_name ] )
          ELSE
            ! Not a substring, might be an array element or function reference.
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefFunctionLike], &
                LOCATIONS=[.LocationOf. pnode%part_refs(1)%part_name] )
          END IF
        ELSE IF (SIZE(pnode%part_refs(1)%subscripts) /= 0) THEN
          ! We have more than one subscript, so we are not a substring-range.
          IF (IsNotFunctionArgList(pnode%part_refs(1)%subscripts)) THEN
            ! Must be an array.
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefDataObject, NatureRefArray],  &
                LOCATIONS=[ .LocationOf. pnode%part_refs(1)%part_name,  &
                    .LocationOf. pnode%part_refs(1)%part_name ] )
          ELSE
            ! Might be an array element or function reference.
            ident_list(1) = IdentifierNature(  &
                NAME=.NameOf. pnode%part_refs(1)%part_name,  &
                NATURES=[NatureRefFunctionLike], &
                LOCATIONS=[.LocationOf. pnode%part_refs(1)%part_name] )
          END IF
        ELSE
          ! We have not subscripts or substring-range.
          ident_list(1) = IdentifierNature(  &
              NAME=.NameOf. pnode%part_refs(1)%part_name,  &
              NATURES=[NatureRefObject], &
              LOCATIONS=[.LocationOf. pnode%part_refs(1)%part_name] )
        END IF
      END IF
    END IF
    
    IF (.NOT. ALLOCATED(ident_list)) ALLOCATE(ident_list(0))
    
  END SUBROUTINE desig_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of Designator%GetLocation - get the starting location 
  !! in source of the designator.
  !!
  !! @param[in]     pnode             The parse node of interest.
  !!
  !! @returns The starting location of the designator in source.
  !!
  !! (A Designator isn't a parse node (yet), but we've kept the same interface.
  
  FUNCTION desig_GetLocation(pnode) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Designator), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%part_refs)) THEN
      loc = pnode%part_refs(1)%GetLocation()
    ELSE
      loc = pnode%literal_location
    END IF
    
  END FUNCTION desig_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a Designator to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the designator to.
  !!
  !! @param[in]     desig             The designator to write out.
  
  RECURSIVE SUBROUTINE WriteDesignator(log_unit, desig)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(Designator), INTENT(IN) :: desig
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    IF (ALLOCATED(desig%part_refs)) THEN
      DO i = 1, SIZE(desig%part_refs) - 1
        CALL WritePartRef(log_unit, desig%part_refs(i))
        WRITE (log_unit, "(' % ')", ADVANCE='NO')
      END DO
      CALL WritePartRef(log_unit, desig%part_refs(i))
    ELSE
      WRITE (log_unit, "('/parent-string/ = ',A)", ADVANCE='NO')
      IF (ALLOCATED(desig%literal_kind)) THEN
        WRITE (log_unit, "(A, ' _ ')", ADVANCE='NO') desig%literal_kind
      END IF
      WRITE (log_unit, "(A)", ADVANCE='NO') desig%literal
    END IF
    
    IF (ALLOCATED(desig%substring)) THEN
      WRITE (log_unit, "(', /substring-range/ = (')", ADVANCE='NO')
      CALL WriteSubstringRange(log_unit, desig%substring)
      WRITE (log_unit, "(')')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE WriteDesignator
  
  
  !*****************************************************************************
  !!
  !> Count the number of parts in a /data-ref/ (or something that looks like 
  !! a /data-ref/).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token in @a tlist from 
  !! which to start the count (the first token of the data-ref).
  !!
  !! @param[in]     finish            Index of the token in @a tlist at which 
  !! to finish the count.
  !!
  !! We cannot distinguish things like binding names and procedure component 
  !! names from the rest of the data-ref - these will be counted as parts.
  
  FUNCTION CountParts(tlist, start, finish) RESULT(num_parts)
    
    USE MatchUtils
    USE Tokens
    
    !----------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    
    ! Function result
    INTEGER :: num_parts
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    INTEGER :: percent        ! Index of the next percent symbol.
    
    !***************************************************************************
    
    num_parts = 1
    it = start
    DO WHILE (it <= finish)
      percent = FindBare(tlist, ittPercent, scck_'%', it)
      IF ((percent == 0) .OR. (percent > finish)) RETURN
      num_parts = num_parts + 1
      it = percent + 1
    END DO
    
  END FUNCTION CountParts
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /structure-component/ (R613).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /structure-component/ in @a tlist.
  !!
  !! @param[out]    part_refs         The results of parsing the 
  !!/structure-component/.
  !!
  !! @param[out]    after             Index of the first token immediately 
  !! after the /structure-component/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseStructureComponent(tlist, start, part_refs, after, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(PartRef), INTENT(OUT), ALLOCATABLE :: part_refs(:)
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a structure component.
    CHARACTER(*), PARAMETER :: comp = 'R613'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! The /part-refs/ up until the last %.
    TYPE(PartRef), ALLOCATABLE :: left_part_refs(:)
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    INTEGER :: i              ! Part ref index - required for workaround.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL parse_left_parts_of_data_ref(  &
        tlist,  &
        start,  &
        left_part_refs,  &
        after,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! The rightmost part should just be a part name.
    
    CALL Expect(tlist, after, '/part-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ALLOCATE(part_refs(SIZE(left_part_refs) + 1))
    
    !===========================================================================
    ! ifort bug to do with array assignment of type with polymorphic component.
    ! The following is a work around for this
    ! part_refs(:SIZE(left_part_refs)) = left_part_refs
    do i = 1, size(left_part_refs)
      part_refs(i) = left_part_refs(i)
    end do
    !===========================================================================
    
    part_refs(SIZE(part_refs))%part_name = IdentifierIfortBug(tlist(after))
    
    after = after + 1
    
    ! We're done!
    
  END SUBROUTINE ParseStructureComponent
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up the left parts of a /data-ref/.
  !!
  !! Left here in terms of left and right.  So we parse a /data-ref/ up until 
  !! the last % separator (@a after will point to the first token after the 
  !! last separator).
  !!
  !! In some cases this my be the entire data ref according to the syntax 
  !! rules, as what comes after the last % is not a part-ref.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index in @a tlist of the first token of 
  !! the /data-ref/.
  !!
  !! @param[out]    part_refs         Results of parsing the left most parts 
  !! of the /data-ref/.
  !!
  !! @param[out]    after             Index of the token in @a tlist that is 
  !! immediately after the last "%" in the token sequence.  If there is no "%: 
  !! then this will be set to @a start and @a part-refs will be allocated to 
  !! zero size.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_left_parts_of_data_ref( tlist, start, part_refs, after,  &
      err_list )
    
    USE Errors
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(PartRef), INTENT(OUT), ALLOCATABLE :: part_refs(:)
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /data-ref/.
    CHARACTER(*), PARAMETER :: comp = 'R611'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    INTEGER :: pass           ! Pass number.
    INTEGER :: percent        ! Index of the next "%" character.
    INTEGER :: part           ! Current part number.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ! First pass we count, second pass we parse.
    pass_loop: DO pass = 1, 2
      
      it = start
      part = 0
      
      token_loop: DO WHILE (it <= SIZE(tlist))
        
        percent = FindBare(tlist, ittPercent, scck_'%', it)
        ! Set finish to point at the token at the end of the part-ref.
        IF (percent == 0) THEN
          IF (pass == 1) THEN
            ! Allocate space for the left parts that we found.
            EXIT token_loop
          ELSE
            ! We're done - set after to point to the position immediately 
            ! after the right most "%" and return.
            after = it
            RETURN
          END IF
        END IF
        
        ! If we get to here, then percent is non-zero, which means that there 
        ! is something that looks like a part-ref to the right of us.  Hence 
        ! what lies between it and percent-1 is one of the left parts.
        
        part = part + 1
        
        IF (pass == 2) THEN
          CALL ParsePartRef( tlist, it, percent - 1, part_refs(part),  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! Continue on error.
        END IF
        
        ! Move over the % that terminated this part.
        it = percent + 1
        
      END DO token_loop
      
      IF (pass == 1) ALLOCATE(part_refs(part))
      
    END DO pass_loop
    
    ! If we made it to here, then the last token in tlist was the percent.  
    ! This is likely a syntax error, but we leave it for client code to sort 
    ! out.
    after = it
    
  END SUBROUTINE parse_left_parts_of_data_ref
  
  
  !*****************************************************************************
  !!
  !> Parse a series of tokens that makes up a /part-ref/ (R612).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index in @a tlist of the first token 
  !! of the /part-ref/.
  !!
  !! @param[out]    finish            Index in @a tlist of the last token 
  !! of the /part-ref/.
  !!
  !! @param[out]    part              Results of parsing the /part-ref/.
  !!
  !! @param[out]    err_list          The list of errors.
  
  SUBROUTINE ParsePartRef(tlist, start, finish, part, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(PartRef), INTENT(OUT) :: part
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /part-ref/.
    CHARACTER(*), PARAMETER :: comp = 'R612'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The part-name
    
    CALL Expect(tlist, start, '/part-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Move across the name stuff
    part%part_name = IdentifierIfortBug(tlist(start))
    
    ! Do we just have the part-name?
    IF (start == finish) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional section-subscript-list in parentheses
    
    CALL Expect( tlist, start + 1, [ittDelimiter, ittDelimiter],  &
        [scck_'(', scck_'['], '"(" or /image-selector/', opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      CALL ExpectClosing(tlist, start + 1, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Go looking for a section-subscript-list
      
      IF (closing == start + 2) THEN
        ! Nothing inside the parentheses.
        CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(closing)),  &
            MSG='Expected /section-subscript-list/.' )
        RETURN
      END IF
      
      ! Parse that list.
      CALL ParseSectionSubscriptList( tlist(start+2:closing-1),  &
          part%subscripts, sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp )
      
      ! Skip over the closing ')'
      it = closing + 1
      
      ! Do we have an /image-selector/?
      IF (it > finish) RETURN
      
      CALL Expect(tlist, it, ittDelimiter, scck_'[', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    ELSE
      it = start + 1
    END IF
    ! it here index of token after name or after closing parenthesis of the 
    ! array subscript.
    
    !---------------------------------------------------------------------------
    ! Optional image-selector in square brackets
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse that list.
    CALL ParseSectionSubscriptList( tlist(it+1:closing-1),  &
        part%image_selector, sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp )
    
    ! Skip over the closing ']'
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! That should be it.
    
    CALL ExpectEnd(tlist(:finish), it, comp, err_list)
    
  END SUBROUTINE ParsePartRef
  
  
  !*****************************************************************************
  !!
  !> Implementation of PartRef%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE part_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PartRef), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WritePartRef(unit, pnode)
    
  END SUBROUTINE part_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of PartRef%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE part_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PartRef), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%subscripts)) THEN
      DO i = 1, SIZE(pnode%subscripts)
        CALL pnode%subscripts(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%image_selector)) THEN
      DO i = 1, SIZE(pnode%image_selector)
        CALL pnode%image_selector(i)%Query(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE part_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PartRef%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE part_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PartRef), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%subscripts)) THEN
      DO i = 1, SIZE(pnode%subscripts)
        CALL pnode%subscripts(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%image_selector)) THEN
      DO i = 1, SIZE(pnode%image_selector)
        CALL pnode%image_selector(i)%Modify(visitor)
      END DO
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE part_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PartRef%GetLocation - get the starting location 
  !! in source of the designator.
  !!
  !! @param[in]     pnode             The parse node of interest.
  !!
  !! @returns The starting location of the designator in source.
  !!
  !! (A PartRef isn't a parse node (yet), but we've kept the same interface.
  
  FUNCTION part_GetLocation(pnode) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(PartRef), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%part_name
    
  END FUNCTION part_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of PartRef%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name that the node represents.
  
  FUNCTION part_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'part-ref'
    
  END FUNCTION part_What
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a PartRef to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the part ref to.
  !!
  !! @param[out]    part_ref          The /part-ref/ to write out.
  
  RECURSIVE SUBROUTINE WritePartRef(log_unit, part_ref)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(PartRef), INTENT(IN) :: part_ref
    
    !***************************************************************************
    
    WRITE (log_unit, "(A)", ADVANCE='NO') .NameOf. part_ref%part_name
    
    IF (ALLOCATED(part_ref%subscripts)) THEN
      WRITE (log_unit, "('(')", ADVANCE='NO') 
      CALL WriteDimSpec(log_unit, part_ref%subscripts)
      WRITE (log_unit, "(')')", ADVANCE='NO')
    END IF
    
    IF (ALLOCATED(part_ref%image_selector)) THEN
      WRITE (log_unit, "('[')", ADVANCE='NO') 
      CALL WriteDimSpec(log_unit, part_ref%image_selector)
      WRITE (log_unit, "(']')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE WritePartRef
  
END MODULE Designators
