! $Id: EndfileStmts.f90 2926 2019-04-27 00:22:04Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the EndFileStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /endfile-stmt/'s (R925).

MODULE EndfileStmts
  
  USE ParseNodes
  USE PositionSpecs
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseEndfileStmt
  
  PUBLIC :: PositionSpecList      ! From PositionSpecs.
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing an /endfile-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: EndfileStmt
    !> The position specification in the statement.
    TYPE(PositionSpecList) :: position_specs
  CONTAINS
    PROCEDURE :: Dump => endfile_Dump
    PROCEDURE :: Query => endfile_Query
    PROCEDURE :: Modify => endfile_Modify
    PROCEDURE, NOPASS :: What => endfile_What
  END TYPE EndfileStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /endfile-stmt/ (R925).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    endfile_stmt      The results of parsing the 
  !! /endfile-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndfileStmt(tlist, endfile_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(EndfileStmt), INTENT(OUT) :: endfile_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /endfile-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R925'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /endfile-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENDFILE... or END FILE...
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'FILE', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! it here the index of the token after ...FILE.
    
    IF (it > SIZE(tlist)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /file-unit-number/ or "(" /position-spec-list/ ")".' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (TestPositionStmtFirstForm(tlist(it:))) THEN
      CALL ParseExpr(  &
          tlist,  &
          it, SIZE(tlist),  &
          endfile_stmt%position_specs%unit,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Failure not tested by StmtParsingTests because a missing 
      ! parenthesis causes the first form to be selected.
      CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, it, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ParsePositionSpecList(  &
          tlist,  &
          it + 1, closing - 1,  &
          endfile_stmt%position_specs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      
      CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    END IF
    
  END SUBROUTINE ParseEndfileStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of EndfileStmt%Dump - write a debugging representation 
  !! of the results of parsing an backspace statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE endfile_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EndfileStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL pnode%position_specs%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE endfile_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of EndfileStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE endfile_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EndfileStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%position_specs%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE endfile_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EndfileStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE endfile_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EndfileStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%position_specs%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE endfile_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of EndfileStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'endfile-stmt'.
  
  FUNCTION endfile_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'endfile-stmt'
    
  END FUNCTION endfile_What
  
END MODULE EndfileStmts
