! $Id: EnumDefStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the EnumDefStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /enum-def-stmt/'s, /enumerator-def-stmt/'s and
!! /end-enum-stmt/'s.

MODULE EnumDefStmts
  
  USE CompilerKinds
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseEnumDefStmt
  PUBLIC :: ParseEnumeratorDefStmt
  PUBLIC :: ParseEndEnumStmt
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a single enumerator definition.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: Enumerator
    !> The identifier for the enumerator.
    TYPE(Identifier) :: name
    !> The value of the enumerator.  May not be allocated when the enumerator
    !! has the default value.
    CLASS(Expression), ALLOCATABLE :: expr
  CONTAINS
    PROCEDURE :: Write => en_Write
    PROCEDURE :: Query => en_Query
    PROCEDURE :: Modify => en_Modify
    PROCEDURE, NOPASS :: What => en_What
    PROCEDURE :: GetLocation => en_GetLocation
    PROCEDURE :: Advertise => en_Advertise
  END TYPE Enumerator
  
  !> Stores the results of parsing an /enumerator-def-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: EnumeratorDefStmt
    !> The list of enumerators.
    TYPE(Enumerator), ALLOCATABLE :: enumerator_list(:)
  CONTAINS
    PROCEDURE :: Dump => enmd_Dump
    PROCEDURE :: Query => enmd_Query
    PROCEDURE :: Modify => enmd_Modify
    PROCEDURE, NOPASS :: What => enmd_What
  END TYPE EnumeratorDefStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /enum-def-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ls                The language spec on the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEnumDefStmt(tlist, ls, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE LanguageBindingSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(LangSpec), INTENT(OUT) :: ls
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /enum-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R459'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /enum-def-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENUM
    
    CALL ExpectKeyword(tlist, 1, scck_'ENUM', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! , BIND(C)
    
    CALL Expect(tlist, 2, ittComma, scck_',', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Parse(tlist, 3, SIZE(tlist), ls, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    IF (ALLOCATED(ls%name)) THEN
      ! If a name specifier is present, then in the absence of other 
      ! syntax errors the NAME token will be token six.
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(6)),  &
          MSG='The "NAME=" specifier is not permitted in a BIND &
            &clause in a /enum-def-stmt/.' )
    END IF
    
  END SUBROUTINE ParseEnumDefStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /enumerator-def-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    enmd_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEnumeratorDefStmt(tlist, enmd_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(EnumeratorDefStmt), INTENT(OUT) :: enmd_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /enumerator-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R460'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.    
    INTEGER :: list_count     ! Number of items in the binding list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.    
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************  
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /enumerator-def-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! ENUMERATOR
    
    CALL ExpectKeyword(tlist, 1, scck_'ENUMERATOR', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or name
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /enumerator/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------    
    ! /enumerator_list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(enmd_stmt%enumerator_list(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_enumerator(  &
          tlist,  &
          item_start, item_finish,  &
          enmd_stmt%enumerator_list(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseEnumeratorDefStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-enum-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndEnumStmt(tlist, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /end-enum-def-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R462'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: after          ! Position after expected token sequence.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-enum-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! END ENUM or ENDENUM
    
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'ENUM', after, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That's it!
    CALL ExpectEOS(tlist, after, comp, err_list)
    
  END SUBROUTINE ParseEndEnumStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumeratorDefStmt%Dump - write a debugging 
  !! representation of the results of parsing a /enumerator-def-stmt/ to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE enmd_Dump(pnode, unit, pfx)
    
    USE LanguageBindingSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumeratorDefStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! enumerator index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/enumerator-list/')") pfx
    DO i = 1, SIZE(pnode%enumerator_list)
      WRITE (unit, "(A,'/named-constant/ = ',A)", ADVANCE='NO')  &
          pfx // '  ', .NameOf. pnode%enumerator_list(i)%name
      IF (ALLOCATED(pnode%enumerator_list(i)%expr)) THEN
        WRITE (unit, "(2X,'/scalar-int-constant-expr/ = ')", ADVANCE='NO')
        CALL WriteExpr(unit, pnode%enumerator_list(i)%expr)
      END IF
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE enmd_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumeratorDefStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE enmd_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumeratorDefStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! enumerator index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%enumerator_list)
      CALL pnode%enumerator_list(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE enmd_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumeratorDefStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE enmd_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumeratorDefStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! enumerator index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%enumerator_list)
      CALL pnode%enumerator_list(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE enmd_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of EnumeratorDefStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule for the statement.
  
  FUNCTION enmd_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'enumerator-def-stmt'
    
  END FUNCTION enmd_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for Enumerator
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /enumerator/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the /enumerator/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the /enumerator/.
  !!
  !! @param[out]    enum              The results of parsing the enumerator.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_enumerator(tlist, start, finish, enum, err_list)
    
    USE CompilerKinds
    USE Errors
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Enumerator), INTENT(OUT) :: enum
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number of an /enumerator/.
    CHARACTER(*), PARAMETER :: comp = 'R461'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/named-constant/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    enum%name = IdentifierIfortBug(tlist(start))
    
    IF (start == finish) RETURN
    
    !---------------------------------------------------------------------------
    ! = expr
    
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ParseExpr(tlist, start + 2, finish, enum%expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_enumerator
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%Write - write a text representation of the 
  !! parse nodeto a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE en_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Enumerator), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%name
    IF (ALLOCATED(pnode%expr)) THEN
      WRITE (unit, "(' = ')", ADVANCE='NO')
      CALL pnode%expr%Write(unit)
    END IF
    
  END SUBROUTINE en_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE en_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Enumerator), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE en_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE en_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Enumerator), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE en_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION en_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'enumerator'
    
  END FUNCTION en_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION en_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(Enumerator), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION en_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of Enumerator%Advertise - advertise the identifiers 
  !! referenced or defined by the node.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE en_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Enumerator), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureDefHasType, NatureDefParameter],  &
        LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    
  END SUBROUTINE en_Advertise
  
END MODULE EnumDefStmts
