! $Id: ExecutionParts.f90 2179 2016-06-03 13:53:17Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ExecutionParts module.


!*******************************************************************************
!!
!> Procedures and types for parsing the execution part of a program unit 
!! or subprogram.

MODULE ExecutionParts
  
  USE ParseNodes
  
  USE Variables
  USE Expressions
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseExecutionPart
  
  PUBLIC :: ParseBlock
  PUBLIC :: DumpBlock
  PUBLIC :: QueryBlock
  PUBLIC :: ModifyBlock
  
  !-----------------------------------------------------------------------------
  
  !> Basic storage type for constructs that have a /block/ inside them - 
  !! i.e. a vector of child executable constructs.
  TYPE, EXTENDS(SCParseNode), ABSTRACT, PUBLIC :: BlockStorage
    !> List of parse nodes for child constructs and statements.
    TYPE(SCParseNodeElement), ALLOCATABLE :: children(:)
    
    ! While it is strongly related to the structure of the storage used by 
    ! the node, we leave the Dump, Query and Modify bindings deferred here 
    ! to force extensions to actually implement it so we don't forget to 
    ! dump/query/modify the extension specific stuff.  Extensions can call 
    ! DumpBlock to actually dump a block, QueryBlock to query a block, and 
    ! ModifyBlock to modify.
  END TYPE BlockStorage
  
  !> Represents the results of parsing an executable part.
  TYPE, EXTENDS(BlockStorage), PUBLIC :: ExecutionPart
  CONTAINS
    PROCEDURE :: Dump => exec_Dump
    PROCEDURE :: Query => exec_Query
    PROCEDURE :: Modify => exec_Modify
    PROCEDURE, NOPASS :: What => exec_What
  END TYPE ExecutionPart
  
  ! Some executable constructs can themselves contain one or more 
  ! execution-part-constructs, which causes a source dependency issue, 
  ! so we use an external procedure to do the actual parsing.
  INTERFACE
    SUBROUTINE ParseExecutionPartConstructListExternal( exec_st, children,  &
        err_list )
      USE ExecStNodes
      USE Errors
      USE ParseNodes
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      TYPE(ExecStNode), INTENT(INOUT) :: exec_st
      TYPE(SCParseNodeElement), INTENT(OUT), ALLOCATABLE :: children(:)
      TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    END SUBROUTINE ParseExecutionPartConstructListExternal
  END INTERFACE
  
CONTAINS
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures related to BlockStorage types.
  
  
  !*****************************************************************************
  !!
  !> Parse an ExecStNode that represents a /block/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! executable part to parse.
  !!
  !! @param[out]    spec              The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  
  RECURSIVE SUBROUTINE ParseBlock(st, children, err_list)
    
    USE Errors
    USE ExecStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExecStNode), INTENT(INOUT) :: st
    TYPE(SCParseNodeElement), INTENT(OUT), ALLOCATABLE :: children(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ! Forward to the external procedure.
    CALL ParseExecutionPartConstructListExternal(st, children, err_list)
    
  END SUBROUTINE ParseBlock
  
  
  !*****************************************************************************
  !!
  !> Dump a vector of ParseNodeElements that supposedly represent a 
  !! /block/ to a logical unit.
  !!
  !! @param[in]     children          The parse node elements to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  RECURSIVE SUBROUTINE DumpBlock(children, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SCParseNodeElement), INTENT(IN) :: children(:)
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(children)
      WRITE (unit, "(A,'/',A,'/')") pfx , children(i)%What()
      CALL children(i)%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE DumpBlock
  
  
  !*****************************************************************************
  !!
  !> Pass the query onto the child nodes of the block.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE QueryBlock(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockStorage), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%children)
      CALL pnode%children(i)%Query(visitor)
    END DO
    
  END SUBROUTINE QueryBlock
  
  
  !*****************************************************************************
  !!
  !> Pass the modify operation onto the child nodes of the block.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ModifyBlock(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(BlockStorage), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%children)
      CALL pnode%children(i)%Modify(visitor)
    END DO
    
  END SUBROUTINE ModifyBlock
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures related to ExecutionPart types.
  
  
  !*****************************************************************************
  !!
  !> Parse an ExecStNode that represents an /execution-part/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! executable part to parse.
  !!
  !! @param[out]    spec              The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This also includes the call to the Init binding to intially define 
  !! the parent component.
  
  RECURSIVE SUBROUTINE ParseExecutionPart(st, exec, err_list)
    
    USE Errors
    USE ExecStNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ExecStNode), INTENT(INOUT) :: st
    TYPE(ExecutionPart), INTENT(OUT) :: exec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL exec%Init(st)
    
    ! Forward to the external procedure.
    CALL ParseExecutionPartConstructListExternal(st, exec%children, err_list)
    
  END SUBROUTINE ParseExecutionPart
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExecutionPart%Dump - write a debugging representation 
  !! of the results of parsing an /execution-part/ to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  RECURSIVE SUBROUTINE exec_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExecutionPart), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    CALL DumpBlock(pnode%children, unit, pfx)
    
  END SUBROUTINE exec_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExecutionPart%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE exec_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ExecutionPart), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL QueryBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE exec_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExecutionPart%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE exec_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ExecutionPart), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL ModifyBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE exec_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExecutionPart%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'executable-part'
  
  FUNCTION exec_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'execution-part'
    
  END FUNCTION exec_What
  
END MODULE ExecutionParts
