! $Id: ExternalStmts.f90 2196 2016-06-07 02:56:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ExternalStmts module.


!*******************************************************************************
!!
!> Procedure for parsing an /external-stmt/ (R1210).

MODULE ExternalStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseExternalStmt
  
  !> Stores the results of parsing an /external-stmt/.
  TYPE, EXTENDS(NameListPNode), PUBLIC :: ExternalStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => ex_GetListName
    PROCEDURE, NOPASS :: What => ex_What
    PROCEDURE :: Advertise => ex_Advertise
  END TYPE ExternalStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /external-stmt/ (R1210).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    ex_stmt           The results of parsing the statement.
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseExternalStmt(tlist, ex_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ExternalStmt), INTENT(OUT) :: ex_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /external-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1210'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /external-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! EXTERNAL ...
    CALL ExpectKeyword(tlist, 1, scck_'EXTERNAL', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional double colon, otherwise a name list.
    
    CALL Expect( tlist, 2, [ittDeclare, ittName], [scck_'::', scck_'  '],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)      ! ::
      it = 3
    CASE (2)      ! A name
      it = 2
    CASE DEFAULT
      STOP 'Unexpected Expect opt result.'
    END SELECT
    
    !---------------------------------------------------------------------------
    ! The /external-name-list/.
    
    CALL ParseNameList( tlist, it, SIZE(tlist), '/external-name/',  &
        ex_stmt%name_list, sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! End of statement check handled above.
    
  END SUBROUTINE ParseExternalStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExternalStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/external-name-list/'
  
  FUNCTION ex_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/external-name-list/'
    
  END FUNCTION ex_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExternalStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'external-stmt'.
  
  FUNCTION ex_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'external-stmt'
    
  END FUNCTION ex_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ExternalStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ex_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE IntentSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExternalStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[NatureDefExternal], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE ex_Advertise
  
END MODULE ExternalStmts
