! $Id: FF08Obfuscate.f90 2323 2016-08-15 03:15:16Z ian $
! ff08 source code copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the FF08Diff program.


!*******************************************************************************
!!
!> Obfuscate Fortran source.
!!
!! STOP codes (which may be the program's exit code):
!! - no code:         Program completed successfully.
!!
!! - 2:               The source code contained errors that prevented parsing.
!!                    (I know what you want me to do, but I couldn't do it.)
!!
!! - 3:               There was a problem with the command line arguments.
!!                    (I haven't got a clue what you want to do.)

PROGRAM FF08Obfuscate
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  
  ! Name of this program - for the --version option.
  CHARACTER(*), PARAMETER :: program_name  &
      = 'FF08Obfuscate'
  
  ! Format specification for an error message.
  CHARACTER(*), PARAMETER :: fmt_error = "(A)"
  
  !*****************************************************************************
  
  CALL run
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Entry point.
  !!
  !! Moved into an internal procedure to avoid unwanted host association.
  
  SUBROUTINE run
    
    USE CmdLine
    USE Debug
    USE FileUtilities
    USE Strings
    USE UnitUtilities
    USE Version
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: OUTPUT_UNIT, ERROR_UNIT
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Command line option description and parsing results.
    !
    ! In order: --help, --version, --dbg, --list.
    TYPE(CmdLineOption) :: cmd_line_options(4)
    
    ! List of non-option arguments.
    TYPE(String), ALLOCATABLE :: arguments(:)
    
    INTEGER :: stat             ! Error code for ParseCmdLine call.
    INTEGER :: stop_code        ! Stop code from procedure calls
    INTEGER :: i                ! Option argument index.
    
    ! Error message
    CHARACTER(:), ALLOCATABLE :: msg
    
    !***************************************************************************
    
    !---------------------------------------------------------------------------
    ! Set up and process command line options.
    
    cmd_line_options(1) = CmdLineOption(  &
        LONG_NAME='help',  &
        CHAR_NAME='?',  &
        DESCRIPTION='Display command line and program usage help, then exit.' )
    
    cmd_line_options(2) = CmdLineOption(  &
        LONG_NAME='version',  &
        DESCRIPTION='Display program version information, then exit.' )
    
    cmd_line_options(3) = CmdLineOption('dbg')
    
    cmd_line_options(4) = CmdLineOption(  &
        LONG_NAME='list',  &
        DESCRIPTION='Specify a list file - each record from the given file &
                    &will be added (in addition to any source file &
                    &specifications provided on the command line) to the &
                    &list of source file specifications to be processed.  &
                    &May be specified multiple times.',  &
        OPTION_TYPE=ioptArg,  &
        CAN_REPEAT=.TRUE. )
    
    CALL ParseCmdLine(cmd_line_options, arguments, stat, msg)
    
    IF (stat /= 0) THEN
      WRITE (ERROR_UNIT, "('Error parsing command line:',A)") msg
      CALL usage_message(OUTPUT_UNIT, cmd_line_options)
      STOP 3
    END IF
    
    ! Check for the --help (or -?) option in priority to all others.
    !
    ! If no arguments or options are provided (i.e. no command line 
    ! arguments) then we have nothing to do.  Interpret that as 
    ! a request for help as well.
    IF ( (cmd_line_options(1)%PresentCount > 0)  &
        .OR. (COMMAND_ARGUMENT_COUNT() == 0) ) THEN
      CALL usage_message(OUTPUT_UNIT, cmd_line_options)
      STOP
    END IF
    
    ! Check for the --version option.
    IF (cmd_line_options(2)%PresentCount > 0) THEN
      CALL VersionMessage(program_name)
      STOP
    END IF
    
    ! Check for the --dbg option.
    IF (cmd_line_options(3)%PresentCount > 0) THEN
      ! The Dbg* variables are global variables from the Debug module.  
      ! Setting the following flag enables debug output for all procedures 
      ! that have such output.  It is also what we use to remember that 
      ! we need to close the debugging log file.
      DbgFlag = .TRUE.
      CALL GetUnit(DbgUnit)
      OPEN( UNIT=DbgUnit,  &
          FILE=program_name //  DbgExtension,  &
          STATUS='REPLACE',  &
          ACTION='WRITE' )
      WRITE (ERROR_UNIT, "('DbgFlag set')")
    ELSE
      DbgFlag = .FALSE.
    END IF
    
    ! If there are any list files specified then expand the list of files 
    ! to accomodate their contents.
    DO i = 1, cmd_line_options(4)%PresentCount
      CALL AddListFile(  &
          arguments,  &
          cmd_line_options(4)%Argument(i)%item,  &
          ERROR_UNIT,  &
          stop_code )
      IF (stop_code /= 0) CALL do_stop(stop_code)
    END DO
    
    !---------------------------------------------------------------------------
    
    CALL execute(  &
        arguments,  &
        OUTPUT_UNIT,  &
        ERROR_UNIT,  &
        stop_code )
    
    !---------------------------------------------------------------------------
    ! Cleanup and shut down.
    
    CALL do_stop(stop_code)
    
  END SUBROUTINE run
  
  
  !*****************************************************************************
  !!
  !> Execute the obfuscation.
  !!
  !! @param[in]     file_specs        The file specifications nominating the 
  !! source files to obfuscate.
  !!
  !! @param[in]     out_unit          Unit connected for formatted output 
  !! for the results of the obfuscation.
  !!
  !! @param[in]     err_unit          Unit connected for formatted output 
  !! for reporting errors.
  !!
  !! @param[out]    stop_code         Stop code for reporting by the main 
  !! program.
  
  SUBROUTINE execute(file_specs, out_unit, err_unit, stop_code)
    
    USE BaseStNodes
    USE BuildTreePass
    USE CompilerHosts
    USE DefaultHosts
    USE Errors
    USE ObfuscateImplementation
    USE ParseNodes
    USE Programs
    USE Strings
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(String), INTENT(IN) :: file_specs(:)
    INTEGER, INTENT(IN) :: out_unit
    INTEGER, INTENT(IN) :: err_unit
    INTEGER, INTENT(OUT) :: stop_code
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Compiler host object to define host behaviour.
    TYPE(DefaultHost) :: host
    
    INTEGER :: ie             ! Error index.
    
    ! File names for the left set.
    TYPE(String), ALLOCATABLE :: file_names(:)
    
    ! The root of the statement tree for the file.
    TYPE(BaseStNode), POINTER :: base
    
    ! The root of the parse tree for the file.
    TYPE(Program) :: program_pnode
    
    ! List of errors from various procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    ! Convert the list of file specifications to a list of file names.  In 
    ! many cases this will be a simple copy.
    CALL ExpandSpecsToNames(host, file_specs, file_names, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Build and check the statement tree for all files.
    
    CALL build_combined_statement_tree(  &
        file_names,  &
        '',  &
        err_unit,  &
        base,  &
        stop_code )
    IF (stop_code /= 0) RETURN
    
    CALL DoBuildTreePassCheck(base, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Convert the statement tree to the parse tree.
    
    CALL ParseProgram(base, program_pnode, err_list)
    DO ie = 1, SIZE(err_list)
      WRITE (err_unit, fmt_error) .Describe. err_list(ie)
    END DO
    IF (Failure(err_list)) THEN
      stop_code = 2
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL Obfuscate(program_pnode, out_unit)
    
    stop_code = 0
    
  END SUBROUTINE execute
  
  
  !*****************************************************************************
  !!
  !> Build the combined statement tree for all files.
  !!
  !! @param[in]     filenames         The names of the source files.
  !!
  !! @param[in]     console           Source form for source read from 
  !! the console.  If the empty string, no source will be read from the 
  !! console, otherwise 'FREE' to read source as free form or 'FIXED' 
  !! to read source as fixed form.
  !!
  !! @param[out]    combined          The combined statement tree for all 
  !! source files.
  !!
  !! @param[in]     err_unit          Unit connected for formatted output 
  !! for reporting errors.
  !!
  !! @param[out]    stop_code         Stop code for reporting by the main 
  !! program.
  
  SUBROUTINE build_combined_statement_tree( filenames, console, err_unit,  &
      combined, stop_code )
    
    USE BaseStNodes
    USE BuildTreePass
    USE DefaultHosts
    USE ScanFixedForm
    USE ScanFreeForm
    USE Sources
    USE Scanner
    USE SourceUnits
    USE Strings
    USE Errors
    
    USE, INTRINSIC :: ISO_FORTRAN_ENV, ONLY: INPUT_UNIT
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(String), INTENT(IN) :: filenames(:)
    CHARACTER(*), INTENT(IN) :: console
    INTEGER, INTENT(IN) :: err_unit
    TYPE(BaseStNode), INTENT(OUT), POINTER :: combined
    INTEGER, INTENT(OUT) :: stop_code
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ie             ! Error index.
    INTEGER :: if             ! File index.
    
    ! Compiler host object to define host behaviour.
    TYPE(DefaultHost) :: host
    
    ! Error object for procedure calls.
    TYPE(Error), ALLOCATABLE :: err_list(:)
    
    ! The statement tree for the current file.
    TYPE(BaseStNode), POINTER :: current
    
    ! Source form for reading from the console.
    CLASS(SourceForm), ALLOCATABLE :: console_form
    
    ! Source object connected to the input unit.
    TYPE(SourceUnit), POINTER :: source_unit
    
    ! Pointer to source_unit.
    CLASS(Source), POINTER :: p_source
    
    !***************************************************************************
    
    ALLOCATE(combined)
    ALLOCATE(combined%prog_units(0))
    
    DO if = 1, SIZE(filenames)
      CALL DoBuildTreePass(  &
          host,  &
          filenames(if)%item,  &
          current,  &
          err_list )
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_error) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END DO
    
    ! If enabled, read statements from the console.
    IF (console /= '') THEN
      CALL ConnectSourceUnit(INPUT_UNIT, source_unit)
      p_source => source_unit
      SELECT CASE (console)
      CASE ('FIXED')
        ALLOCATE(console_form, SOURCE=FixedFormSource(p_source))
      CASE ('FREE')
        ALLOCATE(console_form, SOURCE=FreeFormSource(p_source))
      END SELECT
      
      CALL DoBuildTreePass(  &
          host,  &
          console_form,  &
          current,  &
          err_list)
      DO ie = 1, SIZE(err_list)
        WRITE (err_unit, fmt_error) .Describe. err_list(ie)
      END DO
      IF (.NOT. Failure(err_list)) THEN
        CALL MergeTrees(combined, current, err_list)
        DO ie = 1, SIZE(err_list)
          WRITE (err_unit, fmt_error) .Describe. err_list(ie)
        END DO
      END IF
      ! current is allocated in DoBuildTreePass, so deallocate it here.
      DEALLOCATE(current)
      IF (Failure(err_list)) THEN
        stop_code = 2
        RETURN
      END IF
    END IF
    
    stop_code = 0
    
  END SUBROUTINE build_combined_statement_tree
  
  
  !*****************************************************************************
  !!
  !> Process a stop code.
  !!
  !! @param[in]     stop_code         The stop code to process.
  !!
  !! F2003 doesn't allow an expression as a stop-code (F2008 relaxes this 
  !! to a constant expression, which is still frustrating) so this is 
  !! a work around.
  !!
  !! If the stop code is zero, we return control to the main program so that 
  !! it can end by simply hitting the END PROGRAM statement.
  
  SUBROUTINE do_stop(stop_code)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: stop_code
    
    !***************************************************************************
    
    SELECT CASE (stop_code)
    CASE (0)   ! Fall through to end statement.
    CASE (1) ; STOP 1
    CASE (2) ; STOP 2
    CASE DEFAULT;  STOP 3
    END SELECT
    
  END SUBROUTINE do_stop
  
  
  !*****************************************************************************
  !!
  !> Prints the command line usage message for the program.  This is 
  !! intended as the handler for '--help' or similar.
  !!
  !! @param[in]     out_unit          Logical unit connected for formatted 
  !! output to receieve the usage message.
  !!
  !! @param[in]     cmd_line_options  Command line options defined for the 
  !! program.
  !!
  !! Try and keep this consistent with the description in the lead-in 
  !! comments for the main program, or users and (more importantly) 
  !! programmers will get confused.
  
  SUBROUTINE usage_message(out_unit, cmd_line_options)
    
    USE CmdLine
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: out_unit
    TYPE(CmdLineOption), INTENT(IN) :: cmd_line_options(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Format specification for a lines worth of message.
    CHARACTER(*), PARAMETER :: fmt_a = "(A)"
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! First argument on command line, which is normally the program name.
    CHARACTER(:), ALLOCATABLE :: local_program_name
    
    INTEGER :: stat           ! Error code for command line argument query.
    
    !***************************************************************************
    
    WRITE (out_unit, fmt_a) 'Obfuscate Fortran source.'
    WRITE (out_unit, "()")
    
    ! Use actual program name in description, in case we've been renamed.
    CALL GetCmdLineArgument(0, local_program_name, stat)
    IF (stat /= 0) local_program_name = program_name
    
    WRITE (out_unit, fmt_a) 'Usage: ' // local_program_name  &
        // ' [options] source-file-specs...'
    WRITE (out_unit, "()")
    WRITE (out_unit, fmt_a) 'Valid options are:'
    WRITE (out_unit, "()")
    
    CALL WriteUsageMessage(cmd_line_options, out_unit)
    
    ! Short and simple - we won't elaborate on names versus spec's here.  We 
    ! rely on format reversion to do multiple records for each 80 character 
    ! (approx) line.
    WRITE (out_unit, fmt_a)  &
!        1234567890123456798012345678901234567890
        'source-file-specs is a list of the names&
        & of the Fortran source files to parse.'
    WRITE (out_unit, "()")
    WRITE (out_unit, fmt_a)  &
        'The obfuscated source is written to the console.'
  END SUBROUTINE usage_message
  
END PROGRAM FF08Obfuscate
