! $Id: FlushStmts.f90 2803 2019-03-22 19:36:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the FlushStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /flush-stmt/'s (R904).

MODULE FlushStmts
  
  USE Expressions
  USE Labels
  USE Variables
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseFlushStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /flush-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: FlushStmt
    !> The UNIT specification.
    !!
    !! This is a /file-unit-number/.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: unit
    !> The IOMSG specification
    !!
    !! This is an /iomsg-variable/.
    !!
    !! Not allocated if there was no IOMSG specifier.
    CLASS(Variable), ALLOCATABLE :: iomsg
    !> The IOSTAT specification.
    !!
    !! Not allocated if there was no IOSTAT specifier.
    CLASS(Variable), ALLOCATABLE :: iostat
    !> The ERR specification.
    !!
    !! Not allocated if there was no ERR specifier.
    TYPE(Label), ALLOCATABLE :: err
  CONTAINS
    PROCEDURE :: Dump => flush_Dump
    PROCEDURE :: Query => flush_Query
    PROCEDURE :: Modify => flush_Modify
    PROCEDURE, NOPASS :: What => flush_What
  END TYPE FlushStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /flush-stmt/ (R928).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseFlushStmt(tlist, flush_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE PositionSpecs
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(FlushStmt), INTENT(OUT) :: flush_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /flush-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R928'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the /flush-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.   
    LOGICAL :: no_spec        ! Flag to indicate UNIT= absent.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /flush-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! FLUSH...
    
    CALL ExpectKeyword(tlist, 1, scck_'FLUSH', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (SIZE(tlist) == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /file-unit-number/ or "(" /flush-spec-list/ ")".' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! /file-unit-number/ or ( /flush-spec-list/ )
    
    IF (TestPositionStmtFirstForm(tlist(2:))) THEN
      CALL ParseExpr(tlist, 2, SIZE(tlist), flush_stmt%unit, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      list_count = ListItemCount(tlist(3:closing-1))
      item_finish = 0
      DO item = 1, list_count
        CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
        CALL parse_flush_spec(  &
            tlist,  &
            item_start, item_finish,  &
            flush_stmt,  &
            no_spec,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        ! Continue on error but don't rely upon results from parse_flush_spec.
        IF (.NOT. Failure(sub_err_list)) THEN
          IF ((item /= 1) .AND. no_spec) THEN
            ! No syntax rule.
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C945',  &
                LOCATION=QueryLocation(tlist(item_start)),  &
                MSG='If the optional "UNIT=" is omitted, the &
                  &/file-unit-number/ shall be the first item in the &
                  &/flush-spec-list/.' )
            ! Continue on error.
          END IF
        END IF
      END DO
    
      IF (.NOT. ALLOCATED(flush_stmt%unit)) THEN
        ! No syntax rule.
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C945',  &
            LOCATION=QueryLocation(tlist(item_start)),  &
            MSG='A /file-unit-number/ shall be specified in a &
              &/flush-spec-list/.' )
      END IF
      
      ! That's it.
      CALL ExpectEOS(tlist, closing + 1, comp, err_list)
      
    END IF
    
  END SUBROUTINE ParseFlushStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /flush-spec/ (R929).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist of 
  !! the item.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist of 
  !! the item.
  !!
  !! @param[out]    flush_stmt        The results of parsing the spec.  Note 
  !! that this is an INTENT(INOUT) argument - each spec sets a particular 
  !! component.
  !!
  !! @param[out]    no_spec           Flag to indicate that the "XXX=" part of 
  !! the spec was absent.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /flush-spec/ is identical as far as we can tell to a /position-spec/.
  !! But we've gone to the trouble of writing a dedicated parse procedure 
  !! for it anyway (an alternative would have been to factor out the 
  !! syntax number and rule name independent stuff).
  
  SUBROUTINE parse_flush_spec( tlist, start, finish, flush_stmt,  &
      no_spec, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs, ONLY: ParseFileUnitNumber
    USE MatchUtils
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
      
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(FlushStmt), INTENT(INOUT) :: flush_stmt
    LOGICAL, INTENT(OUT) :: no_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Keywords for each /flush-spec/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(4) = [  &
        scck_'UNIT  ', scck_'IOMSG ', scck_'IOSTAT', scck_'ERR   ' ]
    
    ! Syntax rule number for a /flush-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R929'
  
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option for the keyword test.
    INTEGER :: assign         ! Index of the =.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /flush-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have an  argument keyword?
    
    assign = FindBare(tlist(:finish), ittAssign, scck_'=', start)
    IF (assign == 0) THEN
      ! No.
      no_spec = .TRUE.
      opt = 1
    ELSE
      ! Yes.  What was the keyword?
      no_spec = .FALSE.
      CALL ExpectKeyword(tlist, start, keywords, '/flush-spec/', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    
      CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    SELECT CASE (opt)
    CASE (1)      ! UNIT
      IF (ALLOCATED(flush_stmt%unit)) THEN
        CALL already('UNIT')
      ELSE
        IF (assign == 0) THEN
          CALL ParseFileUnitNumber(  &
              tlist,  &
              start, finish,  &
              flush_stmt%unit,  &
              sub_err_list )
        ELSE
          CALL ParseFileUnitNumber(  &
              tlist,  &
              assign + 1, finish,  &
              flush_stmt%unit,  &
              sub_err_list )
        END IF
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! IOMSG
      IF (ALLOCATED(flush_stmt%iomsg)) THEN
        CALL already('IOMSG')
      ELSE
        ! The syntax term is actually an /iomsg-variable/.
        CALL ParseVariable(  &
            tlist, &
            assign + 1, finish,  &
            flush_stmt%iomsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! IOSTAT
      IF (ALLOCATED(flush_stmt%iostat)) THEN
        CALL already('IOSTAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            assign + 1, finish,  &
            flush_stmt%iostat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (4)      ! ERR
      IF (ALLOCATED(flush_stmt%err)) THEN
        CALL already('ERR')
      ELSE
        ALLOCATE(flush_stmt%err)
        CALL ParseLabel(  &
            tlist,  &
            assign + 1, finish,  &
            flush_stmt%err,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /flush-spec/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C944',  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this statement.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_flush_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of FlushStmt%Dump - write a debugging 
  !! representation of the results of parsing a computed goto statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE flush_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FlushStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%unit)) THEN
      WRITE (unit, "(A,'UNIT = ')", ADVANCE='NO') pfx
      CALL WriteExpr(unit, pnode%unit)
      WRITE (unit, "()")
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      WRITE (unit, "(A,'IOMSG = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iomsg)
      WRITE (unit, "()")
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      WRITE (unit, "(A,'IOSTAT = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%iostat)
      WRITE (unit, "()")
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      WRITE (unit, "(A,'ERR = ',I0)") pfx, .ValueOf. pnode%err
    END IF
    
  END SUBROUTINE flush_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FlushStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE flush_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FlushStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%unit%Query(visitor)
    
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Query(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE flush_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FlushStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE flush_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FlushStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%unit%Modify(visitor)
    
    IF (ALLOCATED(pnode%iomsg)) CALL pnode%iomsg%Modify(visitor)
    IF (ALLOCATED(pnode%iostat)) CALL pnode%iostat%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE flush_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FlushStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'flush-stmt'.
  
  FUNCTION flush_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'flush-stmt'
    
  END FUNCTION flush_What
  
END MODULE FlushStmts
