! $Id: ForallConstructs.f90 2876 2019-04-03 20:51:06Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ForallConstructs module.


!*******************************************************************************
!!
!> Procedures and types for parsing /forall-construct/'s.

MODULE ForallConstructs
  
  USE ForallStmts
  USE ExecutionParts
  USE Identifiers
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseForallConstruct
  PUBLIC :: ParseForallStmtToConstruct
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a forall statement or forall construct.
  TYPE, EXTENDS(BlockStorage), PUBLIC :: ForallConstruct
    !> The construct name, if present.
    TYPE(Identifier), ALLOCATABLE :: construct_name
    !> The results of parsing the /forall-header/.
    TYPE(ForallHeader) :: header
  CONTAINS
    PROCEDURE :: Dump => forall_Dump
    PROCEDURE, NOPASS :: What => forall_What
    PROCEDURE :: Query => forall_Query
    PROCEDURE :: Modify => forall_Modify
    PROCEDURE :: Advertise => forall_Advertise
  END TYPE ForallConstruct
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a statement tree node for a /forall-construct/.
  !!
  !! @param[in]     st                The statement tree node for the 
  !! construct.
  !!
  !! @param[out]    construct         The results of parsing the construct.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseForallConstruct(st, construct, err_list)
    
    USE ForallConstructStNodes
    USE Errors
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ForallConstructStNode), INTENT(INOUT) :: st
    TYPE(ForallConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The opening forall statement.
    
    CALL ParseForallConstructStmt(  &
        st%first%tlist,  &
        construct%header,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Executable children (/forall-body-construct/s).
    
    CALL ParseBlock(st%exec, construct%children, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! The closing end forall statement.
    
    CALL ParseEndForallStmt(  &
        st%last%tlist,  &
        construct%construct_name,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    
  END SUBROUTINE ParseForallConstruct
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallConstruct%Dump - write a debugging 
  !! representation of the results of parsing a forall construct to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE forall_Dump(pnode, unit, pfx)
    
    USE ExecutionParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallConstruct), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%construct_name))  &
        WRITE (unit, "(A,'construct_name = ',A)") pfx,  &
          .NameOf. pnode%construct_name
    
    ! The expression.
    WRITE (unit, "(A,'header = ')", ADVANCE='NO') pfx
    CALL pnode%header%Write(unit)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'child executable statements:')") pfx
    CALL DumpBlock(pnode%children, unit, pfx // '  ')
    
  END SUBROUTINE forall_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallConstruct%Query.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE forall_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallConstruct), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%header%Query(visitor)
    
    CALL QueryBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE forall_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallConstruct%Modify.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE forall_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ForallConstruct), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%header%Modify(visitor)
    
    CALL ModifyBlock(pnode, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE forall_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallConstruct%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'forall-construct'.
  
  FUNCTION forall_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'forall-construct'
    
  END FUNCTION forall_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ForallConstruct%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE forall_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallConstruct), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(0))
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL MergeNatures(  &
          ident_list,  &
          IdentifierNature(  &
            NAME=.NameOf. pnode%construct_name,  &
            NATURES=[NatureDefConstruct], &
            LOCATIONS=[.LocationOf. pnode%construct_name] ) )
    END IF
    
  END SUBROUTINE forall_Advertise
  
  
  !*****************************************************************************
  !!
  !> Parse an IfStmt through to an IfConstruct structure.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    if_construct      The results of parsing the /if-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseForallStmtToConstruct(tlist, construct, err_list)
    
    USE Classifier
    USE Errors
    USE ErrorCodes
    USE ExecStNodes
    USE ExecutionParts
    USE SingleStNodes
    USE Statements
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(INOUT) :: tlist(:)
    TYPE(ForallConstruct), INTENT(OUT) :: construct
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /forall-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R759'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Tokens that make up the child /forall-assignment-stmt/ of the 
    ! /forall-stmt/.
    TYPE(Token), ALLOCATABLE :: child_tlist(:)
    
    ! Action statement tree node.
    TYPE(ExecStNode) :: exec_st
    
    INTEGER :: ist            ! Child statement number.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Parse the leading if-stmt.
    
    CALL ParseForallStmt(  &
        tlist,  &
        construct%header,  &
        child_tlist,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list)
    ! No point continuing if we can't parse the front bit.
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Work out what the action statement is.
    
    CALL Classify(child_tlist, [INTEGER ::], ist)
    SELECT CASE (ist)
    CASE (istUnclassified)
      ! Not covered by ConstructParsingTests.
      CALL Add( err_list,  &
          CODE=errUnclassifiedStatement,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='Unclassifiable /forall-assignment-stmt/.' )
      
    CASE (istAssignment, istPointerAssignment)
      ! Parse it.
      CALL MakeExecStNode(child_tlist, ist, exec_st)
      CALL ParseBlock(exec_st, construct%children, sub_err_list)
      CALL Add(err_list, sub_err_list)
      
    CASE DEFAULT
      CALL Add( err_list,  &
          CODE=errIfBadChildStmt,  &
          COMPONENT='R757 via ' // comp,  &  
          LOCATION=QueryLocation(child_tlist(1)),  &
          MSG='A ' // GetStName(ist) // ' is not a valid &
            &/forall-assignment-stmt/.' )
      RETURN
      
    END SELECT
    
  END SUBROUTINE ParseForallStmtToConstruct
  
END MODULE ForallConstructs
