! $Id: FormatStmts.f90 2791 2019-03-20 07:28:15Z ian $
!> @file
!! Defines the FormatStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /format-stmt/'s and /format-specifications/.

MODULE FormatStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseFormatStmt
  
  !> Stores the results of parsing a format statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: FormatStmt
    !> The string that is the format specification.
    CHARACTER(:), ALLOCATABLE :: format_spec
  CONTAINS
    PROCEDURE :: Dump => fmt_Dump
    PROCEDURE :: Query => fmt_Query
    PROCEDURE :: Modify => fmt_Modify
    PROCEDURE, NOPASS :: What => fmt_What
  END TYPE FormatStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /format-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    fmt_stmt          The results of parsing a format 
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseFormatStmt(tlist, fmt_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(FormatStmt), INTENT(OUT) :: fmt_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /format-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1001'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
  
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /format-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! FORMAT ...
    
    CALL Expect(tlist, 1, ittName, scck_'FORMAT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... /format-specification/    (which includes the parentheses).
    
    CALL Expect(tlist, 2, ittFormatSpecification, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    fmt_stmt%format_spec = QueryValue(tlist(2))      ! Kind conversion.
    
    ! That's all folks...
    CALL ExpectEOS(tlist, 3, comp, err_list)
    
  END SUBROUTINE ParseFormatStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of FormatStmt%Dump - write a debugging representation 
  !! of the results of parsing a format statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE fmt_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FormatStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit,"(A,'/format-specification/ = ',A)") pfx, pnode%format_spec
    
  END SUBROUTINE fmt_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of FormatStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE fmt_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FormatStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fmt_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of FormatStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE fmt_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FormatStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE fmt_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of FormatStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'format-stmt'.
  
  FUNCTION fmt_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'format-stmt'
    
  END FUNCTION fmt_What
  
END MODULE FormatStmts
