! $Id: Formatters.f90 2761 2016-12-21 05:45:54Z ian $
!> @file
!! Defines the Formatters module.


!*******************************************************************************
!!
!> Procedures for formatting render nodes into text.

MODULE Formatters
  
  USE RenderNodes
  USE Statements
  USE Tokens, ONLY: ittEndStatement,  &
      ittName, ittKeyword, ittIntegerLiteral, ittRealLiteral,  &
      ittCharacterLiteral, ittLogicalLiteral, ittBOZLiteral, ittOperator,  &
      ittStatementLabel, ittDelimiter, ittComma, ittAssign, ittAssociate,  &
      ittColon, ittDeclare, ittPercent, ittFormatSpecification
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: FormatFreeForm
  
  !-----------------------------------------------------------------------------
  
  TYPE, PUBLIC :: FormatFreeFormOptions
    !> The maximum line length.
    INTEGER :: MaximumLineLength = 80
    
    !> Number of spaces in indent added when entering a construct.
    !!
    !! The @a IncreaseIndentStatements defines the list of statements 
    !! that this indent applies to.
    INTEGER :: ConstuctIndent = 2
    
    !> Number of spaces in indent added when a statement is continued.
!    INTEGER :: StmtContinueIndent = 4
    
    !> Number of spaces in indent when a general syntax term is continued.
    INTEGER :: TermContinueIndent = 2
    
    !> The maximum indent permitted.
    INTEGER :: MaximumIndent = 40
    
    !> The statements that trigger an increment in construct indent after 
    !! they are formatted.
    !!
    !! If not allocated, then the value of the IncreaseIndentStatements 
    !! named constant is used.
    INTEGER, ALLOCATABLE :: IncreaseIndentStatements(:)
    
    !> The statements that trigger a decrement in construct indent before 
    !! they are formatted.
    !!
    !! If not allocated then the value of the DecreateIndentStatements 
    !! named constant is used.
    INTEGER, ALLOCATABLE :: DecreaseIndentStatements(:)
    
    !> The preferred quote delimiter.
    CHARACTER :: PreferredQuote = "'"
    
    !> The continuation suffix.
    CHARACTER(:), PRIVATE, ALLOCATABLE :: continue
  END TYPE FormatFreeFormOptions
  
  !> Statements that reduce indent.
  INTEGER, PARAMETER :: DecreaseIndentStatements(23) = [  &
      istEndDo,             istEndAssociate,      istCase,  &
      istTypeGuard,         istEndIf,             istElseIf,  &
      istElse,              istEndFunction,       istEndSubroutine,  &
      istEndInterface,      istEndType,           istContains,  &
      istEndEnum,           istElsewhere,         istMaskedElsewhere,  &
      istEndWhere,          istEndForall,         istEndModule,  &
      istEndSubmodule,      istEndProgram,        istEndBlockData,  &
      istEndSelect,         istEndSelectType ]
  
  !> Statements that increase indent.
  INTEGER, PARAMETER :: IncreaseIndentStatements(23) = [  &
      istNonlabelDo,        istAssociate,         istCase,  &
      istTypeGuard,         istIfThen,            istElseIf,  &
      istElse,              istFunction,          istSubroutine,  &
      istInterface,         istDerivedType,       istContains,  &
      istEnumDef,           istWhereConstruct,    istMaskedElsewhere,  &
      istElsewhere,         istForallConstruct,   istModule,  &
      istSubmodule,         istProgram,           istBlockData,  &
      istSelectCase,        istSelectType ]
  
  !> Token types that must be separated by a blank.
  INTEGER, PARAMETER :: separate(8)  = [  &
      ittKeyword,           ittName,              ittIntegerLiteral,  &
      ittRealLiteral,       ittCharacterLiteral,  ittLogicalLiteral,  &
      ittBOZLiteral,        ittStatementLabel ] 
  
  !> State object for FormatFreeForm.
  TYPE :: free_form_state
    !> Previous type of token written to the file.
    INTEGER :: prev_type = ittEndStatement
    !> Current indent stack for statements.
    INTEGER, ALLOCATABLE :: indent_stack(:)
    !> Position in current record of last character written.
    INTEGER :: pos = 0
  END TYPE free_form_state 
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Format a tree of render nodes in free form.

  SUBROUTINE FormatFreeForm(unit, rnode, opts)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    CLASS(RenderNode), INTENT(IN) :: rnode
    TYPE(FormatFreeFormOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    TYPE(FormatFreeFormOptions) :: local_opts
    TYPE(free_form_state) :: state
    
    !***************************************************************************
    
    local_opts = opts
    IF (.NOT. ALLOCATED(local_opts%IncreaseIndentStatements)) THEN
      local_opts%IncreaseIndentStatements = IncreaseIndentStatements
    END IF
    IF (.NOT. ALLOCATED(local_opts%DecreaseIndentStatements)) THEN
      local_opts%DecreaseIndentStatements = DecreaseIndentStatements
    END IF
    local_opts%continue = '  &'
    
    state%indent_stack = [0]
    
    CALL free_form_write_rnode(unit, rnode, local_opts, state)
    
  END SUBROUTINE FormatFreeForm
  
  
  !*****************************************************************************
  !!
  !> Render a sequence of nodes to a formatted file.
  
  RECURSIVE SUBROUTINE free_form_write_rnode(unit, rnode, opts, state)
    
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    CLASS(RenderNode), INTENT(IN) :: rnode
    TYPE(FormatFreeFormOptions), INTENT(IN) :: opts
    TYPE(free_form_state), INTENT(INOUT) :: state
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Statement index.
    INTEGER :: length         ! Sequence length.
    INTEGER :: last_type      ! Last type of token written.
    INTEGER :: prev_type      ! Next type of token.
    INTEGER :: start          ! Starting point in token text.
    
    ! rnode text.
    CHARACTER(:), ALLOCATABLE :: text
    
    !***************************************************************************
    
    SELECT TYPE (rnode)
    TYPE IS (RenderNodeConstruct)
      IF (SIZE(rnode%items) /= 0) THEN
        DO i = 1, SIZE(rnode%items)
          CALL free_form_write_rnode(unit, rnode%items(i)%item, opts, state)
        END DO
      END IF
      
    TYPE IS (RenderNodeStatement)
      ! Certain statements pop a level off the indent stack.
      IF (ANY(rnode%context == opts%DecreaseIndentStatements)) THEN
        CALL decr_indent(state)
      END IF
      CALL free_form_write_indent(state, unit)
      CALL free_form_write_rnode(unit, rnode%RenderNodeSequence, opts, state)
      WRITE (unit, "()")
      state%prev_type = ittEndStatement
      state%pos = 0
      ! Certain statements add a level to the indent stack.
      IF (ANY(rnode%context == opts%IncreaseIndentStatements)) THEN
        CALL incr_indent(state, opts%ConstuctIndent, opts%MaximumIndent)
      END IF
      
    TYPE IS (RenderNodeSequence)
      last_type = state%prev_type
      CALL incr_indent(state, opts%TermContinueIndent, opts%MaximumIndent)
      DO i = 1, SIZE(rnode%items)
        prev_type = last_type
        CALL query_length(  &
            rnode%items(i)%item,  &
            opts,  &
            prev_type,  &
            last_type,  &
            length )
        IF ( ( state%pos + length  &
              <= opts%MaximumLineLength - LEN(opts%continue) )  &
            .OR. ( (i == SIZE(rnode%items))  &
              .AND. (state%pos + length <= opts%MaximumLineLength) ) ) THEN
          ! This item fits - write it.
          CALL query_text(  &
              rnode%items(i)%item,  &
              opts,  &
              prev_type,  &
              last_type,  &
              text )
          WRITE (unit, "(A)", ADVANCE='NO') text
          state%pos = state%pos + length
        ELSE
          ! This item does not fit.  If it is the first item, split it, 
          ! otherwise write it on the next line.
          IF (i == 1) THEN
            CALL free_form_write_rnode(unit, rnode%items(i)%item, opts, state)
          ELSE
            WRITE (unit, "(A)") opts%continue
            state%prev_type = ittEndStatement
            state%pos = 0
            CALL free_form_write_indent(state, unit)
            CALL free_form_write_rnode(unit, rnode%items(i)%item, opts, state)
          END IF
        END IF
      END DO
      CALL decr_indent(state)
      
    TYPE IS (RenderNodeToken)
      CALL incr_indent(state, opts%TermContinueIndent, opts%MaximumIndent)
      CALL query_text(rnode, opts, state%prev_type, last_type, text)
      length = LEN(text)
      IF ( state%pos + length  &
          <= opts%MaximumLineLength - LEN(opts%continue) ) THEN
        ! This token fits - write it.
        WRITE (unit, "(A)", ADVANCE='NO') text
        state%pos = state%pos + length
      ELSE
        ! We need to split this token.
        WRITE (unit, "(A,'&')")  &
            text(:opts%MaximumLineLength-LEN(opts%continue)-state%pos+1)
        start = opts%MaximumLineLength - LEN(opts%continue) - state%pos + 1
        state%pos = 0
        CALL free_form_write_indent(state, unit)
        DO WHILE ( state%pos + LEN(text) - start  &
            > opts%MaximumLineLength - LEN(opts%continue) )
          ! Multiple splits!
          WRITE (unit, "('&',A,'&')")  &
              text( start+1  &
                  : start+opts%MaximumLineLength-LEN(opts%continue)-state%pos )
          start = start + opts%MaximumLineLength - LEN(opts%continue) - state%pos
          state%pos = 0
          CALL free_form_write_indent(state, unit)
        END DO
        WRITE (unit, "('&',A)", ADVANCE='NO') text(start+1:)
        state%pos = state%pos + LEN(text) - start
      END IF
      state%prev_type = rnode%type
      CALL decr_indent(state)
      
    TYPE IS (RenderNodeComment)
      ! We should probably be looking at the last token - see if 
      ! it is end of statement.
      IF (state%pos /= 0) THEN
        IF (ALLOCATED(rnode%prefix)) THEN
          WRITE (unit, "(' & ',A,A)") rnode%prefix, rnode%content
        ELSE
          WRITE (unit, "(' & ! ',A)") rnode%content
        END IF
      ELSE
        CALL free_form_write_indent(state, unit)
        IF (ALLOCATED(rnode%prefix)) THEN
          WRITE (unit, "(A,A)") rnode%prefix, rnode%content
        ELSE
          WRITE (unit, "('! ',A)") rnode%content
        END IF
      END IF
      state%pos = 0
      
    END SELECT
    
  END SUBROUTINE free_form_write_rnode
  
  
  !*****************************************************************************
  !!
  !> Get the length of an rnode.
  
  RECURSIVE SUBROUTINE query_length(rnode, opts, prev_type, last_type, length)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNode), INTENT(IN) :: rnode
    TYPE(FormatFreeFormOptions), INTENT(IN) :: opts
    INTEGER, INTENT(IN) :: prev_type
    INTEGER, INTENT(OUT) :: last_type
    INTEGER, INTENT(OUT) :: length
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Sequence index.
    INTEGER :: item_length    ! Length of child item.
    INTEGER :: local_type     ! prev_type from previous item in sequence.
    
    !***************************************************************************
    
    SELECT TYPE (rnode)
    TYPE IS (RenderNodeToken)
      IF (ANY(rnode%type == separate) .AND. ANY(prev_type == separate)) THEN
        length = 1
      ELSE
        length = 0
      END IF
      length = length + LEN(query_token_text(rnode, opts))
      last_type = rnode%type
      
    TYPE IS (RenderNodeConstruct)
      ! Technically this won't work if the construct represents a program 
      ! unit as a program unit must occupy unique source lines.
      length = 0
      last_type = prev_type
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_length(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_length )
        length = length + item_length
      END DO
      
    TYPE IS (RenderNodeStatement)
      IF (prev_type /= ittEndStatement) THEN
        length = 1
      ELSE
        length = 0
      END IF
      last_type = ittEndStatement
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_length(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_length )
        length = length + item_length
      END DO
      
    CLASS IS (RenderNodeSequence)
      length = 0
      last_type = prev_type
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_length(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_length)
        length = length + item_length
      END DO
      
    CLASS DEFAULT
      STOP 'Internal error in RenderTest%%query_length, &
          &unknown or invalid type of rnode.'
    END SELECT
    
  END SUBROUTINE query_length
  
  
  !*****************************************************************************
  !!
  !> Get the text of an rnode.
  
  RECURSIVE SUBROUTINE query_text(rnode, opts, prev_type, last_type, text)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNode), INTENT(IN) :: rnode
    TYPE(FormatFreeFormOptions), INTENT(IN) :: opts
    INTEGER, INTENT(IN) :: prev_type
    INTEGER, INTENT(OUT) :: last_type
    CHARACTER(:), INTENT(OUT), ALLOCATABLE :: text
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Sequence index.
    INTEGER :: local_type     ! prev_type from previous item in sequence.
    
    ! Text of child item
    CHARACTER(:), ALLOCATABLE :: item_text
    
    !***************************************************************************
    
    SELECT TYPE (rnode)
    TYPE IS (RenderNodeToken)
      IF (ANY(rnode%type == separate) .AND. ANY(prev_type == separate)) THEN
        text = ' '
      ELSE
        text = ''
      END IF
      text = text // query_token_text(rnode, opts)
      last_type = rnode%type
      
    TYPE IS (RenderNodeConstruct)
      ! Technically this won't work if the construct represents a program 
      ! unit as a program unit must occupy unique source lines.
      text = ''
      last_type = prev_type
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_text(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_text)
        text = text // item_text
      END DO
      
    TYPE IS (RenderNodeStatement)
      IF (prev_type /= ittEndStatement) THEN
        text = ';'
      ELSE
        text = ''
      END IF
      last_type = ittEndStatement
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_text(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_text )
        text = text // item_text
      END DO
      
    CLASS IS (RenderNodeSequence)
      text = ''
      last_type = prev_type
      DO i = 1, SIZE(rnode%items)
        local_type = last_type
        CALL query_text(  &
            rnode%items(i)%item,  &
            opts,  &
            local_type,  &
            last_type,  &
            item_text )
        text = text // item_text
      END DO
      
    CLASS DEFAULT
      STOP 'Internal error in RenderTest%%query_text, &
          &unknown type of rnode.'
    END SELECT
    
  END SUBROUTINE query_text
  
  
  !*****************************************************************************
  !!
  !> Add the current indent to the start of the buffer.
  
  SUBROUTINE free_form_write_indent(state, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
  
    TYPE(free_form_state), INTENT(INOUT) :: state
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    state%pos = state%pos + current_indent(state)
    WRITE (unit, "(A)", ADVANCE='NO') REPEAT(' ', current_indent(state))
    
  END SUBROUTINE free_form_write_indent
  
  
  !*****************************************************************************
  !!
  !> Get the text of the given token.
  
  FUNCTION query_token_text(tok, opts) RESULT(txt)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(RenderNodeToken), INTENT(IN) :: tok
    TYPE(FormatFreeFormOptions), INTENT(IN) :: opts
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: txt
    
    !***************************************************************************
    
    SELECT CASE (tok%type)
    CASE (ittName, ittKeyword, ittStatementLabel,  &
        ittDelimiter, ittComma, ittAssign, ittAssociate, ittColon,  &
        ittDeclare, ittPercent, ittFormatSpecification, ittOperator )
      txt = tok%value
      
    CASE (ittIntegerLiteral, ittRealLiteral, ittLogicalLiteral)
      txt = tok%value
      IF (tok%kind /= '') THEN
        txt = txt // '_' // tok%kind
      END IF
      
    CASE (ittBOZLiteral)
      txt = tok%kind // opts%PreferredQuote // tok%value // opts%PreferredQuote
      
    CASE (ittCharacterLiteral)
      IF (tok%kind /= '') THEN
        txt = tok%kind // '_'
      ELSE
        txt = ''
      END IF
      txt = txt // query_character_literal_value_text(  &
          tok%value,  &
          opts%PreferredQuote )
    
    CASE DEFAULT
      STOP 'Unknown token type.'
      
    END SELECT
    
  END FUNCTION query_token_text
  
  
  !*****************************************************************************
  !!
  
  FUNCTION query_character_literal_value_text(value, quote_char) RESULT(txt)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*,ock), INTENT(IN) :: value
    CHARACTER, INTENT(IN) :: quote_char
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: txt
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Character position.
    
    !***************************************************************************
    
    txt = quote_char
    
    DO i = 1, LEN(value)
      IF (value(i:i) == quote_char) THEN
        txt = txt // quote_char
      END IF
      txt = txt // value(i:i)
    END DO
    
    txt = txt // quote_char
    
  END FUNCTION query_character_literal_value_text
  
  
  !*****************************************************************************
  !!
  !> Query the current ident position.
  !!
  !! @param[in]     state             The write state object.
  
  FUNCTION current_indent(state)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(free_form_state), INTENT(IN) :: state
    
    ! Function result.
    INTEGER :: current_indent
    
    !***************************************************************************
    
    IF (ALLOCATED(state%indent_stack)) THEN
      IF (SIZE(state%indent_stack) > 0) THEN
        current_indent = state%indent_stack(SIZE(state%indent_stack))
        RETURN
      END IF
    END IF
    
    current_indent = 0
    
  END FUNCTION current_indent
  
  
  !*****************************************************************************
  !!
  !> Decrement (pop) the indent position.
  !!
  !! @param[in,out] state             The write state object.
  
  SUBROUTINE decr_indent(state)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(free_form_state), INTENT(INOUT) :: state
    
    !***************************************************************************
    
    IF (ALLOCATED(state%indent_stack)) THEN
      IF (SIZE(state%indent_stack) > 0) THEN
        state%indent_stack = state%indent_stack(:SIZE(state%indent_stack)-1)
      END IF
    END IF
    
  END SUBROUTINE decr_indent
  
  
  !*****************************************************************************
  !!
  !> Increment the indent position.
  !!
  !! @param[in,out] state             The write state object.
  !!
  !! @param[in]     amt               The amount to increase the indent by.
  
  SUBROUTINE incr_indent(state, amt, max)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(free_form_state), INTENT(INOUT) :: state
    INTEGER, INTENT(IN) :: amt
    INTEGER, INTENT(IN) :: max
    
    !***************************************************************************
    
    IF (.NOT. ALLOCATED(state%indent_stack)) THEN
      state%indent_stack = [amt]
    ELSE
      state%indent_stack = [  &
          (state%indent_stack),  &
          MIN(max, current_indent(state) + amt) ]
    END IF
    
  END SUBROUTINE incr_indent
  
END MODULE Formatters
