! $Id: IdentifierNatures.f90 2949 2019-10-04 12:52:05Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IdentifierNatures module.


!*******************************************************************************
!!
!> Constants and procedures for working with identifer natures.
!!
!! As the semantics of a scope are progressively analysed, a picture is 
!! progressively constructed as to what each identifier, other than use 
!! associated identifiers, means.
!!
!! By tracking the nature of identifier references and definitions in a 
!! scope, we can issue relatively informative error messages when some 
!! sort of mis-use of an identifier occurs.
!!
!! There have been several false starts to encapsulating identifier natures.  
!! There might be several more.
!!
!! A plausible alternative that might be more maintainable would be to have a 
!! mask that defines what the identifier cannot be in later references or 
!! definitions, and a mask that defines what the identifier must be.  Backing 
!! a suitable error message out in that situation might be a bit tricky 
!! though.
!!
!! The concept of "identifier" here is closer to that in the standard than 
!! the concept in the `Identifiers` module.  Here an identifier includes 
!! something like WRITE(FORMATTED) - an identifier for a defined input/output 
!! procedure.

MODULE IdentifierNatures
  
  USE IdentifierSymbols
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: Add
  PUBLIC :: MergeNatures
  PUBLIC :: GetIdentifierName
  PUBLIC :: GetOperatorSymbolName
  PUBLIC :: GetNatureDescription
  PUBLIC :: GetNatureText
  PUBLIC :: FindIdentifier
  PUBLIC :: IsReferenceNature
  PUBLIC :: IsDefinitionNature
  
  PUBLIC :: OPERATOR(.NameOf.)
  PUBLIC :: OPERATOR(.KeywordOf.)
  PUBLIC :: OPERATOR(.IsOperator.)
  PUBLIC :: OPERATOR(.IsName.)
  
  !-----------------------------------------------------------------------------
  ! Natures for identifier references.
  !
  ! Natures that are references are merged.
  
  !> An identifier is being referenced as a data object.
  !!
  !! Perhaps the identifier has appeared as the base object in a multipart 
  !! designator.
  INTEGER, PARAMETER, PUBLIC :: NatureRefDataObject = 1
  
  !> An identifier is being referenced in a function like reference.
  !!
  !! Syntax along the lines of name(expr-list), where expr-list contains 
  !! nothing that looks like an array section or substring-range.
  !!
  !! The identifier may be a generic interface for a function, a specific 
  !! function, a type name (being used as a structure constructor) or 
  !! an array data object.
  INTEGER, PARAMETER, PUBLIC :: NatureRefFunctionLike = 2
  
  !> An identifier is being referenced as a subroutine.
  !!
  !! Syntax such as CALL name(...)
  !!
  !! Perhaps the identifier is a specific subroutine, or a generic 
  !! subroutine, or a pointer to a subroutine.
  INTEGER, PARAMETER, PUBLIC :: NatureRefSubroutine = 3
  
  !> An identifier is being referenced as a type.
  !!
  !! Perhaps the identifier has appeared as the type name in a type 
  !! specifier, or as the type name in an EXTENDS specifier.
  INTEGER, PARAMETER, PUBLIC :: NatureRefType = 4
  
  !> An identifier is being referenced as a procedure interface.
  !!
  !! The identifier has appeared as the procedure interface in a 
  !! procedure declaration statement.  It must be a specific procedure 
  !! or a type.
  INTEGER, PARAMETER, PUBLIC :: NatureRefProcInterface = 5
  
  !> An identifier is being referenced as a namelist.
  !!
  !! The identifier has appeared in the NML specified of an IO statement.
  INTEGER, PARAMETER, PUBLIC :: NatureRefNamelist = 6
  
  !> An identifier is being referenced as a module.
  !!
  !! The identifier has appeared as the module name in a USE statement or 
  !! a submodule statement.
  INTEGER, PARAMETER, PUBLIC :: NatureRefModule = 7
  
  !> An identifier has been defined as a construct name.
  !!
  !! The identifier has appeared as a construct name in other than the 
  !! opening statement of a construct.
  INTEGER, PARAMETER, PUBLIC :: NatureRefConstruct = 8
  
  !> An identifier is being referenced as a submodule.
  !!
  !! The identifier is the pair of the module name and the submodule name.
  INTEGER, PARAMETER, PUBLIC :: NatureRefSubmodule = 9
  
  !> An identifier is being referenced as an operator.
  !!
  !! The identifier is the operator symbol.
  INTEGER, PARAMETER, PUBLIC :: NatureRefOperator = 10
  
  !> An identifier is being referenced as a procedure.
  !!
  !! The identifier was listed in a /procedure-stmt/ (in an interface block) 
  !! or in a /type-bound-procedure-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureRefProcedure = 11
  
  !> An identifier was referenced in an object like fashion.
  !!
  !! Whether the identifier is a data object or a procedure or whatever 
  !! depends.
  INTEGER, PARAMETER, PUBLIC :: NatureRefObject = 12
  
  !> An identifier was referenced in an object like fashion.
  !!
  !! Whether the identifier is a data object or a procedure or whatever 
  !! depends.
  INTEGER, PARAMETER, PUBLIC :: NatureRefPointer = 13
  
  !> An identifier was referenced as a type parameter.
  !!
  !! The identifier appeared in a /type-param-def-stmt/ or in a 
  !! /type-param-spec/.
  !!
  !! The appearance of a type parameter in a /derived-type-stmt/ is considered 
  !! the definition.
  INTEGER, PARAMETER, PUBLIC :: NatureRefTypeParameter = 14
  
  !> An identifier was referenced as a binding.
  !!
  !! The identifier appeared in a generic binding declaration.
  INTEGER, PARAMETER, PUBLIC :: NatureRefSpecBinding = 15
  
  !> An identifier was referenced as a separate module procedure.
  !!
  !! The identifier appeared in a function or subroutine or separate module 
  !! subprogram with the MODULE prefix.
  INTEGER, PARAMETER, PUBLIC :: NatureRefSeparate = 16
  
  !> An identifier was referenced as an equivalenced object.
  !!
  !! The identifier appeared as the thing nominated by an /equivalence-object/.
  INTEGER, PARAMETER, PUBLIC :: NatureRefEquivalenceObject = 17
  
  !> An identifier was declared as the local name in a use statement.
  !!
  !! If the identifier was renamed in the use statment, then this is the 
  !! local name, otherwise it is just the name.
  INTEGER, PARAMETER, PUBLIC :: NatureRefUseName = 18
  
  !> An identifier was referenced in an /import-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureRefImportName = 19
  
  !> The identifier was referenced in a substring-range like manner.
  !!
  !! After the identifier there was a single subscript, with a single colon.  
  !! The identifier had no following part references.
  !!
  !! The identifier may be a character object, or a rank one array.
  INTEGER, PARAMETER, PUBLIC :: NatureRefSubstringLike = 20
  
  !> The identifier was referenced as an array.
  !!
  !! After the identifier there was either a single subscript with 
  !! two colons (so the triplet included the step), or there were 
  !! multiple subscripts, one of which had at least one colon.
  !!
  !! The identifier has to be an array object.
  INTEGER, PARAMETER, PUBLIC :: NatureRefArray = 21
  
  !> The identifier was referenced as the name of a common block.
  !!
  !! The identifier appeared as an entity in a bind statement or 
  !! a save statment, with surrounding `/` delimiters.
  INTEGER, PARAMETER, PUBLIC :: NatureRefCommon = 22
  
  !> The highest valued reference nature.
  INTEGER, PARAMETER, PUBLIC :: NatureRefLast = 22
  
  !-----------------------------------------------------------------------------
  ! Natures for identifier definitions.
  !
  ! Natures that are definitions are never merged.
  
  !> The first nature that is a definition.
  INTEGER, PARAMETER, PUBLIC :: NatureDefFirst = 101
  
  !> An identifier has been defined as a type.
  !!
  !! The identifier has appeared as the name of the type being defined in a 
  !! type definition construct.
  INTEGER, PARAMETER, PUBLIC :: NatureDefType = 101
  
  !> An identifier has been given a type.
  !!
  !! The identifier appeared as the entity in in an /entity-decl/ in 
  !! a type declaration statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefHasType = 102
  
  !> An identifier has been defined as a generic interface.
  !!
  !! The identifier has appeared in a INTERFACE statement or type bound 
  !! generic statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefGeneric = 103
  
  !> An identifier has been defined as a main program.
  !!
  !! The identifier has appeared in a PROGRAM statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefMainProgram = 104
  
  !> An identifier has been defined as a block data unit.
  !!
  !! The identifier has appeared in a BLOCK DATA statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefBlockData = 105
  
  !> An identifier has been defined as a module.
  !!
  !! The identifier has appeared in a MODULE statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefModule = 106
  
  !> An identifier has been defined as a submodule.
  !!
  !! The identifier has appeared, collectively, in a SUBMODULE statement.  
  !! Note that the identifier is the submodule identifier, which is not 
  !! just the submodule name.
  INTEGER, PARAMETER, PUBLIC :: NatureDefSubmodule = 107
  
  !> An identifier has been defined as a function.
  !!
  !! The identifier has appeared in a function statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefFunction = 108
  
  !> An identifier has been defined as a subroutine.
  !!
  !! The identifier has appeared in a subroutine statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefSubroutine = 109
  
  !> An identifier has been defined as a namelist.
  !!
  !! The identifier has appeared in a NAMELIST statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefNamelist = 110
  
  !> An identifier has been defined as a statement function.
  !!
  !! The identifier has appeared in a /stmt-function-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefStmtFunction = 111
  
  !> An identifier has been defined as a construct name.
  !!
  !! The identifier has appeared as a construct name in the opening 
  !! statement of a construct.
  INTEGER, PARAMETER, PUBLIC :: NatureDefConstruct = 112
  
  !> An identifier has been defined as a dummy argument.
  INTEGER, PARAMETER, PUBLIC :: NatureDefDummy = 113
  
  !> An identifier has been given an accessibility attribute of PUBLIC.
  INTEGER, PARAMETER, PUBLIC :: NatureDefPublic = 121
  
  !> An identifier has been given an accessibility attribute of PRIVATE.
  INTEGER, PARAMETER, PUBLIC :: NatureDefPrivate = 122
  
  !> An identifier has been defined as a function result.
  !!
  !! The identifier was the name in the RESULT specified in a function 
  !! statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefResult = 123
  
  !> An identifier has been defined as a procedure entry.
  !!
  !! The identifier was the entry name in an entry statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefEntry = 124
  
  !> An identifier has been given the allocatable attribute.
  !!
  !! 
  INTEGER, PARAMETER, PUBLIC :: NatureDefAllocatable = 125
  
  !> An identifier has been defined as having the asychronous attribute.
  !!
  !! The identifier was a name in an ASYNCHRONOUS statement or type 
  !! declaration statement with the ASYNCHRONOUS keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefAsynchronous = 126
  
  !> An identifier has been defined as having the language binding attribute.
  !!
  !! The identifier was a name in an BIND statement or type 
  !! declaration statement with the BIND keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefBind = 127
  
  !> An identifier has been given the codimension attribute.
  !!
  !! The identifier was a name in a CODIMENSION statement, or type 
  !! declaration statement with the CODIMENSION keyword or 
  !! with a codimension specification.
  INTEGER, PARAMETER, PUBLIC :: NatureDefCodimension = 128
  
  !> An identifier has been defined as having the contiguous attribute.
  !!
  !! The identifier was a name in a CONTIGUOUS statement or type 
  !! declaration statement with the CONTIGUOUS keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefContiguous = 129
  
  !> An identifier has been given the dimension attribute.
  !!
  !! The identifier was a name in a DIMENSION statement, or type 
  !! declaration statement with the DIMENSION keyword or 
  !! with a dimension specification.
  INTEGER, PARAMETER, PUBLIC :: NatureDefDimension = 130
  
  !> An identifier has been defined as a procedure.
  !!
  !! The identifier has appeared in an EXTERNAL or PROCEDURE declaration 
  !! statement, or in a type declaration statement with the EXTERNAL 
  !! keyword.
  !!
  !! It may be that an interface in the procedure statement identifies 
  !! the interface (or result type) for the identifier as a procedure, and 
  !! hence it is known as a function or a subroutine.  However we don't 
  !! deal with that level of detail here - that requires the procedure 
  !! identifier to be resolved.
  INTEGER, PARAMETER, PUBLIC :: NatureDefExternal = 131
  
  !> An identifier has been defined as having the intrinsic attribute.
  !!
  !! The identifier was a name in an INTRINSIC statement or type 
  !! declaration statement with the INTRINSIC keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefIntrinsic = 132
  
  !> An identifier has been defined as having the optional attribute.
  !!
  !! The identifier was a name in an OPTIONAL statement or type 
  !! declaration statement with the OPTIONAL keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefOptional = 133
  
  !> An identifier has been given the parameter attribute.
  !!
  !! The identifier was a name in a PARAMETER statement or a type 
  !! declaration statement with the PARAMETER keyword, or an 
  !! enumerator statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefParameter = 134
  
  !> An identifier has been defined as having the pointer attribute.
  !!
  !! The identifier was a name in a POINTER statement or type 
  !! declaration statement with the POINTER keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefPointer = 135
  
  !> An identifier has been defined as having the protected attribute.
  !!
  !! The identifier was a name in a PROTECTED statement or type 
  !! declaration statement with the PROTECTED keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefProtected = 136
  
  !> An identifier has been given the save attribute.
  INTEGER, PARAMETER, PUBLIC :: NatureDefSave = 137
  
  !> An identifier has been given the target attribute.
  INTEGER, PARAMETER, PUBLIC :: NatureDefTarget = 138
  
  !> An identifier has been defined as having the value attribute.
  !!
  !! The identifier was a name in a VALUE statement or type 
  !! declaration statement with the VALUE keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefValue = 139 
  
  !> An identifier has been defined as having the volatile attribute.
  !!
  !! The identifier was a name in a VOLATILE statement or type 
  !! declaration statement with the VOLATILE keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefVolatile = 140
  
  !> An identifier has been defined as having the INTENT(IN) attribute.
  !!
  !! The identifier was a name in an INTENT(IN) statement or type 
  !! declaration statement with the INTENT(IN) keywords.
  INTEGER, PARAMETER, PUBLIC :: NatureDefIntentIn = 141
  
  !> An identifier has been defined as having the INTENT(OUT) attribute.
  !!
  !! The identifier was a name in an INTENT(OUT) statement or type 
  !! declaration statement with the INTENT(OUT) keywords.
  INTEGER, PARAMETER, PUBLIC :: NatureDefIntentOut = 142
  
  !> An identifier has been defined as having the INTENT(INOUT) attribute.
  !!
  !! The identifier was a name in an INTENT(INOUT) statement or type 
  !! declaration statement with the INTENT(INOUT) keywords.
  INTEGER, PARAMETER, PUBLIC :: NatureDefIntentInOut = 143
  
  !> An identifier has been defined as an implied do control variable.
  !!
  !! The identifier appeared in the syntax (..., ident=start, stop) in 
  !! an io list item, data value or array constructor value.
  INTEGER, PARAMETER, PUBLIC :: NatureDefImpliedDoVar = 144
  
  !> An identifier has been defined as a type parameter.
  !!
  !! The identifier appeared in a /derived-type-stmt/ as a parameter of the 
  !! type.
  !!
  !! The appearance of the parameter in a /type-param-def-stmt/ is considered 
  !! as a reference to the type parameter.
  INTEGER, PARAMETER, PUBLIC :: NatureDefTypeParameter = 145
  
  !> An identifier has been defined as specific binding of a type.
  !!
  !! The identifier appeared in a /type-bound-procedure-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefSpecBinding = 146
  
  !> An identifier has been defined as an implied do control variable.
  !!
  !! The identifier appeared in the syntax (..., ident=start, stop) in 
  !! an io list item, data value or array constructor value.
  INTEGER, PARAMETER, PUBLIC :: NatureDefGenericBinding = 147
  
  !> An identifier has been defined as an associate name.
  !!
  !! The identifier appeared as the associate name in a SELECT TYPE or 
  !! ASSOCIATE construct.
  INTEGER, PARAMETER, PUBLIC :: NatureDefAssociateName = 148
  
  !> An identifier has been defined as a separate module procedure.
  !!
  !! The identifier appeared in an interface body with the MODULE keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefSeparate = 149
  
  !> An identifier has been defined as being polymorphic.
  !!
  !! The identifier appeared in a type declaration statement with a CLASS 
  !! declaration type spec.
  INTEGER, PARAMETER, PUBLIC :: NatureDefPolymorphic = 150
  
  !> An identifier has been defined as a common block object.
  !!
  !! The identifier appeared as the thing nominated by a /common-block-object/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefCommonObject = 151
  
  !> An identifier has been initialized in data.
  !!
  !! The identifier appeared as the thing nominated by a /data-stmt-object/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefDataObject = 152
  
  !> An identifier has been declared as a kind type parameter.
  !!
  !! The identifier appeared in a /type-param-def-stmt/ with the KIND 
  !! keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefKindParam = 153
  
  !> An identifier has been declared as a kind type parameter.
  !!
  !! The identifier appeared in a /type-param-def-stmt/ with the LEN 
  !! keyword.
  INTEGER, PARAMETER, PUBLIC :: NatureDefLenParam = 154
  
  !> An identifier was declared with an initializer.
  !!
  !! There was an initializer in the /entity-decl/ for the identifier 
  !! in a /type-declaration-stmt/ or /component-def-stmt/ or 
  !! /type-param-def-stmt/.
  !!
  !! This is a separate nature to the identifier appearing in a data 
  !! statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefInitializer = 155
  
  !> An identifier was declared as a data component.
  !!
  !! The identifier appeared in a /data-component-def-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefDataComponent = 156
  
  !> An identifier was declared as a procedure component.
  !!
  !! The identifier appeared in a /proc-component-def-stmt/.
  INTEGER, PARAMETER, PUBLIC :: NatureDefProcComponent = 157
  
  !> An identifier was declared as a common block.
  !!
  !! The identifier appeared in a common statement.
  INTEGER, PARAMETER, PUBLIC :: NatureDefCommon = 158
  
  !> An identifier was declared as a module procedure.
  INTEGER, PARAMETER, PUBLIC :: NatureDefModuleProcedure = 159
  
  !> An identifier was declared as an internal procedure.
  INTEGER, PARAMETER, PUBLIC :: NatureDefInternalProcedure = 160
  
  !> An identifier was used as the procedure name in the enclosing scope.
  !!
  !! This is used for the procedure name inside the subprogram that defines 
  !! the procedure itself.  Because it is a definition it prevent host and 
  !! use association within the scope of the procedure from hiding the 
  !! name, but because it is a separate concept from NatureDefFunction and 
  !! NatureDefSubroutine, we can avoid having the local definition hide 
  !! the "host associated" definition.
  INTEGER, PARAMETER, PUBLIC :: NatureDefProcedureDef = 161
  
  !> The last nature that is a definition.
  INTEGER, PARAMETER, PUBLIC :: NatureDefLast = 161
  
  !-----------------------------------------------------------------------------
  
  !> Provides details of an identifier defined or referenced in a scope and 
  !! the nature of that reference or definition.
  TYPE, PUBLIC :: IdentifierNature
    !> The identifier of the identifier.
    !!
    !! This is called its name, but the identifier for an identifier 
    !! is not a name.
    TYPE(IdentifierSymbol) :: symbol
    
    ! Should the following two components be placed in their own type?  If 
    ! we do that, we could handle existing source by overloading the 
    ! structure constructor.  @todo Consider.
    
    !> The natures of the reference or definition of the identifier.
    INTEGER, ALLOCATABLE :: natures(:)
    
    !> The location in source that best describes where the identifier 
    !! acquired each nature.  Should be the same size as the natures 
    !! component.
    TYPE(SourceLocation), ALLOCATABLE :: locations(:)
  END TYPE
  
  !> Construct an IdentifierNature object.
  INTERFACE IdentifierNature
    MODULE PROCEDURE IdentifierNature_name
    MODULE PROCEDURE IdentifierNature_keyword
  END INTERFACE IdentifierNature
  
  !> Extract the text describing the identifier.
  INTERFACE OPERATOR(.NameOf.)
    MODULE PROCEDURE op_nameof
  END INTERFACE OPERATOR(.NameOf.)
  
  !> Extract the keyword of the identifier.
  INTERFACE OPERATOR(.KeywordOf.)
    MODULE PROCEDURE op_keywordof
  END INTERFACE OPERATOR(.KeywordOf.)
  
  !> Test if the identifier is an operator.
  INTERFACE OPERATOR(.IsOperator.)
    MODULE PROCEDURE op_isoperator
  END INTERFACE OPERATOR(.IsOperator.)
  
  !> Test if the identifier is a name.
  INTERFACE OPERATOR(.IsName.)
    MODULE PROCEDURE op_isname
  END INTERFACE OPERATOR(.IsName.)
  
  !> Add an IdentifierNature object to an existing vector.
  INTERFACE Add
    MODULE PROCEDURE Add_
  END INTERFACE Add
  
  !> Merge a scalar or vector of IdentifierNature objcets with an existing 
  !! vector.
  INTERFACE MergeNatures
    MODULE PROCEDURE MergeNatures_array
    MODULE PROCEDURE MergeNatures_scalar
  END INTERFACE MergeNatures
  
  !> Type for mapping nature codes to a text description.
  TYPE :: nm
    !> The nature part of the map.
    INTEGER :: nature
    !> Preamble appropriate for the description.
    CHARACTER(14) :: preamble
    !> The text description that the nature maps to.
    CHARACTER(40) :: text
  END TYPE nm
  
  !> Texts for nature description preambles.
  !                                  12345678901234
  CHARACTER(*), PARAMETER :: decl = 'declared as a '
  CHARACTER(*), PARAMETER :: used = 'used in a     '
  CHARACTER(*), PARAMETER :: attr = 'given the     '
  
  !> The map of nature codes to text descriptions.
  !!
  !! We use this for generating debugging output, and perhaps in the 
  !! future for error messages.  The nature of the nature descriptions 
  !! is not particularly consistent and perhaps additional text 
  !! is required.  @todo Review.
  TYPE(nm), PARAMETER :: map(75) = [   &
      !          1234567890123456789012345678901234567890
      nm( NatureRefDataObject,  &
          used, 'data object reference                   ' ),  &
      nm( NatureRefFunctionLike,  &
          used, 'function-like reference                 ' ),  &
      nm( NatureRefSubroutine,  &
          used, 'subroutine reference                    ' ),  &
      nm( NatureRefType,  &
          used, 'type reference                          ' ),  &
      nm( NatureRefProcInterface,  &
          used, 'procedure interface reference           ' ),  &
      nm( NatureRefNamelist,  &
          used, 'namelist reference                      ' ),  &
      nm( NatureRefModule,  &
          used, 'module reference                        ' ),  &
      nm( NatureRefConstruct,  &
          used, 'construct name reference                ' ),  &
      nm( NatureRefOperator,  &
          used, 'operator reference                      ' ),  &
      nm( NatureRefSubmodule,  &
          used, 'submodule reference                     ' ),  &
      nm( NatureRefProcedure,  &
          used, 'procedure reference                     ' ),  &
      nm( NatureRefSpecBinding,  &
          used, 'specific binding reference              ' ),  &
      nm( NatureRefSeparate,  &
          used, 'separate module procedure reference     ' ),  &
      nm( NatureRefTypeParameter,  &
          used, 'type parameter reference                ' ),  &
      nm( NatureRefEquivalenceObject,  &
          used, 'equivalence object                      ' ),  &
      nm( NatureRefObject,  &
          used, 'object reference                        ' ),  &
      nm( NatureRefUseName,  &
          decl, 'use associated thing                    ' ),  &
      nm( NatureRefImportName,  &
          decl, 'imported thing                          ' ),  &
      nm( NatureRefSubstringLike,  &
          used, 'substring-like reference                ' ),  &
      nm( NatureRefArray,  &
          used, 'array reference                         ' ),  &
      nm( NatureRefCommon,  &
          used, 'common block                            ' ),  &
      nm( NatureDefType,  &
          decl, 'derived type                            ' ),  &
      nm( NatureDefHasType,  &
          decl, 'entity with type                        ' ),  &
      nm( NatureDefGeneric,  &
          decl, 'generic declaration                     ' ),  &
      nm( NatureDefMainProgram,  &
          decl, 'main program                            ' ),  &
      nm( NatureDefBlockData,  &
          decl, 'block data                              ' ),  &
      nm( NatureDefModule,  &
          decl, 'module                                  ' ),  &
      nm( NatureDefSubmodule,  &
          decl, 'submodule                               ' ),  &
      nm( NatureDefFunction,  &
          decl, 'function                                ' ),  &
      nm( NatureDefSubroutine,  &
          decl, 'subroutine                              ' ),  &
      nm( NatureDefNamelist,  &
          decl, 'namelist                                ' ),  &
      nm( NatureDefStmtFunction,  &
          decl, 'statement function                      ' ),  &
      nm( NatureDefConstruct,  &
          used, 'construct name                          ' ),  &
      nm( NatureDefDummy,  &
          decl, 'dummy argument                          ' ),  &
      nm( NatureDefResult,  &
          decl, 'function result                         ' ),  &
      nm( NatureDefEntry,  &
          decl, 'procedure entry                         ' ),  &
      nm( NatureDefPublic,  &
          attr, 'accessibility attribute of PUBLIC       ' ),  &
      nm( NatureDefPrivate,  &
          attr, 'accessibility attribute of PRIVATE      ' ),  &
      nm( NatureDefAllocatable,  &
          attr, 'ALLOCATABLE attribute                   ' ),  &
      nm( NatureDefAsynchronous,  &
          attr, 'ASYNCHRONOUS attribute                  ' ),  &
      nm( NatureDefBind,  &
          attr, 'BIND attribute                          ' ),  &
      nm( NatureDefCodimension,  &
          attr, 'CODIMENSION attribute                   ' ),  &
      nm( NatureDefContiguous,  &
          attr, 'CONTIGUOUS attribute                    ' ),  &
      nm( NatureDefDimension,  &
          attr, 'DIMENSION attribute                     ' ),  &
      nm( NatureDefExternal,  &
          attr, 'EXTERNAL attribute                      ' ),  &
      nm( NatureDefIntrinsic,  &
          attr, 'INTRINSIC attribute                     ' ),  &
      nm( NatureDefOptional,  &
          attr, 'OPTIONAL attribute                      ' ),  &
      nm( NatureDefParameter,  &
          decl, 'named constant                          ' ),  &
      nm( NatureDefPointer,  &
          attr, 'POINTER attribute                       ' ),  &
      nm( NatureDefProtected,  &
          attr, 'PROTECTED attribute                     ' ),  &
      nm( NatureDefSave,  &
          attr, 'SAVE attribute                          ' ),  &
      nm( NatureDefTarget,  &
          attr, 'TARGET attribute                        ' ),  &
      nm( NatureDefValue,  &
          attr, 'VALUE attribute                         ' ),  &
      nm( NatureDefVolatile,  &
          attr, 'VOLATILE attribute                      ' ),  &
      nm( NatureDefIntentIn,  &
          attr, 'INTENT(IN) attribute                    ' ),  &
      nm( NatureDefIntentOut,  &
          attr, 'INTENT(OUT) attribute                   ' ),  &
      nm( NatureDefIntentInOut,  &
          attr, 'INTENT(INOUT) attribute                 ' ),  &
      nm( NatureDefImpliedDoVar,  &
          decl, 'implied do variable                     ' ),  &
      nm( NatureDefTypeParameter,  &
          decl, 'type parameter                          ' ),  &
      nm( NatureDefSpecBinding,  &
          decl, 'specific binding                        ' ),  &
      nm( NatureDefGenericBinding,  &
          decl, 'generic binding                         ' ),  &
      nm( NatureDefAssociateName,  &
          decl, 'associate name                          ' ),  &
      nm( NatureDefSeparate,  &
          decl, 'separate module procedure               ' ),  &
      nm( NatureDefPolymorphic,  &
          decl, 'polymorphic                             ' ),  &
      nm( NatureDefCommonObject,  &
          decl, 'common block object                     ' ),  &
      nm( NatureDefDataObject,  &
          decl, 'data statement object                   ' ),  &
      nm( NatureDefKindParam, &
          decl, 'kind type parameter                     ' ),  &
      nm( NatureDefLenParam,  &
          decl, 'len type parameter                      ' ),  &
      nm( NatureDefInitializer,  &
          decl, 'entity with an initializer              ' ),  &
      nm( NatureDefDataComponent,  &
          decl, 'data component                          ' ),  &
      nm( NatureDefProcComponent,  &
          decl, 'procedure component                     ' ),  &
      nm( NatureDefCommon,  &
          decl, 'common block                            ' ),  &
      nm( NatureDefModuleProcedure,  &
          decl, 'module procedure                        ' ),  & 
      nm( NatureDefInternalProcedure,  &
          decl, 'internal procedure                      ' ),  &
      nm( NatureDefProcedureDef,  &
          decl, 'procedure being defined                 ' ) ]
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Construct an IdentifierNature object for an identifier that is just 
  !! a name.
  !!
  !! This is a compatibility shim for the previous form of structure 
  !! constructor for the IdentifierNature type.
  !!
  !! @param[in]     name              The name of the identifier.
  !!
  !! @param[in]     natures           The natures for the identifier.
  !!
  !! @param[in]     locations         The locations in source responsible 
  !! for each nature.  This must be the same size as @a natures.
  
  FUNCTION IdentifierNature_name(name, natures, locations) RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: name
    INTEGER, INTENT(IN) :: natures(:)
    TYPE(SourceLocation), INTENT(IN) :: locations(:)
    
    ! Function result.
    TYPE(IdentifierNature) :: r
    
    !***************************************************************************
    
    IF (is_letter_char(name(1:1))) THEN
      r%symbol = NameSymbol(name)
    ELSE IF (name(1:1) == '.') THEN
      r%symbol = OperatorSymbol(name)
    ELSE IF (name == '=') THEN
      r%symbol = AssignmentSymbol()
    ELSE
      ! Assume it is an operator.  Other forms of identifier shouldn't 
      ! be coming via this function anyway.
      r%symbol = OperatorSymbol(name)
    END IF
    r%natures = natures
    r%locations = locations
    
  END FUNCTION IdentifierNature_name
  
  
  !*****************************************************************************
  !!
  !> Test whether a character is a letter.
  !!
  !! @param[in]     ch                The character to test.
  !!
  !! @returns .TRUE. if @a char is a letter (3.1.2), .FALSE. otherwise.
  !!
  !! This duplicates CharacterTypes%%IsLetterChar, but is for default 
  !! character kind.
  
  ELEMENTAL FUNCTION is_letter_char(ch) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER, INTENT(IN) :: ch
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    ! Use ASCII collation sequence to define the ranges for big letters and 
    ! little letters.
    b = (LGE(ch, 'A') .AND. LLE(ch, 'Z'))  &
        .OR. (LGE(ch, 'a') .AND. LLE(ch, 'z'))
    
  END FUNCTION is_letter_char
  
  
  !*****************************************************************************
  !!
  !> Construct an IdentifierNature object for an identifier that is not just 
  !! a name.
  !!
  !! @param[in]     name              The name part of the identifier, 
  !! if the identifier has a name, otherwise not allocated.  This must be 
  !! allocated if keyword is 0 or 1.
  !!
  !! @param[in]     keyword           The form of identifier.  This is 
  !! the same as the keyword component of the GenericSpec type and 
  !! of the IdentifierIdentifier type.
  !> - 0: No keyword - the name is just the name.
  !! - 1: OPERATOR - the name will be the operator.
  !! - 2: ASSIGNMENT - the name will be '='.
  !! - 3: READ(FORMATTED) - name not used.
  !! - 4: WRITE(FORMATTED) - name not used.
  !! - 5: READ(UNFORMATTED) - name not used.
  !! - 6: WRITE(UNFORMATTED) - name not used.
  !!
  !! @param[in]     natures           The natures for the identifier.
  !!
  !! @param[in]     locations         The locations in source responsible 
  !! for each nature.  This must be the same size as @a natures.
  !!
  !! @returns A constructed IdentifierNature object.
  
  FUNCTION IdentifierNature_keyword(symbol, natures, locations)  &
      RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierSymbol), INTENT(IN) :: symbol
    INTEGER, INTENT(IN) :: natures(:)
    TYPE(SourceLocation), INTENT(IN) :: locations(:)
    
    ! Function result.
    TYPE(IdentifierNature) :: r
    
    !***************************************************************************
    
    r%symbol = symbol
    r%natures = natures
    r%locations = locations
    
  END FUNCTION IdentifierNature_keyword
  
  
  !*****************************************************************************
  !!
  !> Extract the name part of an IdentifierNature.
  !!
  !! @param[in]     ident             The identifier nature of interest.
  !!
  !! @returns The name or description of the identifier.
  
  PURE FUNCTION op_nameof(ident) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierNature), INTENT(IN) :: ident
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    str = .NameOf. ident%symbol
    
  END FUNCTION op_nameof
  
  
  !*****************************************************************************
  !!
  !> Extract the name part of an IdentifierNature.
  !!
  !! @param[in]     ident             The identifier nature of interest.
  !!
  !! @returns The keyword of the identifier.
  
  PURE FUNCTION op_keywordof(ident) RESULT(keyword)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierNature), INTENT(IN) :: ident
    
    ! Function result.
    INTEGER :: keyword
    
    !***************************************************************************
    
    keyword = ident%symbol%keyword
    
  END FUNCTION op_keywordof
  
  
  !*****************************************************************************
  !!
  !> Tests whether the identifier is for an operator.
  !!
  !! @param[in]     ident             The identifier nature of interest.
  !!
  !! @returns True if the identifier is an operator, false otherwise.
  
  PURE FUNCTION op_isoperator(ident) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierNature), INTENT(IN) :: ident
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ident%symbol%keyword == 1
    
  END FUNCTION op_isoperator
  
  
  !*****************************************************************************
  !!
  !> Tests whether the identifier is a name.
  !!
  !! @param[in]     ident             The identifier nature of interest.
  !!
  !! @returns True if the identifier is a name, false otherwise.
  
  PURE FUNCTION op_isname(ident) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierNature), INTENT(IN) :: ident
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = ident%symbol%keyword == 0
    
  END FUNCTION op_isname
  
  
  !*****************************************************************************
  !!
  !> Merge two sets of IdentiferNature objects together.
  !!
  !! @param[in,out] base              The base set.
  !!
  !! @param[in]     other             The set to merge in.
  !!
  !! Locations for existing reference natures in the @a base set are always 
  !! maintained, locations for matching reference natures in the @a other set 
  !! may be discarded if the base set already has a reference of that nature.
  !!
  !! Duplicate definition natures are retained.
  
  SUBROUTINE MergeNatures_array(base, other)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: base(:)
    TYPE(IdentifierNature), INTENT(IN) :: other(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: io             ! Index into other.
    
    !***************************************************************************
    
    other_loop: DO io = 1, SIZE(other)
      CALL MergeNatures(base, other(io))
    END DO other_loop
    
  END SUBROUTINE MergeNatures_array
  
  
  !*****************************************************************************
  !!
  !> Merge a single identifier with an existing set of IdentiferNature objects.
  !!
  !! @param[in,out] base              The base set.
  !!
  !! @param[out]    other             The identifier nature definition 
  !! to merge in.
  !!
  !! Locations for existing reference natures in the @a base set are always 
  !! maintained, locations for matching reference natures in the @a other set 
  !! may be discarded if the base set already has a reference of that nature.
  !!
  !! Duplicate definition natures are retained.
  
  SUBROUTINE MergeNatures_scalar(base, other)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: base(:)
    TYPE(IdentifierNature), INTENT(IN) :: other
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ib             ! Index into base.
    
    !***************************************************************************
    
    ib = FindIdentifier(base, other%symbol%name)
    IF (ib == 0) THEN
      ! Not found - add the other identifier to the end of base.
      CALL Add(base, other)
    ELSE
      CALL merge_natures(base(ib), other)
    END IF
    
  END SUBROUTINE MergeNatures_scalar
  
  
  !*****************************************************************************
  !!
  !> Find an identifier in a vector of identifiers.
  !!
  !! @param[in]     vector            The vector of identifiers to search.
  !!
  !! @param[in]     name              The name to search for.
  !!
  !! @returns The index of the relevant identifier in @a vector, or zero if 
  !! there was no such identifier.
  
  FUNCTION FindIdentifier(vector, name) RESULT(idx)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierNature), INTENT(IN) :: vector(:)
    CHARACTER(*), INTENT(IN) :: name
    
    ! Function result.
    INTEGER :: idx
  
    !***************************************************************************
    
    DO idx = 1, SIZE(vector)
      IF (vector(idx)%symbol%name == name) RETURN
    END DO
    
    idx = 0
    
  END FUNCTION FindIdentifier
  
  
  !*****************************************************************************
  !!
  !> Add a scalar IdentifierNature object to an existing vector.
  !!
  !! @param[in,out] base              The existing vector.  Must be allocated.
  !!
  !! @param[in]     other             The object to add.
  !!
  !! The identifier to be added must not already be in @a base.
  
  SUBROUTINE Add_(base, other)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT), ALLOCATABLE :: base(:)
    TYPE(IdentifierNature), INTENT(IN) :: other
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Temporary for growing @a base.
    TYPE(IdentifierNature), ALLOCATABLE :: tmp(:)
    
    !***************************************************************************
    
    ALLOCATE(tmp(SIZE(base)+1))
    tmp(:SIZE(base)) = base
    tmp(SIZE(tmp)) = other
    
    CALL MOVE_ALLOC(tmp, base)
    
  END SUBROUTINE Add_
  
  
  !*****************************************************************************
  !!
  !> Merge two scalar IdentifierNature objects together.
  !!
  !! @param[in,out] base              The base object to merge into.
  !!
  !! @param[in]     other             The object to be merged in.
  !!
  !! Locations for existing reference natures in the @a base set are always 
  !! maintained, locations for matching reference natures in the @a other set 
  !! may be discarded if the base set already has a reference of that nature.
  !!
  !! Duplicate definition natures are retained.
  
  SUBROUTINE merge_natures(base, other)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierNature), INTENT(INOUT) :: base
    TYPE(IdentifierNature), INTENT(IN) :: other
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ib             ! Index into base.
    INTEGER :: io             ! Index into other.
    
    !***************************************************************************
    
    other_loop: DO io = 1, SIZE(other%natures)
      ! We merge references, but always transfer definitions.
      IF (IsReferenceNature(other%natures(io))) THEN
        DO ib = 1, SIZE(base%natures)
          IF (base%natures(ib) == other%natures(io)) CYCLE other_loop
        END DO
      END IF
      ! Nature not previously found.  Add it.
      base%natures = [base%natures, other%natures(io)]
      base%locations = [base%locations, other%locations(io)]
    END DO other_loop
    
  END SUBROUTINE merge_natures
  
  
  !*****************************************************************************
  !!
  !> Get a text description of a nature.
  !!
  !! @param[in]     nature            The nature of interest.
  !!
  !! @returns The trimmed corresponding description from the description map, 
  !! suitable for use in diagnostic output.
  !!
  !! If we cannot find the nature in the map we make some nonsense up.
  
  FUNCTION GetNatureDescription(nature) RESULT(s)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: nature
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: s
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: im             ! Map index.
    
    !***************************************************************************
    
    ! Find the nature in the map.
    DO im = 1, SIZE(map)
      IF (map(im)%nature == nature) THEN
        s = TRIM(map(im)%preamble) // ' ' // TRIM(map(im)%text)
        RETURN
      END IF
    END DO
    
    ALLOCATE(CHARACTER(32) :: s)
    WRITE (s, "('unknown: ',I0)") nature
    s = TRIM(s)
    
  END FUNCTION GetNatureDescription
  
  
  !*****************************************************************************
  !!
  !> Get a text description of a nature.
  !!
  !! @param[in]     nature            The nature of interest.
  !!
  !! @returns The trimmed corresponding description from the description map, 
  !! suitable for use in diagnostic output.
  !!
  !! If we cannot find the nature in the map we make some nonsense up.
  
  FUNCTION GetNatureText(nature) RESULT(s)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: nature
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: s
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: im             ! Map index.
    
    !***************************************************************************
    
    ! Find the nature in the map.
    DO im = 1, SIZE(map)
      IF (map(im)%nature == nature) THEN
        s = TRIM(map(im)%text)
        RETURN
      END IF
    END DO
    
    ALLOCATE(CHARACTER(32) :: s)
    WRITE (s, "('unknown: ',I0)") nature
    s = TRIM(s)
    
  END FUNCTION GetNatureText
  
  
  !*****************************************************************************
  !!
  !> Test whether a nature is considered a reference.
  !!
  !! @param[in]     nature            The nature to test.
  !!
  !! @returns True if the nature is a reference nature.  Identities can 
  !! generally be referred to an unlimited number of times.
  
  PURE FUNCTION IsReferenceNature(nature) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: nature
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = nature <= NatureRefLast
    
  END FUNCTION IsReferenceNature
  
  
  !*****************************************************************************
  !!
  !> Test whether a nature is considered a defintiion.
  !!
  !! @param[in]     nature            The nature to test.
  !!
  !! @returns True if the nature is a definition nature.  Identities should 
  !! only be defined once, bar some exceptions.
  
  PURE FUNCTION IsDefinitionNature(nature) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: nature
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = nature > NatureRefLast
    
  END FUNCTION IsDefinitionNature
  
END MODULE IdentifierNatures
