! $Id: IdentifierSymbols.f90 2391 2016-08-24 13:21:24Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IdentifierSymbols module.


!*******************************************************************************
!!
!> Defines the IdentiferSymbol type and some basic supporting procedures.
!!
!! An IdentifierSymbol object is like an Identifier object (from the 
!! Identifiers module), but:
!! - it does not store a location,
!! - we also store identifiers that are not names or operators,
!! - we remember whether the identifier was a name, operator, or some 
!!   other thing.

MODULE IdentifierSymbols
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: NameSymbol
  PUBLIC :: OperatorSymbol
  PUBLIC :: AssignmentSymbol
  PUBLIC :: ReadFormattedSymbol
  PUBLIC :: WriteFormattedSymbol
  PUBLIC :: ReadUnformattedSymbol
  PUBLIC :: WriteUnformattedSymbol
  PUBLIC :: FindSymbol
  PUBLIC :: OPERATOR(.NameOf.)
  PUBLIC :: OPERATOR(.KeywordOf.)  
  PUBLIC :: OPERATOR(==)
  
  PUBLIC :: GetIdentifierName
  PUBLIC :: GetOperatorSymbolName
  
  !> The identifier for an identifier.
  !!
  !! Historically this was just a character string, but the issue with 
  !! that is that we discard information gained during parsing about 
  !! the identifier when that identifier is a generic identifier that 
  !! is not a name.
  TYPE, PUBLIC :: IdentifierSymbol
    !> If the identifier is just a name, then this is the name.
    !!
    !! Always allocated.  Always length greater than zero.
    !!
    !! If the identifier is a generic identifier that is not a name, then 
    !! the name is a special form, as given by 
    !! GenericSpecs%%GetGenericSpecIdentifierName.
    CHARACTER(:), ALLOCATABLE :: name
    
    !> For identifiers that are generic identifiers that are not names, 
    !! this describes the type of identifier.
    !! 
    !! If the idenfier is a name (including a generic name) then this 
    !! will be zero.  Otherwise this has the same values as for 
    !! the `keyword` component of the `GenericSpec` type in the 
    !! `GenericSpecs` module.
    !!
    !! - 0: No keyword - the name is just the name.
    !! - 1: OPERATOR - the name will be the operator.
    !! - 2: ASSIGNMENT - the name will be '='.
    !! - 3: READ(FORMATTED) - name not used.
    !! - 4: WRITE(FORMATTED) - name not used.
    !! - 5: READ(UNFORMATTED) - name not used.
    !! - 6: WRITE(UNFORMATTED) - name not used.
    INTEGER :: keyword = 0
  END TYPE IdentifierSymbol
  
  !> Compare two symbols.
  INTERFACE OPERATOR(==)
    MODULE PROCEDURE op_equals
  END INTERFACE OPERATOR(==)
  
  !> Extract the name of the symbol.
  INTERFACE OPERATOR(.NameOf.)
    MODULE PROCEDURE op_nameof
  END INTERFACE OPERATOR(.NameOf.)
  
  !> Extract the keyword of the symbol.
  INTERFACE OPERATOR(.KeywordOf.)
    MODULE PROCEDURE op_keywordof
  END INTERFACE OPERATOR(.KeywordOf.)
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is a name.
  
  PURE FUNCTION NameSymbol(name) RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: name
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = name
    r%keyword = 0
    
  END FUNCTION NameSymbol
  
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is a name.
  
  PURE FUNCTION OperatorSymbol(operator) RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: operator
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = GetOperatorSymbolName(operator)
    r%keyword = 0
    
  END FUNCTION OperatorSymbol

  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is a name.
  
  PURE FUNCTION AssignmentSymbol() RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = '='
    r%keyword = 2
    
  END FUNCTION AssignmentSymbol
  
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is for formatted input.
  
  PURE FUNCTION ReadFormattedSymbol() RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = 'READ(FORMATTED)'
    r%keyword = 3
    
  END FUNCTION ReadFormattedSymbol
  
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is for formatted output.
  
  PURE FUNCTION WriteFormattedSymbol() RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = 'WRITE(FORMATTED)'
    r%keyword = 4
    
  END FUNCTION WriteFormattedSymbol
  
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is for unformatted 
  !! input.
  
  PURE FUNCTION ReadUnformattedSymbol() RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = 'READ(UNFORMATTED)'
    r%keyword = 5
    
  END FUNCTION ReadUnformattedSymbol
  
  
  !*****************************************************************************
  !!
  !> Create an IdentifierSymbol for an identifier that is for unformatted 
  !! output.
  
  PURE FUNCTION WriteUnformattedSymbol() RESULT(r)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    TYPE(IdentifierSymbol) :: r
    
    !***************************************************************************
    
    r%name = 'WRITE(UNFORMATTED)'
    r%keyword = 6
    
  END FUNCTION WriteUnformattedSymbol
  
  
  !*****************************************************************************
  !!
  !> Extract the name part of an IdentifierSymbol.
  !!
  !! @param[in]     symbol            The symbol of interest.
  !!
  !! @returns The name or description of the symbol.
  
  PURE FUNCTION op_nameof(symbol) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierSymbol), INTENT(IN) :: symbol
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    str = symbol%name
    
  END FUNCTION op_nameof
  
  
  !*****************************************************************************
  !!
  !> Extract the name part of an IdentifierSymbol.
  !!
  !! @param[in]     symbol            The symbol of interest.
  !!
  !! @returns The keyword of the symbol.
  
  PURE FUNCTION op_keywordof(symbol) RESULT(keyword)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierSymbol), INTENT(IN) :: symbol
    
    ! Function result.
    INTEGER :: keyword
    
    !***************************************************************************
    
    keyword = symbol%keyword
    
  END FUNCTION op_keywordof
  
  
  !*****************************************************************************
  !!
  !> Compare two IdentifierSymbol objects.
  !!
  !! @param[in]     lhs               The left identifier symbol.
  !!
  !! @param[in]     rhs               The right identifier symbol.
  !!
  !! @returns True if the symbols are the same.
  
  PURE FUNCTION op_equals(lhs, rhs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierSymbol), INTENT(IN) :: lhs
    TYPE(IdentifierSymbol), INTENT(IN) :: rhs
    
    ! Function result.
    LOGICAL :: b
    
    !***************************************************************************
    
    b = lhs%keyword == rhs%keyword
    IF (.NOT. b) RETURN
    
    IF (lhs%keyword < 2) THEN
      b = lhs%name == rhs%name
    END IF
    
  END FUNCTION op_equals
  
  
  !*****************************************************************************
  !!
  !> Get a string that describes an identifier
  !!
  !! @param[in]     name              The name part of the identifier, 
  !! if the identifier has a name, otherwise not allocated.  This must be 
  !! allocated if keyword is 0 or 1.
  !!
  !! @param[in]     keyword           The form of identifier.  This is 
  !! the same as the keyword component of the GenericSpec type and 
  !! of the IdentifierIdentifier type.
  !> - 0: No keyword - the name is just the name.
  !! - 1: OPERATOR - the name will be the operator.
  !! - 2: ASSIGNMENT - the name will be '='.
  !! - 3: READ(FORMATTED) - name not used.
  !! - 4: WRITE(FORMATTED) - name not used.
  !! - 5: READ(UNFORMATTED) - name not used.
  !! - 6: WRITE(UNFORMATTED) - name not used.
  
  PURE FUNCTION GetIdentifierName(name, keyword) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), INTENT(IN), ALLOCATABLE :: name
    INTEGER, INTENT(IN) :: keyword
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    SELECT CASE (keyword)
    CASE (0)      ; str = name
    CASE (1)      ; str = GetOperatorSymbolName(name)
    CASE (2)      ; str = '='
    CASE (3)      ; str = 'READ(FORMATTED)'
    CASE (4)      ; str = 'WRITE(FORMATTED)'
    CASE (5)      ; str = 'READ(UNFORMATTED)'
    CASE (6)      ; str = 'WRITE(UNFORMATTED)'
    CASE DEFAULT  ; str = '(invalid)'
    END SELECT
    
  END FUNCTION GetIdentifierName
  
  
  !*****************************************************************************
  !!
  !> Remap an operator name to the consistent entity name.
  !!
  !! @param[in]     name              The operator symbol.
  !!
  !! @returns The consistent entity name for the operator.
  
  PURE FUNCTION GetOperatorSymbolName(name) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: name
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    SELECT CASE (name)
    CASE ('.EQ.')     ; str = '=='
    CASE ('.NE.')     ; str = '/='
    CASE ('.GT.')     ; str = '>'
    CASE ('.GE.')     ; str = '>='
    CASE ('.LT.')     ; str = '<'
    CASE ('.LE.')     ; str = '<='
    CASE DEFAULT      ; str = name
    END SELECT
    
  END FUNCTION GetOperatorSymbolName
  
  
  !*****************************************************************************
  !!
  !> Find a symbol in an array of symbols.
  !!
  !! @param[in]     vector            The vector of symbols to search.
  !!
  !! @param[in]     symbol            The symbol to search for.
  
  PURE FUNCTION FindSymbol(vector, symbol) RESULT(idx)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(IdentifierSymbol), INTENT(IN) :: vector(:)
    TYPE(IdentifierSymbol), INTENT(IN) :: symbol
    
    ! Function result.
    INTEGER :: idx
    
    !***************************************************************************
    
    IF (symbol%keyword <= 1) THEN
      DO idx = 1, SIZE(vector)
        IF (symbol%keyword /= vector(idx)%keyword) CYCLE
        IF (vector(idx)%name == symbol%name) RETURN
      END DO
    ELSE
      DO idx = 1, SIZE(vector)
        IF (symbol%keyword == vector(idx)%keyword) RETURN
      END DO
    END IF
    
    idx = 0
    
  END FUNCTION FindSymbol
  
END MODULE IdentifierSymbols
