! $Id: ImplicitStmts.f90 2928 2019-04-27 06:21:03Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @ file
!! Defines the ImplicitStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing implicit statements.

MODULE ImplicitStmts
  
  USE DeclarationTypeSpecs
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseImplicitStmt
  PUBLIC :: TestSpecsOverlap
  PUBLIC :: TestLetterSpecsOverlap
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /letter-spec/.
  TYPE, PUBLIC :: LetterSpec
    !> The starting character of the /letter-spec/.  Should always be an
    !! alphabetic character.
    CHARACTER :: start
    !> If no finish character was specified then this is a blank.
    CHARACTER :: finish
    !> The location of the letter spec.
    TYPE(SourceLocation), PRIVATE :: location
  CONTAINS
    PROCEDURE :: GetLocation => letterspec_GetLocation
  END TYPE LetterSpec
  
  !> Results of parsing an /implicit-spec/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ImplicitSpec
    !> The /declaration-type-spec/ designated by the /implicit-spec/.
    TYPE(DeclTypeSpec) :: dts
    !> The contents of the /letter-spec-list/.
    TYPE(LetterSpec), ALLOCATABLE :: letters(:)
  CONTAINS
    PROCEDURE :: Write => is_Write
    PROCEDURE :: Query => is_Query
    PROCEDURE :: Modify => is_Modify
    PROCEDURE, NOPASS :: What => is_What
    PROCEDURE :: GetLocation => is_GetLocation
  END TYPE ImplicitSpec
  
  !> The results of parsing an implicit statement.
  TYPE, PUBLIC, EXTENDS(SCParseNode) :: ImplicitStmt
    !> The implicit specs in the statement.  If the statement specifies 
    !! IMPLICIT NONE then this is allocated to zero size.
    TYPE(ImplicitSpec), ALLOCATABLE :: specs(:)
  CONTAINS
    PROCEDURE :: Check => implicit_Check
    PROCEDURE :: Dump => implicit_Dump
    PROCEDURE :: Query => implicit_Query
    PROCEDURE :: Modify => implicit_Modify
    PROCEDURE, NOPASS :: What => implicit_What
  END TYPE ImplicitStmt
  
  ! Syntax rule number for a /letter-spec/.
  CHARACTER(*), PARAMETER :: letterspec_comp = 'R562'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /implicit-stmt/ (R560).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    implicit_stmt     The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseImplicitStmt(tlist, implicit_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ImplicitStmt), INTENT(OUT) :: implicit_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number of an /implicit-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R560'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Match function list for Expect call.
    TYPE(MatchProcList) :: procs(2)
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect call option result.    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: ispec          ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for parse_implicit_spec call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /implicit-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'IMPLICIT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Do we have IMPLICIT NONE or IMPLICIT /implicit-spec-list/ ?
    
    procs(1)%item => match_none
    procs(2)%item => MatchDeclarationTypeSpecStart
    CALL Expect(tlist, 2, procs, '"NONE" or /declaration-type-spec/', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN      ! IMPLICIT NONE
      ALLOCATE(implicit_stmt%specs(0))
      CALL ExpectEnd(tlist, 3, comp, err_list)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! IMPLICIT /implicit-spec-list/
    
    list_count = ListItemCount(tlist(2:))
    ALLOCATE(implicit_stmt%specs(list_count))
    item_finish = 0
    DO ispec = 1, list_count
      CALL GetNextListItem(tlist, 2, SIZE(tlist), item_start, item_finish)
      CALL parse_implicit_spec(  &
          tlist(item_start:item_finish),  &
          implicit_stmt%specs(ispec),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseImplicitStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a /letter-spec/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    letter_spec       The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_letter_spec(tlist, letter_spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(LetterSpec), INTENT(OUT) :: letter_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! The first letter
    
    CALL parse_letter(tlist, 1, letter_spec%start, err)
    CALL Add(err_list, err, COMPONENT=letterspec_comp)
    IF (Failure(err)) RETURN
    
    ! Save the location.
    letter_spec%location = QueryLocation(tlist(1))
    
    IF (SIZE(tlist) == 1) THEN
      letter_spec%finish = ''
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The minus
    
    CALL Expect(tlist, 2, ittOperator, scck_'-', err) 
    CALL Add(err_list, err, COMPONENT=letterspec_comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The second letter
    
    CALL parse_letter(tlist, 3, letter_spec%finish, err)
    CALL Add(err_list, err, COMPONENT=letterspec_comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! All done.
    
    CALL ExpectEnd(tlist, 4, letterspec_comp, err_list)
    
  END SUBROUTINE parse_letter_spec
  
  
  !*****************************************************************************
  !!
  !> Validate a letter spec.
  !!
  !! @param[in]     letter_spec       The spec to validate.
  !!
  !! @param[in,out] err_list          List of errors.  Note this is INOUT.
  !!
  !! We check that the letters are in the right order.
  
  SUBROUTINE check_letter_spec(letter_spec, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(LetterSpec), INTENT(IN) :: letter_spec
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    IF (letter_spec%finish == '') RETURN
    
    ! The ascii order gives us what the standard regards as being alphabetical.
    IF (letter_spec%finish <= letter_spec%start) THEN
      CALL Add( err_list,  &
          CODE=errNotAlphabetical,  &
          COMPONENT='C583 via ' // letterspec_comp,  &
          LOCATION=letter_spec%GetLocation(),  &
          MSG='The second letter in a /letter-spec/ must follow the first &
            &letter alphabetically.' )
    END IF
    
  END SUBROUTINE check_letter_spec
  
  
  !*****************************************************************************
  !!
  !> Parse the letter part of a /letter-spec/
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     pos               Position of the token of interest in 
  !! @a tlist.
  !!
  !! @param[out]    letter            The letter.
  !!
  !! @param[out]    err               Error object.
  
  SUBROUTINE parse_letter(tlist, pos, letter, err)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: pos
    CHARACTER, INTENT(OUT) :: letter
    TYPE(Error), INTENT(OUT) :: err
    
    !***************************************************************************
    
    CALL Expect(tlist, pos, '/letter/', err)
    IF (Failure(err)) RETURN
    
    IF (LEN(QueryValue(tlist(pos))) /= 1) THEN
      err = ConstructError(  &
          CODE=errSyntax,  &
          LOCATION=QueryLocation(tlist(pos)),  &
          MSG='A single /letter/ was expected but "' // QueryRaw(tlist(pos))  &
            // '" was encountered.' )
      RETURN
    END IF
    
    ! Only the first character gets copied over.  Kind conversion here.
    letter = QueryValue(tlist(pos))
    
  END SUBROUTINE parse_letter
  
  
  !*****************************************************************************
  !!
  !> Implementation of LetterSpec%GetLocation - get the source location of the 
  !! letter spec.
  !!
  !! @param[in]     spec              The letter spec of interest.
  !!
  !! @returns The location of the spec.
  
  FUNCTION letterspec_GetLocation(spec) RESULT(loc)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(LetterSpec), INTENT(IN) :: spec
    
    ! Function result
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = spec%location
    
  END FUNCTION letterspec_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implmentation of ImplictStmt%Check - check the parsed 
  !! representation of an implicit statement against the relevant 
  !! language rules.
  !!
  !! @param[in]     pnode             The parsed representation.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! We check that the letter ranges don't overlap.
  
  SUBROUTINE implicit_Check(pnode, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImplicitStmt), INTENT(IN) :: pnode
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ispec          ! Primary specification index.
    INTEGER :: ip             ! Secondary specification index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    DO ispec = 1, SIZE(pnode%specs)
      ! Adds to existing err_list.
      CALL check_implicit_spec(pnode%specs(ispec), err_list)
    END DO
    
    ! Faulty ordering will confuse the overlapping tests.
    IF (Failure(err_list)) RETURN
    
    DO ispec = 1, SIZE(pnode%specs) - 1
      DO ip = ispec + 1, SIZE(pnode%specs)
        IF (TestSpecsOverlap(  &
            pnode%specs(ispec),  &
            pnode%specs(ip) )) THEN
          CALL Add( err_list,  &
              CODE=errOverlappingImplicitSpecs,  &
              COMPONENT='5.5p2',  &
              LOCATION=pnode%GetLocation(),  &
              MSG='There is an overlap between the letters in the &
                &/letter-spec/''s within this /implicit-stmt/.' )
          RETURN
        END IF
      END DO
    END DO
    
  END SUBROUTINE implicit_Check
  
  
  !*****************************************************************************
  !!
  !> Test whether two /implicit-spec/'s overlap.
  !!
  !! @param[in]     lhs               One of the /implicit-spec/'s.
  !!
  !! @param[in]     rhs               The other /implicit-spec/.
  !!
  !! @return .TRUE. if there is a common letter or letters between the 
  !! two specifications, .FALSE. otherwise.
  
  FUNCTION TestSpecsOverlap(lhs, rhs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(ImplicitSpec), INTENT(IN) :: lhs
    TYPE(ImplicitSpec), INTENT(IN) :: rhs
    
    ! Function result
    LOGICAL :: b
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: il
    INTEGER :: ir
    
    !***************************************************************************
    
    DO il = 1, SIZE(lhs%letters)
      DO ir = 1, SIZE(rhs%letters)
        b = TestLetterSpecsOverlap(lhs%letters(il), rhs%letters(ir))
        IF (b) RETURN
      END DO
    END DO
    
    b = .FALSE.
    
  END FUNCTION TestSpecsOverlap
  
  
  !*****************************************************************************
  !!
  !> Test whether two /letter-spec/'s overlap.
  !!
  !! @param[in]     lhs               One of the /letter-spec/'s.
  !!
  !! @param[in]     rhs               The other /letter-spec/.
  !!
  !! @return .TRUE. if there is a common letter or letters between the 
  !! two specifications, .FALSE. otherwise.
  !!
  !! The lexical parsing routines have already made the letters upper case.
  
  FUNCTION TestLetterSpecsOverlap(lhs, rhs) RESULT(b)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(LetterSpec), INTENT(IN) :: lhs
    TYPE(LetterSpec), INTENT(IN) :: rhs
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    IF (lhs%finish == '') THEN      ! lhs single character
      IF (rhs%finish == '') THEN    ! rhs single character
        b = lhs%start == rhs%start
      ELSE                          ! rhs range
        b = LGE(lhs%start, rhs%start) .AND. LLE(lhs%start, rhs%finish)
      END IF
    ELSE                            ! lhs range
      IF (rhs%finish == '') THEN    ! rhs single character.
        b = LGE(rhs%start, lhs%start) .AND. LLE(rhs%start, lhs%finish)
      ELSE                          ! rhs range
        b = LGE(lhs%finish, rhs%start) .AND. LGE(rhs%finish, lhs%start)
      END IF
    END IF
    
  END FUNCTION TestLetterSpecsOverlap
  
  
  !*****************************************************************************
  !!
  !> Test for the keyword NONE.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Position in @a tlist to test.
  !!
  !! @returns .TRUE. if the token at @a tlist ( @a start ) is a name token 
  !! with value NONE.
  
  FUNCTION match_none(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'NONE')
    
  END FUNCTION match_none
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitStmt%Dump - write a debugging representation 
  !! of the results of parsing an implicit statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE implicit_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImplicitStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: ii             ! Implicit spec.
    
    !***************************************************************************
    
    IF (SIZE(pnode%specs) == 0) THEN
      WRITE (unit, "(A,'IMPLICIT NONE!')") pfx
    ELSE
      WRITE (unit, "(A,'/implicit-specs/')") pfx
      DO ii = 1, SIZE(pnode%specs)
        WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
        CALL pnode%specs(ii)%Write(unit)
        WRITE (unit, "()")
      END DO
    END IF
    
  END SUBROUTINE implicit_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE implicit_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImplicitStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%specs)
      CALL pnode%specs(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE implicit_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE implicit_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImplicitStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Spec index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%specs)
      CALL pnode%specs(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE implicit_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'implicit-stmt'
  
  FUNCTION implicit_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'implicit-stmt'
    
  END FUNCTION implicit_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ImplicitSpec.
  
  
  !*****************************************************************************
  !!
  !> Parse an /implicit-spec/ (R561).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    implicit_spec     Results of parsing the /implicit-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_implicit_spec(tlist, implicit_spec, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ImplicitSpec), INTENT(OUT) :: implicit_spec
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /implicit-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R561'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error object for Expect calls.
    TYPE(Error) :: err
    
    INTEGER :: it   ! Position of the end of the declaration-type-spec.
    
    ! Error list for procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: closing        ! Index of closing parenthesis.s
    
    INTEGER :: list_count     ! Number of items in the /letter-spec-list/.
    INTEGER :: ispec          ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    ! Work out where the declaration-type-spec ends.
    
    ! This is easier that the case for /type-declaration-stmt/'s, as we 
    ! have the parentheses around the letter list that we can use as a 
    ! reference.  For this reason we've not factored the common code together.
    ! Perhaps this was a mistake.
    !    
    ! Our options (where INTEGER stands for any intrinsic type):
    !
    ! INTEGER ( /letter-list/ )
    ! DOUBLE PRECISION ( /letter-list/ )
    ! DOUBLEPRECISION ( /letter-list/ )
    ! CHARACTER xxxx ( /letter-list/ )   ! Note xxxx could include *(yyy)
    ! INTEGER ( kinds... ) ( /letter-list/ )
    ! TYPE ( type-name... ) ( /letter-list/ )
    ! CLASS ( type-name... ) ( /letter-list/ )
    !
    ! At the end of this block we want it to be the index of the last token
    ! in the declaration-type-spec.
    
    ! Find the location of the last '('.  If we can't find one we parse the 
    ! whole thing as a /declaration-type-spec/ and then cop an error later.
    DO it = SIZE(tlist), 1, -1
      IF (tlist(it) == ittDelimiter) THEN
        IF (tlist(it) == scck_'(') EXIT
      END IF
    END DO
    IF (it <= 1) THEN
      it = SIZE(tlist)
    ELSE
      it = it - 1
    END IF
    
    CALL ParseDeclarationTypeSpec(  &
        tlist(:it),  &
        .FALSE.,  &
        implicit_spec%dts,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! The ( /letter-spec-list/ ) part.
    
    it = it + 1
    ! Failure not covered by StmtParsingTests due to last parenthesis 
    ! logic above.
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err_list)) RETURN
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err_list)) RETURN
    
    ! Parse the /letter-spec-list/.
    list_count = ListItemCount(tlist(it+1:closing-1))
    ALLOCATE(implicit_spec%letters(list_count))
    item_finish = 0
    DO ispec = 1, list_count
      CALL GetNextListItem(tlist, it + 1, closing - 1, item_start, item_finish)
      CALL parse_letter_spec(  &
          tlist(item_start:item_finish),  &
          implicit_spec%letters(ispec),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    CALL ExpectEnd(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE parse_implicit_spec
  
  
  !*****************************************************************************
  !!
  !> Check the parsed representation of an /implicit-spec/ against the 
  !! relevant language rules.
  !!
  !! @param[in]     spec              The implicit spec to check.
  !!
  !! @param[in,out] err_list          List of errors.  Note this is INOUT.
  
  SUBROUTINE check_implicit_spec(spec, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ImplicitSpec), INTENT(IN) :: spec
    TYPE(Error), INTENT(INOUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Letter spec index.
    INTEGER :: i2             ! Other letter spec index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(spec%letters)
      ! Adds to existing err_list.
      CALL check_letter_spec(spec%letters(i), err_list)
    END DO
    
    IF (Failure(err_list)) RETURN
    
    ! Check letters within this spec.
    DO i = 1, SIZE(spec%letters) - 1
      DO i2 = i+1, SIZE(spec%letters)
        IF (TestLetterSpecsOverlap(spec%letters(i), spec%letters(i2))) THEN
          CALL Add( err_list,  &
              CODE=errOverlappingImplicitSpecs,  &
              COMPONENT='5.5p2',  &
              LOCATION=spec%GetLocation(),  &
              MSG='There is an overlap between the letters in the &
                &/letter-spec/''s within this /implicit-spec/.' )
        END IF
      END DO
    END DO
    
  END SUBROUTINE check_implicit_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitSpec%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE is_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImplicitSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: il             ! Letter spec.
    
    !***************************************************************************
    
    CALL WriteDeclTypeSpec(unit, pnode%dts)
    WRITE (unit, "(1X)", ADVANCE='NO')
    DO il = 1, SIZE(pnode%letters)
      ASSOCIATE(ls => pnode%letters(il))
        IF (ls%finish /= '') THEN
          WRITE (unit, "('(',A,' - ',A,')',1X)", ADVANCE='NO') ls%start, ls%finish
        ELSE
          WRITE (unit, "('(',A,')',1X)", ADVANCE='NO') ls%start
        END IF
      END ASSOCIATE
    END DO
    
  END SUBROUTINE is_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitSpec%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE is_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImplicitSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%dts%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE is_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitSpec%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE is_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImplicitSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%dts%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE is_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitSpec%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the parse node.
  
  FUNCTION is_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'implicit-spec'
    
  END FUNCTION is_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImplicitSpec%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION is_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ImplicitSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%dts
    
  END FUNCTION is_GetLocation
  
END MODULE ImplicitStmts
