! $Id: ImpliedDos.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ImpliedDos module.


!*******************************************************************************
!!
!> Procedures to assist with parsing of implied do syntax rules.
!!
!! For data statements:
!!
!! R537 data-implied-do is ( data-i-do-object-list , data-i-do-variable = 
!!        scalar-int-constant-expr , scalar-int-constant-expr 
!!        [ , scalar-int-constant-expr ] )
!!
!! For array constructors:
!!
!! R473 ac-implied-do is ( ac-value-list , ac-implied-do-control )
!! 
!! R474 ac-implied-do-control is ac-do-variable = scalar-int-expr , 
!!        scalar-int-expr [ , scalar-int-expr ]
!!
!! For input/output statements:
!!
!! R918 io-implied-do is ( io-implied-do-object-list , io-implied-do-control )
!!
!! R920 io-implied-do-control is do-variable = scalar-int-expr , 
!!        scalar-int-expr [ , scalar-int-expr ]
!!
!! All of these rules have some sort of object that appears in the list part 
!! of the implied do and then the same syntax for the loop control (though 
!! for some reason the data-stmt doesn't split its loop control parts out 
!! into a separate rule).


MODULE ImpliedDos
  
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseImpliedDo
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing the common parts of an implied do.
  TYPE, EXTENDS(TermParseNode), ABSTRACT, PUBLIC :: ImpliedDoControl
    !> The name of the /do-variable/.
    TYPE(Identifier) :: do_variable
    
    !> The expression that designatest the start of the implied do.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: start
    
    !> The expression that designates the finish of the implied do.
    !!
    !! Always allocated.
    CLASS(Expression), ALLOCATABLE :: finish
    
    !> The expression that designates the step of the implied do.
    !!
    !! Not allocated if there is no step.
    CLASS(Expression), ALLOCATABLE :: step
  CONTAINS
    PROCEDURE :: Write => idc_Write
    PROCEDURE :: Query => idc_Query
    PROCEDURE :: Modify => idc_Modify
    PROCEDURE :: Advertise => idc_Advertise
    PROCEDURE :: GetLocation => idc_GetLocation
  END TYPE ImpliedDoControl
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse sequence of tokens that is the contents of an implied-do, 
  !! extracting the do-control part.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The starting point in the sequence 
  !! of the contents of the implied do.  This is typically the token that 
  !! is immediately after the opening parenthesis.
  !!
  !! @param[in]     finish            The finishing point in the sequence of 
  !! the contents of the implied do.  This is typically the token that is 
  !! immediately before the closing parenthesis.
  !!
  !! @param[in]     list_rule_name    The syntax rule of the list part 
  !! of the implied do.  Do not include the slashes!
  !!
  !! @param[in]     implied_do_name   The syntax rule for the entire 
  !! implied do part.  Include the slashes.
  !!
  !! @param[in]     implied_do_comp   The syntax rule number for the entire 
  !! implied do part.
  !!
  !! @param[out]    list_count        The number of value items in the 
  !! implied do.
  !!
  !! @param[out]    do_control        The control parts of the implied do.
  !!
  !! @param[out]    err_list          List of errors.
  
  ! Recursive because some psychopath could have an implied do nested inside 
  ! the control sequence (versus the list out the front) of an implied do, 
  ! so you might end up back here via the expression parsing.
  RECURSIVE SUBROUTINE ParseImpliedDo( tlist, start, finish, list_rule_name,  &
      implied_do_name, implied_do_comp, list_count, do_control, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: list_rule_name
    CHARACTER(*), INTENT(IN) :: implied_do_name
    CHARACTER(*), INTENT(IN) :: implied_do_comp
    INTEGER, INTENT(OUT) :: list_count
    CLASS(ImpliedDoControl), INTENT(OUT) :: do_control
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: assign         ! Index of the assign token.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! - 0: Yet to see the start expression of the io-implied-do-control.
    ! - 1: Got the start but yet to see the finish of the do control.
    ! - 2: Got the start and the finish.
    ! - 3: Got the start, the finish and the step.
    INTEGER :: state
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! List processing would pick this up, but just for consistency.
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=implied_do_comp,  &
          MSG='Expected /' // list_rule_name // '-list/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    item_finish = 0
    list_count = 0
    state = 0
    DO
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      IF (item_start == 0) EXIT   ! test for list finish
      
      assign = FindBare(tlist(:item_finish), ittAssign, scck_'=', start)
      SELECT CASE (state)
      CASE (0)    ! Yet to strike the variable=start part of the implied do.
        IF (assign /= 0) THEN   ! = implies that we have the start part.
          ! /do-variable/, which must be /name/.
          CALL Expect(tlist, item_start, ittName, err)
          CALL Add(err_list, err, COMPONENT=implied_do_comp)
          IF (Failure(err)) RETURN
          do_control%do_variable = IdentifierIfortBug(tlist(item_start))
          
          ! /do-variable/ = ...
          CALL Expect(tlist, item_start + 1, ittAssign, scck_'=', err)
          CALL Add(err_list, err, COMPONENT=implied_do_comp)
          IF (Failure(err)) RETURN
          
          ! /do_variable/ = expr
          CALL ParseExpr(  &
              tlist,  &
              item_start + 2, item_finish,  &
              do_control%start, &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=implied_do_comp)
          IF (Failure(err)) RETURN
          
          state = 1
        ELSE
          ! Just an object in the list.
          list_count = list_count + 1
        END IF
        
      CASE (1)    ! Looking for the finish part of the implied do.
        CALL ParseExpr(  &
            tlist,  &
            item_start, item_finish,  &
            do_control%finish,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=implied_do_comp)
        IF (Failure(err)) RETURN
        state = 2
        
      CASE (2)    ! Looking for the optional step
        CALL ParseExpr(  &
            tlist,  &
            item_start, item_finish,  &
            do_control%step,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=implied_do_comp)
        IF (Failure(err)) RETURN
        state = 3
        
      CASE (3)    ! Extra item after loop control.
        CALL Add(  &
            err_list,  &
            CODE=errSyntax,  &
            COMPONENT=implied_do_comp,  &
            LOCATION=QueryLocation(tlist(item_start)),  &
            MSG='Additional items after the stride of a '  &
              // implied_do_name // '.' )
        RETURN
        
      END SELECT
    END DO
    
    IF (state == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=implied_do_comp,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          MSG=implied_do_name // ' missing the finish expression part of &
            &the loop control.' )
      RETURN
    ELSE IF (state == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=implied_do_comp,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          MSG=implied_do_name // ' missing the loop control parts.' )
      RETURN
    ELSE IF (list_count == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=implied_do_comp,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          MSG='Expected /' // list_rule_name // '-list/.' )
      RETURN
    END IF
    
  END SUBROUTINE ParseImpliedDo
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImpliedDoControl%Write - write a text representation 
  !! of an implied do to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE idc_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImpliedDoControl), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "('variable = ',A)", ADVANCE='NO')   &
        .NameOf. pnode%do_variable
    WRITE (unit, "(', start = ')", ADVANCE='NO')
    CALL pnode%start%Write(unit)
    WRITE (unit, "(', finish = ')", ADVANCE='NO')
    CALL pnode%finish%Write(unit)
    
  END SUBROUTINE idc_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImpliedDoControl%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE idc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImpliedDoControl), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%start%Query(visitor)
    CALL pnode%finish%Query(visitor)
    
    IF (ALLOCATED(pnode%step)) CALL pnode%Step%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE idc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImpliedDoControl%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE idc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImpliedDoControl), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%start%Modify(visitor)
    CALL pnode%finish%Modify(visitor)
    
    IF (ALLOCATED(pnode%step)) CALL pnode%Step%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE idc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImpliedDoControl%Advertise - advertise 
  !! the identifiers referenced or defined by the array value.
  !!
  !! @param[in]     pnode             The implied do control to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  !!
  !! The do variable is given a special definition nature, because it may 
  !! be a construct local variable.
  
  ! Recursive because an implied do control can contain another implied do 
  ! control if the programmer is being ridiculous.
  RECURSIVE SUBROUTINE idc_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImpliedDoControl), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%do_variable,  &
        NATURES=[NatureDefImpliedDoVar],  &
        LOCATIONS=[.LocationOf. pnode%do_variable] )
    
  END SUBROUTINE idc_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImpliedDoControl%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION idc_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ImpliedDoControl), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%do_variable
    
  END FUNCTION idc_GetLocation
  
END MODULE ImpliedDos
