! $Id: ImportStmts.f90 2948 2019-09-30 01:49:37Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ImportStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /import-stmt/'s.

MODULE ImportStmts
  
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseImportStmt
  
  !> Stores the results of parsing an /import-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: ImportStmt
    !> The identifiers in the import statement.
    !!
    !! Always allocated, may be zero sized, in which case the statement 
    !! was just IMPORT without any names.
    TYPE(Identifier), ALLOCATABLE :: import_names(:)
  CONTAINS
    PROCEDURE :: Dump => import_Dump
    PROCEDURE, NOPASS :: What => import_What
    PROCEDURE :: Advertise => import_Advertise
    PROCEDURE :: Query => import_Query
    PROCEDURE :: Modify => import_Modify
  END TYPE ImportStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /import-stmt/ (R1209).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    import_stmt       The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseImportStmt(tlist, import_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE ErrorCodes
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(ImportStmt), INTENT(OUT) :: import_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /import-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1209'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err    ! Error object for Expect calls.
    INTEGER :: opt        ! Option return from Expect.
    INTEGER :: it         ! Current token index.
    
    ! Error list for ParseNameList call.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /import-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------    
    ! IMPORT ...
    
    CALL ExpectKeyword(tlist, 1, scck_'IMPORT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! That can be it.
    IF (SIZE(tlist) == 1) THEN
      ALLOCATE(import_stmt%import_names(0))
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Optional ::
    
    CALL Expect(  &
        tlist,  &
        2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /import-name/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! /import-name-list/
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/import-name/',  &
        import_stmt%import_names,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseImportStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImportStmt%Dump - write a debugging representation 
  !! of the results of parsing a use statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE import_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImportStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Item index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%import_names)
      WRITE (unit, "(A,A)") pfx, .NameOf. pnode%import_names(i)
    END DO
    
  END SUBROUTINE import_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImportStmt%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE import_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImportStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE import_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImportStmt%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE import_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ImportStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE import_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImportStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE import_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ImportStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Import name index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%import_names)))
    DO i = 1, SIZE(pnode%import_names)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%import_names(i),  &
          NATURES=[NatureRefImportName],  &
          LOCATIONS=[.LocationOf. pnode%import_names(i)] )
    END DO
    
  END SUBROUTINE import_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of ImportStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'import-stmt'
  
  FUNCTION import_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'import-stmt'
    
  END FUNCTION import_What
  
END MODULE ImportStmts
