! $Id: IntrinsicStmts.f90 2196 2016-06-07 02:56:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the IntrinsicStmts module.


!*******************************************************************************
!!
!> Procedure for parsing an /intrinsic-stmt/ (R1218).

MODULE IntrinsicStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseIntrinsicStmt
  
  !> Stores the results of parsing an /intrinsic-stmt/.
  TYPE, EXTENDS(NameListPNode), PUBLIC :: IntrinsicStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => intr_GetListName
    PROCEDURE, NOPASS :: What => intr_What
  END TYPE IntrinsicStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /intrinsic-stmt/ (R1242).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    intr_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseIntrinsicStmt(tlist, intr_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Strings
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(IntrinsicStmt), INTENT(OUT) :: intr_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /intrinsic-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1218'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /intrinsic-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! INTRINSIC ...
    CALL ExpectKeyword(tlist, 1, scck_'INTRINSIC', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional double colon, otherwise a name list.
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    SELECT CASE (opt)
    CASE (1)      ! ::
      it = 3
    CASE (2)      ! A name
      it = 2
    CASE DEFAULT
      STOP 'Unexpected Expect opt result.'
    END SELECT
    
    !---------------------------------------------------------------------------
    ! The /intrinsic-procedure-name-list/.
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/intrinsic-procedure-name/',  &
        intr_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! End of statement check handled above.
    
  END SUBROUTINE ParseIntrinsicStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntrinsicStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/intrinsic-procedure-name-list/'
  
  FUNCTION intr_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/intrinsic-procedure-name-list/'
    
  END FUNCTION intr_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntrinsicStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'intrinsic-stmt'.
  
  FUNCTION intr_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'intrinsic-stmt'
    
  END FUNCTION intr_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of IntrinsicStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE intr_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IntrinsicStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[NatureDefIntrinsic], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE intr_Advertise
  
END MODULE IntrinsicStmts
