! $Id: LanguageBindingSpecs.f90 2363 2016-08-22 01:29:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the LanguageBindingSpecs module.


!*******************************************************************************
!!
!> Types and procedures for parsing and storing language bindings: "BIND(xx)".
!!
!! This module handles both /language-binding-spec/'s and 
!! /proc-language-binding-spec/'s, the latter being a more constrained 
!! version of the former (those constraints are not relevant to parsing).

MODULE LanguageBindingSpecs
  
  USE CompilerKinds
  USE Expressions
  USE ParseNodes
  USE SourceLocations
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: Parse
  PUBLIC :: ParseLanguageBindingSpec
  PUBLIC :: WriteLangSpec
  PUBLIC :: ParseProcLanguageBindingSpec
  
  !-----------------------------------------------------------------------------
  
  !> Object describing the results of parsing a /language-binding-spec/ or 
  !! a /proc-language-binding-spec/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: LangSpec
    !> The language that the entity is bound to.
    !!
    !! - 0: Not specified (no language binding).  Not particularly possible.
    !! - 1: C (ISO/IEC 9899:1999).
    INTEGER :: lang = 0
    
    !> The external binding-name for the entity.  Not allocated if no 
    !! name is specified.
    !!
    !! This should resolve to a scalar-default-char-constant-expr.
    CLASS(Expression), ALLOCATABLE :: name
    
    !> The location of the language binding spec.
    TYPE(SourceLocation), PRIVATE :: location
  CONTAINS
    PROCEDURE :: Write => ls_Write
    PROCEDURE :: Query => ls_Query
    PROCEDURE :: Modify => ls_Modify
    PROCEDURE, NOPASS :: What => ls_What
    PROCEDURE :: GetLocation => ls_GetLocation
  END TYPE LangSpec
  
  !-----------------------------------------------------------------------------
  
  !> Parse a language binding spec.
  INTERFACE Parse
    MODULE PROCEDURE ParseLanguageBindingSpec
    MODULE PROCEDURE Parse_progressive
  END INTERFACE Parse
  
  !-----------------------------------------------------------------------------
  ! Local constants
  
  ! Syntax rule number for a /language-binding-spec/.
  CHARACTER(*), PARAMETER :: lbs_comp = 'R508'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /language-binding-spec/ (R508).
  !!
  !! A language-binding-spec is syntactically the same as a 
  !! /proc-language-binding-spec/, but the latter has more constraints.
  !!
  !! @param[in]     tlist             The sequence of tokens for the enclosing 
  !! statement.
  !!
  !! @param[in]     start             The index of the token that is the 
  !! BIND keyword.
  !!
  !! @param[in]     finish            The index of the token that ends the 
  !! /language-binding-spec/.  This would normally be the closing parenthesis.
  !!
  !! @param[out]    ls                The language specification.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseLanguageBindingSpec(tlist, start, finish, ls, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(LangSpec), INTENT(OUT) :: ls
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Component for error reporting.
    CHARACTER(*), PARAMETER :: comp = lbs_comp
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error code from Expect.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: after          ! Index of token after the actual language-spec.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) < start) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /language-binding-spec/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! BIND keyword
    
    CALL ExpectKeyword(tlist, start, scck_'BIND', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! The rest.
    
    CALL Parse(tlist, start + 1, ls, after, sub_err_list)
    CALL Add(err_list, sub_err_list)
    IF (Failure(sub_err_list)) RETURN
    
    CALL ExpectEnd(tlist(:finish), after, comp, err_list)
    
  END SUBROUTINE ParseLanguageBindingSpec
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /language-binding-spec/ (R508).
  !!
  !! A /language-binding-spec/ is syntactically the same as a 
  !! /proc-language-binding-spec/, but the latter has more constraints.
  !!
  !! @param[in]     tlist             The sequence of tokens for the enclosing 
  !! statement.
  !!
  !! @param[in]     start             The index of the token following the 
  !! BIND keyword.  This should be an opening parenthesis.
  !!
  !! @param[out]    ls                The language specification.
  !!
  !! @param[out]    after             Index of the token following the 
  !! /language-binding-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE Parse_progressive(tlist, start, ls, after, err_list)
    
    USE Errors
    USE Expressions
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(LangSpec), INTENT(OUT) :: ls
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Component for error reporting.
    CHARACTER(*), PARAMETER :: comp = lbs_comp
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error code from Expect.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing delimiter.
    
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Opening parentheses.
    CALL Expect(tlist, start, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Language spec
    
    CALL ExpectKeyword(tlist, start + 1, scck_'C', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! A language of C, which is our only option.
    ls%lang = 1
    
    ! Comma or closing parentheses.
    CALL Expect(tlist, start + 2, [ittComma, ittDelimiter],  &
        [scck_',', scck_')'], opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! Name keyword.
      CALL ExpectKeyword(tlist, start + 3, scck_'NAME', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! =
      CALL Expect(tlist, start + 4, ittAssign, scck_'=', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! The name expression.  This runs up to the terminating ).  Finding 
      ! a terminating ) guarantees that tlist is at least of size 6.
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ParseExpr(  &
          tlist,  &
          start + 5, closing - 1,  &
          ls%name,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
      
      after = closing + 1
    ELSE
      after = start + 3
    END IF
    
  END SUBROUTINE Parse_progressive
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /proc-language-binding-spec/ 
  !! (R1231).
  !!
  !! A /language-binding-spec/ is syntactically the same as a 
  !! /proc-language-binding-spec/, but the latter has more constraints.
  !!
  !! @param[in]     tlist             The sequence of tokens for the enclosing 
  !! statement.
  !!
  !! @param[in]     start             The index of the token that is the 
  !! BIND keyword.
  !!
  !! @param[in]     finish            The index of the token that ends the 
  !! /language-binding-spec/.  This would normally be the closing parenthesis.
  !!
  !! @param[out]    ls                The language specification.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This is simply a wrapper around ParseLanguageBindingSpec that sets 
  !! an additional component level in any error messages.
  
  SUBROUTINE ParseProcLanguageBindingSpec(tlist, start, finish, ls, err_list)
    
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(LangSpec), INTENT(OUT) :: ls
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /proc-language-binding-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R2131'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    CALL ParseLanguageBindingSpec(tlist, start, finish, ls, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseProcLanguageBindingSpec
  
  
  !*****************************************************************************
  !!
  !> Write a text representation of a LangSpec to a logical unit.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]    ls                 The LangSpec to write out.
  !!
  !! Uses non-advancing output.
  
  SUBROUTINE WriteLangSpec(unit, ls)
    
    USE Expressions
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    TYPE(LangSpec), INTENT(IN) :: ls
    
    !***************************************************************************
    
    WRITE (unit, "('language = ')", ADVANCE='NO')
    
    SELECT CASE (ls%lang)
    CASE (1)      ; WRITE (unit, "(A)", ADVANCE='NO') 'C'
    CASE DEFAULT
      STOP 'Internal error in LanguageBindingSpecs%%WriteLangSpec, &
          &unknown language specifier.'
    END SELECT 
    
    IF (ALLOCATED(ls%name)) THEN
      WRITE (unit, "(', name = ')", ADVANCE='NO')
      CALL WriteExpr(unit, ls%name)
    END IF
    
  END SUBROUTINE WriteLangSpec
  
  
  !*****************************************************************************
  !!
  !> Implementation of LangSpec%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE ls_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(LangSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL WriteLangSpec(unit, pnode)
    
  END SUBROUTINE ls_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of LangSpec%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE ls_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(LangSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%name)) CALL pnode%name%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ls_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of LangSpec%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ls_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(LangSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%name)) CALL pnode%name%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ls_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of LangSpec%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'language-binding-spec'.
  
  FUNCTION ls_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'language-binding-spec'
    
  END FUNCTION ls_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of LangSpec%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ls_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(LangSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = pnode%location
    
  END FUNCTION ls_GetLocation
  
END MODULE LanguageBindingSpecs
