! $Id: LockStmts.f90 2923 2019-04-26 23:01:54Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SyncStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /lock-stmt/'s and /unlock-stmt/'s.
!!
!! Relies on /sync-stat/ parsing code in the SyncStmts module.

MODULE LockStmts
  
  USE Expressions
  USE SyncStmts
  USE Variables
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseLockStmt
  PUBLIC :: ParseUnlockStmt
  
  !> The results of parsing a /lock-stmt/.
  TYPE, EXTENDS(SyncStat), PUBLIC :: LockStmt
    !> The /scalar-logical-variable/ for the ACQUIRED_LOCK specified, 
    !! if present, not allocated if not.
    CLASS(Variable), ALLOCATABLE :: acquired_lock
    !> The results of parsing the lock variable.
    CLASS(Variable), ALLOCATABLE :: lock_variable
  CONTAINS
    PROCEDURE :: Dump => lock_Dump
    PROCEDURE :: Query => lock_Query
    PROCEDURE :: Modify => lock_Modify
    PROCEDURE, NOPASS :: What => lock_What
  END TYPE LockStmt
  
  !> The results of parsing an /unlock-stmt/.
  TYPE, EXTENDS(SyncStat), PUBLIC :: UnlockStmt
    !> The results of parsing the lock variable.
    CLASS(Variable), ALLOCATABLE :: lock_variable
  CONTAINS
    PROCEDURE :: Dump => unlock_Dump
    PROCEDURE :: Query => unlock_Query
    PROCEDURE :: Modify => unlock_Modify
    PROCEDURE, NOPASS :: What => unlock_What
  END TYPE UnlockStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a /lock-stmt/ (R863).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    lock_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseLockStmt(tlist, lock_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(LockStmt), INTENT(OUT) :: lock_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /lock-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R864'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the /lock-stat-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: got_variable   ! Flag for the lock-variable.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /lock-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! LOCK ...
    
    CALL ExpectKeyword(tlist, 1, scck_'LOCK', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ( ... )
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /lock-variable/ [ , /lock-stat-list/ ]
    
    got_variable = .FALSE.
    list_count = ListItemCount(tlist(3:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      IF (.NOT. got_variable) THEN
        CALL ParseVariable(  &
            tlist,  &
            item_start, item_finish,  &
            lock_stmt%lock_variable,  &
            sub_err_list )
        got_variable = .TRUE.
      ELSE
        CALL parse_lock_stat(  &
            tlist,  &
            item_start, item_finish,  &
            lock_stmt,  &
            sub_err_list )
      END IF
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      ! Continue on error.
    END DO
    
    ! Nothing else!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseLockStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a /sync-all-stmt/ (R858).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[out]    lock_variable     The results of parsing the lock 
  !! variable.
  !!
  !! @param[out]    sync__all_stmt    The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseUnlockStmt(tlist, unlock_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(UnlockStmt), INTENT(OUT) :: unlock_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /sync-images-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R860'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    INTEGER :: list_count     ! Number of items in the /connect-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    LOGICAL :: got_variable   ! Flag for the lock-variable.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list, CODE=errSyntax, COMPONENT=comp,  &
          MSG='Expected /sync-images-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! UNLOCK ...
    
    CALL ExpectKeyword(tlist, 1, scck_'UNLOCK', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ... ( /image-set/ ... )
    
    CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    got_variable = .FALSE.
    
    list_count = ListItemCount(tlist(3:closing-1))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 3, closing - 1, item_start, item_finish)
      IF (.NOT. got_variable) THEN
        CALL ParseVariable(  &
            tlist,  &
            item_start, item_finish,  &
            unlock_stmt%lock_variable,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        got_variable = .TRUE.
      ELSE
        CALL ParseSyncStat(  &
            tlist,  &
            item_start, item_finish,  &
            unlock_stmt,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      ! Continue on error.
    END DO
    
    ! Nothing else!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseUnlockStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /lock-stat/ (R859).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /sync-stat/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /sync-stat/ in @a tlist.
  !!
  !! @param[in,out] lock_stat         Results of parsing the /lock-stat/.  
  !! Note that this is an INTENT(INOUT) argument - each spec sets a 
  !! particular component
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! For control over error reporting we duplicate much of the content 
  !! of the SyncStmts::ParseSyncStat procedure.  @todo Rethink this.
  
  SUBROUTINE parse_lock_stat( tlist, start, finish, lock_stat,  &
      err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE IoControlSpecs
    USE ParseUtils
    USE Scanner
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(LockStmt), INTENT(INOUT) :: lock_stat
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /lock-stat/.
    CHARACTER(*), PARAMETER :: comp = 'R864'
    
    ! Keywords for each /lock-stat/.
    CHARACTER(*,KIND=scck), PARAMETER :: keywords(3) = [  &
        scck_'STAT         ',  &      ! Also a /sync-stat/
        scck_'ERRMSG       ',  &      ! Also a /sync-stat/
        scck_'ACQUIRED_LOCK' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /lock-stat/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! keyword ...
    
    CALL ExpectKeyword(tlist, start, keywords, '/lock-stat/', opt, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! ... = ...
    CALL Expect(tlist, start + 1, ittAssign, scck_'=', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! stat specific thing (all variables).
    
    SELECT CASE (opt)
    CASE (1)      ! STAT
      IF (ALLOCATED(lock_stat%stat)) THEN
        CALL already('STAT')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            lock_stat%stat,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (2)      ! ERRMSG
      IF (ALLOCATED(lock_stat%errmsg)) THEN
        CALL already('ERRMSG')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            lock_stat%errmsg,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE (3)      ! ACQUIRED_LOCK
      ! F08/0098 TC# adds...
      IF (ALLOCATED(lock_stat%acquired_lock)) THEN
        CALL already('ACQUIRED_LOCK')
      ELSE
        CALL ParseVariable(  &
            tlist,  &
            start + 2, finish,  &
            lock_stat%acquired_lock,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
      END IF
      
    CASE DEFAULT
      STOP 'Unknown /lock-stat/.'
      
    END SELECT
    
  CONTAINS
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C852a on ' // comp,  &
          LOCATION=QueryLocation(tlist(start)),  &
          MSG=text // ' has already been specified for this statement.' )
    END SUBROUTINE already
    
  END SUBROUTINE parse_lock_stat
  
  
  !*****************************************************************************
  !!
  !> Implementation of LockStmt%Dump - write a debugging representation 
  !! of the results of parsing a lock statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE lock_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(LockStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/sync-stat-list/ = ')", ADVANCE='NO') pfx
    CALL WriteSyncStat(unit, pnode)
    WRITE (unit, "()")
    
    IF (ALLOCATED(pnode%acquired_lock)) THEN
      WRITE (unit, "(A,'/acquired-lock/ = ')", ADVANCE='NO') pfx
      CALL WriteVariable(unit, pnode%acquired_lock)
      WRITE (unit, "()")
    END IF
    
    WRITE (unit, "(A,'/lock-variable/ = ')", ADVANCE='NO') pfx
    CALL WriteVariable(unit, pnode%lock_variable)
    WRITE (unit, "()")
    
  END SUBROUTINE lock_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of LockStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE lock_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(LockStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Query(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Query(visitor)
    IF (ALLOCATED(pnode%acquired_lock)) THEN
      CALL pnode%acquired_lock%Query(visitor)
    END IF
    IF (ALLOCATED(pnode%lock_variable)) THEN
      CALL pnode%lock_variable%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE lock_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of LockStmt%Modif - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE lock_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(LockStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Modify(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Modify(visitor)
    IF (ALLOCATED(pnode%acquired_lock)) THEN
      CALL pnode%acquired_lock%Modify(visitor)
    END IF
    IF (ALLOCATED(pnode%lock_variable)) THEN
      CALL pnode%lock_variable%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE lock_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of LockStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'lock-stmt'.
  
  FUNCTION lock_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'lock-stmt'
    
  END FUNCTION lock_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for UnlockStmt.
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnlockStmt%Dump - write a debugging representation 
  !! of the results of parsing an unlock statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE unlock_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnlockStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/sync-stat-list/ = ')", ADVANCE='NO') pfx
    CALL WriteSyncStat(unit, pnode)
    WRITE (unit, "()")
    
    WRITE (unit, "(A,'/lock-variable/ = ')", ADVANCE='NO') pfx
    CALL WriteVariable(unit, pnode%lock_variable)
    WRITE (unit, "()")
    
  END SUBROUTINE unlock_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnlockStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE unlock_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnlockStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Query(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Query(visitor)
    IF (ALLOCATED(pnode%lock_variable)) THEN
      CALL pnode%lock_variable%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE unlock_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnlockStmt%Modif - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE unlock_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UnlockStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%stat)) CALL pnode%stat%Modify(visitor)
    IF (ALLOCATED(pnode%errmsg)) CALL pnode%errmsg%Modify(visitor)
    IF (ALLOCATED(pnode%lock_variable)) THEN
      CALL pnode%lock_variable%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE unlock_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of UnlockStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'unlock-stmt'.
  
  FUNCTION unlock_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'unlock-stmt'
    
  END FUNCTION unlock_What
  
END MODULE LockStmts
