! $Id: ObfuscateImplementation.f90 2957 2019-11-24 07:26:13Z ian $
! ff08 source code copyright 2016 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ObfuscateImplementation module.


!*******************************************************************************
!!
!> Basic implementation of FF08Obfuscate.

MODULE ObfuscateImplementation
  
  USE IdentifierMapMod
  USE ParseNodes
  USE Strings
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  
  !> Parent type that encapsulates an operation that visits each identifier 
  !! in the source.
  TYPE, ABSTRACT :: identifier_modifier
  CONTAINS
    PROCEDURE(im_Modify), DEFERRED :: Modify
  END TYPE identifier_modifier
  
  
  ABSTRACT INTERFACE
    !> The interface of the Modify binding in identifier_modifier.s
    !!
    !! @param[in,out] modifier        The identifier modifier object.
    !!
    !! @param[in,out] ident           The particular identifier being visited.
    SUBROUTINE im_Modify(modifier, ident)
      USE Identifiers
      IMPORT :: identifier_modifier
      IMPLICIT NONE
      !-------------------------------------------------------------------------
      CLASS(identifier_modifier), INTENT(INOUT) :: modifier
      TYPE(Identifier), INTENT(INOUT) :: ident
    END SUBROUTINE im_Modify
  END INTERFACE
  
  !> A collection of strings.
  TYPE :: string_list
    TYPE(String), ALLOCATABLE :: items(:)
  END TYPE string_list
  
  INTERFACE string_list
    MODULE PROCEDURE string_list_char
  END INTERFACE string_list
  
  !> An identifier modifier that obfuscates the identifiers.
  TYPE, EXTENDS(identifier_modifier) :: obfuscator
    !> The map from original identifier to obfuscated identifier.
    TYPE(IdentifierMap) :: forward
    !> The map from obfuscated identifier to original.
    TYPE(IdentifierMap) :: reverse
    !> Current identifier number
    INTEGER :: num = 0
  CONTAINS
    PROCEDURE :: Modify => obfuscator_Modify
  END TYPE obfuscator
  
  !> The parse node visitor that carries out the parse node modification.
  TYPE, EXTENDS(ParseNodeModifyVisitor) :: identifier_visitor
    !> The identifier visitor.
    CLASS(identifier_modifier), ALLOCATABLE :: modifier
    !> Stack of names that suppress keyword renaming.
    !!
    !! The stack is added to when renames of known compiler supplied 
    !! procedures are detected.
    TYPE(string_list), ALLOCATABLE :: name_stack(:)
  CONTAINS
    PROCEDURE :: Start => iv_Start
  END TYPE identifier_visitor
  
  !> The list of generic intrinsic procedure names.
  !!
  !! This is used to suppress keyword renaming and to suppress renaming of 
  !! the procedure identifier.
  CHARACTER(*), PARAMETER :: intrinsic_generic_names(173) = [  &
      'ABS                           ', 'ACHAR                         ',  &
      'ACOS                          ', 'ACOSH                         ',  &
      'ADJUSTL                       ', 'ADJUSTR                       ',  &
      'AIMAG                         ', 'AINT                          ',  &
      'ALL                           ', 'ALLOCATED                     ',  &
      'ANINT                         ', 'ANY                           ',  &
      'ASIN                          ', 'ASINH                         ',  &
      'ASSOCIATED                    ', 'ATAN                          ',  &
      'ATAN2                         ', 'ATANH                         ',  &
      'ATOMIC_DEFINE                 ', 'ATOMIC_REF                    ',  &
      'BESSEL_J0                     ', 'BESSEL_J1                     ',  &
      'BESSEL_JN                     ', 'BESSEL_Y0                     ',  &
      'BESSEL_Y1                     ', 'BESSEL_YN                     ',  &
      'BGE                           ', 'BGT                           ',  &
      'BLE                           ', 'BLT                           ',  &
      'BIT_SIZE                      ', 'BTEST                         ',  &
      'CEILING                       ', 'CHAR                          ',  &
      'CMPLX                         ', 'COMMAND_ARGUMENT_COUNT        ',  &
      'CONJG                         ', 'COS                           ',  &
      'COSH                          ', 'COUNT                         ',  &
      'CPU_TIME                      ', 'CSHIFT                        ',  &
      'DATE_AND_TIME                 ', 'DBLE                          ',  &
      'DIGITS                        ', 'DIM                           ',  &
      'DOT_PRODUCT                   ', 'DPROD                         ',  &
      'DSHIFTL                       ', 'DSHIFTR                       ',  &
      'EOSHIFT                       ', 'EPSILON                       ',  &
      'ERF                           ', 'ERFC                          ',  &
      'ERFC_SCALED                   ', 'EXECUTE_COMMAND_LINE          ',  &
      'EXP                           ', 'EXPONENT                      ',  &
      'EXTENDS_TYPE_OF               ', 'FINDLOC                       ',  &
      'FLOOR                         ', 'FRACTION                      ',  &
      'GAMMA                         ', 'GET_COMMAND                   ',  &
      'GET_COMMAND_ARGUMENT          ', 'GET_ENVIRONMENT_VARIABLE      ',  &
      'HUGE                          ', 'HYPOT                         ',  &
      'IACHAR                        ', 'IALL                          ',  &
      'IAND                          ', 'IANY                          ',  &
      'IBCLR                         ', 'IBITS                         ',  &
      'IBSET                         ', 'ICHAR                         ',  &
      'IEOR                          ', 'IMAGE_INDEX                   ',  &
      'INDEX                         ', 'INT                           ',  &
      'IOR                           ', 'IPARITY                       ',  &
      'ISHFT                         ', 'ISHFTC                        ',  &
      'IS_CONTIGUOUS                 ', 'IS_IOSTAT_END                 ',  &
      'IS_IOSTAT_EOR                 ', 'KIND                          ',  &
      'LBOUND                        ', 'LCOBOUND                      ',  &
      'LEADZ                         ', 'LEN                           ',  &
      'LEN_TRIM                      ', 'LGE                           ',  &
      'LGT                           ', 'LLE                           ',  &
      'LLT                           ', 'LOG                           ',  &
      'LOG_GAMMA                     ', 'LOG10                         ',  &
      'LOGICAL                       ', 'MASKL                         ',  &
      'MASKR                         ', 'MATMUL                        ',  &
      'MAX                           ', 'MAXEXPONENT                   ',  &
      'MAXLOC                        ', 'MAXVAL                        ',  &
      'MERGE                         ', 'MERGE_BITS                    ',  &
      'MIN                           ', 'MINEXPONENT                   ',  &
      'MINLOC                        ', 'MINVAL                        ',  &
      'MOD                           ', 'MODULO                        ',  &
      'MOVE_ALLOC                    ', 'MVBITS                        ',  &
      'NEAREST                       ', 'NEW_LINE                      ',  &
      'NINT                          ', 'NORM2                         ',  &
      'NOT                           ', 'NULL                          ',  &
      'NUM_IMAGES                    ', 'PACK                          ',  &
      'PARITY                        ', 'POPCNT                        ',  &
      'POPPAR                        ', 'PRECISION                     ',  &
      'PRESENT                       ', 'PRODUCT                       ',  &
      'RADIX                         ', 'RANDOM_NUMBER                 ',  &
      'RANDOM_SEED                   ', 'RANGE                         ',  &
      'REAL                          ', 'REPEAT                        ',  &
      'RESHAPE                       ', 'RRSPACING                     ',  &
      'SAME_TYPE_AS                  ', 'SCALE                         ',  &
      'SCAN                          ', 'SELECTED_CHAR_KIND            ',  &
      'SELECTED_INT_KIND             ', 'SELECTED_REAL_KIND            ',  &
      'SET_EXPONENT                  ', 'SHAPE                         ',  &
      'SHIFTA                        ', 'SHIFTL                        ',  &
      'SHIFTR                        ', 'SIGN                          ',  &
      'SIN                           ', 'SINH                          ',  &
      'SIZE                          ', 'SPACING                       ',  &
      'SPREAD                        ', 'SQRT                          ',  &
      'STORAGE_SIZE                  ', 'SUM                           ',  &
      'SYSTEM_CLOCK                  ', 'TAN                           ',  &
      'TANH                          ', 'THIS_IMAGE                    ',  &
      'TINY                          ', 'TRAILZ                        ',  &
      'TRANSFER                      ', 'TRANSPOSE                     ',  &
      'TRIM                          ', 'UBOUND                        ',  &
      'UCOBOUND                      ', 'UNPACK                        ',  &
      'VERIFY                        ' ]
  
  !> List of the intrinsic specific procedure names.
  !! 
  !! This is used to suppress keyword renaming and to suppress renaming of 
  !! the procedure identifier.
  !!
  !! This list includes specific names that are the same as the relevant 
  !! generic.
  CHARACTER(*), PARAMETER :: intrinsic_specific_names(80) = [  &
      'ABS                           ', 'ACOS                          ',  &
      'AIMAG                         ', 'AINT                          ',  &
      'ALOG                          ', 'ALOG10                        ',  &
      'AMAX0                         ', 'AMAX1                         ',  &
      'AMIN0                         ', 'AMIN1                         ',  &
      'AMOD                          ', 'ANINT                         ',  &
      'ASIN                          ', 'ATAN                          ',  &
      'ATAN2                         ', 'CABS                          ',  &
      'CCOS                          ', 'CEXP                          ',  &
      'CHAR                          ', 'CLOG                          ',  &
      'CONJG                         ', 'COS                           ',  &
      'COSH                          ', 'CSIN                          ',  &
      'CSQRT                         ', 'DABS                          ',  &
      'DACOS                         ', 'DASIN                         ',  &
      'DATAN                         ', 'DATAN2                        ',  &
      'DCOS                          ', 'DCOSH                         ',  &
      'DDIM                          ', 'DEXP                          ',  &
      'DIM                           ', 'DINT                          ',  &
      'DINT                          ', 'DLOG                          ',  &
      'DLOG10                        ', 'DMAX1                         ',  &
      'DMIN1                         ', 'DMOD                          ',  &
      'DNINT                         ', 'DPROD                         ',  &
      'DSIGN                         ', 'DSIN                          ',  &
      'DSINH                         ', 'DSQRT                         ',  &
      'DTAN                          ', 'DTANH                         ',  &
      'EXP                           ', 'FLOAT                         ',  &
      'IABS                          ', 'ICHAR                         ',  &
      'IDIM                          ', 'IDINT                         ',  &
      'IDNINT                        ', 'IFIX                          ',  &
      'INDEX                         ', 'INT                           ',  &
      'ISIGN                         ', 'LEN                           ',  &
      'LGE                           ', 'LGT                           ',  &
      'LLE                           ', 'LLT                           ',  &
      'MAX0                          ', 'MAX1                          ',  &
      'MIN0                          ', 'MIN1                          ',  &
      'MOD                           ', 'NINT                          ',  &
      'REAL                          ', 'SIGN                          ',  &
      'SIN                           ', 'SINH                          ',  &
      'SNGL                          ', 'SQRT                          ',  &
      'TAN                           ', 'TANH                          ' ]
  
  !> List of procedure names from ISO_FORTRAN_ENV
  !! 
  !! This is used to suppress keyword renaming and to suppress renaming of 
  !! the procedure identifier.
  CHARACTER(*), PARAMETER :: isofortran_procedure_names(2) = [  &
      'COMPILER_OPTIONS              ', 'COMPILER_VERSION              ' ]
  
  !> List of procedure names from IEEE_* intrinsic modules.
  !! 
  !! This is used to suppress keyword renaming and to suppress renaming of 
  !! the procedure identifier.
  CHARACTER(*), PARAMETER :: ieee_procedure_names(36) = [  &
      'IEEE_CLASS                    ', 'IEEE_COPY_SIGN                ',  &
      'IEEE_GET_ROUNDING_MODE        ', 'IEEE_GET_UNDERFLOW_MODE       ',  &
      'IEEE_IS_FINITE                ', 'IEEE_IS_NAN                   ',  &
      'IEEE_IS_NEGATIVE              ', 'IEEE_IS_NORMAL                ',  &
      'IEEE_LOGB                     ', 'IEEE_NEXT_AFTER               ',  &
      'IEEE_REM                      ', 'IEEE_RINT                     ',  &
      'IEEE_SCALB                    ', 'IEEE_SELECTED_REAL_KIND       ',  &
      'IEEE_SET_ROUNDING_MODE        ', 'IEEE_SET_UNDERFLOW_MODE       ',  &
      'IEEE_SUPPORT_DATATYPE         ', 'IEEE_SUPPOR_DENORMAL          ',  &
      'IEEE_SUPPORT_DIVIDE           ', 'IEEE_SUPPORT_INF              ',  &
      'IEEE_SUPPORT_IO               ', 'IEEE_SUPPORT_NAN              ',  &
      'IEEE_SUPPORT_ROUNDING         ', 'IEEE_SUPPORT_SQRT             ',  &
      'IEEE_SUPPORT_STANDARD         ', 'IEEE_SUPPORT_UNDERFLOW_CONTROL',  &
      'IEEE_UNORDERED                ', 'IEEE_VALUE                    ',  &
      'IEEE_GET_FLAG                 ', 'IEEE_GET_HALTING_MODE         ',  &
      'IEEE_GET_STATUS               ', 'IEEE_SET_FLAG                 ',  &
      'IEEE_SET_HALTING_MODE         ', 'IEEE_SET_STATUS               ',  &
      'IEEE_SUPPORT_FLAG             ', 'IEEE_SUPPORT_HALTING          ' ]
      
  !> List of procedure names from ISO_C_BINDING.
  !! 
  !! This is used to suppress keyword renaming and to suppress renaming of 
  !! the procedure identifier.
  CHARACTER(*), PARAMETER :: isocbinding_procedure_names(6) = [  &
      'C_ASSOCIATED                  ', 'C_F_POINTER                   ',  &
      'C_F_PROC_POINTER              ', 'C_FUNLOC                      ',  &
      'C_LOC                         ', 'C_SIZEOF                      ' ]
      
CONTAINS
  
  !*****************************************************************************
  !!
  !> Render an obfuscated variant of a parse tree.
  !!
  !! @param[in]     program_pnode     The program node at the top of the 
  !! parse tree.
  !!
  !! @param[in]     out_unit          Unit for output of the obfuscated code.  
  !! Must be connected for formatted output.
  
  SUBROUTINE Obfuscate(program_pnode, out_unit)
    
    USE Formatters
    USE Programs
    USE RenderNodes
    USE RenderParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Program), INTENT(INOUT) :: program_pnode
    INTEGER, INTENT(IN) :: out_unit
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The obfuscator object.
    TYPE(obfuscator), ALLOCATABLE :: ob
    
    ! The object that obfuscates the identifiers in each parse node.
    TYPE(identifier_visitor) :: visitor
    
    ! The options for the rendering to tokens.
    TYPE(RenderOptions) :: opts
    
    ! The rendered tokens.
    CLASS(RenderNode), ALLOCATABLE :: program_rnode
    
    !***************************************************************************
    
    ! Set up the identifier modifier object, adding intrinsic identifiers 
    ! (and similar) to the mapping.
    ALLOCATE(ob)
    CALL add_intrinsics(ob%forward, ob%reverse)
    
    ! Set up the obfuscator parse node visitor.  The visitor is given the 
    ! list of intrinsic-like procedures that must not have their 
    ! argument keywords mangled.  This is a subset of the list of the list 
    ! of identity mappings added to the identifier modifier object - it 
    ! doesn't include data objects.
    CALL MOVE_ALLOC(ob, visitor%modifier)
    ALLOCATE(visitor%name_stack(1))
    visitor%name_stack(1) = string_list(intrinsic_generic_names)
    visitor%name_stack(1) = string_list_union(  &
        visitor%name_stack(1),  &
        string_list(intrinsic_specific_names) )
    visitor%name_stack(1) = string_list_union(  &
        visitor%name_stack(1),  &
        string_list(isofortran_procedure_names) )
    visitor%name_stack(1) = string_list_union(  &
        visitor%name_stack(1),  &
        string_list(ieee_procedure_names) )
    visitor%name_stack(1) = string_list_union(  &
        visitor%name_stack(1),  &
        string_list(isocbinding_procedure_names) )
    
    ! Modify the identifiers in the parse tree.
    CALL program_pnode%Modify(visitor)
    
    ! Convert the modified parse tree back into a sequence of render tokens.
    CALL RenderParseTree(program_pnode, program_rnode, opts)
    
    ! Output the render tokens to the console.
    CALL FormatFreeForm(  &
        out_unit,  &
        program_rnode,  &
        FormatFreeFormOptions() )
    
  END SUBROUTINE Obfuscate
  
  
  !*****************************************************************************
  !!
  !> Create a name_list object from a character array.
  !!
  !! @param[in]     array             The character array.
  !!
  !! @returns A string list of the trimmed elements of @a array.
  
  FUNCTION string_list_char(array) RESULT(sl)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*), INTENT(IN) :: array(:)
    
    ! Function result.
    TYPE(string_list) :: sl
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! @a array index.
    
    !***************************************************************************
    
    ALLOCATE(sl%items(SIZE(array)))
    DO i = 1, SIZE(array)
      sl%items(i) = String(TRIM(array(i)))
    END DO
    
  END FUNCTION string_list_char
  
  
  !*****************************************************************************
  !!
  !> Return the union of two string lists.
  !!
  !! @param[in]     left              The left string list.  All items in 
  !! this list are assumed to be unique and are copied to the result.
  !!
  !! @param[in]     right             The right string list.  All items 
  !! this list that are not already in @a left are copied to the result.
  !!
  !! @returns The union of the items in @a left and @a right.
  
  FUNCTION string_list_union(left, right) RESULT(sl)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(string_list), INTENT(IN) :: left
    TYPE(string_list), INTENT(IN) :: right
    
    ! Function result.
    TYPE(string_list) :: sl
  
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Array index.
    
    !***************************************************************************
    
    sl = left
    DO i = 1, SIZE(right%items)
      CALL AppendIfUnique(sl%items, right%items(i))
    END DO
    
  END FUNCTION string_list_union
  
  
  !*****************************************************************************
  !!
  !> Test whether a character value is in a list of strings.
  !!
  !! @param[in]     sl                The string list to test against.
  !!
  !! @param[in]     str               The string to test.
  !!
  !! @returns True if @a str is in @a sl, false otherwise.
  
  FUNCTION in_list(sl, str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(string_list), INTENT(IN) :: sl
    CHARACTER(*), INTENT(IN) :: str
    
    ! Function result.
    LOGICAL :: in_list
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! items index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(sl%items)
      in_list = sl%items(i)%item == str
      IF (in_list) RETURN
    END DO
    
    in_list = .FALSE.
    
  END FUNCTION 
  
  
  !*****************************************************************************
  !!
  !> Add intrinsic identifiers to the map as identity transformations.
  
  SUBROUTINE add_intrinsics(forward, reverse)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierMap), INTENT(INOUT) :: forward
    TYPE(IdentifierMap), INTENT(INOUT) :: reverse
    
    !---------------------------------------------------------------------------
    ! Locals
    
    CHARACTER(*), PARAMETER :: intrinsic_type_names(5) = [  &
        'INTEGER                     ',  &
        'REAL                        ',  &
        'COMPLEX                     ',  &
        'LOGICAL                     ',  &
        'CHARACTER                   ' ]
    
    CHARACTER(*), PARAMETER :: intrinsic_operators(23) = [  &
        '+                           ', '-                           ',  &
        '*                           ', '/                           ',  &
        '**                          ', '//                          ',  &
        '==                          ', '/=                          ',  &
        '>                           ', '<                           ',  &
        '>=                          ', '<=                          ',  &
        '.EQ.                        ', '.NE.                        ',  &
        '.GT.                        ', '.LT.                        ',  &
        '.GE.                        ', '.LE.                        ',  &
        '.EQV.                       ', '.NEQV.                      ',  &
        '.AND.                       ', '.OR.                        ',  &
        '.NOT.                       ' ]
    
    CHARACTER(*), PARAMETER :: intrinsic_modules(5) = [  &
        'ISO_FORTRAN_ENV             ', 'ISO_C_BINDING               ',  &
        'IEEE_ARITHMETIC             ', 'IEEE_FEATURES               ',  &
        'IEEE_EXCEPTIONS             ' ]
    
    ! Does not include the intrinsic module procedures.
    CHARACTER(*), PARAMETER :: intrinsic_module_entities(106) = [  &
        'ATOMIC_INT_KIND             ', 'ATOMIC_LOGICAL_KIND         ',  &
        'CHARACTER_KINDS             ', 'CHARACTER_STORAGE_SIZES     ',  &
        'ERROR_UNIT                  ', 'FILE_STORAGE_SIZE           ',  &
        'INPUT_UNIT                  ', 'INT8                        ',  &
        'INT16                       ', 'INT32                       ',  &
        'INT64                       ', 'INTEGER_KINDS               ',  &
        'IOSTAT_END                  ', 'IOSTAT_EOR                  ',  &
        'IOSTAT_INQUIRE_INTERNAL_UNIT', 'LOCK_TYPE                   ',  &
        'LOGICAL_KINDS               ', 'NUMERIC_STORAGE_SIZE        ',  &
        'OUTPUT_UNIT                 ', 'REAL_KINDS                  ',  &
        'REAL32                      ', 'REAL64                      ',  &
        'REAL128                     ', 'STAT_LOCKED                 ',  &
        'STAT_LOCKED_OTHER_IMAGE     ', 'STAT_STOPPED_IMAGE          ',  &
        'STAT_UNLOCKED               ', 'IEEE_FLAG_TYPE              ',  &
        'IEEE_INVALID                ', 'IEEE_OVERFLOW               ',  &
        'IEEE_DIVIDE_BY_ZERO         ', 'IEEE_UNDERFLOW              ',  &
        'IEEE_INEXACT                ', 'IEEE_USUAL                  ',  &
        'IEEE_ALL                    ', 'IEEE_STATUS_TYPE            ',  &
        'IEEE_CLASS_TYPE             ', 'IEEE_SIGNALING_NAN          ',  &
        'IEEE_QUIET_NAN              ', 'IEEE_NEGATIVE_INF           ',  &
        'IEEE_NEGATIVE_NORMAL        ', 'IEEE_NEGATIVE_DENORMAL      ',  &
        'IEEE_NEGATIVE_ZERO          ', 'IEEE_POSITIVE_ZERO          ',  &
        'IEEE_POSITIVE_DENORMAL      ', 'IEEE_POSITIVE_NORMAL        ',  &
        'IEEE_POSITIVE_INF           ', 'IEEE_OTHER_VALUE            ',  &
        'IEEE_ROUND_TYPE             ', 'IEEE_NEAREST                ',  &
        'IEEE_TO_ZERO                ', 'IEEE_UP                     ',  &
        'IEEE_DOWN                   ', 'IEEE_OTHER                  ',  &
        'IEEE_FEATURES_TYPE          ', 'IEEE_DATATYPE               ',  &
        'IEEE_DENORMAL               ', 'IEEE_DIVIDE                 ',  &
        'IEEE_HALTING                ', 'IEEE_INEXACT_FLAG           ',  &
        'IEEE_INF                    ', 'IEEE_INVALID_FLAG           ',  &
        'IEEE_NAN                    ', 'IEEE_ROUNDING               ',  &
        'IEEE_SQRT                   ', 'IEEE_UNDERFLOW              ',  &
        'C_INT                       ', 'C_SHORT                     ',  &
        'C_LONG                      ', 'C_LONG_LONG                 ',  &
        'C_SIGNED_CHAR               ', 'C_SIZE_T                    ',  &
        'C_INT8_T                    ', 'C_INT16_T                   ',  &
        'C_INT32_T                   ', 'C_INT64_T                   ',  &
        'C_INT_LEAST8_T              ', 'C_INT_LEAST16_T             ',  &
        'C_INT_LEAST32_T             ', 'C_INT_LEAST64_T             ',  &
        'C_INT_FAST8_T               ', 'C_INT_FAST16_T              ',  &
        'C_INT_FAST32_T              ', 'C_INT_FAST64_T              ',  &
        'C_INTMAX_T                  ', 'C_INTPTR_T                  ',  &
        'C_FLOAT                     ', 'C_DOUBLE                    ',  &
        'C_LONG_DOUBLE               ', 'C_FLOAT_COMPLEX             ',  &
        'C_DOUBLE_COMPLEX            ', 'C_LONG_DOUBLE_COMPLEX       ',  &
        'C_BOOL                      ', 'C_CHAR                      ',  &
        'C_NULL_CHAR                 ', 'C_ALERT                     ',  &
        'C_BACKSPACE                 ', 'C_FORM_FEED                 ',  &
        'C_NEW_LINE                  ', 'C_CARRIAGE_RETURN           ',  &
        'C_HORIZONTAL_TAB            ', 'C_VERTICAL_TAB              ',  &
        'C_PTR                       ', 'C_FUNPTR                    ',  &
        'C_NULL_PTR                  ', 'C_NULL_FUNPTR               ' ]
    
    INTEGER :: i              ! Name index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(intrinsic_type_names)
      CALL add_self(forward, reverse, TRIM(intrinsic_type_names(i)))
    END DO
    
    DO i = 1, SIZE(intrinsic_generic_names)
      CALL add_self(forward, reverse, TRIM(intrinsic_generic_names(i)))
    END DO
    
    DO i = 1, SIZE(intrinsic_specific_names)
      CALL add_self(forward, reverse, TRIM(intrinsic_specific_names(i)))
    END DO
    
    DO i = 1, SIZE(intrinsic_operators)
      CALL add_self(forward, reverse, TRIM(intrinsic_operators(i)))
    END DO
    
    DO i = 1, SIZE(intrinsic_modules)
      CALL add_self(forward, reverse, TRIM(intrinsic_modules(i)))
    END DO
    
    DO i = 1, SIZE(intrinsic_module_entities)
      CALL add_self(forward, reverse, TRIM(intrinsic_module_entities(i)))
    END DO
    
  END SUBROUTINE add_intrinsics
  
  
  !*****************************************************************************
  !!
  !> Add an identity identifier mapping to the forward and reverse maps.
  !!
  !! @param[in,out] forward           The forward mapping.
  !!
  !! @param[in,out] reverse           The reverse mapping.
  !!
  !! @param[in]     name              The name to add.
  
  SUBROUTINE add_self(forward, reverse, name)
    
    USE Strings
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierMap), INTENT(INOUT) :: forward
    TYPE(IdentifierMap), INTENT(INOUT) :: reverse
    CHARACTER(*), INTENT(IN) :: name
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The name as a string.
    TYPE(String) :: name_str
    
    ! Existing value for the name in the forward mapping.
    TYPE(String), POINTER :: value
    
    !***************************************************************************
    
    name_str = String(name)
    
    ! Don't add a mapping that already exists.
    CALL Get(forward, name_str, value)
    IF (ASSOCIATED(value)) RETURN
    
    CALL Insert(forward, name_str, name_str)
    CALL Insert(reverse, name_str, name_str)
    
  END SUBROUTINE add_self
  
  
  !*****************************************************************************
  !!
  !> Carry out the modification action on the identifiers that are not 
  !! in child parse nodes.
  
  SUBROUTINE iv_Start(visitor, pnode)
    
    USE AllocatableStmts
    USE AllocateStmts
    USE ArrayConstructors
    USE AssociateConstructs
    USE AsynchronousStmts
    USE BindStmts
    USE BlockConstructs
    USE BlockDataStmts
    USE CallStmts
    USE CaseConstructs
    USE CodimensionStmts
    USE CommonStmts
    USE ComponentDefStmts
    USE ContiguousStmts
    USE CriticalConstructs
    USE DataStmts
    USE DeclarationTypeSpecs
    USE DerivedTypeStmts
    USE Designators
    USE DimensionStmts
    USE DoConstructs
    USE DoStmts
    USE EntryStmts
    USE EnumDefStmts
    USE EquivalenceStmts
    USE ExitStmts
    USE Expressions
    USE ExternalStmts
    USE FinalProcedureStmts
    USE ForallConstructs
    USE ForallStmts
    USE FunctionStmts
    USE GenericSpecs
    USE ImpliedDos
    USE ImportStmts
    USE InputOutputItems
    USE IntentStmts
    USE InterfaceBlocks
    USE IntrinsicStmts
    USE IoControlSpecs
    USE ModuleStmts
    USE MpSubprogramStmts
    USE NamelistStmts
    USE OptionalStmts
    USE ParameterStmts
    USE PointerStmts
    USE PrimaryExpressions
    USE ProcedureDeclarationStmts
    USE ProgramUnits
    USE ProgramStmts
    USE ProtectedStmts
    USE SaveStmts
    USE SelectTypeConstructs
    USE StmtFunctionStmts
    USE StructureConstructors
    USE SubmoduleStmts
    USE SubroutineStmts
    USE TargetStmts
    USE TypeBoundGenericStmts
    USE TypeBoundProcedureStmts
    USE TypeDeclarationStmts
    USE TypeParamDefStmts
    USE TypeParamInquiries
    USE UseStmts
    USE ValueStmts
    USE VolatileStmts
    USE WhereConstructs
    
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(identifier_visitor), INTENT(INOUT) :: visitor
    CLASS(ParseNode), INTENT(INOUT) :: pnode
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Miscellaneous index.
    INTEGER :: i2             ! Nested miscellaneous index.
    LOGICAL :: is_intrinsic   ! Intrinsic procedure call flag.
    
    ! Token kind text.
    CHARACTER(:), ALLOCATABLE :: text
    
    ! Temporary identifier for processing tokens.
    TYPE(Identifier) :: tmp
    
    !***************************************************************************
    
    ! Prefer non-polymorphic type guards in the following, to avoid issues 
    ! with components added in extensions being ignored.
    
    SELECT TYPE (pnode)
    TYPE IS (ACImpliedDoControl)
      CALL visitor%modifier%Modify(pnode%do_variable)
      
    TYPE IS (AllocatableDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (AllocateObjectVariable)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (AssociateConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      DO i = 1, SIZE(pnode%associations)
        CALL visitor%modifier%Modify(pnode%associations(i)%name)
      END DO
      
    TYPE IS (AsynchronousStmt)
      DO i = 1, SIZE(pnode%object_name_list)
        CALL visitor%modifier%Modify(pnode%object_name_list(i))
      END DO
      
    TYPE IS (BinaryOpExpr)
      CALL visitor%modifier%Modify(pnode%op)
      
    TYPE IS (BindStmt)
      ! If we have a single entity and no binding name specification, 
      ! then add a binding name specification.
      IF ( (SIZE(pnode%entities) == 1)  &
          .AND. (.NOT. ALLOCATED(pnode%ls%name)) ) THEN
        CALL set_binding_name(pnode%ls, pnode%entities(1)%name)
      END IF
      DO i = 1, SIZE(pnode%entities)
        CALL visitor%modifier%Modify(pnode%entities(i)%name)
      END DO
      
    TYPE IS (BlockConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (BlockDataStmt)
      IF (ALLOCATED(pnode%block_data_name)) THEN
        CALL visitor%modifier%Modify(pnode%block_data_name)
      END IF
      
    TYPE IS (CallStmt)
      ! Test whether something is intrinsic before we mangle it.
      is_intrinsic = test_is_intrinsic(visitor, pnode%name)
      CALL visitor%modifier%Modify(pnode%name)
      ! If the function is intrinsic-like, we don't mangle the argument 
      ! keywords.
      IF (.NOT. is_intrinsic) THEN
        DO i = 1, SIZE(pnode%args)
          IF (ALLOCATED(pnode%args(i)%keyword)) THEN
            CALL visitor%modifier%Modify(pnode%args(i)%keyword)
          END IF
        END DO
      END IF
      
    TYPE IS (CaseConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (CodimensionDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (CommonBlockSet)
      IF (ALLOCATED(pnode%name)) THEN
        CALL visitor%modifier%Modify(pnode%name)
      END IF
      
    TYPE IS (CommonBlockObject)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (ComponentSpec)
      IF (ALLOCATED(pnode%keyword)) THEN
        CALL visitor%modifier%Modify(pnode%keyword)
      END IF
      
    TYPE IS (ConstantExpr)
      ! Do we have a kind, and is it a name?
      IF (HasKind(pnode%tok)) THEN
        text = QueryKind(pnode%tok)
        IF ( (IACHAR(text(1:1)) >= IACHAR('A'))  &
            .AND. (IACHAR(text(1:1)) <= IACHAR('Z')) ) THEN
          tmp = IdentifierIfortBug(text, QueryLocation(pnode%tok))
          CALL visitor%modifier%Modify(tmp)
          CALL SetKind(pnode%tok, .NameOf. tmp) 
        END IF
      END IF
      
    TYPE IS (ContiguousStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (CriticalConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (CycleStmt)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (DataImpliedDoControl)
      CALL visitor%modifier%Modify(pnode%do_variable)
      
    TYPE IS (DerivedTypeStmt)
      CALL visitor%modifier%Modify(pnode%name)
      IF (ALLOCATED(pnode%extends)) THEN
        CALL visitor%modifier%Modify(pnode%extends)
      END IF
      DO i = 1, SIZE(pnode%params)
        CALL visitor%modifier%Modify(pnode%params(i))
      END DO
      
    TYPE IS (DimensionDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (DoConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (EntityDetails)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (EntryStmt)
      ! Fix up binding name if it was provided.
      IF (ALLOCATED(pnode%ls)) THEN
        IF (.NOT. ALLOCATED(pnode%ls%name)) THEN
          ! Create a scalar Expression that represents the lower case 
          ! variant of the original identifier.
          CALL set_binding_name(pnode%ls, pnode%entry_name)
        END IF
      END IF
      CALL visitor%modifier%modify(pnode%entry_name)
      DO i = 1, SIZE(pnode%args)
        IF (ALLOCATED(pnode%args(i)%name)) THEN
          CALL visitor%modifier%Modify(pnode%args(i)%name)
        END IF
      END DO
      IF (ALLOCATED(pnode%result_name)) THEN
        CALL visitor%modifier%Modify(pnode%result_name)
      END IF
      
    TYPE IS (Enumerator)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (ExitStmt)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (ExternalStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (FinalProcedureStmt)
      DO i = 1, SIZE(pnode%finals)
        CALL visitor%modifier%Modify(pnode%finals(i))
      END DO
      
    TYPE IS (ForallConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    TYPE IS (ForallTripletSpec)
      CALL visitor%modifier%Modify(pnode%index)
      
    TYPE IS (FunctionStmt)
      ! Fix up binding name if it was provided.
      IF (ALLOCATED(pnode%ls)) THEN
        IF (.NOT. ALLOCATED(pnode%ls%name)) THEN
          ! Create a scalar Expression that represents the lower case 
          ! variant of the original identifier.
          CALL set_binding_name(pnode%ls, pnode%function_name)
        END IF
      END IF
      CALL visitor%modifier%Modify(pnode%function_name)
      DO i = 1, SIZE(pnode%args)
        CALL visitor%modifier%Modify(pnode%args(i))
      END DO
      IF (ALLOCATED(pnode%result_name)) THEN
        CALL visitor%modifier%Modify(pnode%result_name)
      END IF
      
    TYPE IS (FunctionRefExpr)
      ! Test whether something is intrinsic before we mangle it.
      is_intrinsic = test_is_intrinsic(visitor, pnode%name)
      CALL visitor%modifier%Modify(pnode%name)
      ! If the function is intrinsic-like, we don't mangle the argument 
      ! keywords.
      IF (.NOT. is_intrinsic) THEN
        DO i = 1, SIZE(pnode%args)
          IF (ALLOCATED(pnode%args(i)%keyword)) THEN
            CALL visitor%modifier%Modify(pnode%args(i)%keyword)
          END IF
        END DO
      END IF
      
    TYPE IS (GenericSpec)
      IF (ALLOCATED(pnode%name)) THEN
        CALL visitor%modifier%Modify(pnode%name)
      END IF
    
    TYPE IS (ImportStmt)
      DO i = 1, SIZE(pnode%import_names)
        CALL visitor%modifier%Modify(pnode%import_names(i))
      END DO
      
    TYPE IS (IndexLoopControl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (IntentStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (IntrinsicStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (IoControlSpecList)
      IF (ALLOCATED(pnode%fmt_or_nml)) THEN
        CALL visitor%modifier%Modify(pnode%fmt_or_nml)
      END IF
      
    TYPE IS (IOImpliedDoControl)
      CALL visitor%modifier%Modify(pnode%do_variable)
      
    TYPE IS (ModuleStmt)
      CALL visitor%modifier%Modify(pnode%module_name)
      
    TYPE IS (MpSubprogramStmt)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (NamedConstantDef)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (NamelistStmt)
      DO i = 1, SIZE(pnode%groups)
        CALL visitor%modifier%Modify(pnode%groups(i)%name)
        DO i2 = 1, SIZE(pnode%groups(i)%objects)
          CALL visitor%modifier%Modify(pnode%groups(i)%objects(i2))
        END DO
      END DO
      
    TYPE IS (NullInit)
      ! Identifier should always be NULL, and should never be changed.
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (OptionalStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (PartRef)
      CALL visitor%modifier%Modify(pnode%part_name)
      
    TYPE IS (PointerDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (ProcComponentAttrs)
      IF (ALLOCATED(pnode%pass_arg_name)) THEN
        CALL visitor%modifier%Modify(pnode%pass_arg_name)
      END IF
      
    TYPE IS (ProcDecl)
      CALL visitor%modifier%Modify(pnode%name)
      IF (ALLOCATED(pnode%init_target)) THEN
        CALL visitor%modifier%Modify(pnode%init_target)
      END IF
      
    TYPE IS (ProcInterface)
      IF (ALLOCATED(pnode%name)) THEN
        CALL visitor%modifier%Modify(pnode%name)
      END IF
      
    TYPE IS (ProcedureStmt)
      DO i = 1, SIZE(pnode%procedure_names)
        CALL visitor%modifier%Modify(pnode%procedure_names(i))
      END DO
      
    TYPE IS (ProgramStmt)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (ProtectedStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (Rename)
      IF (ALLOCATED(pnode%use)) THEN
        CALL visitor%modifier%Modify(pnode%use)
      END IF
      IF (ALLOCATED(pnode%name)) THEN
        CALL visitor%modifier%Modify(pnode%name)
      END IF
      
    TYPE IS(SavedEntity)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (SelectTypeConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      IF (ALLOCATED(pnode%associate_name)) THEN
        CALL visitor%modifier%Modify(pnode%associate_name)
      END IF
      
    TYPE IS (StmtFunctionStmt)
      CALL visitor%modifier%Modify(pnode%function_name)
      DO i = 1, SIZE(pnode%args)
        CALL visitor%modifier%Modify(pnode%args(i))
      END DO
      
    TYPE IS (SubmoduleStmt)
      CALL visitor%modifier%Modify(pnode%ancestor_module_name)
      IF (ALLOCATED(pnode%parent_submodule_name)) THEN
        CALL visitor%modifier%Modify(pnode%parent_submodule_name)
      END IF
      CALL visitor%modifier%Modify(pnode%submodule_name)
      
    TYPE IS (SubroutineStmt)
      ! Fix up binding name if it was provided.
      IF (ALLOCATED(pnode%ls)) THEN
        IF (.NOT. ALLOCATED(pnode%ls%name)) THEN
          ! Create a scalar Expression that represents the lower case 
          ! variant of the original identifier.
          CALL set_binding_name(pnode%ls, pnode%subroutine_name)
        END IF
      END IF
      CALL visitor%modifier%Modify(pnode%subroutine_name)
      DO i = 1, SIZE(pnode%args)
        IF (ALLOCATED(pnode%args(i)%name)) THEN
          CALL visitor%modifier%Modify(pnode%args(i)%name)
        END IF
      END DO
      
    TYPE IS (TargetDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (TypeBoundGenericStmt)
      DO i = 1, SIZE(pnode%binding_list)
        CALL visitor%modifier%Modify(pnode%binding_list(i))
      END DO
      
    TYPE IS (TypeBoundProcedureStmt)
      IF (ALLOCATED(pnode%interface_name)) THEN
        CALL visitor%modifier%Modify(pnode%interface_name)
      END IF
      IF (ALLOCATED(pnode%attrs%pass_arg_name)) THEN
        CALL visitor%modifier%Modify(pnode%attrs%pass_arg_name)
      END IF
      DO i = 1, SIZE(pnode%decls)
        CALL visitor%modifier%Modify(pnode%decls(i)%binding)
        IF (ALLOCATED(pnode%decls(i)%specific)) THEN
          CALL visitor%modifier%Modify(pnode%decls(i)%specific)
        END IF
      END DO
      
    TYPE IS (TypeParamDecl)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (TypeParamInquiry)
      CALL visitor%modifier%Modify(pnode%type_param_name)
      
    TYPE IS (TypeParamSpec)
      IF (ALLOCATED(pnode%name)) THEN
        CALL visitor%modifier%Modify(pnode%name)
      END IF
      
    TYPE IS (TypeSpec)
      CALL visitor%modifier%Modify(pnode%type_name)
      
    TYPE IS (UnaryOpExpr)
      CALL visitor%modifier%Modify(pnode%op)
      
    TYPE IS (UseStmt)
      CALL visitor%modifier%Modify(pnode%name)
      
    TYPE IS (ValueStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (VolatileStmt)
      DO i = 1, SIZE(pnode%name_list)
        CALL visitor%modifier%Modify(pnode%name_list(i))
      END DO
      
    TYPE IS (WhereConstruct)
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL visitor%modifier%Modify(pnode%construct_name)
      END IF
      
    END SELECT
    
  END SUBROUTINE iv_Start
  
  
  !*****************************************************************************
  !!
  !> Explicitly set the binding name for a Fortran identifier to the 
  !! lower case variant of the provided identifier.
  
  SUBROUTINE set_binding_name(ls, ident)
    
    USE CharUtils
    USE CompilerKinds
    USE Expressions
    USE Identifiers
    USE LanguageBindingSpecs
    USE PrimaryExpressions
    USE SourceLocations
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(LangSpec), INTENT(INOUT) :: ls
    TYPE(Identifier), INTENT(IN) :: ident
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Lower case, scck kind of the identifier name.
    CHARACTER(:,KIND=scck), ALLOCATABLE :: lower_name
    
    ! The token for the character expression for the binding name.
    TYPE(Token) :: name_token
    
    ! A temporary for constructing the relevant name expression.
    TYPE(ConstantExpr), ALLOCATABLE :: tmp
    
    !***************************************************************************
    
    lower_name = .NameOf. ident
    CALL LowerCase(lower_name)      ! Could have kind issues.
    
    name_token = Tkn(  &
        ittCharacterLiteral,  &
        SourceLocationIfortBug(),  &
        .FALSE.,  &
        .FALSE. )
    CALL AppendRaw(name_token, lower_name)
    CALL SetValue(name_token, lower_name)
    
    tmp = ConstantExpr(name_token)
    
    CALL MOVE_ALLOC(tmp, ls%name)
    
  END SUBROUTINE set_binding_name
  
  
  !*****************************************************************************
  !!
  !> Update the intrinsic procedure flag stack based on the given identifier.
  !!
  !! @param[in,out] visitor           The parse node visitor.
  !!
  !! @param[in]     ident             The identifier to check.
  
  FUNCTION test_is_intrinsic(visitor, ident)
    
    USE Identifiers
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(identifier_visitor), INTENT(INOUT) :: visitor
    TYPE(Identifier), INTENT(IN) :: ident
    
    ! Function result
    LOGICAL :: test_is_intrinsic
    
    !***************************************************************************
    
    test_is_intrinsic = in_list(  &
        visitor%name_stack(SIZE(visitor%name_stack)),  &
        .NameOf. ident )
    
  END FUNCTION test_is_intrinsic
  
  
  !*****************************************************************************
  !!
  
  SUBROUTINE obfuscator_Modify(modifier, ident)
    
    USE Identifiers
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(obfuscator), INTENT(INOUT) :: modifier
    TYPE(Identifier), INTENT(INOUT) :: ident
    
    !***************************************************************************
    
    CALL process_identifier(  &
          modifier%forward,  &
          modifier%reverse,  &
          ident,  &
          modifier%num )
    
  END SUBROUTINE obfuscator_Modify
  
  
  !*****************************************************************************
  !!
  
  SUBROUTINE process_identifier(forward, reverse, ident, num)
    
    USE Identifiers
    USE SourceLocations
    USE Strings
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(IdentifierMap), INTENT(INOUT) :: forward
    TYPE(IdentifierMap), INTENT(INOUT) :: reverse
    TYPE(Identifier), INTENT(INOUT) :: ident
    INTEGER, INTENT(INOUT) :: num
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Current value of the identifier.
    CHARACTER(:), ALLOCATABLE :: key
    
    ! Existing value from the mapping.
    TYPE(String), POINTER :: value
    
    ! Obfuscated name for an identifier not seen before.
    CHARACTER(:), ALLOCATABLE :: new_value
    
    !***************************************************************************
    
    key = .NameOf. ident
    CALL Get(forward, String(key), value)
    
    ! If a forward mapping exists, then set the identifier name to the value 
    ! from the mapping.
    IF (ASSOCIATED(value)) THEN
      ident = IdentifierIfortBug(value%item, SourceLocation())
      RETURN
    END IF
    
    ! Generate a new string for the identifier.
    DO
      num = num + 1
      IF (key(1:1) == '.') THEN
        new_value = '.' // alpha(num) // '.'
      ELSE
        ! Preserve initial character in case of implicit typing.
        new_value = key(1:1) // alphanumeric(num)
      END IF
      ! See if the chosen identifier already exists.  If not, 
      ! try again.
      CALL Get(reverse, String(new_value), value)
      IF (.NOT. ASSOCIATED(value)) EXIT
    END DO
    
    ! Add the new mapping.
    CALL Insert(forward, String(.NameOf. ident), String(new_value))
    CALL Insert(reverse, String(new_value), String(.NameOf. ident))
    
    ! Alter the current identifier.
    ident = IdentifierIfortBug(new_value, SourceLocation())
    
  END SUBROUTINE process_identifier
  
  
  !*****************************************************************************
  !!
  !> Get an alphabetic ([a-z]*) representation of an integer
  !!
  !! @param[in]     in                The value.  Must be greater than zero.
  !!
  !! @returns A string in the series A, B, C, ..., Z, AA, AB, AC, ...
  
  FUNCTION alpha(in) RESULT(out)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: in
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: out
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: rem            ! Remainder.
    INTEGER :: m              ! Modulo of the current value.
    
    !***************************************************************************
    
    rem = in
    out = ''
    
    DO
      m = MOD(rem - 1, 26)
      out = ACHAR(m + IACHAR('A')) // out
      rem = (rem - 1) / 26
      IF (rem == 0) EXIT
    END DO
    
  END FUNCTION alpha
  
  
  !*****************************************************************************
  !!
  !> Get an alpha-alphanumeric ([a-z][a-z0-9_]*) representation of an integer
  !!
  !! @param[in]     in                The value.  Must be greater than zero.
  !!
  !! @returns A string in the series A, B, C, ..., Z, AA, AB, AC, ..., 
  !! A1, A2, A3, ..., BA, BB, ... (the first character is always alphabetic).
  
  FUNCTION alphanumeric(in) RESULT(out)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: in
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: out
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: rem            ! Remainder.
    INTEGER :: m              ! Modulo of the current value.
    
    CHARACTER(*), PARAMETER :: alphanum  &
        = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ01234567890_'
    
    !***************************************************************************
    
    rem = in
    out = ''
    
    DO WHILE (rem > 26)
      m = MOD(rem - 27, LEN(alphanum))
      out = alphanum(m+1:m+1) // out
      rem = (rem - 27) / LEN(alphanum) + 1
    END DO
    
    DO
      m = MOD(rem - 1, 26)
      out = ACHAR(m + IACHAR('A')) // out
      rem = (rem - 1) / 26

      IF (rem == 0) EXIT
    END DO
    
  END FUNCTION alphanumeric
  
END MODULE ObfuscateImplementation
