! $Id: ParseExpressions.f90 2803 2019-03-22 19:36:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the ParseExprExternal subroutine and an internal module 
!! expression_parse_utils.


!*******************************************************************************
!!
!> Utilities to support the ParseExpressionsExternal subroutine.

MODULE expression_parse_utils
  
  USE Errors
  USE Expressions
  USE Tokens
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseExprInternal
  PUBLIC :: WriteExpressionInternal
  
  !-----------------------------------------------------------------------------
  
  ABSTRACT INTERFACE
    !***************************************************************************
    !> Interface for a function that searches for a particular operator
    !! or class of operators.
    !!
    !! @param[in]     tlist             The sequence of tokens.
    !!
    !! @returns The position of the token, or zero if no matching token was 
    !! found.
    FUNCTION find_op_intf(tlist) RESULT(pos)
      IMPORT :: Token
      IMPLICIT NONE
      !-----------------------------------------------------------------------
      TYPE(Token), INTENT(IN) :: tlist(:)
      ! Function result
      INTEGER :: pos
    END FUNCTION find_op_intf
  END INTERFACE
  
  ABSTRACT INTERFACE
    !***************************************************************************
    !!
    !> Interface for a procedure that parses a child syntax rule.
    !!
    !! @param[in]     tlist             The sequence of tokens.
    !!
    !! @param[in]     start             Index of the first token of the term.
    !!
    !! @param[in]     finish            Index of the last token of the term.
    !!
    !! @param[out]    ex                An expression object.  May not be 
    !! allocated in the event of an error that prevented parsing.
    !!
    !! @param[out]    err_list          List of errors.
    SUBROUTINE parse_child_intf(tlist, start, finish, ex, err_list)
      IMPORT :: Token
      IMPORT :: Expression
      IMPORT :: Error
      IMPLICIT NONE
      !-----------------------------------------------------------------------
      TYPE(Token), INTENT(IN) :: tlist(:)
      INTEGER, INTENT(IN) :: start
      INTEGER, INTENT(IN) :: finish
      CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
      TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    END SUBROUTINE parse_child_intf
  END INTERFACE
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an expression (R722 /expr/).
  !!
  !! This is the public entry point for all expression parsing.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /expression/.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /expression/.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An /expr/ is:
  !!  [ /expr/ /defined-binary-op/ ] /level-5-expr/
  
  RECURSIVE SUBROUTINE ParseExprInternal(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start,  &
        finish,  &
        'R722',  &
        '/expr/',  &
        find_defined_binary_op,  &
        parse_level_5_expr,  &
        ex,  &
        err_list )
    
  END SUBROUTINE ParseExprInternal
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an expression that involves a 
  !! binary operator of some form, with grouping running from left to right.
  !!
  !! Whether this procedure correctly groups from left to right depends 
  !! on the behaviour of the find_op procedure - that procedure must search 
  !! backwards to find the operator token of interest.
  !!
  !! Left to right grouping is what is required for things like the or 
  !! operator, where the expression "a .or. b .or. c" should be evaluated as 
  !! "(a .or. b) .or. c".  Syntax rules for these operators have the optional 
  !! parts of the rule on the left of the mandatory parts.  Another 
  !! requirement for this particular parsing procedure is that the optional 
  !! part be a repeat of the syntax rule being defined... for example:
  !!
  !!   rule_a := [ rule_a operator ] rule_b
  !!
  !! Binary expression syntax rules that are exceptions to this include 
  !! /level-4-expr/ and /mult-operand/.  /level-2-expr/ is also a little 
  !! different, because it can also represent a unary operator.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[in]     comp              The syntax rule number being parsed.
  !!
  !! @param[in]     rule              The syntax rule name being parsed.
  !!
  !! @param[in]     reverse_find_op   Function to search for the relevant 
  !! operator, starting from the end of the token sequence and working 
  !! backwards.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  
  RECURSIVE SUBROUTINE parse_binary_expr( tlist, start, finish, comp,  &
      rule, find_op, parse_child, ex, err_list )
    
    USE CompilerKinds
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CHARACTER(*), INTENT(IN) :: comp
    CHARACTER(*), INTENT(IN) :: rule
    PROCEDURE(find_op_intf) :: find_op
    PROCEDURE(parse_child_intf) :: parse_child
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index (end of term).
    INTEGER :: op_idx         ! Index of next op.
    INTEGER :: term_finish    ! Index of the finish of the current term.
    
    ! Temporary for constructing an object if there are two terms.
    CLASS(BinaryOpExpr), ALLOCATABLE :: bin_op
    
    ! Error list for child parsing procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected ' // rule // '.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Summary of approach.
    !
    ! original:                              t1  op  t2  op  t3  op  t4 
    ! find first operator:                        ^                     
    ! parse up to first op into child:      {t1} op  t2  op  t3  op  t4 
    ! find next operator:                                ^              
    !                                      {{t1} op  t2} op  t3  op  t4 
    ! find next operator:                                        ^      
    !                                     {{{t1} op  t2} op  t3} op  t4 
    ! no more operators:                                                
    ! remainder into final binop:        {{{{t1} op  t2} op  t3} op  t4}
    
    it = start
    
    token_loop: DO
      op_idx = find_op(tlist(it:finish))
      ! If there's no operator, then the entire token sequence is treated as 
      ! a child expression.  Otherwise we break the token sequence in two at 
      ! the operator, parse the first part as a child, advance to the 
      ! operator and then repeat.
      IF (op_idx == 0) THEN
        term_finish = finish
      ELSE
        term_finish = op_idx + it - 2
      END IF
      
      ! The first iteration of token loop will not have ex allocated.
      IF (ALLOCATED(ex)) THEN
        ALLOCATE(bin_op)
        ! What we had previously is the left hand side.
        CALL MOVE_ALLOC(ex, bin_op%lhs)
        ! Operator that terminated the previous term is the relevant binary.
        bin_op%op = IdentifierIfortBug(tlist(it - 1))
        ! Parse up until the end of the current term into the right hand side.
        CALL parse_child(tlist, it, term_finish, bin_op%rhs, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(bin_op, ex)
      ELSE
        ! First term for this particular syntax rule.
        CALL parse_child(tlist, it, term_finish, ex, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
      ! The test here needs to consider both no operator found and a matching 
      ! operator found as the first token, as per comments above.
      IF (op_idx == 0) EXIT
      
      ! Skip over the operator.
      it = term_finish + 2
    END DO token_loop
    
  END SUBROUTINE parse_binary_expr
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /level-5-expr/ (R717)
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /level-5-expr/ /equiv-op/ ] /equiv-operand/
  
  RECURSIVE SUBROUTINE parse_level_5_expr(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start, finish,  &
        'R717',  &
        '/level-5-expr/',  &
        find_equiv_op,  &
        parse_equiv_operand,  &
        ex,  &
        err_list )
    
  END SUBROUTINE parse_level_5_expr
  
  
  !*****************************************************************************
  !!
  !> Parse an /equiv-operand/ (R716)
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /equiv-operand/ /or-op/ ] /or-operand/
  
  RECURSIVE SUBROUTINE parse_equiv_operand(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start, finish,  &
        'R716',  &
        '/equiv-operand/',  &
        find_or_op,  &
        parse_or_operand,  &
        ex,  &
        err_list )
      
  END SUBROUTINE parse_equiv_operand
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /or-operand/ (R715).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /or-operand/ /and-op/ ] /and-operand/
  
  RECURSIVE SUBROUTINE parse_or_operand(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start, finish,  &
        'R715',  &
        '/or-operand/',  &
        find_and_op,  &
        parse_and_operand,  &
        ex,  &
        err_list )
    
  END SUBROUTINE parse_or_operand
  
  
  !*****************************************************************************
  !!
  !> Parse an and-operand (R714).
  !!
  !! @param[in]     tlist             The list of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ not-op ] level-4-expr
  
  RECURSIVE SUBROUTINE parse_and_operand(tlist, start, finish, ex, err_list)
    
    USE CompilerKinds
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /and-operand/.
    CHARACTER(*), PARAMETER :: comp = 'R714'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for constructing an object if the term has a unary operator.
    CLASS(UnaryOpExpr), ALLOCATABLE :: un_op
    
    ! Error list for child parsing procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /and-operand/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (IsToken(tlist(start), ittOperator, scck_'.NOT.')) THEN
      ALLOCATE(un_op)
      un_op%op = IdentifierIfortBug(tlist(start))
      CALL parse_level_4_expr(  &
          tlist,  &
          start + 1, finish,  &
          un_op%rhs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(un_op, ex)
    ELSE
      CALL parse_level_4_expr(  &
          tlist,  &
          start, finish,  &
          ex,  &
          sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
  END SUBROUTINE parse_and_operand
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /level-4-expr/ (R712).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /level-3-expr/ /rel-op/ ] /level-3-expr/
  !!
  !! Note how this is not the normal binary expression pattern - it does 
  !! not reference /level-4-expr/ on the optional left hand side of the 
  !! /rel-op/.
  
  RECURSIVE SUBROUTINE parse_level_4_expr(tlist, start, finish, ex, err_list)
    
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /level-4-expr/.
    CHARACTER(*), PARAMETER :: comp = 'R712'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: op_idx         ! Index of next /concat-op/.
    
    ! Temporary for constructing an object if there are two terms.
    TYPE(BinaryOpExpr), ALLOCATABLE :: bin_op
    
    ! Error list for child parsing procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /level-4-expr/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Do we have a single term or two terms?
    
    op_idx = find_rel_op(tlist(start:finish), 1)
    
    IF (op_idx == 0) THEN
      ! Single term.
      CALL parse_level_3_expr(tlist, start, finish, ex, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Two terms.
      ALLOCATE(bin_op)
      bin_op%op = IdentifierIfortBug(tlist(start+op_idx-1))
      CALL parse_level_3_expr(  &
          tlist,  &
          start, start + op_idx - 2,  &
          bin_op%lhs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      CALL parse_level_3_expr(  &
          tlist,  &
          start + op_idx, finish,  &
          bin_op%rhs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      CALL MOVE_ALLOC(bin_op, ex)
    END IF
    
  END SUBROUTINE parse_level_4_expr
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /level-3-expr/ (R710).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /level-3-expr/ /concat-op/ ] /level-2-expr/
  
  RECURSIVE SUBROUTINE parse_level_3_expr(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start, finish,  &
        'R710',  &
        '/level-3-expr/',  &
        find_concat_op,  &
        parse_level_2_expr, ex, err_list )
    
  END SUBROUTINE parse_level_3_expr
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /level-2-expr/ (R706).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! The syntax rule is:
  !!
  !! [ [ /level-2-expr/ ] /add-op/ ] /add-operand/
  !!
  !! If the /level-2-expr/ is simply an /add-operand/, then we return 
  !! that /add-operand/ in @a ex.  The dynamic type of @a ex will be 
  !! whatever results from the parsing of the /add-operand/.
  !!
  !! If the /level-2-expr/ is something like "/add-op/ /add-operand/", then 
  !! the dynamic type of @a ex will be a UnaryOp, with the operator in 
  !! the op component of that type and the /add-operand/ in the rhs component.
  !!
  !! If the /level-2-expr/ is something like A + B + C then the 
  !! top level ex object will be a BinaryOpExpr, with a rhs of C and a 
  !! lhs that is another BinaryOpExpr that is A + B.
  
  RECURSIVE SUBROUTINE parse_level_2_expr(tlist, start, finish, ex, err_list)
     
    USE CompilerKinds
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /level-2-expr/.
    CHARACTER(*), PARAMETER :: comp = 'R706'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index (start of term).
    INTEGER :: term_finish    ! Index of the end of the current operand.
    INTEGER :: op_idx         ! Index of the next 
    
    ! Temporary for constructing an object for the far left term if that term
    ! is something like /add-op/ /add-operand/
    CLASS(UnaryOpExpr), ALLOCATABLE :: un_op
    
    ! Temporary for constructing an object for any /add-operand/ 
    ! /add-op/ /add-operand/ pairs.
    CLASS(BinaryOpExpr), ALLOCATABLE :: bin_op
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (start > finish) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /level-2-expr/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    it = start
    
    ! Can have /add-op/ /add-operand/.
    ! SIZE(tlist) guaranteed to be >= 1 by opening test.
    IF ( IsToken(tlist(it), ittOperator, scck_'+')  &
        .OR. IsToken(tlist(it), ittOperator, scck_'-') ) THEN
      ALLOCATE(un_op)
      un_op%op = IdentifierIfortBug(tlist(it))   ! Should be + or -
      it = it + 1
      IF (it > finish) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=SafeLocation(tlist, it),  &
            MSG='Expected /add-operand/.' )
        RETURN
      END IF
    END IF
    
    token_loop: DO
      
      op_idx = find_add_op(tlist(start:finish), it - start + 1)
      IF (op_idx == 0) THEN
        term_finish = finish
      ELSE
        term_finish = start + op_idx - 2
      END IF
      
      IF (ALLOCATED(un_op)) THEN      ! Possible first iteration only.
        CALL parse_add_operand(  &
            tlist,  &
            it, term_finish,  &
            un_op%rhs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(un_op, ex)
      ELSE IF (ALLOCATED(ex)) THEN    ! Not possible on first iteration.
        ALLOCATE(bin_op)
        CALL MOVE_ALLOC(ex, bin_op%lhs)
        bin_op%op = IdentifierIfortBug(tlist(it-1)) ! Should be + or -
        CALL parse_add_operand(  &
            tlist,  &
            it, term_finish,  &
            bin_op%rhs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(bin_op, ex)
      ELSE                            ! Possible first iteration only.
        CALL parse_add_operand(  &
            tlist,  &
            it, term_finish,  &
            ex,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
      IF (op_idx == 0) EXIT
      
      ! Skip over the operator.
      it = start + op_idx
      
    END DO token_loop
    
  END SUBROUTINE parse_level_2_expr
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /add-operand/ (R705).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ /add-operand/ /mult-op/ ] /mult-operand/
  
  RECURSIVE SUBROUTINE parse_add_operand(tlist, start, finish, ex, err_list)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL parse_binary_expr(  &
        tlist,  &
        start, finish,  &
        'R705',  &
        '/add-operand/',  &
        find_mult_op,  &
        parse_mult_operand,  &
        ex,  &
        err_list )
    
  END SUBROUTINE parse_add_operand
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /mult-operand/ (R704).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! /level-1-expr/ [ /power-op/ /mult-operand/ ]
  !!
  !! The precedence of this one is the other way around to the others, so 
  !! we work backwards.
  
  RECURSIVE SUBROUTINE parse_mult_operand(tlist, start, finish, ex, err_list)
    
    USE CompilerKinds
    USE ErrorCodes
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /mult-operand/.
    CHARACTER(*), PARAMETER :: comp = 'R704'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token index (end of term).
    INTEGER :: op_idx         ! Index of previous /power-op/.
    INTEGER :: term_start     ! Start of the previous term.
    
    ! Temporary for constructing an object if there are two terms.
    CLASS(BinaryOpExpr), ALLOCATABLE :: bin_op
    
    ! Error list for child parsing procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add(err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /mult-operand/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    it = finish
    
    token_loop: DO
    
      op_idx = FindBareReverse(  &
          tlist(start:finish),  &
          ittOperator,  &
          scck_'**',  &
          it - start + 1 )
      IF (op_idx == 0) THEN
        term_start = start
      ELSE
        term_start = start + op_idx
      END IF
      
      IF (ALLOCATED(ex)) THEN
        ALLOCATE(bin_op)
        CALL MOVE_ALLOC(ex, bin_op%rhs)
        bin_op%op = IdentifierIfortBug(tlist(it+1))      ! Should be **
        CALL parse_level_1_expr(  &
            tlist,  &
            term_start, it,  &
            bin_op%lhs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(bin_op, ex)
      ELSE
        CALL parse_level_1_expr(  &
            tlist,  &
            term_start, it,  &
            ex,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
      END IF
      
      IF (op_idx == 0) EXIT
      
      ! Skip (backwards) over the operator.
      it = start + op_idx - 2
      
    END DO token_loop
    
  END SUBROUTINE parse_mult_operand
  
  
  !*****************************************************************************
  !!
  !> Parse a /level-1 expr/ (R702).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! [ defined-unary-op ] primary
  
  RECURSIVE SUBROUTINE parse_level_1_expr(tlist, start, finish, ex, err_list)
    
    USE CompilerKinds
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /level-1-expr/.
    CHARACTER(*), PARAMETER :: comp = 'R702'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Temporary for constructing an object if the term has a unary operator.
    CLASS(UnaryOpExpr), ALLOCATABLE :: un_op
    
    ! Error list for child parsing procedures.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /level-1-expr/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (tlist(start) == ittOperator) THEN
      IF (.NOT. is_user_defined(QueryValue(tlist(start)))) THEN
        ! Something like "*2" or "// 2".
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='C703 on ' // comp,  &
            LOCATION=QueryLocation(tlist(start)),  &
            MSG='"' // QueryValue(tlist(start)) // '" is not a valid &
              &/defined-unary-op/.' )
        RETURN
      END IF
      ! We assume that we've not missed any other operators in the stages 
      ! above.
      ALLOCATE(un_op)
      un_op%op = IdentifierIfortBug(tlist(start))
      CALL parse_primary(  &
          tlist,  &
          start+1, finish,  &
          un_op%rhs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(un_op, ex)
    ELSE
      CALL parse_primary(  &
          tlist,  &
          start, finish,  &
          ex,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
  END SUBROUTINE parse_level_1_expr
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /primary/ (R701) - a primary 
  !! expression.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token of the term.
  !!
  !! @param[in]     finish            Index of the last token of the term.
  !!
  !! @param[out]    ex                An expression object.  May not be 
  !! allocated in the event of an error that prevented parsing.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This is the tricky one.
  !!
  !! A /primary/ is a:
  !! - /constant/: if it is a REAL, INTEGER or LOGICAL literal constant then 
  !! we can pick it up straight away.  If it is a CHARACTER constant then 
  !! we need to make sure that it is not actually part of a substring (which 
  !! by rights is a designator).
  !!
  !! - /designator/: Several complicated options, most of which are data-refs.
  !!
  !! - /array-constructor/: Reasonably unique.
  !!
  !! - /structure-constructor/: Looks like a function reference, and may also 
  !! look like a /designator/ that is an array reference.
  !!
  !! - /function-reference/: Not distinguishable from structure constructors 
  !! or array references in some cases - ends up being called a designator.
  !!
  !! - /type-param-inquiry/: Looks like a structure component reference so 
  !! ends up as a designator.
  !!
  !! - /type-param-name/: Looks like a simgple name reference - ends up as 
  !! a designator.
  !!
  !! - ( /expr/ ): Reasonably unique.
  !!
  !! The last parenthesised form may result in recursive calls to this 
  !! procedure - hence the RECURSIVE attribute.
  
  RECURSIVE SUBROUTINE parse_primary(tlist, start, finish, ex, err_list)
    
    USE ArrayConstructors
    USE CallStmts
    USE CompilerKinds
    USE ComplexLiteralConstants
    USE Designators
    USE ErrorCodes
    USE ParseUtils
    USE PrimaryExpressions
    USE StructureConstructors
    USE TypeParamInquiries
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /primary/.
    CHARACTER(*), PARAMETER :: comp = 'R701'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Typically used for the position of a closing parenthesis in a token 
    ! list, relative to the position of the opening parenthesis.
    INTEGER :: closing
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! If we encounter a literal constant or a designator we construct the 
    ! relevant variable below and then MOVE_ALLOC that variable into @a ex.
    TYPE(ConstantExpr), ALLOCATABLE :: const_ex
    TYPE(ComplexLiteralExpr), ALLOCATABLE :: complex_ex
    TYPE(TypeParamInquiry), ALLOCATABLE :: tpi_ex
    TYPE(DesignatorExpr), ALLOCATABLE :: desig_ex
    TYPE(ArrayConstructorExpr), ALLOCATABLE :: ac_ex
    TYPE(FunctionRefExpr), ALLOCATABLE :: func_ex
    TYPE(StructureConstructorExpr), ALLOCATABLE :: struct_ex
    TYPE(ParenthesisedExpr), ALLOCATABLE :: parens_ex
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /primary/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Eliminate the simple constants first
    
    ! Simple constants.
    IF ( (tlist(start) == ittIntegerLiteral)  &
        .OR. (tlist(start) == ittRealLiteral)  &
        .OR. (tlist(start) == ittLogicalLiteral)  &
        .OR. (tlist(start) == ittBOZLiteral) ) THEN
      ALLOCATE(const_ex)
      const_ex%tok = tlist(start)
      CALL MOVE_ALLOC(const_ex, ex)
      
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a parenthesised expression?  Or a complex literal?
    
    IF (IsToken(tlist(start), ittDelimiter, scck_'(')) THEN
      CALL ExpectClosing(tlist, start, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      IF (IsComplexLiteralConstant(tlist(start+1:closing-1))) THEN
        ALLOCATE(complex_ex)
        complex_ex%paren_location = QueryLocation(tlist(start))
        CALL ParseComplexLiteralConstant(  &
            tlist,  &
            start + 1,  &
            closing - 1,  &
            complex_ex%constant,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        CALL MOVE_ALLOC(complex_ex, ex)
        CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
        RETURN
      END IF
      
      ! From the parse tree point of view, there is no need to remember 
      ! that we are dealing with a parenthesised expression, as the 
      ! hierarchy of child expression objects still honours the precedence 
      ! established by the expression syntax rules.  However, the presence 
      ! of parentheses may be important for things like argument association, 
      ! or if we want to regenerate the token sequence for the expression, 
      ! so we still create an intermediate node.
      ALLOCATE(parens_ex)
      CALL ParseExpr(  &
          tlist,  &
          start + 1, closing - 1,  &
          parens_ex%ex,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      CALL MOVE_ALLOC(parens_ex, ex)
      
      CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a character literal, maybe masquerading as a designator?
    
    ! We know SIZE(tlist) > 0 from test at start.
    IF (tlist(start) == ittCharacterLiteral) THEN
      IF (finish - start >= 1) THEN
        ! Character literals can be a part of a designator - they can have 
        ! a sub-string part attached.  Check for that here.
        IF (IsToken(tlist(start+1), ittDelimiter, scck_'(')) THEN
          CALL ExpectClosing(tlist, start + 1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          ! Perhaps bizarre, but the syntax rules permit a 
          ! /type-param-inquiry/ on a substringed character literal.  Such 
          ! a beast will have a percent token following the substring range.
          IF (closing + 1 < finish) THEN
            IF (IsToken(tlist(closing+1), ittPercent, scck_'%')) THEN
              ALLOCATE(tpi_ex)
              CALL ParseTypeParamInquiry(  &
                  tlist,  &
                  start,  &
                  finish,  &
                  tpi_ex,  &
                  sub_err_list )
              CALL Add(err_list, sub_err_list, COMPONENT=comp)
              CALL MOVE_ALLOC(tpi_ex, ex)
              RETURN
            END IF
          END IF
          
          ALLOCATE(desig_ex)
          CALL ParseDesignator(  &
              tlist(start:closing),  &
              desig_ex%desig,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          IF (Failure(sub_err_list)) RETURN
          CALL ExpectEnd(tlist(:finish), closing + 1, comp, err_list)
          CALL MOVE_ALLOC(desig_ex, ex)
          RETURN
        ELSE
          CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
          RETURN
        END IF
      ELSE
        ALLOCATE(const_ex)
        const_ex%tok = tlist(start)
        CALL MOVE_ALLOC(const_ex, ex)
        RETURN
      END IF
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps an /array-constructor/?
    
    IF ( IsToken(tlist(start), ittDelimiter, scck_'(/')  &
        .OR. IsToken(tlist(start), ittDelimiter, scck_'[') ) THEN
      ALLOCATE(ac_ex)
      CALL ParseArrayConstructor(  &
          tlist,  &
          start, finish,  &
          ac_ex%type_spec,  &
          ac_ex%values,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(ac_ex, ex)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a structure constructor?  We only parse things as a structure 
    ! constructor if they can be a structure constructor but cannot otherwise 
    ! be a function reference or designator.
    
    IF (test_structure_constructor(tlist(start:finish))) THEN
      ALLOCATE(struct_ex)
      CALL ParseStructureConstructor(  &
          tlist,  &
          start, finish,  &
          struct_ex%sc,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(struct_ex, ex)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Perhaps a function reference?  We only parse things as a function 
    ! reference if they can be a function reference and cannot be a designator.
    
    IF (test_function_reference(tlist(start:finish))) THEN
      ALLOCATE(func_ex)
      CALL ParseFunctionReference(  &
          tlist,  &
          start, finish,  &
          func_ex%parts,  &
          func_ex%name,  &
          func_ex%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      CALL MOVE_ALLOC(func_ex, ex)
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! The whole thing must look like a designator.  Perhaps it is a badly 
    ! formed structure constructor or function reference, but c'est la vie, 
    ! we'll sort it out in the semantic analysis phase.
    !
    ! See also FunctionRefExprFromDesignator, which fixes the designator 
    ! to function reference case if the wrong choice is detected later.
    
    ALLOCATE(desig_ex)
    CALL ParseDesignator(tlist(start:finish), desig_ex%desig, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    CALL MOVE_ALLOC(desig_ex, ex)
    
  END SUBROUTINE parse_primary
  
  
  !*****************************************************************************
  !!
  !> Tests a sequence of tokens to see whether it could be considered a 
  !! /designator/ or /function-reference/ or whether it must be a structure 
  !! constructor.
  !!
  !! Something which looks like a /designator/ or /function-reference/ could 
  !! also just be a structure constructor, but the reverse does not 
  !! necessarily apply.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns .TRUE. if the sequence is definitely a structure constructor, 
  !! .FALSE. otherwise.
  !!
  !! Similar code in DataStmts::test_structure_constructor.
  !!
  !! A structure constructor can be distinguished from a function reference 
  !! if the type spec for the constructor has type parameters (there are 
  !! two sets of ().  If the type spec bit has multiple parts then it is 
  !! not a structure constructor.
  !!
  !! Both a function reference and a structure constructor can be 
  !! distinguished from a designator if the former have an empty argument 
  !! list or KEYWORD= form of actual argument specifiers.
  
  FUNCTION test_structure_constructor(tlist) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    LOGICAL :: b
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: closing        ! Index of the first closing delimiter.
    INTEGER :: closing_two    ! Index of the second closing delimiter.
    
    !***************************************************************************
    
    ! Structure constructor must start with a name.
    b = Match(tlist, 1, ittName)
    IF (.NOT. b) RETURN
    
    ! Structure constructor cannot have a bare '%'
    b = FindBare(tlist, ittPercent, scck_'%') == 0
    IF (.NOT. b) RETURN
    
    ! Second token of structure constructor must be a (.
    b = Match(tlist, 2, ittDelimiter, scck_'(')
    IF (.NOT. b) RETURN
    
    ! That ( must be matched.
    closing = FindBare(tlist, ittDelimiter, scck_')', 2)
    b = closing /= 0
    IF (.NOT. b) RETURN
    
    ! To be sure this is a structure constructor and not a function 
    ! reference, the type spec must have a parameter list (inside the set of 
    ! parentheses that we parsed above).  After that type spec there must be 
    ! a (possibly empty) ( component-spec-list ) sort of construct.
    b = Match(tlist, closing + 1, ittDelimiter, scck_'(')
    IF (.NOT. b) RETURN
    
    ! To distinguish structure constructors from substringed and subscripted 
    ! array references, the second parenthesised bit must not contain a 
    ! bare colon.  If we can't find a matching closing parenthesis, we 
    ! assume that the reference is a substring.
    closing_two = FindBare(tlist, ittDelimiter, scck_')', closing + 1)
    b = closing_two /= 0
    IF (.NOT. b) RETURN
    
    b = .NOT. ( &
        FindBare(tlist(closing+2:closing_two-1), ittColon, scck_':') /= 0 )
    
    ! That's sufficient for us to decide whether  we definitely have a 
    ! structure constructor.
    
  END FUNCTION test_structure_constructor
  
  
  !*****************************************************************************
  !!
  !> Test  a sequence of tokens to see whether it has to be considered a 
  !! designator or whether it must be treated as a /function-reference/ 
  !! or /structure-constructor/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @return .TRUE. if @a tlist cannot represent a /designator/ and must 
  !! represent a /function-reference/ or /structure-constructor/, .FALSE. 
  !! otherwise.
  
  FUNCTION test_function_reference(tlist) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    LOGICAL :: b
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: percent        ! Index of the next percent character.
    INTEGER :: it             ! Current token index.
    INTEGER :: closing        ! Index of the next closing delimiter.
    
    !***************************************************************************
    
    ! Fast forward to the last part in any procedure-designator that happens 
    ! to involve a multi-part /data-ref/.
    it = 1
    DO
      percent = FindBare(tlist, ittPercent, scck_'%', it)
      IF (percent == 0) EXIT
      it = percent + 1
    END DO
    
    ! The last apparent part of a /function-reference/ always looks like:
    !
    ! /name/ ( [ /actual-arg-spec-list/ ] )
    !
    ! The /name/ part comes from the last syntax part of the options for a 
    ! /procedure-designator/ - obvious for the /procedure-name/ and the
    ! /data-ref/ % /binding-name/ forms, and /proc-component-ref/ similarly 
    ! resolves to /scalar-variable/ % /procedure-component-name/.
    
    b = Match(tlist, it, ittName)
    IF (.NOT. b) RETURN
    
    b = Match(tlist, it + 1, ittDelimiter, scck_'(')
    IF (.NOT. b) RETURN
    
    closing = FindBare(tlist, ittDelimiter, scck_')', it + 1)
    IF (closing == 0) THEN
      b = .FALSE.
      RETURN
    END IF
    
    ! To definitely be a function reference and NOT a designator there needs 
    ! to be a KEYWORD=argument thing in between the parentheses, or there 
    ! needs to be nothing in between the parentheses, or there needs to be 
    ! a bare boz-literal-constant.
    b = closing == it + 2
    IF (b) RETURN
    
    b = FindBare(tlist(it+2:closing-1), ittAssign, scck_'=') /= 0
    IF (b) RETURN
    
    b = FindBare(tlist(it+2:closing-1), ittBozLiteral, 1) /= 0
    
  END FUNCTION test_function_reference
  
  
  !*****************************************************************************
  !!
  !> Find the position of the last bare user defined binary operator (R723).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token.
  !!
  !! A user defined binary operator is anything that isn't a built in 
  !! operator and that also isn't a unary operator.
  !!
  !! To allow for constructions such as "lhs .binary. .unary. rhs" we skip 
  !! back over potential unary operators.
  !!
  !! - `a .op1. .op2. b` - op1 must be a user defined binary operator, 
  !!   while op2 is a user defined unary operator.
  !!
  !! - `a .op1. + b` - op1 must be a user defined binary operator, 
  !!   the `+` is unary addition.
  !!
  !! - `a * .op2. b` - op2 must be a user defined unary operator, 
  !!   the `*` is binary multiplication.
  !!
  !! - `a .op1. * b` - syntax error.  If you treat op1 as a binary operator 
  !!   then the `*` has no argument.  If you treat the `*` as a binary 
  !!   operator then the op1 cannot be second in a term.  This function 
  !!   treats `op1` as a binary, we rely on processing of the other 
  !!   operator to pick up the error.
  !!
  !! - `( .op. a)` - no operator there.
  !!
  !! = `( a .op. b)`- no operator there (it is a parenthesised expression.
  !!
  !! This is used in the parsing of the /expr/ syntax rule.
  
  FUNCTION find_defined_binary_op(tlist) RESULT(it)
    
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    ! The FindBare search must start from one or we might miss parentheses.
    it = 1
    DO WHILE (it <= SIZE(tlist))
      it = FindBare(tlist, ittOperator, it)
      IF (it == 0) EXIT     ! No operator found.
      ! First token cannot be a binary operator (it might be a unary).
      IF (it /= 1) THEN
        IF (is_user_defined(QueryValue(tlist(it)))) THEN
          ! Check that the previous token isn't an operator.  If it is then 
          ! the token at it can't be a valid binary operator.  Note that we 
          ! only start searching from it == 2 on, so we don't have to worry 
          ! about the following going out of bounds.
          IF (tlist(it - 1) /= ittOperator) RETURN
        END IF
      END IF
      ! Skip over and continue searching
      it = it + 1
    END DO
    
  END FUNCTION find_defined_binary_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of a equiv-op (R721).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token.
  !!
  !! An equiv-op is .EQV. or .NEQV.
  !!
  !! This is used in parsing the /level-5-expr/ syntax rule.
  
  FUNCTION find_equiv_op(tlist) RESULT(it)
    
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    ! Scan backwards looking for operator tokens.
    it = 1
    DO WHILE (it <= SIZE(tlist))
      it = FindBare(tlist, ittOperator, it)
      IF (it == 0) EXIT       ! No such operator.
      ! If the operator token is .EQV. or .NEQV., we've found what we are 
      ! looking for.  Potentaially it could be 1 here, but latter parsing 
      ! should identify that as problematic.
      IF ((tlist(it) == '.EQV.') .OR. (tlist(it) == '.NEQV.')) RETURN
      ! Skip over and continue searching.
      it = it + 1
    END DO
    
    it = 0
    
  END FUNCTION find_equiv_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of an /or-op/ (R720) - ".OR.".
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token in @a tlist.
  !!
  !! This is used in parsing the /equiv-operand/ syntax rule.
  
  FUNCTION find_or_op(tlist) RESULT(it)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = FindBare(tlist, ittOperator, scck_'.OR.')
    
  END FUNCTION find_or_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of an /and-op/ (R719) - ".AND.".
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token in @a tlist.
  !!
  !! This is used in parsing the /or-operand/ syntax rule.
  
  FUNCTION find_and_op(tlist) RESULT(it)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = FindBare(tlist, ittOperator, scck_'.AND.')
    
  END FUNCTION find_and_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of a rel-op (R713).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index of the token from which 
  !! to start searching.
  !!
  !! @returns The position of the token relative to @a start.
  !!
  !! An equiv-op is .EQV. or .NEQV.
  
  FUNCTION find_rel_op(tlist, start) RESULT(it)
    
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = start - 1
    DO
      it = FindBare(tlist, ittOperator, it + 1)
      IF (it == 0) EXIT
      IF ( (tlist(it) == '.EQ.')  &
          .OR. (tlist(it) == '.NE.')  &
          .OR. (tlist(it) == '.LT.')  &
          .OR. (tlist(it) == '.LE.')  &
          .OR. (tlist(it) == '.GT.')  &
          .OR. (tlist(it) == '.GE.')  &
          .OR. (tlist(it) == '==')  &
          .OR. (tlist(it) == '/=')  &
          .OR. (tlist(it) == '<')  &
          .OR. (tlist(it) == '<=')  &
          .OR. (tlist(it) == '>')  &
          .OR. (tlist(it) == '>=') ) RETURN
    END DO
    
  END FUNCTION find_rel_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of an /concat-op/ (R711) - "//".
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token in @a tlist.
  !!
  !! This is used in parsing the /level-3-expr/ syntax rule.
  
  FUNCTION find_concat_op(tlist) RESULT(it)
    
    USE CompilerKinds
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = FindBare(tlist, ittOperator, scck_'//')
    
  END FUNCTION find_concat_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of a mult-op (R709).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @returns The position of the token in @a tlist.
  !!
  !! An add-op is * or /.
  !!
  !! This is used in parsing the /add-operand/ syntax rule.
  
  FUNCTION find_mult_op(tlist) RESULT(it)
    
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = 1
    DO WHILE (it <= SIZE(tlist))
      it = FindBare(tlist, ittOperator, it)
      IF (it == 0) EXIT
      IF ((tlist(it) == '*') .OR. (tlist(it) == '/')) RETURN
      ! Skip over and continue.
      it = it + 1
    END DO
    
    it = 0
    
  END FUNCTION find_mult_op
  
  
  !*****************************************************************************
  !!
  !> Find the position of an add-op (R709).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The index of the token from which 
  !! to start searching.
  !!
  !! @returns The position of the token relative to @a start.
  !!
  !! An add-op is + or -.
  
  FUNCTION find_add_op(tlist, start) RESULT(it)
    
    USE MatchUtils
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    INTEGER :: it
    
    !***************************************************************************
    
    it = start - 1
    DO 
      it = FindBare(tlist, ittOperator, it + 1)
      IF (it == 0) EXIT
      IF ((tlist(it) == '+') .OR. (tlist(it) == '-')) RETURN
    END DO
    
  END FUNCTION find_add_op
  
  
  !*****************************************************************************
  !!
  !> Test an operator name to see if it is user defined.
  !!
  !! @param[in]     op_name           Name of the operator.
  !!
  !! @returns .TRUE. if @a name the name of a built in operator, .FALSE. 
  !! otherwise.
  !!
  !! A user defined operator is any operator that we don't recognise as 
  !! being built in.  It must start with a dot.
  !!
  !! The program may extend the intrinsic operators, but that doesn't change 
  !! the precedence rules, which is what this test is for.
  !!
  !! This procedure helps us enforce C703 and C704.
  
  FUNCTION is_user_defined(op_name) RESULT(b)
    
    USE CompilerKinds
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(*,scck), INTENT(IN) :: op_name
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = .NOT. ( (op_name(1:1) /= scck_'.')  &
        .OR. (op_name == scck_'.EQ.')  &
        .OR. (op_name == scck_'.NE.')  &
        .OR. (op_name == scck_'.EQV.')  &
        .OR. (op_name == scck_'.NEQV.')  &
        .OR. (op_name == scck_'.OR.')  &
        .OR. (op_name == scck_'.AND.')  &
        .OR. (op_name == scck_'.LT.')  &
        .OR. (op_name == scck_'.LE.')  &
        .OR. (op_name == scck_'.GT.')  &
        .OR. (op_name == scck_'.GE.')  &
        .OR. (op_name == scck_'.NOT.') )
    
  END FUNCTION is_user_defined
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of an expression to a unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the expression to.
  !!
  !! @param[in]     expr              The expression to write out.
  !!
  !! This is on its way to becoming defunct.  It has been replaced by 
  !! the Write binding of the Expression type.  Some call 
  !! paths remain because code is yet to be updated.
  !!
  !! Recursive because it is recursive.
  
  RECURSIVE SUBROUTINE WriteExpressionInternal(log_unit, ex)
    
    USE ArrayConstructors
    USE CallStmts
    USE Designators
    USE PrimaryExpressions
    USE StructureConstructors
    USE TypeParamInquiries
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    CLASS(Expression), INTENT(IN) :: ex
    
    !***************************************************************************
    
    SELECT TYPE (ex)
    TYPE IS (UnaryOpExpr)
      WRITE (log_unit, "(A,' {')", ADVANCE='NO') .NameOf. ex%op
      CALL ex%rhs%Write(log_unit)
      WRITE (log_unit, "('}')", ADVANCE='NO')
      
    TYPE IS (BinaryOpExpr)
      WRITE (log_unit, "('{')", ADVANCE='NO')
      CALL ex%lhs%Write(log_unit)
      WRITE (log_unit, "('} ',A,' {')", ADVANCE='NO') .NameOf. ex%op
      CALL ex%rhs%Write(log_unit)
      WRITE (log_unit, "('}')", ADVANCE='NO')
      
    TYPE IS (ConstantExpr)
      SELECT CASE (QueryType(ex%tok))
      CASE (ittIntegerLiteral, ittRealLiteral, ittLogicalLiteral)
        IF (HasKind(ex%tok)) THEN
          WRITE (log_unit, "(A,'_',A)", ADVANCE='NO')  &
              QueryValue(ex%tok), QueryKind(ex%tok)
        ELSE
          WRITE (log_unit, "(A)", ADVANCE='NO') QueryValue(ex%tok)
        END IF
      CASE (ittCharacterLiteral)
        IF (HasKind(ex%tok)) THEN
          WRITE (log_unit, "(A,'_''',A,'''')", ADVANCE='NO')  &
              QueryKind(ex%tok), QueryValue(ex%tok)
        ELSE
          WRITE (log_unit, "('''',A,'''')", ADVANCE='NO') QueryValue(ex%tok)
        END IF
      CASE DEFAULT
        ! Not expectecting this.
        WRITE (log_unit, "(A)", ADVANCE='NO') QueryRaw(ex%tok)
      END SELECT
      
    TYPE IS (ComplexLiteralExpr)
      CALL ex%constant%Write(log_unit)
      
    TYPE IS (DesignatorExpr)
      CALL WriteDesignator(log_unit, ex%desig)
      
    TYPE IS (ArrayConstructorExpr)
      CALL WriteArrayConstructor(log_unit, ex%type_spec, ex%values)
      
    TYPE IS (StructureConstructorExpr)
      CALL ex%sc%Write(log_unit)
      
    TYPE IS (FunctionRefExpr)
      CALL WriteProcedureReference(  &
          log_unit,  &
          ex%parts,  &
          ex%name,  &
          ex%args )
      
    TYPE IS (ParenthesisedExpr)
       WRITE (log_unit, "('(')", ADVANCE='NO')
       CALL ex%ex%Write(log_unit)
       WRITE (log_unit, "(')')", ADVANCE='NO')
      
    TYPE IS (TypeParamInquiry)
      CALL ex%Write(log_unit)
      
    CLASS DEFAULT
      WRITE (log_unit, "('?expression?')", ADVANCE='NO')
      
    END SELECT
    
  END SUBROUTINE WriteExpressionInternal
  
END MODULE expression_parse_utils


!*****************************************************************************
!!
!> External procedure that implements ParseExpr - parse a sequence of 
!! tokens that is an /expr/ (R722).
!!
!! @param[in]     tlist             The sequence of tokens that make up the 
!! expression.  Note that tokens that are part of @a tlist but not part of 
!! the expression will result in an error.
!!
!! @param[in]     start             Index of the first token of the 
!! /expression/.
!!
!! @param[in]     finish            Index of the last token of the 
!! /expression/.
!!
!! @param[out]    ex                Allocated to the expression if one was 
!! successfully parsed.
!!
!! @param[out]    err_list          List of errors.
!!
!! This is an external subprogram because everything involves expressions, 
!! therefore everything needs to parse an expression, but expressions involve 
!! a lot of other things (arrays, variables, substrings, constants, functions 
!! etc) - so you end up with this horrendous cycle.  When submodules are 
!! supported widely we can make this a submodule.

RECURSIVE SUBROUTINE ParseExprExternal(tlist, start, finish, ex, err_list)
  
  USE Errors
  USE Expressions, ONLY: Expression   ! Must exclude self interface!
  USE Tokens
  
  USE expression_parse_utils
  
  IMPLICIT NONE
  
  !---------------------------------------------------------------------------
  ! Arguments
  
  TYPE(Token), INTENT(IN) :: tlist(:)
  INTEGER, INTENT(IN) :: start
  INTEGER, INTENT(IN) :: finish
  CLASS(Expression), INTENT(OUT), ALLOCATABLE :: ex
  TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
  
  !***************************************************************************
  
  CALL ParseExprInternal(tlist, start, finish, ex, err_list)
  
END SUBROUTINE ParseExprExternal


!*******************************************************************************
!!
!> External procedure that implements WriteExpr - write a debugging 
!! representation of an expression to a logical unit.
!!
!! @param[in]     unit              Logical unit connected for formatted 
!! output.
!!
!! @param[in]     ex                The expression to write.
!!
!! To avoid problems with us accidentally making visible our own interface 
!! inside this procedure we forward the call onto an implementation 
!! procedure inside the expression_parse_utils module.

RECURSIVE SUBROUTINE WriteExprExternal(unit, ex)
  
  USE Expressions, ONLY: Expression
  
  USE expression_parse_utils
  
  IMPLICIT NONE
  
  !-----------------------------------------------------------------------------
  ! Arguments
  
  INTEGER, INTENT(IN) :: unit
  CLASS(Expression), INTENT(IN) :: ex
  
  !*****************************************************************************
  
  CALL WriteExpressionInternal(unit, ex)
  
END SUBROUTINE WriteExprExternal
