! $Id: PrintStmts.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PrintStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /print-stmt/'s.

MODULE PrintStmts
  
  USE Formats, ONLY: Format
  USE InputOutputItems
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParsePrintStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /print-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: PrintStmt
    !> The format specification of the /print-stmt/.
    TYPE(Format) :: fmt
    !> The output items of the /print-stmt/.
    !!
    !! Always allocated, may be zero sized.
    TYPE(IOItemList), ALLOCATABLE :: items(:)
  CONTAINS
    PROCEDURE :: Dump => print_Dump
    PROCEDURE :: Query => print_Query
    PROCEDURE :: Modify => print_Modify
    PROCEDURE, NOPASS :: What => print_What
  END TYPE PrintStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that make up a /print-stmt/ (R912).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    print_stmt        The results of parsing the /print-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParsePrintStmt(tlist, print_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ErrorLevels
    USE Formats
    USE InputOutputItems
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(PrintStmt), INTENT(OUT) :: print_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /print-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R912'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: comma          ! Comma position after the /format/.
    INTEGER :: it             ! Current token index.
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    ! Error list for child procedure calls
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /print-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! PRINT ...
    
    CALL ExpectKeyword(tlist, 1, scck_'PRINT', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! /format/
    
    IF (SIZE(tlist) == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,   &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(1)),  &
          MSG='Expected /format/.' )
      RETURN
    END IF
    
    ! Work out where the format spec runs to.
    comma = FindBare(tlist, ittComma, scck_',', 2)
    IF (comma == 0) THEN
      it = SIZE(tlist)
    ELSE
      it = comma - 1
    END IF
    
    CALL ParseFormat(tlist, 2, it, print_stmt%fmt, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ! Keep going on error.
    
    IF (comma == 0) THEN
      ALLOCATE(print_stmt%items(0))
      RETURN
    END IF
    
    it = comma + 1      ! Index of token after the comma.
    
    !---------------------------------------------------------------------------
    ! /output-item-list/
    
    CALL ParseIOItemList(  &
        tlist,  &
        it,  &
        SIZE(tlist),  &
        .TRUE.,  &            ! Specify an /output-item-list/.
        print_stmt%items,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParsePrintStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of PrintStmt%Dump - write a debugging representation 
  !! of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  !!
  !! @param[in]     pfx               Prefix to write before each record.
  
  SUBROUTINE print_Dump(pnode, unit, pfx)
    
    USE Formats
    USE InputOutputItems
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PrintStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/format/ = ')", ADVANCE='NO') pfx
    CALL pnode%fmt%Write(unit)
    WRITE (unit, "()")
      
    WRITE (unit, "(A,'/output-item-list/')") pfx
    CALL DumpIOItemList(pnode%items, unit, pfx // '  ')
    
  END SUBROUTINE print_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of PrintStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE print_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PrintStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fmt%Query(visitor)
    
    CALL QueryIOItemList(pnode%items, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE print_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of PrintStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE print_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PrintStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%fmt%Modify(visitor)
    
    CALL ModifyIOItemList(pnode%items, visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE print_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of PrintStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'print-stmt'
  
  FUNCTION print_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'print-stmt'
    
  END FUNCTION print_What
  
END MODULE PrintStmts
