! $Id: Programs.f90 2363 2016-08-22 01:29:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the Programs module.


!*******************************************************************************
!!
!> Types and procedures for parsing the program as a whole.
!!
!! To parse a program, build the statement tree (see BuildTreePass), then 
!! pass the top level tree node to the ParseProgram procedure.  Results of 
!! a successful parse are stored in an object of type Program.
!!
!! This is pretty straight-forward from a client code point of view, so there 
!! is no "ParsePass" entry point.  General documentation for the conceptual 
!! ParsePass entry point is below.
!!
!! It is plausible that the construct level parsing routines may not actually 
!! be used in anger in programs that have later passes - it may make more 
!! sense for the advertise pass to do the same job itself.  This is a work 
!! in progress - we have completed the construct parsing bits anyway (there 
!! aren't that many of them - most of the work is in the statement parsing 
!! procedures) to assist in testing.
!!
!! In a very similar (too similar, perhaps) manner to the way things are 
!! stored in the statement tree, an StNode is generally mapped across to a 
!! ParseNode by a ParseXxxx procedure (i.e. the ParseXxxx procedure 
!! understands both the relevant StNode and the resulting ParseNode).
!!
!! Child ParseNodes are not generally created for source elements that are 
!! just structural - for example the opening and closing statements of 
!! constructs.  Procedures that deal with the parsing of these structural 
!! statements do not return ParseNode extensions and the information 
!! returned from parsing those statements (if any) is simply stored in 
!! components in the parent construct.

MODULE Programs
  
  USE ParseNodes
  USE ProgramUnits
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseProgram
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing [part of] a program.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: Program
    !> The program units that make up the program.
    TYPE(ProgramUnitElement), ALLOCATABLE :: program_units(:)
  CONTAINS
    PROCEDURE :: Dump => program_Dump
    PROCEDURE, NOPASS :: What => program_What
    PROCEDURE :: Query => program_Query
    PROCEDURE :: Modify => program_Modify
  END TYPE Program
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse the top level statement tree node for the entire program (or 
  !! some collection of program units that make up a program).
  !!
  !! @param[in]     st                The statement tree node for the 
  !! program
  !!
  !! @param[out]    prog              The results of parsing the 
  !! program.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Program units that fail to parse correctly as not stored, so the 
  !! program_units component is still usable even if errors are returned.
  !!
  !! We don't enforce uniqueness of names, - that is left to the Advertise 
  !! pass when attempts to add the program unit name to the entity table 
  !! will cause an error due to the name clash (or perhaps even later, 
  !! in linking).
  !!
  !! We don't enforce the requirement that there be one and only one main 
  !! program.  We don't enforce the requriement that there be at most one 
  !! unnamed block data.  These aspects are left for later passes (perhaps 
  !! linking).
  !!
  !! This also includes the call to the Init binding to intially define 
  !! the parent component.
  
  SUBROUTINE ParseProgram(st, prog, err_list)
    
    USE BaseStNodes
    USe BlockDataStNodes
    USE Errors
    USE ModuleStNodes
    USE ProgUnitStNodes
    USE ProgramUnits
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(BaseStNode), INTENT(INOUT) :: st
    TYPE(Program), INTENT(OUT) :: prog
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Program unit tre index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Objects that are constructed, then moved into the program_units vector.
    
    ! Object for parsing the main program.
    !
    ! Technically there should only ever be one of these in a conforming 
    ! program, but we allow for more here at this stage.
    TYPE(MainProgram), ALLOCATABLE :: main_program
    
    ! Object for parsing an external subroutine subprogram.
    TYPE(SubroutineSubprogram), ALLOCATABLE :: subroutine_subprogram
    
    ! Object for parsing an external function subprogram.
    TYPE(FunctionSubprogram), ALLOCATABLE :: function_subprogram
    
    ! Object for parsing a block data unit.
    TYPE(BlockData), ALLOCATABLE :: block_data_unit
    
    ! Object for parsing a module.
    TYPE(Module), ALLOCATABLE :: module_unit
    
    ! Object for parsing a submodule.
    TYPE(Submodule), ALLOCATABLE :: submodule_unit
    
    ! Object for a general program unit.
    CLASS(ProgramUnit), ALLOCATABLE :: the_unit
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ! Initially define the ParseNode parent component.
    CALL prog%Init(st)
    
    !---------------------------------------------------------------------------
    
    ! Loop over child statement tree nodes, and parse them.
    ALLOCATE(prog%program_units(0))
    
    DO it = 1, SIZE(st%prog_units)
      ! The pattern is to allocate an object of the program unit specific 
      ! type, parse the relevant StNode into that object, on success move the 
      ! object into an object of declared type ProgramUnit, then at the end 
      ! of the select type construct add that ProgramUnit object to the list.
      SELECT TYPE (item => st%prog_units(it)%item)
      TYPE IS (MainProgramStNode)
        ALLOCATE(main_program)
        CALL ParseMainProgram(item, main_program, sub_err_list)
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(main_program)
        ELSE
          CALL MOVE_ALLOC(main_program, the_unit)
        END IF
        
      TYPE IS (FunctionSubprogramStNode)
        ALLOCATE(function_subprogram)
        CALL ParseFunctionSubprogram(  &
            item,  &
            ispntExternal,  &
            function_subprogram,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(function_subprogram)
        ELSE
          CALL MOVE_ALLOC(function_subprogram, the_unit)
        END IF
        
      TYPE IS (SubroutineSubprogramStNode)
        ALLOCATE(subroutine_subprogram)
        CALL ParseSubroutineSubprogram(  &
            item,  &
            ispntExternal,  &
            subroutine_subprogram,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(subroutine_subprogram)
        ELSE
          CALL MOVE_ALLOC(subroutine_subprogram, the_unit)
        END IF
        
      TYPE IS (BlockDataStNode)
        ALLOCATE(block_data_unit)
        CALL ParseBlockData(item, block_data_unit, sub_err_list)
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(block_data_unit)
        ELSE
          CALL MOVE_ALLOC(block_data_unit, the_unit)
        END IF
        
      TYPE IS (ModuleStNode)
        ALLOCATE(module_unit)
        CALL ParseModule(item, module_unit, sub_err_list)
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(module_unit)
        ELSE
          CALL MOVE_ALLOC(module_unit, the_unit)
        END IF
        
      TYPE IS (SubmoduleStNode)
        ALLOCATE(submodule_unit)
        CALL ParseSubmodule(item, submodule_unit, sub_err_list)
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) THEN
          DEALLOCATE(submodule_unit)
        ELSE
          CALL MOVE_ALLOC(submodule_unit, the_unit)
        END IF
        
      CLASS DEFAULT
        STOP 'Internal error in Programs%%ParseProgram, &
            &StNode is the wrong type.'
        
      END SELECT
      
      IF (ALLOCATED(the_unit)) THEN
        CALL the_unit%Init(st%prog_units(it)%item)
        CALL AddProgramUnit(prog%program_units, the_unit)
      END IF
    END DO
    
  END SUBROUTINE ParseProgram
  
  
  !*****************************************************************************
  !!
  !> Implementation of Program%Dump - write a debugging representation 
  !! of the results of parsing a program to a logical unit.
  !!
  !! @param[in]     pnode             The MainProgram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE program_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Program), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Program unit index.
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode%program_units)
      WRITE (unit, "(A,'/',A,'/')") pfx, pnode%program_units(i)%item%What()
      CALL pnode%program_units(i)%item%Dump(unit, pfx // '  ')
    END DO
    
  END SUBROUTINE program_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of Program%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'program'.
  
  FUNCTION program_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'program'
    
  END FUNCTION program_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Program%Query - execute a query traversal operation on 
  !! the node.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE program_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Program), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%program_units)
      CALL pnode%program_units(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE program_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of Program%Modify - execute a modify traversal operation on 
  !! the node.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE program_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(Program), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Child node index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%program_units)
      CALL pnode%program_units(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE program_Modify
  
END MODULE Programs
