! $Id: RenderNodes.f90 2779 2017-02-02 10:38:57Z ian $
!> @file
!! Defines the RenderParseNodes module.


!*******************************************************************************
!!
!> Defines types to help re-create a token stream.
!!
!! We use a new RenderNodeToken type because we are not really sure what 
!! we want to stick in it for now.  Perhaps this could just be an extension 
!! of Tokens%%Token.

MODULE RenderNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: AppendRNode
  PUBLIC :: AppendRNodes
  PUBLIC :: AppendConstruct
  PUBLIC :: AppendStmt
  PUBLIC :: AppendSeq
  PUBLIC :: AppendKeyword
  PUBLIC :: AppendName
  PUBLIC :: AppendLabel
  PUBLIC :: AppendDelimiter
  PUBLIC :: AppendComma
  PUBLIC :: AppendToken
  PUBLIC :: AppendComment
  
  !-----------------------------------------------------------------------------
  
  !> The output character kind.
  INTEGER, PARAMETER, PUBLIC :: ock = KIND('A')
  
  !> A node in the render tree.
  TYPE, PUBLIC :: RenderNode
    !> The source context assocaited with the node.  This is currently defined 
    !! to be the syntax rule number.
    !!
    !! This is a bit of a mess.  Parse nodes currently have a What binding 
    !! that return the syntax rule name.  Perhaps this should be that term.
    INTEGER :: context
  END TYPE RenderNode
  
  !> A comment.
  TYPE, EXTENDS(RenderNode), PUBLIC :: RenderNodeComment
    !> The contents of the comment.
    CHARACTER(:,KIND=ock), ALLOCATABLE :: content
    !> The prefix for the comment.
    !!
    !! If not allocated then the prefix is '! ' (exclamation mark 
    !! followed by a space).
    CHARACTER(:,KIND=ock), ALLOCATABLE :: prefix
    character(8) :: tag = 'rnodcmmt'
  END TYPE RenderNodeComment
  
  !> Describe a token for output.
  TYPE, EXTENDS(RenderNode), PUBLIC :: RenderNodeToken
    !> The type of the token.
    !!
    !! Note that in some cases we classify things to a different type than 
    !! the scanner would classify, based on the additional context that 
    !! is available with the parse tree.  In particular, a `/` is 
    !! classified as a delimiter if it is being used as one in the 
    !! syntax.
    INTEGER :: type
    !> The value part of the token.
    !!
    !! This is case sensitive for some token types.
    !!
    !! For character literal tokens it DOES NOT include the leading and 
    !! trailing quotes, and similarly DOES NOT include things like doubling 
    !! of quotes.
    !!
    !! This is always allocated.  I can't think of any reason why it won't 
    !! always have a size greater than zero.
    CHARACTER(:,KIND=ock), ALLOCATABLE :: value
    !> The kind part of the token.
    !!
    !! This is always allocated, but will be zero length for tokens that 
    !! do not have a kind part.  Some token types (names, keywords, 
    !! operators...) never have a kind part
    CHARACTER(:,KIND=ock), ALLOCATABLE :: kind
    character(8) :: tag = 'rnodtokn'
  END TYPE RenderNodeToken
  
  !> An element in an array of render nodes.
  TYPE, PUBLIC :: RenderNodeElement
    !> The item of the element.
    CLASS(RenderNode), ALLOCATABLE :: item
    character(8) :: tag = 'rnodelem'
  END TYPE RenderNodeElement
  
  !> A render node that is a sequence of other nodes.
  TYPE, EXTENDS(RenderNode), PUBLIC :: RenderNodeSequence
    !> The sequence.
    TYPE(RenderNodeElement), ALLOCATABLE :: items(:)
  END TYPE RenderNodeSequence
  
  !> A render node that is a sequence that makes a statement.
  TYPE, EXTENDS(RenderNodeSequence), PUBLIC :: RenderNodeStatement
    character(8) :: tag = 'rnodstmt'
  END TYPE RenderNodeStatement
  
  !> A render node that is a sequence that makes a construct.
  TYPE, EXTENDS(RenderNodeSequence), PUBLIC :: RenderNodeConstruct
    character(8) :: tag = 'rnodcons'
  END TYPE RenderNodeConstruct
  
  INTERFACE AppendName
    MODULE PROCEDURE AppendName_char
    MODULE PROCEDURE AppendName_identifier
  END INTERFACE AppendName
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Grow a sequence of nodes
  !!
  SUBROUTINE grow_sequence(items, sz)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(RenderNodeElement), INTENT(INOUT), ALLOCATABLE :: items(:)
    INTEGER, INTENT(IN), OPTIONAL :: sz
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Temporary for growing seq%items.
    TYPE(RenderNodeElement), ALLOCATABLE :: tmp(:)
    
    INTEGER :: local_sz       ! Defaulted variant of sz.
    INTEGER :: i              ! Sequence index.
    
    !***************************************************************************
    
    IF (PRESENT(sz)) THEN
      local_sz = sz
    ELSE
      local_sz = 1
    END IF
    
    IF (ALLOCATED(items)) THEN
      ALLOCATE(tmp(SIZE(items) + local_sz))
      DO i = 1, SIZE(items)
        CALL MOVE_ALLOC(items(i)%item, tmp(i)%item)
      END DO
    ELSE
      ALLOCATE(tmp(local_sz))
    END IF
    CALL MOVE_ALLOC(tmp, items)
    
  END SUBROUTINE grow_sequence
  
  
  !*****************************************************************************
  !!
  !> Append another render node to a sequence of nodes.
  
  SUBROUTINE AppendRNode(seq, rnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderNode), INTENT(INOUT), ALLOCATABLE :: rnode
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    CALL MOVE_ALLOC(rnode, seq%items(SIZE(seq%items))%item)
    
  END SUBROUTINE AppendRNode
    
  
  !*****************************************************************************
  !!
  !> Append a sequence of render nodes to another sequence of nodes.
  
  SUBROUTINE AppendRNodes(seq, rnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderNodeSequence), INTENT(INOUT) :: rnode
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: start          ! Starting size of @a seq.
    INTEGER :: i              ! Node index.
    
    !***************************************************************************
    
    IF (ALLOCATED(seq%items)) THEN
      start = SIZE(seq%items)
    ELSE
      start = 0
    END IF
    
    CALL grow_sequence(seq%items, SIZE(rnode%items))
    DO i = 1, SIZE(rnode%items)
      CALL MOVE_ALLOC(  &
          rnode%items(i)%item,  &
          seq%items(start + i)%item )
    END DO
    DEALLOCATE(rnode%items)
    ALLOCATE(rnode%items(0))
    
  END SUBROUTINE AppendRNodes
    
  
  !*****************************************************************************
  !!
  !> Append a statement node to a sequence of nodes.
  
  SUBROUTINE AppendConstruct(seq, rnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    TYPE(RenderNodeConstruct), INTENT(INOUT), ALLOCATABLE :: rnode
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    CALL MOVE_ALLOC(rnode, seq%items(SIZE(seq%items))%item)
    
  END SUBROUTINE AppendConstruct
    
  
  !*****************************************************************************
  !!
  !> Append a statement node to a sequence of nodes.
  
  SUBROUTINE AppendStmt(seq, rnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    TYPE(RenderNodeStatement), INTENT(INOUT), ALLOCATABLE :: rnode
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    CALL MOVE_ALLOC(rnode, seq%items(SIZE(seq%items))%item)
    
  END SUBROUTINE AppendStmt
    
  
  !*****************************************************************************
  !!
  !> Append a sequence node to another sequence of nodes.
  
  SUBROUTINE AppendSeq(seq, rnode)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    TYPE(RenderNodeSequence), INTENT(INOUT), ALLOCATABLE :: rnode
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    CALL MOVE_ALLOC(rnode, seq%items(SIZE(seq%items))%item)
    
  END SUBROUTINE AppendSeq
    
  
  !*****************************************************************************
  !!
  !> Append a keyword token to a sequence of nodes.
  
  SUBROUTINE AppendKeyword(seq, text, context)
    
    USE Tokens, ONLY: ittKeyword
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CHARACTER(*,ock), INTENT(IN) :: text
    INTEGER, INTENT(IN) :: context
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(context, ittKeyword, text, '') )
    
  END SUBROUTINE AppendKeyword
  
  
  !*****************************************************************************
  !!
  !> Append a name token to a sequence of OutTokens.
  
  SUBROUTINE AppendName_char(seq, text, context)
    
    USE Tokens, ONLY: ittName
      
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CHARACTER(*,ock), INTENT(IN) :: text
    INTEGER, INTENT(IN) :: context
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(context, ittName, text, '') )
    
  END SUBROUTINE AppendName_char
  
  
  !*****************************************************************************
  !!
  !> Append a name token to a sequence of OutTokens.
  
  SUBROUTINE AppendName_identifier(seq, ident, context)
    
    USE IdentifierUtils
    USE Tokens, ONLY: ittName
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    TYPE(Identifier), INTENT(IN) :: ident
    INTEGER, INTENT(IN) :: context
    
    !---------------------------------------------------------------------------
    ! Locals
    
    CHARACTER(:), ALLOCATABLE :: ifort_bug_temp
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    !===========================================================================
    ifort_bug_temp = .NameOf. ident
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(context, ittName, ifort_bug_temp, '') )
    !===========================================================================
    
  END SUBROUTINE AppendName_identifier
  
  
  !*****************************************************************************
  !!
  !> Append a label to a sequence of OutTokens.
  
  SUBROUTINE AppendLabel(seq, the_label)
    
    USE LabelsUtils
    USE Tokens, ONLY: ittStatementLabel
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    TYPE(Label), INTENT(IN) :: the_label
    
    !---------------------------------------------------------------------------
    ! Locals
    
    CHARACTER(MaxLabelLength) :: label_text
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    WRITE (label_text, "(I0)") .ValueOf. the_label
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(0, ittStatementLabel, TRIM(label_text), '') )
    
  END SUBROUTINE AppendLabel
  
  
  !*****************************************************************************
  !!
  !> Append a delimiter token to a sequence of OutTokens.
  
  SUBROUTINE AppendDelimiter(seq, text)
    
    USE Tokens, ONLY: ittDelimiter
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CHARACTER(*,ock), INTENT(IN) :: text
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(0, ittDelimiter, text, '') )
    
  END SUBROUTINE AppendDelimiter
  
  
  !*****************************************************************************
  !!
  !> Append a comma token to a sequence of OutTokens.
  
  SUBROUTINE AppendComma(seq)
    
    USE Tokens, ONLY: ittComma
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(0, ittComma, ',', '') )
    
  END SUBROUTINE AppendComma
  
  
  !*****************************************************************************
  !!
  !> Append a general token to a sequence of OutTokens.
  
  SUBROUTINE AppendToken(seq, type, value, kind, context)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    INTEGER, INTENT(IN) :: type
    CHARACTER(*,ock), INTENT(IN) :: value
    CHARACTER(*,ock), INTENT(IN) :: kind
    INTEGER, INTENT(IN) :: context
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE( seq%items(SIZE(seq%items))%item,  &
        SOURCE=RenderNodeToken(context, type, value, kind) )
    
  END SUBROUTINE AppendToken
  
  
  !*****************************************************************************
  !!
  !> Append a comment to a sequence of render nodes.
  
  SUBROUTINE AppendComment(seq, text, prefix)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CHARACTER(*,ock), INTENT(IN) :: text
    CHARACTER(*,ock), INTENT(IN), OPTIONAL :: prefix
    
    !---------------------------------------------------------------------------
    ! Locals
    
    CLASS(RenderNodeComment), ALLOCATABLE :: tmp
    
    !***************************************************************************
    
    CALL grow_sequence(seq%items)
    ALLOCATE(tmp)
    tmp%content = text
    IF (PRESENT(prefix)) tmp%prefix = prefix
    CALL MOVE_ALLOC(tmp, seq%items(SIZE(seq%items))%item)
    
  END SUBROUTINE AppendComment
  
END MODULE RenderNodes
