! $Id: RewindStmts.f90 2892 2019-04-07 00:34:44Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the RewindStmts module.


!*******************************************************************************
!!
!> Types and procedures for parsing /rewind-stmt/'s (R926).

MODULE RewindStmts
  
  USE BackspaceStmts
  USE PositionSpecs
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseRewindStmt
  
  !> Stores the results of parsing a /rewind-stmt/.
  !!
  !! The information for a backspace statement is practically the same, 
  !! so we cheat.
  TYPE, EXTENDS(BackspaceStmt), PUBLIC :: RewindStmt
  CONTAINS
    PROCEDURE, NOPASS :: What => rewind_What
  END TYPE RewindStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /rewind-stmt/ (R926).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    rewind_stmt       The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseRewindStmt(tlist, rewind_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(RewindStmt), INTENT(OUT) :: rewind_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /rewind-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R926'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: closing        ! Index of the closing parenthesis.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /rewind-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! REWIND...
    
    CALL ExpectKeyword(tlist, 1, scck_'REWIND', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (SIZE(tlist) == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(SIZE(tlist))),  &
          MSG='Expected /file-unit-number/ or "(" /position-spec-list/ ")".' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    IF (TestPositionStmtFirstForm(tlist(2:))) THEN
      CALL ParseExpr(  &
          tlist,  &
          2, SIZE(tlist),  &
          rewind_stmt%position_specs%unit,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Failure not tested by StmtParsingTests because a missing 
      ! parenthesis causes the first form to be selected.
      CALL Expect(tlist, 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ParsePositionSpecList(  &
          tlist, 3,  &
          closing - 1,  &
          rewind_stmt%position_specs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      
      CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    END IF
    
  END SUBROUTINE ParseRewindStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of RewindStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'rewind-stmt'.
  
  FUNCTION rewind_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'rewind-stmt'
    
  END FUNCTION rewind_What
  
END MODULE RewindStmts
