! $Id: SaveStmts.f90 2159 2016-05-31 13:08:04Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the PointerStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /save-stmt/'s (R553).

MODULE SaveStmts
  
  USE CompilerKinds
  USE Identifiers
  USE SubscriptLists
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSaveStmt
  
  !-----------------------------------------------------------------------------
  
  !> The results of parsing a /saved-entity/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: SavedEntity
    !> The name of the saved entity.
    TYPE(Identifier) :: name
    !> Flag to indicate that the entity was a common block name.
    LOGICAL :: common = .FALSE.
  CONTAINS
    PROCEDURE :: Write => se_Write
    PROCEDURE :: Query => se_Query
    PROCEDURE :: Modify => se_Modify
    PROCEDURE, NOPASS :: What => se_What
    PROCEDURE :: Advertise => se_Advertise
    PROCEDURE :: GetLocation => se_GetLocation
  END TYPE SavedEntity
  
  !> Stores the results of parsing a save statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: SaveStmt
    !> The list of entities being given the save attribute.
    TYPE(SavedEntity), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => save_Dump
    PROCEDURE :: Query => save_Query
    PROCEDURE :: Modify => save_Modify
    PROCEDURE, NOPASS :: What => save_What
  END TYPE SaveStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /save-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    save_stmt         The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE ParseSaveStmt(tlist, save_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !--------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SaveStmt), INTENT(OUT) :: save_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /save-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R553'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the /save-entity-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /save-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! SAVE...
    
    CALL ExpectKeyword(tlist, 1, scck_'SAVE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /saved-entity/.  A /saved-entity/ starts with a /name/ or a slash.
    
    CALL Expect(  &
        tlist,  &
        2,  &
        [ittDeclare, ittName, ittOperator],  &
        [scck_'::', scck_'  ', scck_'/ '],  &
        '"::" or /save-entity/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /saved-entity-list/ here.
    
    !---------------------------------------------------------------------------
    ! /saved-entity-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(save_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      CALL parse_saved_entity(  &
          tlist,  &
          item_start, item_finish,  &
          save_stmt%decls(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
    ! That's it!
    
  END SUBROUTINE ParseSaveStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /saved-entity/ (R554).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! statement.
  !!
  !! @param[in]     start             Index of the first token of the 
  !! /saved-entity/.
  !!
  !! @param[in]     finish            Index of the last token of the 
  !! /saved-entity/.
  !!
  !! @param[out]    decl              Results of parsing the /saved-entity/.  
  !! Not defined on error.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This procedure has been written assuming that the slash comes back as an 
  !! operator type token.  @todo Consider modifying tlist to have the 
  !! correct token type.
  
  SUBROUTINE parse_saved_entity(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(SavedEntity), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /saved-entity/.
    CHARACTER(*), PARAMETER :: comp = 'R554'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect(  &
        tlist,  &
        start,  &
        [ittName, ittOperator],  &
        [scck_' ', scck_'/'],  &
        '/object-name/, /proc-pointer-name/ or "/" /common-block-name/ "/"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      decl%common = .FALSE.
      decl%name = IdentifierIfortBug(tlist(start))
      CALL ExpectEnd(tlist(:finish), start + 1, comp, err_list)
    ELSE
      decl%common = .TRUE.
      
      CALL Expect(tlist, start + 1, '/common-block-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      decl%name = IdentifierIfortBug(tlist(start+1))
      
      CALL Expect(tlist, start + 2, ittOperator, scck_'/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectEnd(tlist(:finish), start + 3, comp, err_list)
    END IF
    
  END SUBROUTINE parse_saved_entity
  
  
  !*****************************************************************************
  !!
  !> Implementation of SaveStmt%Dump - write a debugging representation 
  !! of the results of parsing a save statement to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE save_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SaveStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Name index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/saved-entity-list/')") pfx
    DO i = 1, SIZE(pnode%decls)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL pnode%decls(i)%Write(unit)
      WRITE (unit, "()")
    END DO
    
  END SUBROUTINE save_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SaveStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE save_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SaveStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Saved entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE save_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SaveStmt%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE save_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SaveStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Saved entity index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%decls)
      CALL pnode%decls(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE save_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SaveStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'save-stmt'.
  
  FUNCTION save_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'save-stmt'
    
  END FUNCTION save_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for bindings in SavedEntity.
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !1 @param[in]     pnode             The node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE se_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SavedEntity), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (pnode%common) THEN
      WRITE (unit, "('/ ',A,' /')", ADVANCE='NO') .NameOf. pnode%name
    ELSE
      WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%name
    END IF
    
  END SUBROUTINE se_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE se_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SavedEntity), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE se_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%Modify - traverse the parse node with 
  !! a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE se_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SavedEntity), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE se_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%Advertise - advertise the 
  !! identifiers in a DimensionDecl that are not child nodes.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE se_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SavedEntity), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    IF (pnode%common) THEN
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefSave, NatureRefCommon],  &
          LOCATIONS=[.LocationOf. pnode%name, .LocationOf. pnode%name] )
    ELSE
      ident_list(1) = IdentifierNature(  &
          NAME=.NameOf. pnode%name,  &
          NATURES=[NatureDefSave],  &
          LOCATIONS=[.LocationOf. pnode%name] )
    END IF
    
  END SUBROUTINE se_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns '/saved-entity/'.
  
  FUNCTION se_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = '/saved-entity/'
    
  END FUNCTION se_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SavedEntity%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION se_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(SavedEntity), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    ! If the saved entity is a common block name, then technically 
    ! the source location is one prior to the name, but I don't care.
    loc = .LocationOf. pnode%name
    
  END FUNCTION se_GetLocation
  
END MODULE SaveStmts
