! $Id: SelectCaseStmts.f90 2159 2016-05-31 13:08:04Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SelectCaseStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /select-case-stmt/'s, /case-stmt/'s and 
!! /end-select-stmt/'s.

MODULE SelectCaseStmts
  
  USE Expressions
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSelectCaseStmt
  PUBLIC :: ParseEndSelectStmt
  PUBLIC :: ParseCaseStmt
  
  PUBLIC :: WriteCaseSelector
  
  !-----------------------------------------------------------------------------
  
  !> Represents the results of parsing a /case-value-range/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: CaseValueRange
    !> Flag to indicate that the value range is a single item.
    !!
    !! If this is .TRUE. then upper will not be allocated and lower will 
    !! be allocated.
    LOGICAL :: is_single
    !> Lower bound of the range, or the single value if it is a single value.
    !!
    !! Not allocated if there was no lower bound specified.
    CLASS(Expression), ALLOCATABLE :: lower
    !> Upper bound of the range.
    !!
    !! Not allocated of there was no upper bound specified.
    CLASS(Expression), ALLOCATABLE :: upper
  CONTAINS
    PROCEDURE :: Write => cv_Write
    PROCEDURE :: Query => cv_Query
    PROCEDURE :: Modify => cv_Modify
    PROCEDURE, NOPASS :: What => cv_What
    PROCEDURE :: GetLocation => cv_GetLocation
  END TYPE CaseValueRange
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /select-case-stmt/ (R838).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /select-case-stmt/.
  !!
  !! @param[out]    expr              The /case-expr/.
  !!
  !! @param[out]    construct_name    The construct name, if provided.  Not 
  !! allocated if the statement had no construct name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSelectCaseStmt(tlist, expr, construct_name, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE Identifiers
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(Expression), INTENT(OUT), ALLOCATABLE :: expr
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /select-case-stmt/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R838'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Current token index.
    INTEGER :: closing        ! Closing delimited index.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /select-case-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Check for construct name
    
    IF (Match(tlist, 1, ittName) .AND. Match(tlist, 2, ittColon)) THEN
      it = 3
      !=========================================================================
      ! ifort bug - assignment of dt with polymorphic component.
      !construct_name = IdentifierIfortBug(tlist(1))
      ALLOCATE(construct_name, SOURCE=IdentifierIfortBug(tlist(1)))
      !=========================================================================
    ELSE
      it = 1
      ! Leave name unallocated.
    END IF
    
    !---------------------------------------------------------------------------
    ! SELECT CASE.. or SELECTCASE... - either form allowed by 3.3.2.2.
    
    CALL ExpectSplittableKeyword(  &
        tlist,  &
        (it),  &
        scck_'SELECT', scck_'CASE',  &
        it,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! SELECT CASE (...)
    
    CALL Expect(tlist, it, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Find closing parenthesis.
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Parse the expression inside the parentheses.
    CALL ParseExpr(tlist, it + 1, closing - 1, expr, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
    ! Thats it!
    CALL ExpectEOS(tlist, closing + 1, comp, err_list)
    
  END SUBROUTINE ParseSelectCaseStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-select-stmt/ (R841).
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /end-select-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /select-case-stmt/ if one was present, not allocated if a construct name 
  !! was not present.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndSelectStmt(tlist, construct_name, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local components
    
    ! Syntax rule number for an /end-select-stmt/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R841'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /end-select-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    ! ENDSELECT or END SELECT
    CALL ExpectSplittableKeyword(tlist, 1, scck_'END', scck_'SELECT', it, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Construct name
    
    CALL CheckConstructNameEnd(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/select-case-stmt/',  &
        '/end-select-stmt/',  &
        'C829 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseEndSelectStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /case-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens that make up the 
  !! /case-stmt/.
  !!
  !! @param[in]     construct_name    The /construct-name/ in the opening 
  !! /select-case-stmt/ if one was present, not allocated if a construct name 
  !! was not present.
  !!
  !! @param[out]    values            The results of parsing the values and 
  !! value ranges nominated by the case statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseCaseStmt(tlist, construct_name, values, err_list)
    
    USE CompilerKinds
    USE ConstructNames
    USE Errors
    USE ErrorCodes
    USE Identifiers
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(IN), ALLOCATABLE :: construct_name
    TYPE(CaseValueRange), INTENT(OUT), ALLOCATABLE :: values(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R840'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: it             ! Token index.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /case-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! CASE ...
    
    CALL ExpectKeyword(tlist, 1, scck_'CASE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! CASE selector
    
    CALL parse_case_selector(tlist, 2, values, it, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Optional construct name.
    
    CALL CheckConstructNameIntermediate(  &
        tlist,  &
        it,  &
        construct_name,  &
        '/select-case-stmt',  &
        '/case-stmt/',  &
        'C829 on ' // comp,  &
        err_list )
    
  END SUBROUTINE ParseCaseStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /case-selector/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /case-value-range/.
  !!
  !! @param[out]    values            The results of parsing the 
  !! selector.  Allocated to zero size if the selector is "DEFAULT".
  !!
  !! @param[out]    after             Index of the token in @a tlist that 
  !! follows the /case-selector/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_case_selector(tlist, start, values, after, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    TYPE(CaseValueRange), INTENT(OUT), ALLOCATABLE :: values(:)
    INTEGER, INTENT(OUT) :: after
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R843'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Index of the closing delimiter.
    INTEGER :: list_count     ! Number of items in the list.
    INTEGER :: item_start     ! Index of the starting token for an item.
    INTEGER :: item_finish    ! Index of the finishing token for an item.
    INTEGER :: item           ! Item index.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL Expect(  &
        tlist,  &
        2,  &
        [ittDelimiter, ittName],  &
        [scck_'(      ', scck_'DEFAULT'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Filter out default, because it's easy.
    IF (opt == 2) THEN
      ALLOCATE(values(0))
      after = start + 1
      RETURN
    END IF
    
    ! Dealing with the value range sort from here on.
    
    ! Match closing parentheses.
    CALL ExpectClosing(tlist, 2, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    list_count = ListItemCount(tlist(3:closing-1))
    ALLOCATE(values(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(  &
          tlist,  &
          3, closing - 1,  &
          item_start,  &
          item_finish )
      CALL parse_case_value_range(  &
          tlist,  &
          item_start, item_finish,  &
          values(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END DO
    
    after = closing + 1
    
  END SUBROUTINE parse_case_selector
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /case-value-range/ (R844).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /case-value-range/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /case-value-range/.
  !!
  !! @param[out]    value             The results of parsing the 
  !! /case-value-range/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_case_value_range(tlist, start, finish, value, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(CaseValueRange), INTENT(OUT) :: value
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /case-value-range/ for error reporting.
    CHARACTER(*), PARAMETER :: comp = 'R844'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: colon          ! Index of the colon.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /case-value-range/.' )
    END IF
    
    !---------------------------------------------------------------------------
    
    ! Go looking for a colon.
    colon = FindBare(tlist, ittColon, scck_':', start)
    
    ! Four cases - no colon, colon at start, colon at end, colon in the middle.
    IF (colon == 0) THEN
      ! No colon.  Set single
      value%is_single = .TRUE.
      CALL ParseExpr(tlist, start, finish, value%lower, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE IF (colon == start) THEN
      ! Colon at the beginning.  Leave lower unallocated.
      value%is_single = .FALSE.
      CALL ParseExpr(tlist, colon + 1, finish, value%upper, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE IF (colon == finish) THEN
      ! Colon at the end.  Leave upper unallocated.
      value%is_single = .FALSE.
      CALL ParseExpr(tlist, start, colon - 1, value%lower, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    ELSE
      ! Colon in the middle.
      value%is_single = .FALSE.
      CALL ParseExpr(tlist, start, colon - 1, value%lower, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      CALL ParseExpr(tlist, colon + 1, finish, value%upper, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
    END IF
    
    ! The expression parse detects any superfluous tokens.
    
  END SUBROUTINE parse_case_value_range
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of an array of CaseValueRange to a 
  !! logical unit.
  !!
  !! @param[in]     log_unit          The unit connected for formatted output 
  !! to write the representation of the CaseValueRange to.
  !!
  !! @param[in]     case_selector     The array of CaseValueRange's.
  
  SUBROUTINE WriteCaseSelector(log_unit, case_selector)
    
    USE Expressions
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(CaseValueRange), INTENT(IN) :: case_selector(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! /case-value-range/ selector.
    
    !***************************************************************************
    
    IF (SIZE(case_selector) == 0) THEN
      WRITE (log_unit, "('DEFAULT')", ADVANCE='NO')
    ELSE
      WRITE (log_unit, "('(')", ADVANCE='NO')
      DO i = 1, SIZE(case_selector)
        CALL cv_Write(case_selector(i), log_unit)
        IF (i == SIZE(case_selector)) EXIT
        WRITE (log_unit, "(',')", ADVANCE='NO')
      END DO
      WRITE (log_unit, "(')')", ADVANCE='NO')
    END IF
    
  END SUBROUTINE WriteCaseSelector
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseValueRange%Write - write a debugging representation 
  !! of the parse node to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE cv_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseValueRange), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (pnode%is_single) THEN
      CALL WriteExpr(unit, pnode%lower)
    ELSE
      IF (ALLOCATED(pnode%lower)) THEN
        CALL WriteExpr(unit, pnode%lower)
      END IF
      WRITE (unit, "(':')", ADVANCE='NO')
      IF (ALLOCATED(pnode%upper)) THEN
        CALL WriteExpr(unit, pnode%upper)
      END IF
    END IF
    
  END SUBROUTINE cv_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseValueRange%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE cv_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseValueRange), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%lower)) THEN
      CALL pnode%lower%Query(visitor)
    END IF
    
    IF (ALLOCATED(pnode%upper)) THEN
      CALL pnode%upper%Query(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cv_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseValueRange%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE cv_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseValueRange), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%lower)) THEN
      CALL pnode%lower%Modify(visitor)
    END IF
    
    IF (ALLOCATED(pnode%upper)) THEN
      CALL pnode%upper%Modify(visitor)
    END IF
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cv_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseValueRange%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'case-value-range'
  
  FUNCTION cv_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'case-value-range'
    
  END FUNCTION cv_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of CaseValueRange%GetLocation - get the start of the source 
  !! that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION cv_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(CaseValueRange), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%lower)) THEN
      loc = .LocationOf. pnode%lower
    ELSE
      loc = .LocationOf. pnode%upper
    END IF
    
  END FUNCTION cv_GetLocation
  
END MODULE SelectCaseStmts
