! $Id: StructureConstructors.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the StructureConstructors module.


!*******************************************************************************
!!
!> Procedures and types for parsing /structure-constructors/.

MODULE StructureConstructors
  
  USE CompilerKinds
  USE DeclarationTypeSpecs
  USE Expressions
  USE Identifiers
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseStructureConstructor
  
  PUBLIC :: TypeSpec    ! From DeclarationTypeSpecs.
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a /component-spec/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: ComponentSpec
    !> Keyword for the component.
    !!
    !! Not allocated if there was no keyword.
    TYPE(Identifier), ALLOCATABLE :: keyword
    !> The /component-data-source/ or /actual-argument/.
    !!
    !! Should always be allocated after a successful parse.
    CLASS(Expression), ALLOCATABLE :: data_source
  CONTAINS
    PROCEDURE :: Write => cs_Write
    PROCEDURE :: Query => cs_Query
    PROCEDURE :: Modify => cs_Modify
    PROCEDURE, NOPASS :: What => cs_What
    PROCEDURE :: GetLocation => cs_GetLocation
  END TYPE ComponentSpec
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a /stsructure-constructor/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: StructureConstructor
    !> The type specifier at the start of the /structure-constructor/ 
    !! (the thing that can look like a /function-reference/
    TYPE(TypeSpec) :: type_spec
    !> The component specifications.
    !!
    !! Always allocated, may be size zero.
    TYPE(ComponentSpec), ALLOCATABLE :: components(:)
  CONTAINS
    PROCEDURE :: Write => sc_Write
    PROCEDURE :: Query => sc_Query
    PROCEDURE :: Modify => sc_Modify
    PROCEDURE, NOPASS :: What => sc_What
    PROCEDURE :: GetLocation => sc_GetLocation
  END TYPE StructureConstructor
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /structure-constructor/ (R455).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /structure-constructor/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /structure-constructor/.
  !!
  !! @param[out]    sc                The results of parsing the structure 
  !! constructor.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! You can embed a structure constructor inside a structure constructor, 
  !! and that will be parsed recursively, hence the RECURSIVE attribute.
  
  RECURSIVE SUBROUTINE ParseStructureConstructor( tlist, start, finish,  &
      sc, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(StructureConstructor), INTENT(OUT) :: sc
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /structure-constructor/.
    CHARACTER(*), PARAMETER :: comp = 'R455'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for procedure calls.
    INTEGER :: arg_closing    ! Index of closing parenthesis of "arguments".
    INTEGER :: spec_finish    ! Index of the last token of the type spec.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: list_count     ! Number of items in the structure constructor.
    INTEGER :: item_start     ! Starting index of an item.
    INTEGER :: item_finish    ! Finishing index of an item.
    INTEGER :: item           ! Item index.
    
    ! Flag that indicates that at least one KEYWORD= style component 
    ! specification has been seen.
    LOGICAL :: seen_keyword
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /structure-constructor/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! derived-type-spec
    
    CALL ExpectName(tlist, start, '/type-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! the syntax can look like abc() or abc()().  To find out go looking for
    ! parentheses.  Initially we'll assume that there is no parameter list 
    ! as part of the type spec.
    CALL Expect(tlist, start + 1, ittDelimiter, scck_'(', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL ExpectClosing(tlist, start + 1, arg_closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! Do we have the second pair?
    IF (Match(tlist, arg_closing + 1, ittDelimiter, scck_'(')) THEN
      ! It looks like we do.
      spec_finish = arg_closing
      CALL ExpectClosing(tlist, spec_finish + 1, arg_closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
    ELSE
      ! Our assumption about there being no type parameter list was correct.
      spec_finish = start
    END IF
    
    ! Parse the type spec
    CALL ParseTypeSpec(tlist(start:spec_finish), sc%type_spec, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Parse the components bit.
    
    ! There don't have to be any component specifications.
    IF (arg_closing - 1 < spec_finish + 2) THEN
      ALLOCATE(sc%components(0))
    ELSE
      seen_keyword = .FALSE.
      list_count = ListItemCount(tlist(spec_finish+2:arg_closing-1))
      ALLOCATE(sc%components(list_count))
      item_finish = 0
      DO item = 1, list_count
        CALL GetNextListItem(  &
            tlist,  &
            spec_finish + 2,  &
            arg_closing - 1,  &
            item_start,  &
            item_finish )
        CALL parse_component_spec(  &
            tlist,  &
            item_start,  &
            item_finish,  &
            sc%components(item),  &
            sub_err_list )
        CALL Add(err_list, sub_err_list)
        IF (Failure(sub_err_list)) RETURN
      
        IF (seen_keyword) THEN
          IF (.NOT. ALLOCATED(sc%components(item)%keyword)) THEN
            CALL Add( err_list,  &
                CODE=errSyntax,  &
                COMPONENT='C493 on ' // comp,  &
                LOCATION=QueryLocation(tlist(item_start)),  &
                MSG='The keyword= may be omitted from a /component-spec/ &
                  &only if the keyword= has been omitted from each &
                  &preceding /component-spec/.' )
          END IF
        ELSE IF (ALLOCATED(sc%components(item)%keyword)) THEN
          seen_keyword = .TRUE.
        END IF
      END DO
    END IF
    
    !---------------------------------------------------------------------------
    ! That's it
    
    CALL ExpectEnd(tlist(:finish), arg_closing + 1, comp, err_list)
    
  END SUBROUTINE ParseStructureConstructor
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructor%Write - write a text representation of 
  !! the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  ! Recursive because the component data sources can be expressions that 
  ! are themselves structure constructors.
  RECURSIVE SUBROUTINE sc_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StructureConstructor), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Component index.
    
    !***************************************************************************
    
    CALL WriteTypeSpec(unit, pnode%type_spec)
    WRITE (unit, "('(')", ADVANCE='NO')
    
    DO i = 1, SIZE(pnode%components)
      CALL pnode%components(i)%Write(unit)
      IF (i == SIZE(pnode%components)) EXIT
      WRITE (unit, "(',')", ADVANCE='NO')
    END DO
    WRITE (unit, "(')')", ADVANCE='NO')
    
  END SUBROUTINE sc_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructor%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE sc_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(StructureConstructor), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Component index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%type_spec%Query(visitor)
    
    DO i = 1, SIZE(pnode%components)
      CALL pnode%components(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sc_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructor%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE sc_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(StructureConstructor), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Component index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%type_spec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%components)
      CALL pnode%components(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sc_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructor%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION sc_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'component-spec'
    
  END FUNCTION sc_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of StructureConstructor%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION sc_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(StructureConstructor), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%type_spec
    
  END FUNCTION sc_GetLocation
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for ComponentSpec.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /component-spec/ (R456).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the /component-spec/.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the /component-spec/.
  !!
  !! @param[out]    component         Results of parsing the 
  !! /component-spec/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A component spec is:
  !!
  !!    [ /keyword/ = ] /component-data-source/
  !!
  !! The component data source could contain a structure constructor with 
  !! component spec's, hence RECURSIVE.
  
  RECURSIVE SUBROUTINE parse_component_spec( tlist, start, finish,  &
      component, err_list )
    
    USE CompilerKinds
    USE Errors
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(ComponentSpec), INTENT(OUT) :: component
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /component-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R456'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: assign         ! Index of = if present.
    TYPE(Error) :: err        ! Error object for expect calls.
    INTEGER :: it             ! Index of the current token.
    
    ! Error list for chlid parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    assign = FindBare(tlist, ittAssign, scck_'=', start)
    
    IF ((assign /= 0) .AND. (assign <= finish)) THEN
      CALL Expect(tlist, start, '/keyword/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      component%keyword = IdentifierIfortBug(tlist(start))
      
      CALL ExpectEnd(tlist(:assign-1), start + 1, comp, err_list)
      it = assign + 1
    ELSE
      it = start
    END IF
    ! it here the index of the start of the data source.
    
    CALL ParseExpr(tlist, it, finish, component%data_source, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE parse_component_spec
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComponentSpec%Write - write a text representation of 
  !! the parse node to a logical unit.
  !!
  !! @param[in]     pnode             The format.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output for the text represenation.
  
  SUBROUTINE cs_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComponentSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%keyword)) THEN
      WRITE (unit, "(A,'=')", ADVANCE='NO')  &
          .NameOf. pnode%keyword
    END IF
    CALL pnode%data_source%Write(unit)
    
  END SUBROUTINE cs_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComponentSpec%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE cs_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ComponentSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%data_source%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cs_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComponentSpec%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE cs_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(ComponentSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%data_source%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE cs_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComponentSpec%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION cs_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'component-spec'
    
  END FUNCTION cs_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of ComponentSpec%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION cs_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(ComponentSpec), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%keyword)) THEN
      loc = .LocationOf. pnode%keyword
    ELSE
      loc = .LocationOf. pnode%data_source
    END IF
    
  END FUNCTION cs_GetLocation
  
END MODULE StructureConstructors
