! $Id: SubroutineStmts.f90 2883 2019-04-06 04:15:51Z ian $
! ff08 source code copyright 2014 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the SubroutineStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /subroutine-stmt/'s and 
!! /end-subroutine-stmt/'s.
!!
!! We borrow some of the prefix parsing code from the FunctionStmts module.

MODULE SubroutineStmts
  
  USE FunctionStmts
  USE Identifiers
  USE LanguageBindingSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseSubroutineStmt
  
  PUBLIC :: ParseEndSubroutineStmt
  PUBLIC :: ParseDummyArgList
  
  !-----------------------------------------------------------------------------
  
  !> Results of parsing a dummy argument.
  TYPE, PUBLIC :: DummyArg
    !> Indicates if the dummy argument is a alternate return.
    LOGICAL :: is_star = .FALSE.
    !> Name of the dummy argument if @a is_star is false.
    !!
    !! Not allocated if @a is_star is true.
    TYPE(Identifier), ALLOCATABLE :: name
  END TYPE DummyArg
  
  !> The results of parsing a /subroutine-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: SubroutineStmt
    !> Prefixes (ELEMENTAL, RECURSIVE, etc).
    TYPE(OtherPrefixSpecs) :: prefixes
    !> Subroutine name.
    TYPE(Identifier) :: subroutine_name
    !> Dummy arguments.
    !!
    !! Always allocated but may be zero length.
    TYPE(DummyArg), ALLOCATABLE :: args(:)
    !> Language binding spec.  May not be allocated.
    TYPE(LangSpec), ALLOCATABLE :: ls
  CONTAINS
    PROCEDURE :: Dump => sub_Dump
    PROCEDURE :: Query => sub_Query
    PROCEDURE :: Modify => sub_Modify
    PROCEDURE, NOPASS :: What => sub_What
    PROCEDURE :: Advertise => sub_Advertise
  END TYPE SubroutineStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /subroutine-stmt/ (R1234).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    sub_stmt          Results of parsing the /subroutine-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseSubroutineStmt(tlist, sub_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(SubroutineStmt), INTENT(OUT) :: sub_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /subroutine-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1234'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(MatchProcList) :: procs(2)   ! Expect match procedure list.
    INTEGER :: opt                    ! Expect option.
    TYPE(Error) :: err                ! Error object for Expect call.
    INTEGER :: closing                ! Position of closing parentheses.
    INTEGER :: it                     ! Current token index.
    
    ! Error list for child Parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /subroutine-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    procs(1)%item => MatchOtherPrefixSpec
    procs(2)%item => match_subroutine
    it = 1
    DO
      CALL Expect(tlist, it, procs, '/prefix/ or "SUBROUTINE"', opt, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      SELECT CASE (opt)
      CASE (1)
        ! An "other prefix spec" is one of the allowable prefixes for 
        ! a subroutine statement (the same as the allowable prefixes 
        ! for a function statement LESS the possibility of a 
        ! /declaration-type-spec/ that can give the function's type).
        CALL ParseOtherPrefixSpec(  &
            tlist(it:it),  &
            sub_stmt%prefixes,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=comp)
        IF (Failure(sub_err_list)) RETURN
        it = it + 1
        
      CASE (2)
        it = it + 1
        EXIT
        
      END SELECT
    END DO
    
    !---------------------------------------------------------------------------
    ! /subroutine-name/
    
    CALL ExpectName(tlist, it, '/subroutine-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    sub_stmt%subroutine_name = IdentifierIfortBug(tlist(it))
    it = it + 1
    
    !---------------------------------------------------------------------------
    ! [ ( [ /dummy-arg-list/ ] ) ]
    
    ! Note that unlike the function-stmt the parentheses are optional.
    
    CALL Expect(  &
        tlist, it,  &
        [ittEndStatement, ittDelimiter],  &
        [scck_' ', scck_'('],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! No parentheses means we are done (means we cannot have a 
    ! /proc-language-binding-spec/ either.)
    IF (opt == 1) THEN
      ALLOCATE(sub_stmt%args(0))
      RETURN
    END IF
    
    CALL ExpectClosing(tlist, it, closing, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The argument list is optional.
    IF (closing - 1 >= it + 1) THEN
      CALL ParseDummyArgList(  &
          tlist,  &
          it + 1,  &
          closing - 1,  &
          sub_stmt%args,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
    ELSE
      ALLOCATE(sub_stmt%args(0))
    END IF
    it = closing + 1
    
    !---------------------------------------------------------------------------
    ! ... [ /proc-language-binding-spec/ ]
    
    CALL Expect(  &
        tlist, it,  &
        [ittEndStatement, ittName],  &
        [scck_'    ', scck_'BIND'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    ! The language spec is optional.
    IF (opt == 1) RETURN
    
    ALLOCATE(sub_stmt%ls)
    CALL ParseProcLanguageBindingSpec(  &
        tlist, it,  &
        SIZE(tlist),  &
        sub_stmt%ls,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    
  END SUBROUTINE ParseSubroutineStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /end-subroutine-stmt/ (R1236).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     name              The /subroutine-name/ from the statement.
  !! or not allocated if there was no name.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseEndSubroutineStmt(tlist, name, err_list)
    
    USE CompilerKinds
    USE EndProgramUnitStmts
    USE Errors
    USE Identifiers
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Identifier), INTENT(OUT), ALLOCATABLE :: name
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !***************************************************************************
    
    CALL ParseEndProgramUnitStmt(  &
        tlist,  &
        scck_'SUBROUTINE',  &
        'R1236',  &
        '/end-subroutine-stmt/',  &
        '/subroutine-name/',  &
        name,  &
        err_list )
    
  END SUBROUTINE ParseEndSubroutineStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a /dummy-arg-list/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the start of the dummy arg list.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the end of the dummy arg list.
  !!
  !! @param[out]    args              The results of parsing the dummy 
  !! arg list.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! A /dummy-arg/ can be either a name or the * symbol.
  !!
  !! Entry statement parsing also uses this procedure.
  
  SUBROUTINE ParseDummyArgList(tlist, start, finish, args, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ErrorLevels
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(DummyArg), INTENT(OUT), ALLOCATABLE :: args(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constant
    
    ! Syntax rule number for a /dummy-arg/ (not a /dummy-arg-list/).
    CHARACTER(*), PARAMETER :: comp = 'R1235'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: item_count     ! Number of items in the list.
    INTEGER :: iarg           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    TYPE(Error) :: err        ! Error code for procedure calls.    
    INTEGER :: opt            ! Expect option.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    item_count = ListItemCount(tlist(start:finish))
    ALLOCATE(args(item_count))
    item_finish = 0
    DO iarg = 1, item_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL Expect(  &
          tlist, item_start,  &
          [ittOperator, ittName],  &
          [scck_'*', scck_' '],  &
          '"*" or /dummy-arg-name/',  &
          opt,  &
          err ) 
      CALL Add(err_list, err)
      IF (Failure(err)) RETURN
      
      IF (opt == 1) THEN
        ! Mark the dummy argument as an alternate return.
        args(iarg)%is_star = .TRUE.
        ! Add a warning for use of an obsolescent feature.
        CALL Add( err_list,  &
            CODE=errObsolescentAlternateReturn,  &
            LEVEL=errLevelWarning,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(item_start)),  &
            MSG='Using a * as a /dummy-arg-spec/ is an obsolescent feature.' )
      ELSE
        ! Copy across the name
        ALLOCATE(args(iarg)%name)     ! ifort bug?
        args(iarg)%name = IdentifierIfortBug(tlist(item_start))
        ! That should be it.
        CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
      END IF
    END DO
    
  END SUBROUTINE ParseDummyArgList
  
  
  !*****************************************************************************
  !!
  !> Match the SUBROUTINE keyword at a position in a sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The position at which to start the 
  !! match.
  !!
  !! @returns True if the given token is the SUBROUTINE keyword.
  
  FUNCTION match_subroutine(tlist, start) RESULT(b)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    LOGICAL :: b
    
    !***************************************************************************
    
    b = Match(tlist, start, ittName, scck_'SUBROUTINE')
    
  END FUNCTION match_subroutine
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of a dummy argument to a unit.
  !!
  !! @param[in]     unit              Unit connected for fomatted output.
  !!
  !! @param[in]     arg               The argument to write.
  
  SUBROUTINE write_dummy_arg(unit, arg)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: unit
    TYPE(DummyArg), INTENT(IN) :: arg
    
    !***************************************************************************
    
    IF (arg%is_star) THEN
      WRITE (unit, "('*')", ADVANCE='NO')
    ELSE
      WRITE (unit, "(A)", ADVANCE='NO') .NameOf. arg%name
    END IF
    
  END SUBROUTINE write_dummy_arg
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a logical unit.
  !!
  !! @param[in]     pnode             The FunctionSubprogram to dump.
  !!
  !! @param[in]     unit              Logical unit connected for 
  !! formatted output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  ! As separate module procedures cannot contain separate module procedures, 
  ! this doesn't need to be recursive.
  SUBROUTINE sub_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Argument index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/subroutine-name/ = ',A)") pfx,  &
        .NameOf. pnode%subroutine_name
    
    CALL DumpOtherPrefixSpec(pnode%prefixes, unit, pfx)
    
    ! Dummy arguments.
    WRITE (unit, "(A,'/dummy-arg-list/')") pfx
    DO i = 1, SIZE(pnode%args)
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL write_dummy_arg(unit, pnode%args(i))
      WRITE (unit, "()")
    END DO
    
    ! Language spec (BIND attribute).
    IF (ALLOCATED(pnode%ls)) THEN
      WRITE ( unit,  &
          "(A,'/proc-language-binding-spec/ = ')",  &
          ADVANCE='NO' ) pfx
      CALL WriteLangSpec(unit, pnode%ls)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE sub_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineStmt%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE sub_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineStmt%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE sub_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(SubroutineStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%ls)) CALL pnode%ls%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE sub_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'separate-module-subprogram'
  
  FUNCTION sub_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'subroutine-stmt'
    
  END FUNCTION sub_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of SubroutineStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The statement to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE sub_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! arg index.
    INTEGER :: id             ! identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1 + COUNT(.NOT. pnode%args%is_star)))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%subroutine_name,  &
        NATURES=[NatureDefProcedureDef, NatureDefSubroutine],  &
        LOCATIONS=[  &
          .LocationOf. pnode%subroutine_name,  &
          .LocationOf. pnode%subroutine_name ] )
    
    ! Identifiers in the arguments.
    id = 2
    DO i = 1, SIZE(pnode%args)
      IF (.NOT. pnode%args(i)%is_star) THEN
        ident_list(id) = IdentifierNature(  &
            NAME=.NameOf. pnode%args(i)%name,  &
            NATURES=[NatureDefDummy],  &
            LOCATIONS=[.LocationOf. pnode%args(i)%name] )
        id = id + 1
      END IF
    END DO
    
  END SUBROUTINE sub_Advertise
  
END MODULE SubroutineStmts
