! $Id: TypeBoundGenericStmts.f90 2870 2019-03-30 23:54:45Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the TypeBoundGenericStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /type-bound-generic-stmt/'s.

MODULE TypeBoundGenericStmts
  
  USE CompilerKinds
  USE Identifiers
  USE GenericSpecs
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseTypeBoundGenericStmt
  
  PUBLIC :: GenericSpec       ! From GenericSpecs.
  
  !> Stores the results of parsing a /type-bound-generic-stmt/.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: TypeBoundGenericStmt
    !> The access specification.
    INTEGER :: access
    !> The generic spec.
    TYPE(GenericSpec) :: gs
    !> The specific bindings.
    !!
    !! Always allocated, always size > 0.
    TYPE(Identifier), ALLOCATABLE :: binding_list(:)
  CONTAINS
    PROCEDURE :: Dump => tbg_Dump
    PROCEDURE :: Query => tbg_Query
    PROCEDURE :: Modify => tbg_Modify
    PROCEDURE, NOPASS :: What => tbg_What
    PROCEDURE :: Advertise => tbg_Advertise
  END TYPE
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-bound-generic-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    tpg_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeBoundGenericStmt(tlist, tpg_stmt, err_list)
    
    USE AccessStmts
    USE Strings
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeBoundGenericStmt), INTENT(OUT) :: tpg_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-bound-generic-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R450'
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    TYPE(Error) :: err    ! Error object from Expect.
    INTEGER :: opt        ! Expect option.
    INTEGER :: it         ! Index of the first token of the current clause.
    INTEGER :: it_end     ! Index of the last token of the current clause.
    INTEGER :: associate  ! Index of the associate token.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-bound-procedure-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'GENERIC', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    
    CALL Expect(  &
        tlist, 2,  &
        [ittComma, ittDeclare],  &
        [scck_', ', scck_'::'],  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      ! An /access-spec/.
      CALL ExpectKeyword(  &
          tlist, 3,  &
          AccessKeywords,  &
          '/access-spec/',  &
          tpg_stmt%access,  &
          err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Now the double colon
      CALL Expect(tlist, 4, ittDeclare, scck_'::', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = 5    ! Token after the double colon.
    ELSE
      tpg_stmt%access = 0
      it = 3    ! Token after the double colon.
    END IF
    
    !---------------------------------------------------------------------------
    ! The generic spec.
    
    ! The generic spec finishes with the associate.
    associate = FindBare(tlist, ittAssociate, scck_'=>', it)
    IF (associate == 0) THEN
      it_end = SIZE(tlist)
    ELSE
      it_end = associate - 1
    END IF
    
    CALL ParseGenericSpec(tlist(it:it_end), tpg_stmt%gs, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    IF (associate == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=QueryLocation(tlist(it_end)),  &
          MSG='Expected "=>".' )
      RETURN
    END IF
    it = associate + 1
    
    !---------------------------------------------------------------------------
    ! The /binding-name-list/
    
    CALL ParseNameList(  &
        tlist, it,  &
        SIZE(tlist),  &
        '/binding-name/',  &
        tpg_stmt%binding_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! All done!  The namelist parse above picks up any extraneous tokens.
    
  END SUBROUTINE ParseTypeBoundGenericStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundGenericStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE tbg_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundGenericStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Binding list index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/generic-spec/ = ')", ADVANCE='NO') pfx // '  '
    CALL WriteGenericSpec(unit, pnode%gs)
    WRITE (unit, "()")
    
    IF (pnode%access /= 0) THEN
      WRITE (unit, "(A,'/access-spec/ = ')", ADVANCE='NO') pfx // '    '
      CALL WriteAccess(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    
    WRITE (unit, "(A,'/binding-name-list/')") pfx // '    '
    DO i = 1, SIZE(pnode%binding_list)
      WRITE (unit, "(A,A)") pfx // '      ', .NameOf. pnode%binding_list(i)
    END DO
    
  END SUBROUTINE tbg_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundGenericStmt%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE tbg_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundGenericStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%gs%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tbg_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundGenericStmt%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE tbg_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundGenericStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%gs%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tbg_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundGenericStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'type-bound-generic-stmt'.
  
  FUNCTION tbg_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-bound-generic-stmt'
    
  END FUNCTION tbg_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundGenericStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE tbg_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundGenericStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: cnt            ! Count of the number of identifiers.
    INTEGER :: i              ! Specific binding index.
    
    ! ifort 16.0 bug work-around - ticket 42.
    character(:), allocatable :: tmp
    
    !***************************************************************************
    
    cnt = SIZE(pnode%binding_list)
    IF (pnode%gs%keyword <= 2) cnt = cnt + 1
    
    ALLOCATE(ident_list(cnt))
    
    IF (pnode%gs%keyword <= 2) THEN
      ident_list(1) = IdentifierNature(  &
          SPEC=pnode%gs,  &
          NATURES=[NatureDefGeneric],  &
          LOCATIONS=[.LocationOf. pnode%gs] )
      cnt = 2
    ELSE
      cnt = 1
    END IF
    
    DO i = 1, SIZE(pnode%binding_list)
      tmp = .NameOf. pnode%binding_list(i)
      ident_list(cnt) = IdentifierNature(  &
          NAME=tmp,  &
          NATURES=[NatureRefSpecBinding],  &
          LOCATIONS=[.Locationof. pnode%binding_list(i)] )
      cnt = cnt + 1
    END DO
    
  END SUBROUTINE tbg_Advertise
  
END MODULE TypeBoundGenericStmts

