! $Id: TypeBoundProcedureStmts.f90 2869 2019-03-30 23:44:32Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the TypeBoundProcedureStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /type-bound-procedure-stmt/'s.

MODULE TypeBoundProcedureStmts
  
  USE ParseNodes
  USE Identifiers
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseTypeBoundProcedureStmt
  
  !-----------------------------------------------------------------------------
  
  !> Stores the result of parsing the binding attributes.
  TYPE, PUBLIC :: BindingAttrs
    !> The accessibility of the binding.
    INTEGER :: access = 0
    !> Indicates the pass attribute was provided.
    !!
    !! If /arg-name/ was provided within the PASS attribute specification, 
    !! then pass_arg_name will be allocated as well.
    LOGICAL :: pass = .FALSE.
    !> Allocated if PASS was specified and an /arg-name/ was specified within 
    !! the pass attribute specification.
    !!
    !! Otherwise (no PASS attribute, or no /arg-name/ specification) not 
    !! allocated.
    TYPE(Identifier), ALLOCATABLE :: pass_arg_name
    !> .TRUE. if NOPASS was specified.
    LOGICAL :: nopass = .FALSE.
    !> .TRUE. if NON_OVERRIDABLE was specified.
    LOGICAL :: non_overridable = .FALSE.
    !> .TRUE. if DEFERRED was specified.
    LOGICAL :: deferred = .FALSE.
  END TYPE BindingAttrs
  
  !> Stores the result of parsing either a /type-bound-decl-list/ (which is
  !! a binding and an optional specific procedure) or a /binding-name-list/
  !! (which is just a list of binding names).
  TYPE, PUBLIC :: BindingDecl
    !> The binding name for the type bound procedure declaration.
    TYPE(Identifier) :: binding
    !> The specific procedure for the type bound procedure declaration.
    !!
    !! Not allocated if there was no specific procedure.
    !!
    !! Check tests that this is not allocated if the DEFERRED attribute was 
    !! present.
    TYPE(Identifier), ALLOCATABLE :: specific
  END TYPE BindingDecl
  
  !> Stores the results of parsing a type bound procedure statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: TypeBoundProcedureStmt
    !> The interface name, if present, in the parentheses immediately after 
    !! the PROCEDURE keyword.  Not allocated if there was no interface name.
    TYPE(Identifier), ALLOCATABLE :: interface_name
    !> The binding attributes (from the /binding-attr-list/).
    TYPE(BindingAttrs) :: attrs
    !> The /type-bound-procedure-decl-list/.
    !!
    !! Always allocated, always size > 0.
    TYPE(BindingDecl), ALLOCATABLE :: decls(:)
  CONTAINS
    PROCEDURE :: Dump => tbp_Dump
    PROCEDURE :: Query => tbp_Query
    PROCEDURE :: Modify => tbp_Modify
    PROCEDURE, NOPASS :: What => tbp_What
    PROCEDURE :: Advertise => tbp_Advertise
  END TYPE
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-bound-procedure-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    tbp_stmt          The results of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeBoundProcedureStmt(tlist, tbp_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeBoundProcedureStmt), INTENT(OUT) :: tbp_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /type-bound-procedure-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R448'
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    TYPE(Error) :: err        ! Error object from Expect.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Index of current token.
    INTEGER :: closing        ! Index of closing parenthesis.
    INTEGER :: finish         ! Index of the end of a syntax component.
    INTEGER :: declare        ! Index of the double colon.
    
    INTEGER :: list_count     ! Number of items in the binding list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for child parse calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-bound-procedure-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    CALL ExpectKeyword(tlist, 1, scck_'PROCEDURE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(  &
        tlist, 2,  &
        [ittDelimiter, ittComma, ittDeclare, ittName],  &
        [scck_'( ', scck_', ', scck_'::', scck_'  '],  &
        opt, err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! ( /interface-name/ )
    
    IF (opt == 1) THEN        ! "PROCEDURE ("
      
      CALL ExpectClosing(tlist, 2, closing, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL ExpectName(tlist, 3, '/interface-name/', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      tbp_stmt%interface_name = IdentifierIfortBug(tlist(3))
      
      IF (closing /= 4) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(4)),  &
            MSG='"' // QueryRaw(tlist(4)) // '" was unexpected.' )
        RETURN
      END IF
      
      CALL Expect(tlist, 5, ittComma, scck_',', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      it = 6    ! Index after the comma.
      opt = 2
    ELSE 
      it = 3    ! Index after the comma, assuming a comma is present.
    END IF
    
    !---------------------------------------------------------------------------
    ! , /binding-attr-list/
    
    IF (opt == 2) THEN          ! "PROCEDURE ," or a follow on from above.
      
      ! it has been set above to be the index after the comma.  If opt is 
      ! 2 then there definitely is a comma, either following the 
      ! ( /interface-name/ ) bit or as detected by the original Expect test.
      
      ! If we have attributes then we must have a bare :: somewhere in the 
      ! statement.  Find it.  
      declare = FindBare(tlist, ittDeclare, scck_'::', it)
      IF (declare == 0) THEN
        ! If we don't find a :: then trying to parse the entire statement as 
        ! the attribute list should flush out an error.  Failing that, we 
        ! specifically add an error below.
        finish = SIZE(tlist)
      ELSE
        finish = declare - 1
      END IF
      CALL parse_binding_attr_list(  &
          tlist, it,  &
          finish,  &
          tbp_stmt%attrs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (declare == 0) THEN
        ! Not covered by StmtParsingTests.
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(finish)),  &
            MSG='Expected "::".' )
        RETURN
      END IF
      it = declare + 1
      
      ! Check C477 - DEFERRED shall appear if and only if /interface-name/ 
      ! appears.
      IF ( tbp_stmt%attrs%deferred  &
          .NEQV. ALLOCATED(tbp_stmt%interface_name) ) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(1)),  &
            MSG='DEFERRED shall appear if and only if /interface-name/ &
              &appears.' )
      END IF
      
    ELSE IF (opt == 3) THEN     ! "PROCEDURE ::"
      it = 3
    ELSE                        ! "PROCEDURE name"
      it = 2
    END IF
    
    ! At this point it should be the index of the first token of the 
    ! declaration list.
    
    !---------------------------------------------------------------------------
    ! /type-bound-proc-decl-list/ or /binding-name-list/
    
    list_count = ListItemCount(tlist(it:))
    ALLOCATE(tbp_stmt%decls(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, it, SIZE(tlist), item_start, item_finish)
      IF (ALLOCATED(tbp_stmt%interface_name)) THEN
        CALL parse_binding_name(  &
            tlist,  &
            item_start, item_finish,  &
            tbp_stmt%decls(item),  &
            sub_err_list )
      ELSE
        CALL parse_type_bound_proc_decl(  &
            tlist,  &
            item_start,  &
            item_finish,  &
            tbp_stmt%decls(item),  &
            sub_err_list )
      END IF
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseTypeBoundProcedureStmt
  
  
  !****************************************************************************
  !!
  !> Parse a /binding-attr-list/.
  
  SUBROUTINE parse_binding_attr_list(tlist, start, finish, attrs, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(BindingAttrs), INTENT(OUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !-------------------------------------------------------------------------
    ! Local constants
    
    ! Rule for each attribute, not the list.
    CHARACTER(*), PARAMETER :: comp = 'R451'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the binding list.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    INTEGER :: opt            ! Expect option result.
    TYPE(Error) :: err        ! Error object from Expect.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist(start:finish))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      CALL ExpectKeyword( tlist,  &
          item_start,  &
          [ scck_'PASS           ', scck_'NOPASS         ',  &
            scck_'NON_OVERRIDABLE', scck_'DEFERRED       ',  &
            scck_'PRIVATE        ', scck_'PUBLIC         ' ],  &
            '/binding-attr/',  &
            opt,  &
            err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      CALL process_attribute
    END DO
    
  CONTAINS
    
    !***************************************************************************
    !!
    !>
    
    SUBROUTINE process_attribute
      
      USE AccessStmts
      
      !-------------------------------------------------------------------------
      ! Local variables
      
      INTEGER :: closing    ! Index of the closing parenthesis
      
      !*************************************************************************
      
      SELECT CASE (opt)
      CASE (1)    ! PASS
        IF (attrs%pass) THEN
          CALL already('"PASS"')
        ELSE IF (attrs%nopass) THEN
          CALL pass_and_nopass
        ELSE
          attrs%pass = .TRUE.
          IF (item_finish > item_start) THEN
            CALL Expect(tlist, item_start + 1, ittDelimiter, scck_'(', err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (Failure(err)) RETURN
            
            CALL ExpectClosing(tlist, item_start + 1, closing, err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (Failure(err)) RETURN
            
            CALL Expect(tlist, item_start + 2, '/arg-name/', err)
            CALL Add(err_list, err, COMPONENT=comp)
            IF (Failure(err)) RETURN
            attrs%pass_arg_name = IdentifierIfortBug(tlist(item_start+2))
            
            CALL ExpectEnd(tlist(:closing-1), item_start + 3, comp, err_list)
            CALL ExpectEnd(tlist(:item_finish), closing + 1, comp, err_list)
          END IF
        END IF
        
      CASE (2)    ! NOPASS
        IF (attrs%nopass) THEN
          CALL already('"NOPASS"')
        ELSE IF (attrs%pass) THEN
          CALL pass_and_nopass
        ELSE
          attrs%nopass = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
      
      CASE (3)    ! NON_OVERRIDABLE
        IF (attrs%non_overridable) THEN
          CALL already('"NON_OVERRIDABLE"')
        ELSE IF (attrs%deferred) THEN
          CALL non_overridable_and_deferred
        ELSE
          attrs%non_overridable = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
          
      CASE (4)    ! DEFERRED
        IF (attrs%deferred) THEN
          CALL already('"DEFERRED"')
        ELSE IF (attrs%non_overridable) THEN
          CALL non_overridable_and_deferred
        ELSE
          attrs%deferred = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (5)  ! PRIVATE
        CALL do_access(iaccPrivate)
        
      CASE (6)  ! PUBLIC
        CALL do_access(iaccPublic)
        
      END SELECT
      
    END SUBROUTINE process_attribute
    
    SUBROUTINE do_access(iacc)
      !-------------------------------------------------------------------------
      ! Arguments
      INTEGER, INTENT(IN) :: iacc
      !*************************************************************************
      IF (attrs%access /= 0) THEN
        CALL already('An /access-spec/')
      ELSE
        attrs%access = iacc
      END IF
      CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
    END SUBROUTINE do_access
    
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C472 via ' // comp,  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG=text // ' has already been specified for this &
            &/type-bound-procedure-stmt/.' )
    END SUBROUTINE already
    
    SUBROUTINE pass_and_nopass
      USE ErrorCodes
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C475 via ' // comp,  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='A binding shall not have both the PASS and NOPASS attributes.' )
    END SUBROUTINE pass_and_nopass
    
    SUBROUTINE non_overridable_and_deferred
      USE ErrorCodes
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C476 via ' // comp,  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='A binding shall not have both the DEFERRED and &
            &NON_OVERRIDABLE attributes.' )
    END SUBROUTINE non_overridable_and_deferred
    
  END SUBROUTINE parse_binding_attr_list
  
  
  !****************************************************************************
  !!
  !> Parse a /binding-name/.  
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token that starts the
  !! /binding-name/.
  !!
  !! @param[in]     finish            Index of the token that ends the 
  !! /binding-name/.
  !!
  !! @param[out]    decl              Results of parsing the /binding-name/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! This doesn't have a specific syntax rule as it is covered by the general 
  !! /xyz-name/ syntax rule (R102).
  
  SUBROUTINE parse_binding_name(tlist, start, finish, decl, err_list)
    
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(BindingDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/binding-name/', err)
    CALL Add(err_list, err)
    IF (Failure(err)) RETURN
    decl%binding = IdentifierIfortBug(tlist(start))
    
    ! Check and error if there is extra stuff.
    IF (finish == start) RETURN
    ! Can't use the ParseUtils variant as we have no component.
    CALL Add( err_list,  &
        CODE=errSyntax,  &  
        LOCATION=QueryLocation(tlist(start+1)),  &
        MSG='"' // QueryRaw(tlist(start+1)) // '" was unexpected.' )
    
  END SUBROUTINE parse_binding_name
  
  
  !*****************************************************************************
  !!
  !> Parse a /type-bound-proc-decl/ (R449).
  
  SUBROUTINE parse_type_bound_proc_decl(tlist, start, finish, decl, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(BindingDecl), INTENT(OUT) :: decl
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /typre-bound-proc-decl/.
    CHARACTER(*), PARAMETER :: comp = 'R449'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error list for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    CALL ExpectName(tlist, start, '/binding-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    decl%binding = IdentifierIfortBug(tlist(start))
    
    ! That can be it - the specific procedure bit is optional.
    IF (start == finish) RETURN
    
    !---------------------------------------------------------------------------
    ! => /procedure-name/
    
    CALL Expect(tlist, start + 1, ittAssociate, scck_'=>', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    CALL Expect(tlist, start + 2, '/procedure-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    decl%specific = IdentifierIfortBug(tlist(start+2))
    
    !---------------------------------------------------------------------------
    ! That's all folks...
    
    CALL ExpectEnd(tlist(:finish), start + 3, comp, err_list)
    
  END SUBROUTINE parse_type_bound_proc_decl
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundProcedureStmt%Dump - write a debugging 
  !! representation of the results of parsing the statement to a 
  !! logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE tbp_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundProcedureStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! binding name index.
    
    !***************************************************************************
    
    IF (ALLOCATED(pnode%interface_name)) THEN
      WRITE (unit, "(A,'/interface-name/ = ',A)") pfx,  &
          .NameOf. pnode%interface_name
    END IF
      
    WRITE (unit, "(A,'attributes:')") pfx
    IF (pnode%attrs%nopass) THEN
      WRITE (unit, "(A,'NOPASS')") pfx // '  '
    ELSE IF (pnode%attrs%pass) THEN
      IF (ALLOCATED(pnode%attrs%pass_arg_name)) THEN
        WRITE (unit, "(A,'PASS(',A,')')") pfx // '  ',  &
            .NameOf. pnode%attrs%pass_arg_name
      ELSE
        WRITE (unit, "(A,'PASS')") pfx // '  '
      END IF
    END IF
    IF (pnode%attrs%access /= 0) THEN
      WRITE (unit, "(A)", ADVANCE='NO') pfx // '  '
      CALL WriteAccess(unit, pnode%attrs%access)
      WRITE (unit, "()")
    END IF
    IF (pnode%attrs%deferred) WRITE (unit, "(A,'DEFERRED')") pfx // '  '
    IF (pnode%attrs%non_overridable)  &
        WRITE (unit, "(A,'NON_OVERRIDABLE')") pfx // '  '
      
    WRITE (unit, "(A,'/binding-name-list/')") pfx
    DO i = 1, SIZE(pnode%decls)
      IF (ALLOCATED(pnode%decls(i)%specific)) THEN
        WRITE (unit, "(A,A,' => ',A)") pfx // '  ',  &
            .NameOf. pnode%decls(i)%binding,  &
            .NameOf. pnode%decls(i)%specific
      ELSE
        WRITE (unit, "(A,A)") pfx // '  ', .NameOf. pnode%decls(i)%binding
      END IF
    END DO
    
  END SUBROUTINE tbp_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundProcedureStmt%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE tbp_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundProcedureStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tbp_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundProcedureStmt%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[in,out] pnode             The parse node element to modify.
  !!
  !! @param[in,out] visitor           The object executing the operation.
  
  SUBROUTINE tbp_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundProcedureStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tbp_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundProcedureStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'type-bound-procedure-stmt'.
  
  FUNCTION tbp_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-bound-procedure-stmt'
    
  END FUNCTION tbp_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeBoundProcedureStmt%Advertise - advertise the 
  !! identifiers referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE tbp_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundProcedureStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: ipass          ! First pass count, second pass assign.
    INTEGER :: cnt            ! Number of identifiers.
    INTEGER :: id             ! Declaration index.
    
    !***************************************************************************
    
    DO ipass = 1, 2
      cnt = 0
      IF (ALLOCATED(pnode%interface_name)) THEN
        cnt = cnt + 1
        IF (ipass == 2) THEN
          ident_list(cnt) = IdentifierNature(  &
              NAME=.NameOf. pnode%interface_name,  &
              NATURES=[NatureRefProcInterface],  &
              LOCATIONS=[.LocationOf. pnode%interface_name] )
        END IF
      END IF
      DO id = 1, SIZE(pnode%decls)
        ! There is always a binding name.
        cnt = cnt + 1
        IF (ipass == 2) THEN
          ! A binding name without a specific procedure and without 
          ! an interface name (i.e. DEFERRED) is also a specific 
          ! procedure reference.
          IF ( ALLOCATED(pnode%decls(id)%specific)  &
              .OR. ALLOCATED(pnode%interface_name) ) THEN
            ident_list(cnt) = IdentifierNature(  &
                NAME=.NameOf. pnode%decls(id)%binding,  &
                NATURES=[NatureDefSpecBinding],  &
                LOCATIONS=[.LocationOf. pnode%decls(id)%binding] )
          ELSE
            ident_list(cnt) = IdentifierNature(  &
                NAME=.NameOf. pnode%decls(id)%binding,  &
                NATURES=[NatureDefSpecBinding, NatureRefProcedure],  &
                LOCATIONS=[ .LocationOf. pnode%decls(id)%binding,  &
                    .LocationOf. pnode%decls(id)%binding ] )
          END IF
        END IF
        ! There may also be a specific procedure reference.
        IF (ALLOCATED(pnode%decls(id)%specific)) THEN
          cnt = cnt + 1
          IF (ipass == 2) THEN
            ident_list(cnt) = IdentifierNature(  &
                NAME=.NameOf. pnode%decls(id)%specific,  &
                NATURES=[NatureRefProcedure],  &
                LOCATIONS=[.Locationof. pnode%decls(id)%specific] )
          END IF
        END IF
      END DO
      IF (ipass == 1) ALLOCATE(ident_list(cnt))
    END DO
    
  END SUBROUTINE tbp_Advertise
  
END MODULE TypeBoundProcedureStmts
