! $Id: TypeDeclarationStmts.f90 2867 2019-03-30 23:09:43Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the TypeDeclarationStmts module.


!*******************************************************************************
!!
!> Procedures and types for parsing /type-declaration-stmt/'s.
!!
!! Note that this module depends on the ProcedureDeclarationStmts module.

MODULE TypeDeclarationStmts
  
  USE DeclarationTypeSpecs
  USE Designators
  USE Expressions
  USE Identifiers
  USE LanguageBindingSpecs
  USE ParseNodes
  USE ProcedureDeclarationStmts
  USE SubscriptLists
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------  
  
  PUBLIC :: ParseTypeDeclarationStmt
  PUBLIC :: ParseDeclList
  
  PUBLIC :: DumpEntityDetails
  
  !-----------------------------------------------------------------------------  
  
  !> Type to capture the attributes that might be directly attached to 
  !! each entity in the /type-declaration-stmt/.
  TYPE, EXTENDS(TermParseNode), PUBLIC :: EntityDetails
    !> The name of the entity
    TYPE(Identifier) :: name
    
    !> Dimension specifications, if present.
    !!
    !! If not allocated then the entity is a scalar.  These override 
    !! any DIMENSION attribute specified in the body of the 
    !! type-declaration-stmt.
    TYPE(DimSpec), ALLOCATABLE :: dim_specs(:)
    
    !> Codimension specification, if present.
    !!
    !! If not allocated then the entity is not a coarray.  These override 
    !! any CODIMENSION attribute specified in the body of the 
    !! type-declaration-stmt.
    TYPE(DimSpec), ALLOCATABLE :: codim_specs(:)
    
    !> Character length specifier, if present.
    CLASS(Expression), ALLOCATABLE :: len
    
    !> An initializer of the /constant-expr/ variety, if present.
    CLASS(Expression), ALLOCATABLE :: expr
    
    !> An initialiser of the /null-init/ variety was present.
    TYPE(NullInit), ALLOCATABLE :: null_init
    
    !> An initialiser of the /initial-data-target/ variety, if present.
    TYPE(Designator), ALLOCATABLE :: initial_data_target
    
    !> The attributes specified by the hosting statement.
    CLASS(AttrBase), POINTER :: attrs => NULL()
  CONTAINS
    PROCEDURE :: Write => ed_Write
    PROCEDURE :: Query => ed_Query
    PROCEDURE :: Modify => ed_Modify
    PROCEDURE, NOPASS :: What => ed_What
    PROCEDURE :: Advertise => ed_Advertise
    PROCEDURE :: GetLocation => ed_GetLocation
  END TYPE EntityDetails
  
  !-----------------------------------------------------------------------------
  
  !> Attributes that may be nominated in a /type-declaration-stmt/.
  TYPE, EXTENDS(AttrBase), PUBLIC :: AttrSpec
    !> The ALLOCATABLE attribute was specified.
    LOGICAL :: allocatable = .FALSE.
    !> The ASYNCHRONOUS attribute was specified.
    LOGICAL :: asynchronous = .FALSE.
    !> The CONTIGUOUS attribute was specified.
    LOGICAL :: contiguous = .FALSE.
    !> The CONTIGUOUS attribute was specified.
    LOGICAL :: external = .FALSE.
    !> The INTRINSIC attribute was specified.
    LOGICAL :: intrinsic = .FALSE.
    !> The OPTIONAL attribute was specified.
    LOGICAL :: optional = .FALSE.
    !> The PARAMETER (constant) attribute was specified.
    LOGICAL :: parameter = .FALSE.
    !> The POINTER attribute was specified.
    LOGICAL :: pointer = .FALSE.
    !> The PROTECTED attribute was specified.
    LOGICAL :: protected = .FALSE.
    !> SAVE attribute explicitly specified.
    LOGICAL :: save = .FALSE.
    !> TARGET attribute specified.
    LOGICAL :: target = .FALSE.
    !> VALUE attribute specified.
    LOGICAL :: value = .FALSE.
    !> Volatile attribute specified.
    LOGICAL :: volatile = .FALSE.
    
    !> Access specification - see iacc* constants defined in the AccessStmts 
    !! module.
    !!
    !! - 0: Not specified.
    !! - 1: Private.
    !! - 2: Public.
    INTEGER :: access = 0
    
    !> Dimension specifications.
    !!
    !! If not allocated then the entity is a scalar.
    TYPE(DimSpec), ALLOCATABLE :: dim_specs(:)
    
    !> Codimension specification.
    !!
    !! If not allocated then the entity is not a coarray.
    TYPE(DimSpec), ALLOCATABLE :: codim_specs(:)
    
    !> Intent specification - see the iint* constants defined in IntentSpecs.
    !!
    !! - 0: Intent not specified.
    !! - 1: INTENT(IN)
    !! - 2: INTENT(OUT)
    !! - 3: INTENT(INOUT)
    INTEGER :: intent = 0
    
    !> Language binding specification.
    !!
    !! If not allocated then there was no binding specification.
    TYPE(LangSpec), ALLOCATABLE :: binding
  CONTAINS
    PROCEDURE :: Dump => as_Dump
    PROCEDURE :: Query => as_Query
    PROCEDURE :: Modify => as_Modify
    PROCEDURE, NOPASS :: What => as_What
    PROCEDURE :: GetNatures => attr_GetNatures
  END TYPE AttrSpec
  
  !> Type to capture the results of parsing a /type-declaration-stmt/.
  !!
  !! Components that are logical either have had the attribute specified 
  !! (.TRUE.) or have not (.FALSE.).  All components are default initialised 
  !! to .FALSE..
  !!
  !! Components that are integer are used for attributes that might have 
  !! an unspecified state as well as other states.
  !!
  !! Note that a lot of the components that are listed here duplicate the 
  !! information that is in the Entity type.  But this type is for the 
  !! statement processing stage, while the Entity type is used for later 
  !! stages.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: TypeDeclarationStmt
    !> The /declaration-type-spec/ that appears at the start of the statement.
    TYPE(DeclTypeSpec) :: decl_spec
    
    !> The attributes nominated in the statement.
    !!
    !! This is managed through a pointer so that we can reliably share it 
    !! with the EntityDetails objects.  The attrs object is owned by the 
    !! TypeDeclarationStmt object.
    !!
    !! Always associated after construction.
    TYPE(AttrSpec), POINTER :: attrs => NULL()
    
    !> Details for each entity defined by the statement.
    TYPE(EntityDetails), ALLOCATABLE :: names(:)
  CONTAINS
    PROCEDURE :: Dump => type_Dump
    PROCEDURE :: Query => type_Query
    PROCEDURE :: Modify => type_Modify
    PROCEDURE, NOPASS :: What => type_What
    
    FINAL :: type_Final_scalar
  END TYPE TypeDeclarationStmt
  
  !-----------------------------------------------------------------------------
  
  LOGICAL, PUBLIC, PARAMETER :: ExplicitShapeSpecListMask(5)  &
      = [.TRUE., .FALSE., .FALSE., .FALSE., .FALSE.]
  
  LOGICAL, PUBLIC, PARAMETER :: DeferredShapeSpecListMask(5)  &
      = [.FALSE., .FALSE., .TRUE., .FALSE., .FALSE.]
  
  LOGICAL, PUBLIC, PARAMETER :: AllShapeSpecListMask(5)  &
      = [.TRUE., .TRUE., .TRUE., .TRUE., .TRUE.]
  
CONTAINS
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for TypeDeclarationStmt.
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-declaration-stmt/ (R501).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    typedec_stmt      Results of parsing the 
  !! /type-declaration-stmt/.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseTypeDeclarationStmt(tlist, typedec_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(TypeDeclarationStmt), INTENT(OUT) :: typedec_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for a /type-declaration-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R501'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: declare        ! Location of the :: token.
    INTEGER :: it             ! Current token index
    INTEGER :: finish         ! Syntax element final token.
    INTEGER :: i              ! Declarator index.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    ALLOCATE(typedec_stmt%attrs)
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /type-declaration-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Locate the tokens associated with the declaration-type-spec.  This 
    ! always starts from token 1, but we need to work out where it finishes.
      
    declare = FindBare(tlist, ittDeclare, scck_'::', 1)
    
    finish = GetDeclarationTypeSpecFinish(tlist, declare == 0)
    IF (finish == 0) finish = SIZE(tlist)
    
    CALL ParseDeclarationTypeSpec(  &
        tlist(:finish),  &
        declare == 0,  &
        typedec_stmt%decl_spec,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(err_list)) RETURN
    
    it = finish + 1
    
    !---------------------------------------------------------------------------
    ! Go looking for attributes.
    
    ! If we have attributes then there will be a comma after the type-spec.
    IF (Match(tlist, it, ittComma, scck_',')) THEN
      IF (declare == 0) THEN
        ! There's a syntax error somewhere - processing the remainder of the 
        ! statement as the list of /attr-spec/'s should flush it out.
        finish = SIZE(tlist)
      ELSE
        finish = declare - 1
      END IF
      CALL parse_attributes( &
          tlist,  &
          it + 1,  &
          finish,  &
          typedec_stmt%attrs,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      
      IF (declare == 0) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=QueryLocation(tlist(finish)),  &
            MSG='Expected "::".' )
        RETURN
      END IF
      it = declare + 1
    ELSE
      ! The double colon is optional.
      IF (Match(tlist, it, ittDeclare, scck_'::')) it = it + 1
    END IF
    
    !---------------------------------------------------------------------------
    ! The /entity-decl-list/.
    
    ! it here should point to the token that is the start of the 
    ! variable name list.
    IF (typedec_stmt%attrs%allocatable .OR. typedec_stmt%attrs%pointer) THEN
      CALL ParseDeclList(  &
          tlist(it:),  &
          typedec_stmt%attrs,  &
          .FALSE.,  &
          DeferredShapeSpecListMask,  &
          typedec_stmt%names,  &
          sub_err_list )
    ELSE
      CALL ParseDeclList(  &
          tlist(it:),  &
          typedec_stmt%attrs,  &
          .FALSE.,  &
          AllShapeSpecListMask,  &
          typedec_stmt%names,  &
          sub_err_list )
    END IF
    CALL Add(err_list, sub_err_list)
    IF (Failure(sub_err_list)) RETURN
    
    !---------------------------------------------------------------------------
    ! Simple sanity checks.
    !
    ! We assume here that the CHARACTER intrinsic type cannot be renamed.
    
    IF (ALLOCATED(typedec_stmt%decl_spec%type_spec)) THEN
      ASSOCIATE(type_spec => typedec_stmt%decl_spec%type_spec)
        IF (.NameOf. type_spec%type_name /= 'CHARACTER') THEN
          DO i = 1, SIZE(typedec_stmt%names)
            IF (ALLOCATED(typedec_stmt%names(i)%len)) THEN
              CALL Add( err_list,  &
                  CODE=errSyntax,  &
                  COMPONENT='C504 on R503',  &
                  LOCATION=typedec_stmt%names(i)%len%GetLocation(),  &
                  MSG='A /char-length/ specifier is only permitted if the &
                    &entity being declared is of type CHARACTER.' )
            END IF
          END DO
        END IF
      END ASSOCIATE
    END IF
    
  END SUBROUTINE ParseTypeDeclarationStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeDeclarationStmt%Dump - write a debugging 
  !! representation of the results of parsing a type declaration statement 
  !! to a logical unit.
  !!
  !! @param[in]     pnode             The TypeDeclarationStmt parse node.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE type_Dump(pnode, unit, pfx)
    
    USE DeclarationTypeSpecs
    USE Designators
    USE Expressions
    USE SubscriptLists
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeDeclarationStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: i              ! Entity index.
    
    !***************************************************************************
    
    WRITE (unit, "(A,'/declaration-type-spec/ = ')", ADVANCE='NO') pfx
    CALL WriteDeclTypeSpec(unit, pnode%decl_spec)
    WRITE (unit, "()")
    
    CALL pnode%attrs%Dump(unit, pfx)
    
    WRITE (unit, "(A,'/entity-decl-list/')") pfx
    DO i = 1, SIZE(pnode%names)
      CALL DumpEntityDetails(pnode%names(i), unit, pfx // '  ')
    END DO
    
  END SUBROUTINE type_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeDeclarationStmt%What - return a descriptive 
  !! string for the parse node.
  !!
  !! @returns 'type-declaration-stmt'.
  
  FUNCTION type_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-declaration-stmt'
    
  END FUNCTION type_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeDeclarationStmt%Query - traverse the parse node with 
  !! a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE type_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeDeclarationStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! entity-decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%decl_spec%Query(visitor)
    
    DO i = 1, SIZE(pnode%names)
      CALL pnode%names(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE type_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeDeclarationStmt%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  SUBROUTINE type_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeDeclarationStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! entity-decl index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    CALL pnode%decl_spec%Modify(visitor)
    
    DO i = 1, SIZE(pnode%names)
      CALL pnode%names(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE type_Modify
  
  
  !*****************************************************************************
  !!
  !> Scalar finalizer for a TypeDeclarationStmt object.
  !!
  !! @param[in,out] typedec_stmt      The object to be finalized.
  
  SUBROUTINE type_Final_scalar(typedec_stmt)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeDeclarationStmt), INTENT(INOUT) :: typedec_stmt
    
    !***************************************************************************
    
    IF (ASSOCIATED(typedec_stmt%attrs)) DEALLOCATE(typedec_stmt%attrs)
    
  END SUBROUTINE type_Final_scalar
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AttrSpec.
  
  
  !*****************************************************************************
  !!
  !> Parse the attributes part of a type declaration statement.
  !!
  !! @param[in]     tlist             The sequence of tokens for the 
  !! type-declaration-stmt.
  !!
  !! @param[in]     start             Index of the first token in @a tlist 
  !! of the list of /attr-spec/'s.
  !!
  !! @param[in]     finish            Index of the last token in @a tlist 
  !! of the list of /attr-spec/'s.
  !!
  !! @param[in,out] attrs             Result of parsing the list of 
  !! /attr-spec/'s.  This is INTENT(OUT) as the attributes are stored 
  !! in the parent TypeDeclarationStmt object.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Note that some attributes could be reasonably complicated (dimension 
  !! for example).
  !!
  !! This is similar in nature to the attribute parsing of derived type 
  !! components (see ComponentDefStmts::parse_dcd_attributes and 
  !! parse_pcd_attributes) and procedure declarations (see 
  !! ProcedureDeclarationStmts::parse_proc_attr_spec).  We've been somewhat 
  !! consistent, but we could do better.
  
  SUBROUTINE parse_attributes(tlist, start, finish, attrs, err_list)
    
    USE AccessStmts
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ErrorLevels
    USE Expressions
    USE IntentSpecs
    USE SourceLocations
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(AttrSpec), INTENT(INOUT) :: attrs
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a single /attr-spec/.
    CHARACTER(*), PARAMETER :: comp = 'R502'
    
    ! Keywords for each /attr-spec/.  The order is important - it needs to 
    ! correspond to the case selectors in the process_attribute internal 
    ! procedure.
    CHARACTER(*,KIND=scck), PARAMETER :: attr_names(19) = [  &
        scck_'PUBLIC      ', scck_'PRIVATE     ', scck_'ALLOCATABLE ',  &
        scck_'ASYNCHRONOUS', scck_'CODIMENSION ', scck_'CONTIGUOUS  ',  &
        scck_'DIMENSION   ', scck_'EXTERNAL    ', scck_'INTENT      ',  &
        scck_'INTRINSIC   ', scck_'BIND        ', scck_'OPTIONAL    ',  &
        scck_'PARAMETER   ', scck_'POINTER     ', scck_'PROTECTED   ',  &
        scck_'SAVE        ', scck_'TARGET      ', scck_'VALUE       ',  &
        scck_'VOLATILE    ' ]
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: list_count     ! Number of items in the /connect-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Location of dimension attribute for error reporting.
    TYPE(SourceLocation) :: dim_loc
    
    ! Location of codimension attribute for error reporting.
    TYPE(SourceLocation) :: codim_loc
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save starting location of the list of attributes.
    attrs%location = QueryLocation(tlist(start))
    
    list_count = ListItemCount(tlist(start:finish))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, start, finish, item_start, item_finish)
      
      CALL ExpectKeyword(  &
          tlist,  &
          item_start,  &
          attr_names,  &
          '/attr-spec/',  &
          opt, err )
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      
      ! Process the attribute into the flags of the typedec_stmt object.
      CALL process_attribute
    END DO
    
    !---------------------------------------------------------------------------
    ! In the component declaration case, we can check that if the ALLOCATABLE 
    ! or POINTER attribute appears then all /component-array-spec/'s in the 
    ! statement must be /deferred-shape-spec-list/.  The wording for the 
    ! general entity declaration is a little different - it refers to the 
    ! entities' /array-spec/, which is not necessarily the array spec 
    ! attached to the DIMENSION attribute.
    !
    ! But we'll emit a warning.
    
    IF (attrs%pointer .OR. attrs%allocatable) THEN
      IF (ALLOCATED(attrs%dim_specs)) THEN
        IF (.NOT. IsDeferredShapeSpecList(attrs%dim_specs))  &
            CALL Add( err_list,  &
                CODE=errSillyAttributes,  &
                LEVEL=errLevelWarning,  &
                COMPONENT='C532',  &
!                LOCATION=dim_loc, ??
                MSG='If an entity has either the ALLOCATABLE or POINTER &
                  &attribute then it shall have an /array-spec/ that is &
                  &a /deferred-shape-spec-list/.  Consequently, this &
                  & /array-spec/ can never be used.' )
      END IF
    END IF
    
  CONTAINS
    
    ! Just to make the structure of the main loop clearer.
    SUBROUTINE process_attribute
      INTEGER :: closing      ! Closing parenthesis position.
      !*************************************************************************
      SELECT CASE (opt)
      CASE (1)  ! PUBLIC
        CALL do_access(iaccPublic)
        
      CASE (2)  ! PRIVATE
        CALL do_access(iaccPrivate)
        
      CASE (3)  ! ALLOCATABLE
        IF (attrs%allocatable) THEN
          CALL already('"ALLOCATABLE"')
        ELSE
          IF (attrs%pointer) CALL pointer_and_allocatable
          attrs%allocatable = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (4)  ! ASYNCHRONOUS
        IF (attrs%asynchronous) THEN
          CALL already('"ASYNCHRONOUS"')
        ELSE
          attrs%asynchronous = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (5)  ! CODIMENSION
        IF (ALLOCATED(attrs%codim_specs)) THEN
          CALL already('"CODIMENSION"')
        ELSE
          IF (attrs%pointer) CALL pointer_and_coarray
          
          CALL Expect(tlist, item_start + 1, ittDelimiter, scck_'[', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ExpectClosing(tlist, item_start + 1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ParseCoarraySpec(  &
              tlist(item_start+2:closing-1),  &
              attrs%codim_specs,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! This location points to the CODIMENSION keyword.
          IF (.NOT. Failure(sub_err_list))  &
              codim_loc = QueryLocation(tlist(item_start))
          
          CALL ExpectEnd(tlist(:item_finish), closing + 1, comp, err_list)
        END IF
        
      CASE (6)  ! CONTIGUOUS
        IF (attrs%contiguous) THEN
          CALL already('"CONTIGUOUS"')
        ELSE
          attrs%contiguous = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (7)  ! DIMENSION
        IF (ALLOCATED(attrs%dim_specs)) THEN
          CALL already('"DIMENSION"')
        ELSE
          CALL Expect(tlist, item_start + 1, ittDelimiter, scck_'(', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ExpectClosing(tlist, item_start + 1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ParseArraySpec(  &
              tlist(item_start+2:closing-1),  &
              .FALSE.,  &
              attrs%dim_specs,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ! This location points to the DIMENSION keyword.
          IF (.NOT. Failure(sub_err_list))  &
              dim_loc = QueryLocation(tlist(item_start))
          
          CALL ExpectEnd(tlist(:item_finish), closing + 1, comp, err_list)
        END IF
        
      CASE (8)  ! EXTERNAL
        IF (attrs%external) THEN
          CALL already('"EXTERNAL"')
        ELSE
          attrs%external = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (9)  ! INTENT
        IF (attrs%intent /= iintNone) THEN
          CALL already('"INTENT"')
        ELSE
          CALL Expect(tlist, item_start + 1, ittDelimiter, scck_'(', err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ExpectClosing(tlist, item_start + 1, closing, err)
          CALL Add(err_list, err, COMPONENT=comp)
          IF (Failure(err)) RETURN
          
          CALL ParseIntentSpec(  &
              tlist(item_start+2:closing-1),  &
              attrs%intent,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list, COMPONENT=comp)
          IF (Failure(sub_err_list)) RETURN
          
          ! Make sure nothing superfluous follows
          CALL ExpectEnd(tlist(:item_finish), closing + 1, comp, err_list)
        END IF
        
      CASE (10) ! INTRINSIC
        IF (attrs%intrinsic) THEN
          CALL already('"INTRINSIC"')
        ELSE
          IF (attrs%pointer) CALL pointer_and_intrinsic
          attrs%intrinsic = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (11) ! BIND
        IF (ALLOCATED(attrs%binding)) THEN
          CALL already('"BIND"')
        ELSE
          ALLOCATE(attrs%binding)
          CALL ParseLanguageBindingSpec(  &
              tlist,  &
              item_start,  &
              item_finish,  &
              attrs%binding,  &
              sub_err_list )
          CALL Add(err_list, sub_err_list)
          IF (Failure(sub_err_list)) RETURN
        END IF
        
      CASE (12) ! OPTIONAL
        IF (attrs%optional) THEN
          CALL already('"OPTIONAL"')
        ELSE
          attrs%optional = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (13) ! PARAMETER
        IF (attrs%parameter) THEN
          CALL already('"PARAMETER"')
        ELSE
          attrs%parameter = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (14) ! POINTER 
        IF (attrs%pointer) THEN
          CALL already('"POINTER"')
        ELSE
          IF (attrs%allocatable) CALL pointer_and_allocatable
          IF (attrs%intrinsic) CALL pointer_and_intrinsic
          IF (attrs%target) CALL pointer_and_target
          IF (ALLOCATED(attrs%codim_specs)) CALL pointer_and_coarray
          attrs%pointer = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (15) ! PROTECTED
        IF (attrs%protected) THEN
          CALL already('"PROTECTED"')
        ELSE
          attrs%protected = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (16) ! SAVE
        IF (attrs%save) THEN
          CALL already('"SAVE"')
        ELSE
          attrs%save = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (17) ! TARGET
        IF (attrs%target) THEN
          CALL already('"TARGET"')
        ELSE
          IF (attrs%pointer) CALL pointer_and_target
          attrs%target = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (18) ! VALUE
        IF (attrs%value) THEN
          CALL already('"VALUE"')
        ELSE
          attrs%value = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE (19) ! VOLATILE
        IF (attrs%volatile) THEN
          CALL already('"VOLATILE"')
        ELSE
          attrs%volatile = .TRUE.
          CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
        END IF
        
      CASE DEFAULT
        STOP 'Unhandled keyword in TypeDeclarationStmts::parse_attributes.'
        
      END SELECT
      
    END SUBROUTINE process_attribute
    
    ! Utility internal procedure to process an /acces-spec/ (PUBLIC or 
    ! PRIVATE).
    SUBROUTINE do_access(iacc)
      !-------------------------------------------------------------------------
      ! Arguments
      INTEGER, INTENT(IN) :: iacc
      !*************************************************************************
      IF (attrs%access /= 0) THEN
        CALL already('An /access-spec/')
      ELSE
        attrs%access = iacc
      END IF
      CALL ExpectEnd(tlist(:item_finish), item_start + 1, comp, err_list)
    END SUBROUTINE do_access
    
    ! Utility internal procedure to handle repeat specification of an 
    ! attribute.
    SUBROUTINE already(text)
      USE ErrorCodes
      !-------------------------------------------------------------------------
      ! Arguments
      CHARACTER(*), INTENT(IN) :: text
      !*************************************************************************
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C501 on' // comp,  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG=text // ' has already been specified for this &
            &/component-def-stmt/.' )
    END SUBROUTINE already
    
    ! Utility internal procedure to handle inconsistent specification of both 
    ! the POINTER and the ALLOCATABLE attributes.
    SUBROUTINE pointer_and_allocatable
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C546',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='An entity shall not have both the POINTER and ALLOCATABLE &
            &attributes.' )
    END SUBROUTINE pointer_and_allocatable
    
    ! Utility internal procedure to handle inconsistent specification of both 
    ! the POINTER and the INTRINSIC attributes.
    SUBROUTINE pointer_and_intrinsic
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C546',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='An entity shall not have both the POINTER and INTRINSIC &
            &attributes.' )
    END SUBROUTINE pointer_and_intrinsic
    
    ! Utility internal procedure to handle inconsistent specification of both 
    ! the POINTER and the TARGET attributes.
    SUBROUTINE pointer_and_target
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C546',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='An entity shall not have both the POINTER and TARGET &
            &attributes.' )
    END SUBROUTINE pointer_and_target
    
    ! Utility internal procedure to handle inconsistent specification of both 
    ! the POINTER and the CODIMENSION attributes.
    SUBROUTINE pointer_and_coarray
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT='C546',  &
          LOCATION=QueryLocation(tlist(item_start)),  &
          MSG='An entity with the POINTER attribute shall not be a coarray.' )
    END SUBROUTINE pointer_and_coarray
    
  END SUBROUTINE parse_attributes
  
  
  !*****************************************************************************
  !!
  !> Write a debugging representation of the /entity-attr/'s in a 
  !! TypeDeclarationStmt to a logical unit.
  !!
  !! @param[in]     log_unit          Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     typedec_stmt      The TypeDeclarationStmt object.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE write_entity_attrs(log_unit, attrs, pfx)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: log_unit
    TYPE(AttrSpec), INTENT(IN) :: attrs
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (attrs%allocatable)   WRITE (log_unit, "(A,'ALLOCATABLE')") pfx
    IF (attrs%asynchronous)  WRITE (log_unit, "(A,'ASYNCHRONOUS')") pfx
    IF (attrs%contiguous)    WRITE (log_unit, "(A,'CONTIGUOUS')") pfx
    IF (attrs%external)      WRITE (log_unit, "(A,'EXTERNAL')") pfx
    IF (attrs%intrinsic)     WRITE (log_unit, "(A,'INTRINSIC')") pfx
    IF (attrs%optional)      WRITE (log_unit, "(A,'OPTIONAL')") pfx
    IF (attrs%parameter)     WRITE (log_unit, "(A,'PARAMETER')") pfx
    IF (attrs%pointer)       WRITE (log_unit, "(A,'POINTER')") pfx
    IF (attrs%protected)     WRITE (log_unit, "(A,'PROTECTED')") pfx
    IF (attrs%save)          WRITE (log_unit, "(A,'SAVE')") pfx
    IF (attrs%target)        WRITE (log_unit, "(A,'TARGET')") pfx
    IF (attrs%value)         WRITE (log_unit, "(A,'VALUE')") pfx
    IF (attrs%volatile)      WRITE (log_unit, "(A,'VOLATILE')") pfx
    
    IF (attrs%access /= 0) THEN
      WRITE (log_unit, "(A,'/access-spec/ = ')", ADVANCE='NO') pfx
      CALL WriteAccess(log_unit, attrs%access)
      WRITE (log_unit, "()")
    END IF
    
    IF (ALLOCATED(attrs%dim_specs)) THEN
      WRITE (log_unit, "(A,'DIMENSION (')",  ADVANCE='NO') pfx
      CALL WriteDimSpec(log_unit, attrs%dim_specs)
      WRITE (log_unit, "(')')")
    END IF
    
    IF (ALLOCATED(attrs%codim_specs)) THEN
      WRITE (log_unit, "(A,'CODIMENSION (')",  ADVANCE='NO') pfx
      CALL WriteDimSpec(log_unit, attrs%codim_specs)
      WRITE (log_unit, "(')')")
    END IF
    
    SELECT CASE (attrs%intent)
    CASE (0)  ! Do nothing
    CASE (1) ; WRITE (log_unit, "(A,'INTENT(IN)')") pfx
    CASE (2) ; WRITE (log_unit, "(A,'INTENT(OUT)')") pfx
    CASE (3) ; WRITE (log_unit, "(A,'INTENT(INOUT)')") pfx
    CASE DEFAULT ; STOP 'Invalid typedec_stmt%intent'
    END SELECT
    
    IF (ALLOCATED(attrs%binding)) THEN
      WRITE (log_unit, "(A,'BIND ')", ADVANCE='NO') pfx
      CALL WriteLangSpec(log_unit, attrs%binding)
      WRITE (log_unit, "()")
    END IF
    
  END SUBROUTINE write_entity_attrs
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%GetNatures - get the natures specified by 
  !! the attributes in the statement.
  !!
  !! @param[in]     attrs             The attributes of the statement.
  !!
  !! @param[out]    natures           The natures specified by the attributes.
  
  SUBROUTINE attr_GetNatures(attrs, natures)
    
    USE AccessStmts
    USE IntentSpecs
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AttrSpec), INTENT(IN) :: attrs
    INTEGER, INTENT(OUT), ALLOCATABLE :: natures(:)
    
    !***************************************************************************
    
    natures = [NatureDefHasType]
    IF (attrs%allocatable)    natures = [natures, NatureDefAllocatable]
    IF (attrs%asynchronous)   natures = [natures, NatureDefAsynchronous]
    IF (attrs%contiguous)     natures = [natures, NatureDefContiguous]
    IF (attrs%external)       natures = [natures, NatureDefExternal]
    IF (attrs%intrinsic)      natures = [natures, NatureDefIntrinsic]
    IF (attrs%optional)       natures = [natures, NatureDefOptional]
    IF (attrs%parameter)      natures = [natures, NatureDefParameter]
    IF (attrs%pointer)        natures = [natures, NatureDefPointer]
    IF (attrs%protected)      natures = [natures, NatureDefProtected]
    IF (attrs%target)         natures = [natures, NatureDefTarget]
    IF (attrs%value)          natures = [natures, NatureDefValue]
    IF (attrs%volatile)       natures = [natures, NatureDefVolatile]
    IF (attrs%save)           natures = [natures, NatureDefSave]
    
    SELECT CASE (attrs%access)
    CASE (iaccPublic)       ; natures = [natures, NatureDefPublic]
    CASE (iaccPrivate)      ; natures = [natures, NatureDefPrivate]
    END SELECT
    
    IF (ALLOCATED(attrs%dim_specs))  &
        natures = [natures, NatureDefDimension]
    IF (ALLOCATED(attrs%codim_specs))  &
        natures = [natures, NatureDefCodimension]
    
    SELECT CASE (attrs%intent)
    CASE (iintIn)           ; natures = [natures, NatureDefIntentIn]
    CASE (iintOut)          ; natures = [natures, NatureDefIntentOut]
    CASE (iintInOut)        ; natures = [natures, NatureDefIntentInOut]
    END SELECT
    
    IF (ALLOCATED(attrs%binding))  &
        natures = [natures, NatureDefBind]
    
  END SUBROUTINE attr_GetNatures
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for EntityDetails
  
  
  !*****************************************************************************
  !!
  !> Parse the /entity-decl-list/ part of a /type-declaration-stmt/, or 
  !! the /component-decl-list/ part of a /data-component-def-stmt/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     component         Flag to indicate that a 
  !! /component-decl-list/ is being parsed rather than a /entity-decl-list/.  
  !! This affects the error messages that may be generated.
  !!
  !! @param[in]     array_spec_mask   Mask of the types of array 
  !! specifications that should be considered valid by this parse.  In element 
  !! order:
  !! - /explicit-shape-spec-list/.
  !! - /assumed-shape-spec-list/.
  !! - /deferred-shape-spec-list/.
  !! - /assumed-size-spec/.
  !! - /implied-shape-spec-list/.
  !!
  !! @param[out]    entities          The list of entities declared by the 
  !! statement.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Note that the name list can include names, dimension specifications, 
  !! length specifications and initialisers.
  !!
  !! The first token in tlist should be the token that is immediately after 
  !! the double colon (or whatever the equivalent is in forms that don't have 
  !! a double colon).
  
  SUBROUTINE ParseDeclList( tlist, attrs, component, array_spec_mask,  &
      entities, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Tokens
    USE ParseUtils
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(AttrBase), INTENT(IN), TARGET :: attrs
    LOGICAL, INTENT(IN) :: component
    LOGICAL, INTENT(IN) :: array_spec_mask(:)
    TYPE(EntityDetails), INTENT(OUT), ALLOCATABLE :: entities(:)
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: list_count     ! Number of items in the /implicit-spec-list/.
    INTEGER :: item           ! List item index.
    INTEGER :: item_start     ! Start index in tlist of a particular item.
    INTEGER :: item_finish    ! Finish index in tlist of a particular item.
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      IF (component) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='R436',  &
            MSG='Expected /component-decl-list/.' )
      ELSE
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT='R501',  &
            MSG='Expected /entity-decl-list/.' )
      END IF
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    
    list_count = ListItemCount(tlist)
    ALLOCATE(entities(list_count))
    item_finish = 0
    DO item = 1, list_count
      CALL GetNextListItem(tlist, 1, SIZE(tlist), item_start, item_finish)
      CALL parse_name(  &
          tlist(item_start:item_finish),  &
          attrs,  &
          component,  &
          array_spec_mask,  &
          entities(item),  &
          sub_err_list )
      CALL Add(err_list, sub_err_list)
      IF (Failure(sub_err_list)) RETURN
    END DO
    
  END SUBROUTINE ParseDeclList
  
  
  !*****************************************************************************
  !!
  !> Parse an /entity-decl/ (R503) or a /component-decl/ (R438).
  !!
  !! @param[in]     tlist             The list of tokens that make up the 
  !! entity-decl.
  !!
  !! @param[in]     component         Flag to indicate that we are parsing 
  !! a /component-decl/.  This alters error messages.
  !!
  !! @param[in]     array_spec_mask   Mask to indicate which array 
  !! specifications are valid.  In element order:
  !! - /explicit-shape-spec-list/.
  !! - /assumed-shape-spec-list/.
  !! - /deferred-shape-spec-list/.
  !! - /assumed-size-spec/.
  !! - /implied-shape-spec-list/.
  !!
  !! @param[out]    entity            The results of parsing the decl.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_name( tlist, attrs, component, array_spec_mask,  &
      entity, err_list )
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE Expressions
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    CLASS(AttrBase), INTENT(IN), TARGET :: attrs
    LOGICAL, INTENT(IN) :: component
    LOGICAL, INTENT(IN) :: array_spec_mask(:)
    TYPE(EntityDetails), INTENT(OUT) :: entity
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: it             ! Current token position.
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: closing        ! Position of closing parenthesis.
    
    ! End of the part of the entity-decl that is being processed.
    INTEGER :: part_end
    
    ! Error list for child procedure calls. 
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    ! Option to indicate what part of the entity-decl was last parsed.
    ! - 0: Nothing yet!
    ! - 1: The name.
    ! - 2: The array-spec.
    ! - 3: The coarray-spec.
    ! - 4: The character length.
    ! - 5: The initialisation.
    INTEGER :: part
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! Save the given attrs pointer.
    entity%attrs => attrs
    
    ! Start from the beginning!
    it = 1
    part = 0
    
    ! The name of the entity will be the first token.  It is required.
    IF (component) THEN
      CALL ExpectName(tlist, it, '/entity-name/', err)
    ELSE
      CALL ExpectName(tlist, it, '/entity-name/ or /function-name/', err)
    END IF
    CALL Add(err_list, err, COMPONENT=my_comp())
    IF (Failure(err)) RETURN
    
    entity%name = IdentifierIfortBug(tlist(it))
    it = it + 1
    part = 1
    
    IF (it > SIZE(tlist)) RETURN
    
    !---------------------------------------------------------------------------
    ! Try for array specification.
    
    IF (IsToken(tlist(it), ittDelimiter, scck_'(')) THEN
      ! Find the end of the list.
      CALL ExpectClosing(tlist, it, part_end, err)
      CALL Add(err_list, err, COMPONENT=my_comp())
      IF (Failure(err)) RETURN
      
      ! Previously specified attributes (or context) may allow the type
      ! of array specification to be narrowed down.  If that is the
      ! case then we take advantage of it.
      IF (ALL(array_spec_mask .EQV. DeferredShapeSpecListMask)) THEN
        CALL ParseDeferredShapeSpecList(  &
            tlist(it+1:part_end-1),  &
            entity%dim_specs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        ! Try and continue on failure
      ELSE IF (ALL(array_spec_mask .EQV. ExplicitShapeSpecListMask)) THEN
        CALL ParseExplicitShapeSpecList(  &
            tlist(it+1:part_end-1),  &
            entity%dim_specs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        ! Try and continue on failure
      ELSE
        CALL ParseArraySpec(  &
            tlist(it+1:part_end-1),  &
            component, &
            entity%dim_specs,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        ! Try and continue on failure
      END IF
      
      ! Skip over what we parsed plus closing ")".
      it = part_end + 1
      IF (it > SIZE(tlist)) RETURN
      part = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! Try for coarray specification.
    
    IF (IsToken(tlist(it), ittDelimiter, scck_'[')) THEN
      ! Find the end of the list.
      CALL ExpectClosing(tlist, it, part_end, err)
      CALL Add(err_list, err, COMPONENT=my_comp())
      IF (Failure(err)) RETURN
      
      IF (component) THEN
        ! Must be a /deferred-coshape-spec-list/.
        CALL ParseDeferredCoshapeSpecList(  &
            tlist(it+1:part_end-1),  &
            entity%codim_specs,  &
            sub_err_list )
      ELSE
        CALL ParseCoarraySpec(  &
            tlist(it+1:part_end-1),  &
            entity%codim_specs,  &
            sub_err_list )
      END IF
      CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
      ! Try and continue on failure
      
      ! Skip over what we parsed plus closing ].
      it = part_end + 1
      IF (it > SIZE(tlist)) RETURN
      part = 3
    END IF
    
    !---------------------------------------------------------------------------
    ! Character length specification
    
    IF (IsToken(tlist(it), ittOperator, scck_'*')) THEN
      IF (SIZE(tlist) == it) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=my_comp(),  &
            LOCATION=QueryLocation(tlist(it)),  &
            MSG='Expected /char-length/ after "*"' )
        RETURN
      END IF
      
      ! Options are things like name * (expr), or name*123.
      !
      ! Test above ensures that tlist has more than it elements.
      CALL Expect(  &
          tlist, it + 1,  &
          [ittDelimiter, ittIntegerLiteral],  &
          [scck_'(', scck_' '],  &
          opt,  &
          err )
      CALL Add(err_list, err, COMPONENT=my_comp())
      IF (Failure(err)) RETURN
      
      IF (opt == 1) THEN
        ! name*(expr)
        CALL ExpectClosing(tlist, it + 1, closing, err)
        CALL Add(err_list, err, COMPONENT=my_comp())
        IF (Failure(err)) RETURN
        
        CALL ParseExpr(tlist, it + 2, closing - 1, entity%len, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        
        ! If there's crap after the closing ), the tests at the end of this 
        ! procedure will pick that up.
        it = closing + 1
      ELSE
        ! Lazy parse of the integer literal into an expression.
        CALL ParseExpr(tlist, it + 1, it + 1, entity%len, sub_err_list)
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        it = it + 2
      END IF
      IF (it > SIZE(tlist)) RETURN
      part = 4
    END IF
    
    !---------------------------------------------------------------------------
    ! Initialiser
    
    IF (IsToken(tlist(it), ittAssign, scck_'=')) THEN
      IF (SIZE(tlist) == it) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=my_comp(),  &
            LOCATION=QueryLocation(tlist(it)),  &
            MSG='Expected /constant-expr/ after "=".' )
        RETURN
      END IF
      
      ! Parse it
      CALL ParseExpr(tlist, it + 1, SIZE(tlist), entity%expr, sub_err_list)
      CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        
      part = 5
      it = SIZE(tlist) + 1
    ELSE IF (IsToken(tlist(it), ittAssociate, scck_'=>')) THEN
      ! The initial pointer target can either be NULL() or an 
      ! /initial-data-target/.  An /initial-data-target/ is a /designator/
      ! with a heap of restrictions (TARGET, SAVE, variable, etc).  We manage 
      ! this by a process of elimination.
      IF (SIZE(tlist) == it) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=my_comp(),  &
            LOCATION=QueryLocation(tlist(it)),  &
            MSG='Expected /null-init/ or /initial-data-target/ after "=>".' )
        RETURN
      END IF
      
      IF (MatchNullInit(tlist, it+1)) THEN
        ALLOCATE(entity%null_init)
        CALL ParseNullInit(  &
            tlist,  &
            it + 1, SIZE(tlist),  &
            entity%null_init,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        IF (Failure(sub_err_list)) RETURN
      ELSE
        ! Try for an /initial-data-target/.
        ALLOCATE(entity%initial_data_target)
        CALL parse_initial_data_target(  &
            tlist,  &
            it + 1, SIZE(tlist),  &
            entity%initial_data_target,  &
            sub_err_list )
        CALL Add(err_list, sub_err_list, COMPONENT=my_comp())
        IF (Failure(sub_err_list)) RETURN
      END IF
      
      part = 5
      it = SIZE(tlist) + 1
    END IF
    
    IF (it <= SIZE(tlist)) THEN
      SELECT CASE (part)
      CASE (1)
        CALL Add( err_list,  &
            MSG='Syntax error - got "'  &
              // QueryRaw(tlist(it)) // '" when expecting ( /array-spec/ ), &
              &[ /coarray-spec/ ], * /char-length/ or /initialization/.',  &
            LOCATION=QueryLocation(tlist(it)),  &
            COMPONENT=my_comp() )
      CASE (2)
        CALL Add( err_list,  &
            MSG='Syntax error - got "'  &
              // QueryRaw(tlist(it)) // '" when expecting [ /coarray-spec/ ],  &
              &* /char-length/ or /initialization/.',  &
            LOCATION=QueryLocation(tlist(it)),  &
            COMPONENT=my_comp() )
      CASE (3)
        CALL Add( err_list,  &
            MSG='Syntax error - got "'  &
              // QueryRaw(tlist(it)) // '" when expecting * /char-length/ or &
              &/initialization/.',  &
            LOCATION=QueryLocation(tlist(it)),  &
            COMPONENT=my_comp() )
      CASE (4)
        CALL Add( err_list,  &
            MSG='Syntax error - got "'  &
              // QueryRaw(tlist(it)) // '" when expecting /initialization/.',  &
            LOCATION=QueryLocation(tlist(it)),  &
            COMPONENT=my_comp() )
      CASE (5)
        ! Not covered by tests in TypeDeclarationStmtTests.
        CALL Add( err_list,  &
            MSG='Syntax error - "'  &
              // QueryRaw(tlist(it)) // '" was unexpected.',  &
            LOCATION=QueryLocation(tlist(it)),  &
            COMPONENT=my_comp() )
      END SELECT
    END IF
    
  CONTAINS
    
    !> Utility internal function to return the right syntax rule for use
    !! in error reporting.
    FUNCTION my_comp()
      CHARACTER(4) :: my_comp
      !*************************************************************************
      IF (component) THEN
        my_comp = 'R438'
      ELSE
        my_comp = 'R503'
      END IF
    END FUNCTION my_comp
    
  END SUBROUTINE parse_name
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /intial-data-target/ (R443).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             The starting index of the sequence
  !! to parse.
  !!
  !! @param[in]     finish            The end index of the sequence to parse.
  !!
  !! @param[out]    data_target       The results of the parse.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An /initial-data-target/ is just a designator.
  
  SUBROUTINE parse_initial_data_target( tlist, start, finish, data_target,  &
      err_list )
    
    USE Designators
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(Designator), INTENT(OUT) :: data_target
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /initial-data-target/.
    CHARACTER(*), PARAMETER :: comp = 'R443'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    ! Error list for procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
      
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------  
    
    CALL ParseDesignator(tlist(start:finish), data_target, sub_err_list)
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
  END SUBROUTINE parse_initial_data_target
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for AttrSpec.
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%Dump - write a debugging 
  !! representation of the results of parsing the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix to use before all output 
  !! records.
  
  SUBROUTINE as_Dump(pnode, unit, pfx)
    
    USE AccessStmts
    USE Expressions
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AttrSpec), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    IF (pnode%allocatable)   WRITE (unit, "(A,'ALLOCATABLE')") pfx
    IF (pnode%asynchronous)  WRITE (unit, "(A,'ASYNCHRONOUS')") pfx
    IF (pnode%contiguous)    WRITE (unit, "(A,'CONTIGUOUS')") pfx
    IF (pnode%external)      WRITE (unit, "(A,'EXTERNAL')") pfx
    IF (pnode%intrinsic)     WRITE (unit, "(A,'INTRINSIC')") pfx
    IF (pnode%optional)      WRITE (unit, "(A,'OPTIONAL')") pfx
    IF (pnode%parameter)     WRITE (unit, "(A,'PARAMETER')") pfx
    IF (pnode%pointer)       WRITE (unit, "(A,'POINTER')") pfx
    IF (pnode%protected)     WRITE (unit, "(A,'PROTECTED')") pfx
    IF (pnode%save)          WRITE (unit, "(A,'SAVE')") pfx
    IF (pnode%target)        WRITE (unit, "(A,'TARGET')") pfx
    IF (pnode%value)         WRITE (unit, "(A,'VALUE')") pfx
    IF (pnode%volatile)      WRITE (unit, "(A,'VOLATILE')") pfx
    
    IF (pnode%access /= 0) THEN
      WRITE (unit, "(A,'/access-spec/ = ')", ADVANCE='NO') pfx
      CALL WriteAccess(unit, pnode%access)
      WRITE (unit, "()")
    END IF
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      WRITE (unit, "(A,'DIMENSION (')",  ADVANCE='NO') pfx
      CALL WriteDimSpec(unit, pnode%dim_specs)
      WRITE (unit, "(')')")
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      WRITE (unit, "(A,'CODIMENSION (')",  ADVANCE='NO') pfx
      CALL WriteDimSpec(unit, pnode%codim_specs)
      WRITE (unit, "(')')")
    END IF
    
    SELECT CASE (pnode%intent)
    CASE (0)  ! Do nothing
    CASE (1) ; WRITE (unit, "(A,'INTENT(IN)')") pfx
    CASE (2) ; WRITE (unit, "(A,'INTENT(OUT)')") pfx
    CASE (3) ; WRITE (unit, "(A,'INTENT(INOUT)')") pfx
    CASE DEFAULT ; STOP 'Invalid typedec_stmt%intent'
    END SELECT
    
    IF (ALLOCATED(pnode%binding)) THEN
      WRITE (unit, "(A,'BIND ')", ADVANCE='NO') pfx
      CALL WriteLangSpec(unit, pnode%binding)
      WRITE (unit, "()")
    END IF
    
  END SUBROUTINE as_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%Query - traverse the node with a 
  !! query operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE as_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AttrSpec), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      DO i = 1, SIZE(pnode%dim_specs)
        CALL pnode%dim_specs(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      DO i = 1, SIZE(pnode%codim_specs)
        CALL pnode%codim_specs(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%binding)) CALL pnode%binding%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE as_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%Modify - traverse the node with a 
  !! modify operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE as_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(AttrSpec), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Subscript index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      DO i = 1, SIZE(pnode%dim_specs)
        CALL pnode%dim_specs(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      DO i = 1, SIZE(pnode%codim_specs)
        CALL pnode%codim_specs(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%binding)) CALL pnode%binding%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE as_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of AttrSpec%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns The syntax rule name for the nodes.
  
  FUNCTION as_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'attr-spec'
    
  END FUNCTION as_What
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for EntityDetails.
  
  
  !*****************************************************************************
  !!
  !> Write a debugging represetnation of an EntityDetails object to a 
  !! logical unit.
  !!
  !! @param[in]     entity            The object to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE DumpEntityDetails(entity, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(EntityDetails), INTENT(IN) :: entity
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') pfx
    CALL entity%Write(unit)
    WRITE (unit, "()")
    
  END SUBROUTINE DumpEntityDetails
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%Write - write a debugging representation 
  !! of the node to a logical unit.
  !!
  !! @param[in]     pnode             The parse node to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE ed_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntityDetails), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    WRITE (unit, "(A)", ADVANCE='NO') .NameOf. pnode%name
    IF (ALLOCATED(pnode%dim_specs)) THEN
      WRITE (unit, "('(')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%dim_specs)
      WRITE (unit, "(')')", ADVANCE='NO')
    END IF
    IF (ALLOCATED(pnode%codim_specs)) THEN
      WRITE (unit, "('[')", ADVANCE='NO')
      CALL WriteDimSpec(unit, pnode%codim_specs)
      WRITE (unit, "(']')", ADVANCE='NO')
    END IF
    IF (ALLOCATED(pnode%len)) THEN
      WRITE (unit, "('*')", ADVANCE='NO')
      CALL pnode%len%Write(unit)
    END IF
    IF (ALLOCATED(pnode%expr)) THEN
      WRITE (unit, "(' = ')", ADVANCE='NO')
      CALL pnode%expr%Write(unit)
    END IF
    IF (ALLOCATED(pnode%null_init)) THEN
      WRITE (unit, "(' => ')", ADVANCE='NO')
      CALL pnode%null_init%Write(unit)
    END IF
    IF (ALLOCATED(pnode%initial_data_target)) THEN
      WRITE (unit, "(' => ')", ADVANCE='NO')
      CALL pnode%initial_data_target%Write(unit)
    END IF
    
  END SUBROUTINE ed_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'entity-details'
  
  FUNCTION ed_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'entity-details'
    
  END FUNCTION ed_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  RECURSIVE SUBROUTINE ed_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(EntityDetails), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Dimension or codimension index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      DO i = 1, SIZE(pnode%dim_specs)
        CALL pnode%dim_specs(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      DO i = 1, SIZE(pnode%codim_specs)
        CALL pnode%codim_specs(i)%Query(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%len)) CALL pnode%len%Query(visitor)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Query(visitor)
    
    IF (ALLOCATED(pnode%null_init)) CALL pnode%null_init%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ed_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  RECURSIVE SUBROUTINE ed_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(EntityDetails), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Dimension or codimension index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      DO i = 1, SIZE(pnode%dim_specs)
        CALL pnode%dim_specs(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      DO i = 1, SIZE(pnode%codim_specs)
        CALL pnode%codim_specs(i)%Modify(visitor)
      END DO
    END IF
    
    IF (ALLOCATED(pnode%len)) CALL pnode%len%Modify(visitor)
    
    IF (ALLOCATED(pnode%expr)) CALL pnode%expr%Modify(visitor)
    
    IF (ALLOCATED(pnode%null_init)) CALL pnode%null_init%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE ed_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%Advertise - advertise the 
  !! identifiers defined or referenced by the parse node, and the nature 
  !! of the definition or reference.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ed_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntityDetails), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Natures common to all entities.
    INTEGER, ALLOCATABLE :: natures(:)
    
    ! Locations for the natures for each specific entity.
    TYPE(SourceLocation), ALLOCATABLE :: locations(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    
    CALL pnode%attrs%GetNatures(natures)
    
    IF ( ALLOCATED(pnode%dim_specs)  &
        .AND. (.NOT. ANY(natures == NatureDefDimension)) ) THEN
      natures = [natures, NatureDefDimension]
    END IF
      
    IF ( ALLOCATED(pnode%codim_specs)  &
        .AND. (.NOT. ANY(natures == NatureDefCodimension)) ) THEN
      natures = [natures, NatureDefCodimension]
    END IF
    
    IF (ALLOCATED(pnode%expr)) THEN
      natures = [natures, NatureDefInitializer]
    END IF
      
    ALLOCATE(locations(SIZE(natures)))
    locations = .LocationOf. pnode%name
    
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=natures,  &
        LOCATIONS=locations )
    
  END SUBROUTINE ed_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of EntityDetails%GetLocation - get the start of the 
  !! source that corresponds to the ParseNode.
  !!
  !! @param[in]     pnode           The parse node of interest
  !!
  !! @returns The location in source of the start of the ParseNode.
  
  FUNCTION ed_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !-------------------------------------------------------------------------
    ! Characteristics
    
    CLASS(EntityDetails), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%name
    
  END FUNCTION ed_GetLocation
  
END MODULE TypeDeclarationStmts
