! $Id: TypeParamInquiries.f90 2795 2019-03-22 09:15:08Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the StructureConstructors module.


!*******************************************************************************
!!
!> Procedures and types for parsing /type-param-inquiries/.

MODULE TypeParamInquiries
  
  USE Designators
  USE Identifiers
  USE Expressions
  USE PrimaryExpressions
  
  IMPLICIT NONE
  
  PRIVATE
  
  PUBLIC :: ParseTypeParamInquiry
  PUBLIC :: ParseTypeParamInquiryFromDesignator
  
  !-----------------------------------------------------------------------------
  
  !> Stores the results of parsing a type param inquiry.
  TYPE, EXTENDS(Expression), PUBLIC :: TypeParamInquiry
    !> The designator for the data object being inquired about.
    TYPE(Designator) :: desig
    !> The type parameter name.
    TYPE(Identifier) :: type_param_name
  CONTAINS
    PROCEDURE :: Write => tpi_Write
    PROCEDURE :: GetLocation => tpi_GetLocation
    PROCEDURE :: Query => tpi_Query
    PROCEDURE :: Modify => tpi_Modify
    PROCEDURE, NOPASS :: What => tpi_What
  END TYPE TypeParamInquiry
  
  !> Syntax rule number for a /type-param-inquiry/.
  CHARACTER(*), PARAMETER :: comp = 'R616'
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /type-param-inquiry/.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the start of the 
  !! /type-param-inquiry/ in @a tlist.
  !!
  !! @param[in]     finish            Index of the end of the 
  !! /type-param-inquiry/ in @a tlist.
  !!
  !! @param[out]    tpi               Results of parsing the 
  !! /type-param-inquiry/.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! Typically we are unable to pick between a type parameter inquiry and 
  !! any general designator.  The one exception is if the type parameter 
  !! inquiry is on a substring of a character literal, but we still handle 
  !! the general case in the following.
  
  SUBROUTINE ParseTypeParamInquiry(tlist, start, finish, tpi, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    INTEGER, INTENT(IN) :: finish
    TYPE(TypeParamInquiry), INTENT(OUT) :: tpi
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: last_percent   ! Index of the last percent charcter
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    TYPE(Error) :: err        ! Error object for Expect* calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (finish < start) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=SafeLocation(tlist, finish),  &
          MSG='Expected /type-param-inquiry/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Find the end of the designator.
    
    last_percent = FindBareReverse(tlist(start:), ittPercent, scck_'%', finish)
    IF (last_percent == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          LOCATION=QueryLocation(tlist(finish)),  &
          COMPONENT=comp,  &
          MSG='Expected "%".' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! Parse the designator.
    
    CALL ParseDesignator(tlist(start:last_percent-1), tpi%desig, sub_err_list)
    CALL Add(err_list, sub_err_list)
    
    !---------------------------------------------------------------------------
    ! Store the type parameter name.
    
    CALL ExpectName( &
        tlist(:finish),  &
        last_percent + 1,  &
        '/type-parameter-name/',  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    tpi%type_param_name = IdentifierIfortBug(tlist(last_percent+1))
    
    !---------------------------------------------------------------------------
    ! Finish up.
    
    CALL ExpectEnd(tlist(:finish), last_percent + 2, comp, err_list)
    
  END SUBROUTINE ParseTypeParamInquiry
  
  
  !*****************************************************************************
  !!
  !> Convert a previously incorrectly classified designator to a type 
  !! parameter inquiry.
  
  SUBROUTINE ParseTypeParamInquiryFromDesignator(desig, tpi, err_list)
    
    USE Errors
    USE ErrorCodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Designator), INTENT(IN) :: desig
    TYPE(TypeParamInquiry), INTENT(OUT) :: tpi
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Workaround for ifort bug relating to self assignment of allocatables.
    type(partref), allocatable :: tmp_parts(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (ALLOCATED(desig%literal)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=.LocationOf. desig,  &
          MSG='Expected "%".' )
      RETURN
    END IF
    
    IF (SIZE(desig%part_refs) == 1) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=.LocationOf. desig,  &
          MSG='Expected "%".' )
      RETURN
    END IF
    
    IF (ALLOCATED(desig%substring)) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          LOCATION=.LocationOf. desig%substring,  &
          MSG='A substring is not permitted on a /type-param-inquiry/.' )
      RETURN
    END IF
    
    ASSOCIATE(last_part => desig%part_refs(SIZE(desig%part_refs)))
      IF (ALLOCATED(last_part%subscripts)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=.LocationOf. desig%substring,  &
            MSG='A /type-param-inquiry/ may not be subscripted.' )
        RETURN
      END IF
      IF (ALLOCATED(last_part%image_selector)) THEN
        CALL Add( err_list,  &
            CODE=errSyntax,  &
            COMPONENT=comp,  &
            LOCATION=.LocationOf. desig%substring,  &
            MSG='A /type-param-inquiry/ may not be followed by an &
              &/image-selector/.' )
        RETURN
      END IF
      tpi%type_param_name = last_part%part_name
    END ASSOCIATE
    tpi%desig = desig
    
    !===========================================================================
    ! Workaround for ifort bug relating to self assignment of things with 
    ! allocatable components.  The temporary should not be required.
    tmp_parts = tpi%desig%part_refs(:SIZE(tpi%desig%part_refs)-1)
    tpi%desig%part_refs = tmp_parts
    !===========================================================================
    
  END SUBROUTINE ParseTypeParamInquiryFromDesignator
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamInquiry%Write - write a debugging 
  !! representation of the expression to a logical unit.
  !!
  !1 @param[in]     pnode             The expression to write.
  !!
  !! @param[in]     unit              Unit connected for formatted output.
  
  SUBROUTINE tpi_Write(pnode, unit)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamInquiry), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    
    !***************************************************************************
    
    CALL pnode%desig%Write(unit)
    WRITE (unit, "(' % ',A)", ADVANCE='NO') .NameOf. pnode%type_param_name
    
  END SUBROUTINE tpi_Write
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamInquiry%GetLocation - get the location in 
  !! source of the start of the expression.
  !!
  !! @param[in]     pnode             The expression.
  !!
  !! @returns The source location for the start of the expression.
  
  FUNCTION tpi_GetLocation(pnode) RESULT(loc)
    
    USE SourceLocations
    
    !---------------------------------------------------------------------------
    ! Charateristics
    
    CLASS(TypeParamInquiry), INTENT(IN) :: pnode
    
    ! Function result.
    TYPE(SourceLocation) :: loc
    
    !***************************************************************************
    
    loc = .LocationOf. pnode%desig
    
  END FUNCTION tpi_GetLocation
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamInquiry%Query - traverse the parse node 
  !! with a query operation.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  RECURSIVE SUBROUTINE tpi_Query(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamInquiry), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The designator.
    CALL pnode%desig%Query(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpi_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamInquiry%Modify - traverse the parse node 
  !! with a modify operation.
  !!
  !! @param[inout] pnode             The parse node element to modify.
  !!
  !! @param[in]     visitor           The object executing the operation.
  
  RECURSIVE SUBROUTINE tpi_Modify(pnode, visitor)
    
    USE ParseNodes
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamInquiry), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    ! The designator.
    CALL pnode%desig%Modify(visitor)
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE tpi_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of TypeParamInquiry%What - get a text description of the 
  !! relevant syntax rule for the parse node.
  !!
  !! @param[in]     pnode             The parse node element to query.
  !!
  !! @returns 'expr'
  !!
  !! We may specialize this for the extensions of Expression at a later 
  !! date.
  
  FUNCTION tpi_What() RESULT(what)
      
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'type-param-inquiry'
    
  END FUNCTION tpi_What
  
END MODULE TypeParamInquiries
