! $Id: UseStmts.f90 2866 2019-03-29 23:36:31Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the UseStmts module.


!*******************************************************************************
!!
!> Procedures for parsing use statements (/use-stmt/'s).
!!
!! A /use-stmt/ (R1109) is:
!! - USE [ [ , /module-nature/ ] :: ] /module-name/ [ , /rename-list/ ]
!! - USE [ [ , /module-nature/ ] :: ] /module-name/ , ONLY: [ /only-list/ ]
!!
!! where /module-nature/ (R111) is:
!! - INTRINSIC 
!! - NON_INTRINSIC
!!
!! and /rename/ (R1111) is:
!! - /local-name/ => /use-name/
!! - OPERATOR( /local-defined-operator/ ) => OPERATOR( /use-defined-operator/ )
!!
!! /only/ is defined by R1112 as:
!! - /generic-spec/
!! - /only-use-name/ 
!! - /rename/
!!
!! /only-use-name/ is defined by R1113 as /use-name/.
!!
!! /generic-spec/ is defined by R1207 as:
!! - /generic-name/
!! - OPERATOR( /defined-operator/ )
!! - ASSIGNMENT ( = )
!! - /defined-io-generic-spec/
!!
!! /defined-io-generic-spec/ is defined by R1208 as:
!! - READ ( FORMATTED )
!! - READ ( UNFORMATTED ) 
!! - WRITE ( FORMATTED )
!! - WRITE ( UNFORMATTED ) 
!!
!! Parsing of /generic-spec/'s and /defined-io-generic-spec/'s is handled by 
!! the GenericSpecs module.

MODULE UseStmts
  
  USE GenericSpecs
  USE Identifiers
  USE ParseNodes
  USE UseStmtsUtils
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  ! Expose module procedures and interfaces
  
  PUBLIC :: GetModuleName
  
  PUBLIC :: ParseUseStmt
  
  PUBLIC :: GetRenameUseDescription
  PUBLIC :: UseSymbol
  
  !-----------------------------------------------------------------------------
  
  !> Stores the information in a /rename/ term, and some of the other 
  !! related clauses that go in the /rename-list/ or /only-list/.
  !!
  !! The parent GenericSpec type has a keyword component that indicates what 
  !! type of generic-spec we are dealing with and a name component that gives 
  !! the identifying name part of the /generic-spec/.  If we are dealing with 
  !! a rename where we are just renaming a name, then the keyword component 
  !! will be zero and the name component will be the local name to use.  
  !! Similarly, if we are renaming an operator then the keyword component will 
  !! be 1 and the name component will be the local name of the operator.
  !!
  !! @todo Consider whether it is appropriate for this to be an extension 
  !! of GenericSpec, or whether that spec should just be a component of 
  !! this type.
  TYPE, PUBLIC, EXTENDS(GenericSpec) :: Rename
    !> Name (which might be an operator) in the referenced module.  Not 
    !! allocated if keyword is 2 or more.
    !!
    !! Not allocated if the module entity is not being renamed.  Only 
    !! names and operators can be renamed - hence this being irrelevant 
    !! if keyword is 2 or more.
    TYPE(Identifier), ALLOCATABLE :: use
  CONTAINS
    PROCEDURE :: Advertise => ren_Advertise
  END TYPE Rename
  
  ! Values for the UseStmt%mod_nature component.
  
  !> The nature of the module was unspecified.
  INTEGER, PARAMETER, PUBLIC :: imnUnspecified = 0
  !> The nature of the module was intrinsic.
  INTEGER, PARAMETER, PUBLIC :: imnIntrinsic = 1
  !> The nature of the module was non-intrinsic
  INTEGER, PARAMETER, PUBLIC :: imnNonIntrinsic = 2
  
  !> Stores the details from a use statement.
  TYPE, EXTENDS(SCParseNode), PUBLIC :: UseStmt
    !> Name of the module.
    TYPE(Identifier) :: name
    !> Nature of the module - see the imn* constants.
    INTEGER :: mod_nature
    !> Flag to indicate whether the ONLY: clause is present.
    LOGICAL :: only_flag
    !> List of only items or rename item.
    !!
    !! Always allocated, may be zero size.
    TYPE(Rename), ALLOCATABLE :: items(:)
  CONTAINS
    PROCEDURE :: Dump => use_Dump
    PROCEDURE, NOPASS :: What => use_What
    PROCEDURE :: Advertise => use_Advertise
    PROCEDURE :: Query => use_Query
    PROCEDURE :: Modify => use_Modify
  END TYPE UseStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /use-stmt/ (R1109).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    use_stmt          Results of parsing the statement.  Not 
  !! fully defined if there are failures in @a err_list.
  !!
  !! @param[out]    err_list          Error list.
  
  SUBROUTINE ParseUseStmt(tlist, use_stmt, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(UseStmt), INTENT(OUT) :: use_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /use-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R1109'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    ! Error list for child parsing calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    INTEGER :: opt            ! Expect option.
    INTEGER :: declare        ! Location of the :: token.
    INTEGER :: it_end         ! End of current sub-sequence.
    
    INTEGER :: it             ! Current token index.
    INTEGER :: pass           ! Pass number.
    INTEGER :: ir             ! Rename index.
    
    INTEGER :: start          ! Start of current item
    INTEGER :: comma          ! Comma position
    INTEGER :: finish         ! Finish of current item
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /use-stmt/.' )
      RETURN
    END IF
    
    ! Set components to reasonably safe values so that in the event of error
    ! we can just bail.
    ALLOCATE(use_stmt%items(0))
    use_stmt%mod_nature = imnUnspecified
    use_stmt%only_flag = .FALSE.
    
    !---------------------------------------------------------------------------
    ! USE...
    
    CALL ExpectKeyword(tlist, 1, scck_'USE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! Maybe a comma followed by a module nature keyword, maybe the ::, 
    ! maybe the module name!
    
    CALL Expect(  &
        tlist,  &
        2,  &
        [ittComma, ittDeclare, ittName],  &
        [scck_'  ', scck_'::', scck_'  '],  &
        '/module-name/, "," or "::"',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      declare = FindBare(tlist, ittDeclare, scck_'::', 2)
      IF (declare == 0) THEN
        it_end = SIZE(tlist)
      ELSE
        it_end = declare - 1
      END IF
      
      CALL parse_module_nature(  &
          tlist(3:it_end),  &
          use_stmt%mod_nature,  &
          sub_err_list )
      CALL Add(err_list, sub_err_list, COMPONENT=comp)
      IF (Failure(sub_err_list)) RETURN
      it = it_end + 1
      
      CALL Expect(tlist, it, ittDeclare, scck_'::', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      it = it + 1
    ELSE IF (opt == 2) THEN
      it = 3
    ELSE
      it = 2
    END IF
    
    !---------------------------------------------------------------------------
    ! Module name
    
    CALL Expect(tlist, it, '/module-name/', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    use_stmt%name = IdentifierIfortBug(tlist(it))
    
    IF (SIZE(tlist) == it) RETURN
    
    !---------------------------------------------------------------------------
    ! Test for only.
    
    ! Whether ONLY is there or not, we expect a comma.
    CALL Expect(tlist, it + 1, ittComma, err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    it = it + 2
    IF (Match(tlist, it, ittName, scck_'ONLY')) THEN
      IF (Match(tlist, it + 1, ittColon)) THEN
        use_stmt%only_flag = .TRUE.
        it = it + 2
      END IF
      
      ! The /only-list/ is optional.  items component already zero allocated 
      ! above.
      IF (it > SIZE(tlist)) RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! only-list or rename-list
    
    ! @todo Rewrite using list parsing routines.
    
    ! First pass count, second pass store.
    DO pass = 1, 2
      start = it
      ir = 0
      DO 
        ! Get the location of the end of the item.
        comma = FindBare(tlist, ittComma, scck_',', start)
        IF (comma == 0) THEN
          finish = SIZE(tlist)
        ELSE
          finish = comma - 1
        END IF
              
        ir = ir + 1
        IF (pass == 2) THEN
          IF (use_stmt%only_flag) THEN
            CALL parse_only(  &
                tlist(start:finish),  &
                use_stmt%items(ir),  &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
          ELSE
            CALL parse_rename(  &
                tlist(start:finish),  &
                use_stmt%items(ir),  &
                sub_err_list )
            CALL Add(err_list, sub_err_list, COMPONENT=comp)
          END IF
        END IF
        IF (comma == 0) EXIT
        ! Skip the comma
        start = comma + 1
      END DO
      
      IF (pass == 1) THEN
        IF ((.NOT. use_stmt%only_flag) .AND. (ir == 0)) THEN
          ! No list followed the comma after the module name.
          CALL Add( err_list,  &
              CODE=errSyntax,  &
              COMPONENT=comp,  &
              LOCATION=QueryLocation(tlist(it-1)),  &
              MSG='Expected rename-list after ",".' )
        END IF
        ! We allocated items on entry, so we need to deallocate and reallocate 
        ! here.
        DEALLOCATE(use_stmt%items)
        ALLOCATE(use_stmt%items(ir))
      END IF
    END DO
    
  END SUBROUTINE ParseUseStmt
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens (which is one token normally) that is a 
  !! /module-nature/ (R1110).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    nature            Index describing the nature.  Not 
  !! defined if @a err_list contains failures.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE parse_module_nature(tlist, nature, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(OUT) :: nature
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /module-nature/.
    CHARACTER(*), PARAMETER :: comp = 'R1110'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    !---------------------------------------------------------------------------
    
    ! These are ordered as per the imn* constants.
    CALL Expect(  &
        tlist, 1,  &
        [ittName, ittName],  &
        [scck_'INTRINSIC    ', 'NON_INTRINSIC'],  &
        nature,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    
    CALL ExpectEnd(tlist, 2, comp, err_list)
    
  END SUBROUTINE parse_module_nature
  
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is a /rename/ (R1111).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    rename_item       Results of parsing the /rename/.  Not 
  !! defined if @a err_list contains failures.
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! @todo Reorder arguments.
  
  SUBROUTINE parse_rename(tlist, rename_item, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Rename), INTENT(OUT) :: rename_item
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for a /rename/.
    CHARACTER(*), PARAMETER :: comp = 'R1111'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    INTEGER :: next           ! Index of the next token after the rename.
    TYPE(Error) :: err        ! Error object for Expect calls.
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /rename/.' )
      RETURN
    END IF
    
    ! The location is always the first token.
    rename_item%location = QueryLocation(tlist(1))
    
    next = match_operator_spec(tlist, 1)
    IF (next /= 0) THEN
      ! We have something that starts with OPERATOR(xxx)
      CALL Expect(tlist, next, ittAssociate, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      ! Need another operator.
      CALL Expect(tlist, next + 1, ittName, scck_'OPERATOR', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      CALL Expect(tlist, next + 2, ittDelimiter, scck_'(', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      CALL Expect(tlist, next + 3, ittOperator, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      CALL Expect(tlist, next + 4, ittDelimiter, scck_')', err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      rename_item%keyword = 1       ! OPERATOR keyword
      rename_item%name = IdentifierIfortBug(tlist(3))
      rename_item%use = IdentifierIfortBug(tlist(next+3))
      next = next + 5
    ELSE
      ! The other option is a name.
      CALL Expect(tlist, 1, ittName, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      CALL Expect(tlist, 2, ittAssociate, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      ! Need another name.
      CALL Expect(tlist, 3, ittName, err)
      CALL Add(err_list, err, COMPONENT=comp)
      IF (Failure(err)) RETURN
      rename_item%keyword = 0       ! No keyword
      rename_item%name = IdentifierIfortBug(tlist(1))
      rename_item%use = IdentifierIfortBug(tlist(3))
      next = 4
    END IF
    
    CALL ExpectEnd(tlist, next, comp, err_list)
    
  END SUBROUTINE parse_rename
  
  
  !*****************************************************************************
  !!
  !> Parse an /only/ item (R1112).
  !!
  !! @param[in]     tlist             The list of tokens that supposedly 
  !! make up the only item.
  !!
  !! @param[out]    only_item         The only list item.  The storage object
  !!
  !! @param[out]    err_list          List of errors.
  !!
  !! An only item can be:
  !! - a generic spec.
  !! - a only-use-name.
  !! - a rename (see parse_rename).
  !!
  !! A /generic-spec/ can also just be a name, so we just use the parsing 
  !! for /generic-spec/'s to handle everything that is not a /rename/
  !!
  !! @todo Reorder arguments.
  
  SUBROUTINE parse_only(tlist, only_item, err_list)
    
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(Rename), INTENT(OUT) :: only_item
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule number for an /only/.
    CHARACTER(*), PARAMETER :: comp_only = 'R1112'
    
    !***************************************************************************
    
    IF (HasBare(tlist, ittAssociate, scck_'=>')) THEN
      CALL parse_rename(tlist, only_item, err_list)
    ELSE
      CALL ParseGenericSpec(tlist, only_item%GenericSpec, err_list)
    END IF
    
  END SUBROUTINE parse_only
  
  
  !*****************************************************************************
  !!
  !> Match a sequence of tokens that is of the form OPERATOR(op).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Starting point in the sequence at 
  !! which to look for the match.
  !!
  !! @returns The index of the token after the end of the operator spec.
  
  FUNCTION match_operator_spec(tlist, start) RESULT(pos)
    
    USE CompilerKinds
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    INTEGER :: pos
    
    !***************************************************************************
    
    ! Assume failure
    pos = 0
    IF (.NOT. Match(tlist, start, ittName, scck_'OPERATOR')) RETURN
    IF (.NOT. Match(tlist, start + 1, ittDelimiter, scck_'(')) RETURN
    IF (.NOT. Match(tlist, start + 2, ittOperator)) RETURN
    IF (.NOT. Match(tlist, start + 3, ittDelimiter, scck_')')) RETURN
    
    pos = start + 4
    
  END FUNCTION match_operator_spec
  
  
  !*****************************************************************************
  !!
  !> Match a /rename/ in a sequence of tokens.
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[in]     start             Index of the token at which to 
  !! start the match test.
  !!
  !! @returns The index of the next token that would appear after the 
  !! operator specification.  That token may not exist if the operator
  !! specification was terminated by the end of tlist.  If no operator
  !! specification was found returns zero.
  
  FUNCTION match_rename(tlist, start) RESULT(next)
    
    USE MatchUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    INTEGER, INTENT(IN) :: start
    
    ! Function result
    INTEGER :: next
    
    !***************************************************************************
    
    ! Assume failure
    next = 0
    IF (Match(tlist, start, ittName)) RETURN
    IF (Match(tlist, start + 1, ittAssociate)) RETURN
    IF (Match(tlist, start + 2, ittName)) RETURN
    
    next = start + 3
    
  END FUNCTION match_rename
  
  
  !*****************************************************************************
  !!
  !> Get a text description of a module nature.
  !!
  !! @param[in]    nature             The module nature.
  !!
  !! @returns An appropriate text string.
  
  FUNCTION get_nature_text(nature)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    INTEGER, INTENT(IN) :: nature
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: get_nature_text
    
    !***************************************************************************
    
    SELECT CASE (nature)
    CASE (0)      ; get_nature_text = '(not specified)'
    CASE (1)      ; get_nature_text = 'INTRINSIC'
    CASE (2)      ; get_nature_text = 'NON_INTRINSIC'
    CASE DEFAULT  ; get_nature_text = '(unknown)'
    END SELECT
    
  END FUNCTION get_nature_text
  
  
  !*****************************************************************************
  !!
  !> Implementation of UseStmt%Dump - write a debugging representation 
  !! of the results of parsing a use statement to a logical unit.
  !!
  !! @param[in]     pnode             The ParseNode to dump.
  !!
  !! @param[in]     unit              Logical unit connected for formatted 
  !! output.
  !!
  !! @param[in]     pfx               Prefix for each record.
  
  SUBROUTINE use_Dump(pnode, unit, pfx)
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UseStmt), INTENT(IN) :: pnode
    INTEGER, INTENT(IN) :: unit
    CHARACTER(*), INTENT(IN) :: pfx
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Item index.
    
    ! Prefix for child items.
    CHARACTER(:), ALLOCATABLE :: local_pfx
    
    !***************************************************************************
    
    local_pfx = pfx
    
    WRITE (unit, "(A,'/module-nature/ = ',A)") local_pfx,  &
        get_nature_text(pnode%mod_nature)
    WRITE (unit, "(A,'/module_name/ = ',A)") local_pfx, .NameOf. pnode%name
    WRITE (unit, "(A,'only_flag = ',L1)") local_pfx, pnode%only_flag
    IF (pnode%only_flag) THEN
      WRITE (unit, "(A,'/only-list/')") local_pfx
      DO i = 1, SIZE(pnode%items)
        IF (ALLOCATED(pnode%items(i)%use)) THEN
          WRITE (unit, "(A,I0,' keyword = ',I0,' name = ',A,&
              &' use = ',A)") local_pfx // '  ',  &
              i, pnode%items(i)%keyword,  &
              .NameOf. pnode%items(i)%name,  &
              .NameOf. pnode%items(i)%use
        ELSE IF (ALLOCATED(pnode%items(i)%name)) THEN
          WRITE (unit, "(A,I0,' keyword = ',I0,' name = ',A)")  &
              local_pfx // '  ',  &
              i,  &
              pnode%items(i)%keyword,  &
              .NameOf. pnode%items(i)%name
        ELSE
          WRITE (unit, "(A,I0,' keyword = ',I0)")  &
              local_pfx // '  ', i, pnode%items(i)%keyword
        END IF
      END DO
    ELSE
      WRITE (unit, "(A,'/rename-list/')") local_pfx
      DO i = 1, SIZE(pnode%items)
        IF (ALLOCATED(pnode%items(i)%use)) THEN
          WRITE (unit, "(A,I0,' keyword = ',I0,' name = ',A,&
                &' use = ',A)") local_pfx // '  ',  &
              i, pnode%items(i)%keyword,  &
              .NameOf. pnode%items(i)%name,  &
              .NameOf. pnode%items(i)%use
        ELSE IF (ALLOCATED(pnode%items(i)%name)) THEN
          WRITE (unit, "(A,I0,' keyword = ',I0,' name = ',A)")  &
              local_pfx // '  ',  &
              i,  &
              pnode%items(i)%keyword,  &
              .NameOf. pnode%items(i)%name
        ELSE
          WRITE (unit, "(A,I0,' keyword = ',I0)")  &
              local_pfx // '  ', i, pnode%items(i)%keyword
        END IF
      END DO
    END IF
    
  END SUBROUTINE use_Dump
  
  
  !*****************************************************************************
  !!
  !> Implementation of UseStmt%Query - traverse the node with a query 
  !! operation.
  !!
  !! @param[in]     pnode             The parse node being queried.
  !!
  !! @param[in,out] visitor           The object carrying out the query action.
  
  SUBROUTINE use_Query(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(UseStmt), INTENT(IN) :: pnode
    CLASS(ParseNodeQueryVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! use item index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%items)
      CALL pnode%items(i)%Query(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE use_Query
  
  
  !*****************************************************************************
  !!
  !> Implementation of UseStmt%Modify - traverse the node with a modify 
  !! operation.
  !!
  !! @param[in,out] pnode             The parse node being modified.
  !!
  !! @param[in,out] visitor           The object carrying out the operation.
  
  SUBROUTINE use_Modify(pnode, visitor)
    
    !---------------------------------------------------------------------------
    ! Argument
    
    CLASS(UseStmt), INTENT(INOUT) :: pnode
    CLASS(ParseNodeModifyVisitor), INTENT(INOUT) :: visitor
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! use item index.
    
    !***************************************************************************
    
    CALL visitor%Start(pnode)
    
    DO i = 1, SIZE(pnode%items)
      CALL pnode%items(i)%Modify(visitor)
    END DO
    
    CALL visitor%Finish(pnode)
    
  END SUBROUTINE use_Modify
  
  
  !*****************************************************************************
  !!
  !> Implementation of UseStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the statement.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE use_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(UseStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        NAME=.NameOf. pnode%name,  &
        NATURES=[NatureRefModule],  &
        LOCATIONS=[.LocationOf. pnode%name] )
    
  END SUBROUTINE use_Advertise
  
  
  !*****************************************************************************
  !!
  !> Implementation of UseStmt%What - return a descriptive string 
  !! for the parse node.
  !!
  !! @returns 'use-stmt'
  
  FUNCTION use_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'use-stmt'
    
  END FUNCTION use_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of Rename%Advertise - advertise the identifiers 
  !! referenced or defined by the syntax term.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE ren_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    USE GenericSpecs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Rename), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !***************************************************************************
    
    ALLOCATE(ident_list(1))
    ident_list(1) = IdentifierNature(  &
        pnode%GenericSpec,  &
        NATURES=[NatureRefUseName],  &
        LOCATIONS=[.LocationOf. pnode] )
    
  END SUBROUTINE ren_Advertise
  
  
  !*****************************************************************************
  !!
  !> Get descriptive text for a generic spec.
  !!
  !! @param[in]     spec              The generic spec to describe.
  !!
  !! @returns A string that resembles the original source.
  
  PURE FUNCTION GetRenameUseDescription(spec) RESULT(str)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Rename), INTENT(IN) :: spec
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: str
    
    !***************************************************************************
    
    IF (.NOT. ALLOCATED(spec%use)) THEN
      str = GetGenericSpecDescription(spec%GenericSpec)
    ELSE
      SELECT CASE (spec%keyword)
      CASE (0)      ; str = .NameOf. spec%use
      CASE (1)      ; str = 'OPERATOR(' // .NameOf. spec%use // ')'
      END SELECT
    END IF
    
  END FUNCTION GetRenameUseDescription
  
  
  !*****************************************************************************
  !!
  !> Convenience constructor to create an IdentifierSymbol from a 
  !! rename that is for the symbol in the original module.
  !!
  !! If you always want the symbol in the target module then call 
  !! IdentifierSymbol(spec%GenericSpec).
  !!
  !! The two symbols are the same if there is no rename.
  
  FUNCTION UseSymbol(spec) RESULT(sym)
    
    ! ifort bug - Shouldn't be required, but the overloaded structure 
    ! constructor causes issues.
    USE GenericSpecs
    USE IdentifierSymbols
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    TYPE(Rename), INTENT(IN) :: spec
    
    ! Function result.
    TYPE(IdentifierSymbol) :: sym
    
    !***************************************************************************
    
    IF (ALLOCATED(spec%use)) THEN
      SELECT CASE (spec%keyword)
      CASE (0)      ; sym = NameSymbol(.NameOf. spec%use)
      CASE (1)      ; sym = OperatorSymbol(.NameOf. spec%use)
      CASE (2)      ; sym = AssignmentSymbol()
      CASE (3)      ; sym = ReadFormattedSymbol()
      CASE (4)      ; sym = WriteFormattedSymbol()
      CASE (5)      ; sym = ReadUnformattedSymbol()
      CASE (6)      ; sym = WriteUnformattedSymbol()
      CASE DEFAULT  ; sym = NameSymbol('???')
      END SELECT
    ELSE
      sym = IdentifierSymbol(spec%GenericSpec)
    END IF
    
  END FUNCTION UseSymbol
  
END MODULE UseStmts
