! $Id: VolatileStmts.f90 2196 2016-06-07 02:56:48Z ian $
! ff08 source code copyright 2012 M.E.G.M.S.  See LICENCE.txt for licence.
!> @file
!! Defines the VolatileStmts module.


!*******************************************************************************
!!
!> Procedures for parsing /volatile-stmt/'s.

MODULE VolatileStmts
  
  USE ParseNodes
  
  IMPLICIT NONE
  
  PRIVATE
  
  !-----------------------------------------------------------------------------
  
  PUBLIC :: ParseVolatileStmt
  
  !> Stores the results of parsing a /volatile-stmt/.
  TYPE, EXTENDS(NameListPNode), PUBLIC :: VolatileStmt
  CONTAINS
    PROCEDURE, NOPASS :: GetListName => vol_GetListName
    PROCEDURE, NOPASS :: What => vol_What
    PROCEDURE :: Advertise => vol_Advertise
  END TYPE VolatileStmt
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Parse a sequence of tokens that is an /volatile-stmt/ (R559).
  !!
  !! @param[in]     tlist             The sequence of tokens.
  !!
  !! @param[out]    vol_stmt          The resuts of parsing the statement.
  !!
  !! @param[out]    err_list          List of errors.
  
  SUBROUTINE ParseVolatileStmt(tlist, vol_stmt, err_list)
    
    USE Strings
    USE CompilerKinds
    USE Errors
    USE ErrorCodes
    USE ParseUtils
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Token), INTENT(IN) :: tlist(:)
    TYPE(VolatileStmt), INTENT(OUT) :: vol_stmt
    TYPE(Error), INTENT(OUT), ALLOCATABLE :: err_list(:)
    
    !---------------------------------------------------------------------------
    ! Local constants
    
    ! Syntax rule for an /optional-stmt/.
    CHARACTER(*), PARAMETER :: comp = 'R559'
    
    !---------------------------------------------------------------------------
    ! Local variables
    
    TYPE(Error) :: err        ! Error object for Expect calls.
    INTEGER :: opt            ! Expect option.
    INTEGER :: it             ! Current token index.
    
    ! Error list for child procedure calls.
    TYPE(Error), ALLOCATABLE :: sub_err_list(:)
    
    !***************************************************************************
    
    ALLOCATE(err_list(0))
    
    IF (SIZE(tlist) == 0) THEN
      CALL Add( err_list,  &
          CODE=errSyntax,  &
          COMPONENT=comp,  &
          MSG='Expected /volatile-stmt/.' )
      RETURN
    END IF
    
    !---------------------------------------------------------------------------
    ! OPTIONAL...
    
    CALL Expect(tlist, 1, ittName, scck_'VOLATILE', err)
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    !---------------------------------------------------------------------------
    ! :: or /dummy-arg-name/
    
    CALL Expect( &
        tlist, 2,  &
        [ittDeclare, ittName],  &
        [scck_'::', scck_'  '],  &
        '"::" or /object-name/',  &
        opt,  &
        err )
    CALL Add(err_list, err, COMPONENT=comp)
    IF (Failure(err)) RETURN
    
    IF (opt == 1) THEN
      it = 3
    ELSE
      it = 2
    END IF
    ! it points to the start of the /dummy-arg-name-list/ here.
    
    !---------------------------------------------------------------------------
    ! The /dummy-arg-name-list.
    
    CALL ParseNameList(  &
        tlist,  &
        it, SIZE(tlist),  &
        '/object-name/',  &
        vol_stmt%name_list,  &
        sub_err_list )
    CALL Add(err_list, sub_err_list, COMPONENT=comp)
    IF (Failure(sub_err_list)) RETURN
    
    ! That's it!
    
  END SUBROUTINE ParseVolatileStmt
  
  
  !*****************************************************************************
  !!
  !> Implementation of VolatileStmt%GetListName - return a name for the 
  !! stored list of names.
  !!
  !! @returns '/object-name-list/'
  
  FUNCTION vol_GetListName() RESULT(name)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: name
    
    !***************************************************************************
    
    name = '/object-name-list/'
    
  END FUNCTION vol_GetListName
  
  
  !*****************************************************************************
  !!
  !> Implementation of VolatileStmt%What - return a descriptive string for 
  !! the parse node.
  !!
  !! @returns 'volatile-stmt'.
  
  FUNCTION vol_What() RESULT(what)
    
    !---------------------------------------------------------------------------
    ! Characteristics
    
    ! Function result.
    CHARACTER(:), ALLOCATABLE :: what
    
    !***************************************************************************
    
    what = 'volatile-stmt'
    
  END FUNCTION vol_What
  
  
  !*****************************************************************************
  !!
  !> Implementation of VolatileStmt%Advertise - advertise the identifiers 
  !! referenced or defined by the construct.
  !!
  !! @param[in]     pnode             The parse node to advertise.
  !!
  !! @param[out]    ident_list        Identifiers defined or referenced by 
  !! the node, and the nature of that definition or reference.
  
  SUBROUTINE vol_Advertise(pnode, ident_list)
    
    USE IdentifierNatures
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(VolatileStmt), INTENT(IN) :: pnode
    TYPE(IdentifierNature), INTENT(OUT), ALLOCATABLE :: ident_list(:)
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Identifier index.
    
    !***************************************************************************
    
    ALLOCATE(ident_list(SIZE(pnode%name_list)))
    
    DO i = 1, SIZE(pnode%name_list)
      ident_list(i) = IdentifierNature(  &
          NAME=.NameOf. pnode%name_list(i),  &
          NATURES=[NatureDefVolatile], &
          LOCATIONS=[.LocationOf. pnode%name_list(i)] )
    END DO
    
  END SUBROUTINE vol_Advertise
  
END MODULE VolatileStmts
