! $Id: RenderParseNodes.f90 2957 2019-11-24 07:26:13Z ian $
!> @file
!! Defines the RenderParseNodes module.


!*******************************************************************************
!!
!> Defines procedures and types to help map parse nodes back into a token 
!! stream.
!!
!! We use a new RenderNodeToken type because we are not really sure what 
!! we want to stick in it for now.  Perhaps this could just be an extension 
!! of Tokens%%Token.

MODULE RenderParseNodes
  
  ! We moved and rename some things - the renaming here is to account 
  ! for that.
  USE RenderNodes,  &
      append_rnode => AppendRNode,  &
      append_construct => AppendConstruct,  &
      append_stmt => AppendStmt,  &
      append_seq => AppendSeq,  &
      append_keyword => AppendKeyword,  &
      append_name => AppendName,  &
      append_label => AppendLabel,  &
      append_delimiter => AppendDelimiter,  &
      append_comma => AppendComma,  &
      append_token => AppendToken
  
  ! This list excludes ittEndStatement, as it has no role here.
  USE Tokens, ONLY:  &
      ittName, ittKeyword, ittIntegerLiteral, ittRealLiteral,  &
      ittCharacterLiteral, ittLogicalLiteral, ittBOZLiteral, ittOperator,  &
      ittStatementLabel, ittDelimiter, ittComma, ittAssign, ittAssociate,  &
      ittColon, ittDeclare, ittPercent, ittFormatSpecification
  
  IMPLICIT NONE
  
  PRIVATE
  
  ! All the token types.
  PUBLIC :: ittName, ittKeyword, ittIntegerLiteral, ittRealLiteral,  &
      ittCharacterLiteral, ittLogicalLiteral, ittBOZLiteral, ittOperator,  &
      ittStatementLabel, ittDelimiter, ittComma, ittAssign, ittAssociate,  &
      ittColon, ittDeclare, ittPercent, ittFormatSpecification
  PUBLIC :: RenderParseTree
  
  !-----------------------------------------------------------------------------
  
  !> Options for rendering to tokens.
  TYPE, PUBLIC :: RenderOptions
    ! Very few procedures pay attention to these at the moment.  One day...
    
    !> Always render the optional double colon in statements where 
    !! an optional double colon is permitted.
    LOGICAL :: always_double_colon = .TRUE.
    !> Always render the full keyword form of end statements for 
    !! program units and subprograms, rather than just "END".
    LOGICAL :: always_end_keywords = .TRUE.
    !> Always render the name of a program unit, subprogram, 
    !! interface body, derived type or other construct (if named) 
    !! in the closing and intermediate statements, where permitted.
    LOGICAL :: always_end_name = .TRUE.
    !> Always render the parentheses in a call statement, even 
    !! if it has no actual arguments.
    LOGICAL :: always_call_parens = .TRUE.
    !> Always render the cosntruct name for intermediate statements, 
    !! even if it is not required.
    LOGICAL :: always_construct_name = .TRUE.
    !> Always use the single statement form of the if, forall and 
    !! where constructs where construct content permits.
    LOGICAL :: always_single_stmt = .TRUE.
    !> Always use the F2008 TYPE(xx) form of declaration type 
    !! specifiers, even for intrinsic types.
    LOGICAL :: always_use_type = .FALSE.
    !> Always render do constructs in block form.
    LOGICAL :: always_block_do_construct = .FALSE.
  END TYPE RenderOptions
  
CONTAINS
  
  !*****************************************************************************
  !!
  !> Renders a parse node tree into a sequence of tokens.
  !!
  !! The token sequence is arranged by statement.
  RECURSIVE SUBROUTINE RenderParseTree(pnode, rnode, opts)
    
    USE ParseNodes
    
    USE AssociateConstructs
    USE BlockConstructs
    USE CaseConstructs
    USE CriticalConstructs
    USE DerivedTypeDefs
    USE DoConstructs
    USE EnumDefs
    USE ExecutionParts
    USE ForallConstructs
    USE IfConstructs
    USE InterfaceBlocks
    USE Programs
    USE ProgramUnits
    USE SelectTypeConstructs
    USE SpecificationParts
    USE WhereConstructs
    
    USE AccessStmts
    USE AllocatableStmts
    USE AllocateStmts
    USE AssignmentStmts
    USE AsynchronousStmts
    USE BackspaceStmts
    USE BindStmts
    USE CallStmts
    USE CloseStmts
    USE CodimensionStmts
    USE ContiguousStmts
    USE ContinueStmts
    USE CommonStmts
    USE ComplexLiteralConstants
    USE ComponentDefStmts
    USE DataStmts
    USE DeallocateStmts
    USE DerivedTypeStmts
    USE DimensionStmts
    USE DoStmts
    USE EndFileStmts
    USE EquivalenceStmts
    USE ExitStmts
    USE FinalProcedureStmts
    USE FlushStmts
    USE FormatStmts
    USE GotoStmts
    USE IfStmts
    USE ImplicitStmts
    USE InquireStmts
    USE NamelistStmts
    USE OpenStmts
    USE PointerAssignmentStmts
    USE PointerStmts
    USE PrintStmts
    USE ProcedureDeclarationStmts
    USE ProgramStmts
    USE ProtectedStmts
    USE ReadStmts
    USE ReturnStmts
    USE RewindStmts
    USE SaveStmts
    USE StmtFunctionStmts
    USE StopStmts
    USE StructureConstructors
    USE TypeDeclarationStmts
    USE TypeBoundProcedureStmts
    USE TypeBoundGenericStmts
    USE UseStmts
    USE WriteStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ParseNode), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !***************************************************************************
    
    SELECT TYPE (pnode)
    
    !---------------------------------------------------------------------------
    ! Constructs.
    
    TYPE IS (AssociateConstruct)
      CALL render_associate_construct(pnode, rnode, opts)
    TYPE IS (BlockConstruct)
      CALL render_block_construct(pnode, rnode, opts)
    TYPE IS (BlockData)
      CALL render_block_data(pnode, rnode, opts)
    TYPE IS (CaseConstruct)
      CALL render_case_construct(pnode, rnode, opts)
    TYPE IS (CriticalConstruct)
      CALL render_critical_construct(pnode, rnode, opts)
    TYPE IS (DerivedTypeDef)
      CALL render_derived_type_def(pnode, rnode, opts)
    TYPE IS (DoConstruct)
      CALL render_do_construct(pnode, rnode, opts)
    TYPE IS (EnumDef)
      CALL render_enum_def(pnode, rnode, opts)
    TYPE IS (ExecutionPart)
      CALL render_execution_part(pnode, rnode, opts)
    TYPE IS (ForallConstruct)
      CALL render_forall_construct(pnode, rnode, opts)
    TYPE IS (FunctionInterfaceBody)
      CALL render_function_interface_body(pnode, rnode, opts)
    TYPE IS (FunctionSubprogram)
      CALL render_function_subprogram(pnode, rnode, opts)
    TYPE IS (IfConstruct)
      CALL render_if_construct(pnode, rnode, opts)
    TYPE IS (InterfaceBlock)
      CALL render_interface_block(pnode, rnode, opts)
    TYPE IS (MainProgram)
      CALL render_main_program(pnode, rnode, opts)
    TYPE IS (Module)
      CALL render_module(pnode, rnode, opts)
    TYPE IS (Program)
      CALL render_program(pnode, rnode, opts)
    TYPE IS (SelectTypeConstruct)
      CALL render_select_type_construct(pnode, rnode, opts)
    TYPE IS (SeparateModuleSubprogram)
      CALL render_separate_module_subprogram(pnode, rnode, opts)
    TYPE IS (SpecificationPart)
      CALL render_specification_part(pnode, rnode, opts)
    TYPE IS (Submodule)
      CALL render_submodule(pnode, rnode, opts)
    TYPE IS (SubroutineInterfaceBody)
      CALL render_subroutine_interface_body(pnode, rnode, opts)
    TYPE IS (SubroutineSubprogram)
      CALL render_subroutine_subprogram(pnode, rnode, opts)
    TYPE IS (WhereConstruct)
      CALL render_where_construct(pnode, rnode, opts)
    
    !---------------------------------------------------------------------------
    ! Statements.
    
    TYPE IS (AccessStmt)
      CALL render_access_stmt(pnode, rnode, opts)
    TYPE IS (AllocatableStmt)
      CALL render_allocatable_stmt(pnode, rnode, opts)
    TYPE IS (AllocateStmt)
      CALL render_allocate_stmt(pnode, rnode, opts)
    TYPE IS (ArithmeticIfStmt)
      CALL render_arithmetic_if_stmt(pnode, rnode, opts)
    TYPE IS (AssignmentStmt)
      CALL render_assignment_stmt(pnode, rnode, opts)
    TYPE IS (BackspaceStmt)
      CALL render_backspace_stmt(pnode, rnode, opts)
    TYPE IS (BindStmt)
      CALL render_bind_stmt(pnode, rnode, opts)
    TYPE IS (CallStmt)
      CALL render_call_stmt(pnode, rnode, opts)
    TYPE IS (CloseStmt)
      CALL render_close_stmt(pnode, rnode, opts)
    TYPE IS (CodimensionStmt)
      CALL render_codimension_stmt(pnode, rnode, opts)
    TYPE IS (CommonStmt)
      CALL render_common_stmt(pnode, rnode, opts)
    TYPE IS (ComputedGotoStmt)
      CALL render_computed_goto_stmt(pnode, rnode, opts)
    TYPE IS (ContiguousStmt)
      CALL render_contiguous_stmt(pnode, rnode, opts)
    TYPE IS (ContinueStmt)
      CALL render_continue_stmt(pnode, rnode, opts)
    TYPE IS (CycleStmt)
      CALL render_cycle_stmt(pnode, rnode, opts)
    TYPE IS (DataComponentDefStmt)
      CALL render_data_component_def_stmt(pnode, rnode, opts)
    TYPE IS (DataStmt)
      CALL render_data_stmt(pnode, rnode, opts)
    TYPE IS (DeallocateStmt)
      CALL render_deallocate_stmt(pnode, rnode, opts)
    TYPE IS (DerivedTypeStmt)
      CALL render_derived_type_stmt(pnode, rnode, opts)
    TYPE IS (DimensionStmt)
      CALL render_dimension_stmt(pnode, rnode, opts)
    TYPE IS (EndFileStmt)
      CALL render_endfile_stmt(pnode, rnode, opts)
    TYPE IS (EquivalenceStmt)
      CALL render_equivalence_stmt(pnode, rnode, opts)
    TYPE IS (ExitStmt)
      CALL render_exit_stmt(pnode, rnode, opts)
    TYPE IS (ErrorStopStmt)
      CALL render_error_stop_stmt(pnode, rnode, opts)
    TYPE IS (FinalProcedureStmt)
      CALL render_final_procedure_stmt(pnode, rnode, opts)
    TYPE IS (FlushStmt)
      CALL render_flush_stmt(pnode, rnode, opts)
    TYPE IS (FormatStmt)
      CALL render_format_stmt(pnode, rnode, opts)
    TYPE IS (GotoStmt)
      CALL render_goto_stmt(pnode, rnode, opts)
    TYPE IS (ImplicitStmt)
      CALL render_implicit_stmt(pnode, rnode, opts)
    TYPE IS (InquireStmt)
      CALL render_inquire_stmt(pnode, rnode, opts)
    TYPE IS (NamelistStmt)
      CALL render_namelist_stmt(pnode, rnode, opts)
    TYPE IS (OpenStmt)
      CALL render_open_stmt(pnode, rnode, opts)
    TYPE IS (PrintStmt)
      CALL render_print_stmt(pnode, rnode, opts)
    TYPE IS (PointerAssignmentStmt)
      CALL render_pointer_assignment_stmt(pnode, rnode, opts)
    TYPE IS (PointerStmt)
      CALL render_pointer_stmt(pnode, rnode, opts)
    TYPE IS (ProcComponentDefStmt)
      CALL render_proc_component_def_stmt(pnode, rnode, opts)
    TYPE IS (ProcedureDeclarationStmt)
      CALL render_procedure_declaration_stmt(pnode, rnode, opts)
    TYPE IS (ProcedureStmt)
      CALL render_procedure_stmt(pnode, rnode, opts)
    TYPE IS (ProgramStmt)
      CALL render_program_stmt(pnode, rnode, opts)
    TYPE IS (ProtectedStmt)
      CALL render_protected_stmt(pnode, rnode, opts)
    TYPE IS (ReadStmt)
      CALL render_read_stmt(pnode, rnode, opts)
    TYPE IS (ReturnStmt)
      CALL render_return_stmt(pnode, rnode, opts)
    TYPE IS (RewindStmt)
      CALL render_rewind_stmt(pnode, rnode, opts)
    TYPE IS (SaveStmt)
      CALL render_save_stmt(pnode, rnode, opts)
    TYPE IS (StmtFunctionStmt)
      CALL render_stmt_function_stmt(pnode, rnode, opts)
    TYPE IS (StopStmt)
      CALL render_stop_stmt(pnode, rnode, opts)
    TYPE IS (SubroutineStmt)
      CALL render_subroutine_stmt(pnode, rnode, opts)
    TYPE IS (TypeBoundProcedureStmt)
      CALL render_type_bound_procedure_stmt(pnode, rnode, opts)
    TYPE IS (TypeBoundGenericStmt)
      CALL render_type_bound_generic_stmt(pnode, rnode, opts)
    TYPE IS (TypeDeclarationStmt)
      CALL render_type_declaration_stmt(pnode, rnode, opts)
    TYPE IS (TypeParamDefStmt)
      CALL render_type_param_def_stmt(pnode, rnode, opts)
    TYPE IS (UseStmt)
      CALL render_use_stmt(pnode, rnode, opts)
    TYPE IS (FunctionStmt)
      CALL render_function_stmt(pnode, rnode, opts)
    TYPE IS (WriteStmt)
      CALL render_write_stmt(pnode, rnode, opts)
    
    !---------------------------------------------------------------------------
    ! Other syntax terms.
    
    TYPE IS (StructureConstructor)
      CALL render_structure_constructor(pnode, rnode, opts)
    TYPE IS (ComplexLiteralConstant)
      CALL render_complex_literal_constant(pnode, rnode, opts)
    
    !---------------------------------------------------------------------------
    ! Things we forgot.
    
    CLASS DEFAULT
      PRINT "('The unimplemented ParseNode was for a ',A)", pnode%What()
      STOP 'Rendering not implemented for a ParseNode.'
    END SELECT
    
    IF (.NOT. ALLOCATED(rnode)) THEN
      PRINT "('The failing node was a ',A)", pnode%What()
      STOP 'RenderNode was unallocated.'
    END IF
    
  END SUBROUTINE RenderParseTree
  
  
  !*****************************************************************************
  !!
  !> Render a program.
  
  SUBROUTINE render_program(pnode, rnode, opts)
    
    USE Programs
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Program), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Program unit index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: tmp_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptProgram
    ALLOCATE(seq%items(0))
    
    DO i = 1, SIZE(pnode%program_units)
      CALL RenderParseTree(pnode%program_units(i)%item, tmp_rnode, opts)
      CALL append_rnode(seq, tmp_rnode)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_program
  
  
  !*****************************************************************************
  !!
  !> Render a MODULE
  
  SUBROUTINE render_module(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Module), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! For building first and last statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptModule
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istModule
    CALL append_keyword(stmt_rnode, 'MODULE', 0)
    CALL append_name(stmt_rnode, pnode%mod_stmt%module_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_module_common(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndModule
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'MODULE', 0)
    CALL append_name(stmt_rnode, pnode%mod_stmt%module_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_module
  
  
  !*****************************************************************************
  !!
  !> Render a MODULE
  
  SUBROUTINE render_submodule(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Submodule), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! For building first and last statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptSubmodule
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istSubmodule
    CALL append_keyword(stmt_rnode, 'SUBMODULE', 0)
    CALL append_delimiter(stmt_rnode, '(')
    
    CALL append_name(stmt_rnode, pnode%submod_stmt%ancestor_module_name, 0)
    IF (ALLOCATED(pnode%submod_stmt%parent_submodule_name)) THEN
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
      CALL append_name(stmt_rnode, pnode%submod_stmt%parent_submodule_name, 0)
    END IF
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_name(stmt_rnode, .NameOf. pnode%submod_stmt%submodule_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_module_common(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istSubmodule
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'SUBMODULE', 0)
    CALL append_name(stmt_rnode, .NameOf. pnode%submod_stmt%submodule_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_submodule
  
  
  !*****************************************************************************
  !!
  !> Render the common bits of a MODULE or SUBMODULE
  
  SUBROUTINE render_module_common(pnode, seq, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ModuleCommon), INTENT(IN) :: pnode
    CLASS(RenderNodeConstruct), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! Module procedure index.
    
    ! Render nodes from child parts.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Render node for the module-subprogram part.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: module_subprogram_rnode
    
    ! Render node for the contains statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    IF (SIZE(pnode%internals) /= 0) THEN
      ALLOCATE(module_subprogram_rnode)
      module_subprogram_rnode%context = iptModuleSubprogramPart
      
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istContains
      CALL append_keyword(stmt_rnode, 'CONTAINS', 0)
      CALL append_stmt(module_subprogram_rnode, stmt_rnode)
    
      DO i = 1, SIZE(pnode%internals)
        CALL RenderParseTree(  &
            pnode%internals(i)%item,  &
            sub_rnode,  &
            opts )
        CALL append_rnode(module_subprogram_rnode, sub_rnode)
      END DO
      
      CALL append_construct(seq, module_subprogram_rnode)
    END IF
    
  END SUBROUTINE render_module_common
  
  
  !*****************************************************************************
  !!
  !> Render a /main-program/.
  
  SUBROUTINE render_main_program(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(MainProgram), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptMainProgram
    
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      CALL render_program_stmt(pnode%prog_stmt, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL render_exec_program_unit(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndProgram
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'PROGRAM', 0)
    IF (ALLOCATED(pnode%prog_stmt)) THEN
      CALL append_name(stmt_rnode, pnode%prog_stmt%name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_main_program
  
  
  !*****************************************************************************
  !!
  !> Render a /function-subprogram/.
  
  SUBROUTINE render_function_subprogram(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionSubprogram), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptFunctionSubprogram
    
    CALL render_function_stmt(pnode%fn_stmt, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_exec_program_unit(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndFunction
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'FUNCTION', 0)
    CALL append_name(stmt_rnode, pnode%fn_stmt%function_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_function_subprogram
  
  
  !*****************************************************************************
  !!
  !> Render a /subroutine-subprogram/.
  
  SUBROUTINE render_subroutine_subprogram(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineSubprogram), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptSubroutineSubprogram
    
    CALL render_subroutine_stmt(pnode%sub_stmt, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_exec_program_unit(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndSubroutine
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'SUBROUTINE', 0)
    CALL append_name(stmt_rnode, pnode%sub_stmt%subroutine_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_subroutine_subprogram
  
  
  !*****************************************************************************
  !!
  !> Render a /separate-module-subprogram/.
  
  SUBROUTINE render_separate_module_subprogram(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SeparateModuleSubprogram), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptSeparateModuleSubprogram
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istMpSubprogram
    CALL append_keyword(stmt_rnode, 'MODULE', 0)
    CALL append_keyword(stmt_rnode, 'PROCEDURE', 0)
    CALL append_name(stmt_rnode, pnode%mps_stmt%name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_exec_program_unit(pnode, seq, opts)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndSubroutine
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'PROCEDURE', 0)
    CALL append_name(stmt_rnode, pnode%mps_stmt%name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_separate_module_subprogram
  
  
  !*****************************************************************************
  !!
  !> Render the common parts of an ExecProgramUnit extension.
  
  SUBROUTINE render_exec_program_unit(pnode, seq, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExecProgramUnit), INTENT(IN) :: pnode
    CLASS(RenderNodeConstruct), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    ! Render node for the internal-subprogram part.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: internal_subprogram_rnode
    
    INTEGER :: i              ! Internal subprogram index.
    
    !***************************************************************************
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_execution_part(pnode%exec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    IF (SIZE(pnode%internals) /= 0) THEN
      ALLOCATE(internal_subprogram_rnode)
      internal_subprogram_rnode%context = iptInternalSubprogramPart
      
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istContains
      CALL append_keyword(stmt_rnode, 'CONTAINS', 0)
      
      CALL append_stmt(internal_subprogram_rnode, stmt_rnode)
    
      DO i = 1, SIZE(pnode%internals)
        CALL RenderParseTree(  &
            pnode%internals(i)%item,  &
            sub_rnode,  &
            opts )
        CALL append_rnode(internal_subprogram_rnode, sub_rnode)
      END DO
      
      CALL append_construct(seq, internal_subprogram_rnode)
    END IF
    
  END SUBROUTINE render_exec_program_unit
  
  
  !*****************************************************************************
  !!
  !> Render a /block-data/ program unit.
  
  SUBROUTINE render_block_data(pnode, rnode, opts)
    
    USE ProgramUnits
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockData), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptBlockData
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istBlockData
    CALL append_keyword(stmt_rnode, 'BLOCK', 0)
    CALL append_keyword(stmt_rnode, 'DATA', 0)
    IF (ALLOCATED(pnode%bd_stmt%block_data_name)) THEN
      CALL append_name(stmt_rnode, pnode%bd_stmt%block_data_name, 0)
    END IF
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndBlockData
    CALL append_keyword(stmt_rnode, 'END', 0)
    IF (opts%always_end_keywords) THEN
      CALL append_keyword(stmt_rnode, 'BLOCK', 0)
      CALL append_keyword(stmt_rnode, 'DATA', 0)
      IF ( opts%always_end_name  &
          .AND. ALLOCATED(pnode%bd_stmt%block_data_name) ) THEN
        CALL append_name(stmt_rnode, pnode%bd_stmt%block_data_name, 0)
      END IF
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_block_data
  
  
  !*****************************************************************************
  !!
  !> Render a specification part.
  
  RECURSIVE SUBROUTINE render_specification_part(pnode, rnode, opts)
    
    USE SpecificationParts
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SpecificationPart), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptSpecificationPart
    
    DO i = 1, SIZE(pnode%use_stmts)
      CALL render_use_stmt(pnode%use_stmts(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    DO i = 1, SIZE(pnode%import_stmts)
      CALL render_import_stmt(pnode%import_stmts(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    DO i = 1, SIZE(pnode%children)
      CALL RenderParseTree(pnode%children(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_specification_part
  
  
  !*****************************************************************************
  !!
  !> Render an /execution-part/.
  
  RECURSIVE SUBROUTINE render_execution_part(pnode, rnode, opts)
    
    USE ExecutionParts
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ExecutionPart), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptExecutionPart
    
    DO i = 1, SIZE(pnode%children)
      CALL RenderParseTree(pnode%children(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_execution_part
  
  
  !*****************************************************************************
  !!
  !> Render a /select-type-construct/.
  
  SUBROUTINE render_select_type_construct(pnode, rnode, opts)
    
    USE SelectTypeConstructs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SelectTypeConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child block index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptSelectTypeConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istSelectType
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'SELECT', 0)
    CALL append_keyword(stmt_rnode, 'TYPE', 0)
    CALL append_delimiter(stmt_rnode, '(')
    IF (ALLOCATED(pnode%associate_name)) THEN
      CALL append_name(stmt_rnode, pnode%associate_name, 0)
      CALL append_token(stmt_rnode, ittAssociate, '=>', '', 0)
    END IF
    CALL render_expression(pnode%selector, sub_rnode, opts)
    CALL append_rnode(stmt_rnode, sub_rnode)
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_stmt(seq, stmt_rnode)
    
    DO i = 1, SIZE(pnode%blocks)
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istTypeGuard
      
      IF (pnode%blocks(i)%polymorphic) THEN
        CALL append_keyword(stmt_rnode, 'CLASS', 0)
      ELSE
        CALL append_keyword(stmt_rnode, 'TYPE', 0)
      END IF
      IF (ALLOCATED(pnode%blocks(i)%type_spec)) THEN
        CALL append_keyword(stmt_rnode, 'IS', 0)
        CALL append_delimiter(stmt_rnode, '(')
        CALL render_type_spec(pnode%blocks(i)%type_spec, sub_rnode, opts)
        CALL append_rnode(stmt_rnode, sub_rnode)
        CALL append_delimiter(stmt_rnode, ')')
      ELSE
        CALL append_keyword(stmt_rnode, 'DEFAULT', 0)
      END IF
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
      
      CALL render_block(pnode%blocks(i)%children, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndSelectType
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'SELECT', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_select_type_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /case-construct/.
  
  SUBROUTINE render_case_construct(pnode, rnode, opts)
    
    USE CaseConstructs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CaseConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child block index.
    INTEGER :: iv             ! Value index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptCaseConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istSelectCase
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'SELECT', 0)
    CALL append_keyword(stmt_rnode, 'CASE', 0)
    CALL append_delimiter(stmt_rnode, '(')
    CALL render_expression(pnode%expr, sub_rnode, opts)
    CALL append_rnode(stmt_rnode, sub_rnode)
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_stmt(seq, stmt_rnode)
    
    DO i = 1, SIZE(pnode%blocks)
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istCase
      
      CALL append_keyword(stmt_rnode, 'CASE ', 0)
      IF (SIZE(pnode%blocks(i)%values) == 0) THEN
        CALL append_keyword(stmt_rnode, 'DEFAULT', 0)
      ELSE
        CALL append_delimiter(stmt_rnode, '(')
        DO iv = 1, SIZE(pnode%blocks(i)%values)
          ASSOCIATE(val => pnode%blocks(i)%values(iv))
            IF (ALLOCATED(val%lower)) THEN
              CALL render_expression(val%lower, sub_rnode, opts)
              CALL append_rnode(stmt_rnode, sub_rnode)
            END IF
            IF (.NOT. val%is_single) THEN
              CALL append_token(stmt_rnode, ittColon, ':', '', 0)
            END IF
            IF (ALLOCATED(val%upper)) THEN
              CALL render_expression(val%upper, sub_rnode, opts)
              CALL append_rnode(stmt_rnode, sub_rnode)
            END IF
          END ASSOCIATE
          IF (iv /= SIZE(pnode%blocks(i)%values)) THEN
            CALL append_comma(stmt_rnode)
          END IF
        END DO
        CALL append_delimiter(stmt_rnode, ')')
      END IF
      IF (ALLOCATED(pnode%construct_name)) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
      
      CALL render_block(pnode%blocks(i)%children, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndSelect
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'SELECT', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_case_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /critical-construct/.
  
  SUBROUTINE render_critical_construct(pnode, rnode, opts)
    
    USE CriticalConstructs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CriticalConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptCriticalConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istCritical
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'CRITICAL', 0)
    CALL append_stmt(seq, stmt_rnode)
      
    CALL render_block(pnode%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndCritical
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'CRITICAL', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_critical_construct
  
  
  !*****************************************************************************
  !!
  !> Render an /associate-construct/.
  
  SUBROUTINE render_associate_construct(pnode, rnode, opts)
    
    USE AssociateConstructs
    USE AssociateStmts
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssociateConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Association index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptAssociateConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istAssociate
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'ASSOCIATE', 0)
    CALL append_delimiter(stmt_rnode, '(')
    DO i = 1, SIZE(pnode%associations)
      CALL append_name(stmt_rnode, pnode%associations(i)%name, 0)
      CALL append_token(stmt_rnode, ittAssociate, '=>', '', 0)
      
      CALL render_expression(pnode%associations(i)%selector, sub_rnode, opts)
      CALL append_rnode(stmt_rnode, sub_rnode)
      IF (i /= SIZE(pnode%associations)) CALL append_comma(stmt_rnode)
    END DO
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_stmt(seq, stmt_rnode)
      
    CALL render_block(pnode%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndAssociate
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'ASSOCIATE', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_associate_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /block-construct/.
  
  SUBROUTINE render_block_construct(pnode, rnode, opts)
    
    USE BlockConstructs
    USE BlockStmts
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BlockConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptBlockConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istAssociate
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'BLOCK', 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_block(pnode%exec_children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndAssociate
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'BLOCK', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_block_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /forall-construct/.
  
  SUBROUTINE render_forall_construct(pnode, rnode, opts)
    
    USE ForallConstructs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptForallConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istForallConstruct
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'FORALL', 0)
    CALL render_forall_header(pnode%header, sub_rnode, opts)
    CALL append_rnode(stmt_rnode, sub_rnode)
    CALL append_stmt(seq, stmt_rnode)
      
    CALL render_block(pnode%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndForall
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'FORALL', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_forall_construct
  
  
  !*****************************************************************************
  !!
  !> Render an /if-construct/.
  
  SUBROUTINE render_if_construct(pnode, rnode, opts)
    
    USE IfConstructs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IfConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child block index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptIfConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istIfThen
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'IF', 0)
    CALL append_delimiter(stmt_rnode, '(')
    CALL render_expression(pnode%blocks(1)%ex, sub_rnode, opts)
    CALL append_rnode(stmt_rnode, sub_rnode)
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_keyword(stmt_rnode, 'THEN', 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_block(pnode%blocks(1)%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    DO i = 2, SIZE(pnode%blocks)
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istElseIf
      CALL append_keyword(stmt_rnode, 'ELSE', 0)
      CALL append_keyword(stmt_rnode, 'IF', 0)
      CALL append_delimiter(stmt_rnode, '(')
      CALL render_expression(pnode%blocks(i)%ex, sub_rnode, opts)
      CALL append_rnode(stmt_rnode, sub_rnode)
      CALL append_delimiter(stmt_rnode, ')')
      CALL append_keyword(stmt_rnode, 'THEN', 0)
      IF ( opts%always_construct_name  &
          .AND. ALLOCATED(pnode%construct_name) ) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
      
      CALL render_block(pnode%blocks(i)%children, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    IF (ALLOCATED(pnode%else_children)) THEN
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istElse
      CALL append_keyword(stmt_rnode, 'ELSE', 0)
      IF ( opts%always_construct_name  &
          .AND. ALLOCATED(pnode%construct_name) ) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
      
      CALL render_block(pnode%else_children, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndIf
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'IF', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_if_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /do-construct/.
  
  SUBROUTINE render_do_construct(pnode, rnode, opts)
    
    USE DoConstructs
    USE DoStmts
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DoConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child expressions.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptDoConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istNonlabelDo
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'DO', 0)
    IF (ALLOCATED(pnode%loop_control)) THEN
      SELECT TYPE (control => pnode%loop_control)
      TYPE IS (IndexLoopControl)
        CALL append_name(stmt_rnode, control%name, 0)
        CALL append_token(stmt_rnode, ittAssign, '=', '', 0)
        CALL render_expression(control%initial, sub_rnode, opts)
        CALL append_rnode(stmt_rnode, sub_rnode)
        CALL append_comma(stmt_rnode)
        CALL render_expression(control%terminal, sub_rnode, opts)
        CALL append_rnode(stmt_rnode, sub_rnode)
        IF (ALLOCATED(control%incrementation)) THEN
          CALL append_comma(stmt_rnode)
          CALL render_expression(control%incrementation, sub_rnode, opts)
          CALL append_rnode(stmt_rnode, sub_rnode)
        END IF
      TYPE IS (WhileLoopControl)
        CALL append_keyword(stmt_rnode, 'WHILE', 0)
        CALL append_delimiter(stmt_rnode, '(')
        CALL render_expression(control%expr, sub_rnode, opts)
        CALL append_rnode(stmt_rnode, sub_rnode)
        CALL append_delimiter(stmt_rnode, ')')
      TYPE IS (ConcurrentLoopControl)
        CALL append_keyword(stmt_rnode, 'CONCURRENT', 0)
        CALL render_forall_header(control%header, sub_rnode, opts)
        CALL append_rnode(stmt_rnode, sub_rnode)
      END SELECT
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL render_block(pnode%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndDo
    IF (ALLOCATED(pnode%end_label)) THEN
      CALL append_label(stmt_rnode, pnode%end_label)
    END IF
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'DO', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_do_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /where-construct/.
  
  SUBROUTINE render_where_construct(pnode, rnode, opts)
    
    USE Statements
    USE SyntaxParts
    USE WhereConstructs
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(WhereConstruct), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: ib             ! Child block index.
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptWhereConstruct
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istWhereConstruct
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(stmt_rnode, pnode%statement_label)
    END IF
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
      CALL append_token(stmt_rnode, ittColon, ':', '', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'WHERE', 0)
    CALL append_delimiter(stmt_rnode, '(')
    CALL render_expression(pnode%blocks(1)%mask, sub_rnode, opts)
    CALL append_rnode(stmt_rnode, sub_rnode)
    CALL append_delimiter(stmt_rnode, ')')
    CALL append_stmt(seq, stmt_rnode)
      
    CALL render_block(pnode%blocks(1)%children, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    DO ib = 2, SIZE(pnode%blocks)
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istMaskedElsewhere
      CALL append_keyword(stmt_rnode, 'ELSE', 0)
      CALL append_keyword(stmt_rnode, 'WHERE', 0)
      CALL append_delimiter(stmt_rnode, '(')
      CALL render_expression(pnode%blocks(ib)%mask, sub_rnode, opts)
      CALL append_rnode(stmt_rnode, sub_rnode)
      CALL append_delimiter(stmt_rnode, ')')
      IF ( opts%always_construct_name  &
          .AND. ALLOCATED(pnode%construct_name) ) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
    END DO
    
    IF (ALLOCATED(pnode%else_children)) THEN
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istElsewhere
      CALL append_keyword(stmt_rnode, 'ELSE', 0)
      IF ( opts%always_construct_name  &
          .AND. ALLOCATED(pnode%construct_name) ) THEN
        CALL append_name(stmt_rnode, pnode%construct_name, 0)
      END IF
      CALL append_stmt(seq, stmt_rnode)
      
      CALL render_block(pnode%else_children, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndWhere
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'WHERE', 0)
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(stmt_rnode, pnode%construct_name, 0)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_where_construct
  
  
  !*****************************************************************************
  !!
  !> Render a /block/.
  !!
  !! A block is stored as a list of SCParseNodeElements, rather than its 
  !! own parse node - though the BlockStorage type in ExecutionParts takes 
  !! a good step towards making it its own ParseNode.
  
  SUBROUTINE render_block(pnodes, rnode, opts)
    
    USE ParseNodes
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(SCParseNodeElement), INTENT(IN) :: pnodes(:)
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child statement index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptBlock
    
    DO i = 1, SIZE(pnodes)
      CALL RenderParseTree(pnodes(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_block
  
  
  !*****************************************************************************
  !!
  !> Render a /derived-type-def/.
  
  SUBROUTINE render_derived_type_def(pnode, rnode, opts)
    
    USE DerivedTypeDefs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeDef), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Child statement index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptDerivedTypeDef
    
    CALL render_derived_type_stmt(pnode%first, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    DO i = 1, SIZE(pnode%params)
      CALL render_type_param_def_stmt(pnode%params(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    IF (pnode%sequence) THEN
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istSequence
      CALL append_keyword(stmt_rnode, 'SEQUENCE', 0)
      CALL append_stmt(seq, stmt_rnode)
    END IF
    
    IF (pnode%private_components) THEN
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istPrivateComponents
      CALL append_keyword(stmt_rnode, 'PRIVATE', 0)
      CALL append_stmt(seq, stmt_rnode)
    END IF
    
    DO i = 1, SIZE(pnode%comps)
      SELECT TYPE (item => pnode%comps(i)%item)
      TYPE IS (ProcComponentDefStmt)
        CALL render_proc_component_def_stmt(item, sub_rnode, opts)
      TYPE IS (DataComponentDefStmt)
        CALL render_data_component_def_stmt(item, sub_rnode, opts)
      CLASS DEFAULT
        STOP 'Internal error in RenderParseNodes::render_derived_type_def, &
            &unknown component stmt.'
      END SELECT
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    IF ( (SIZE(pnode%tbps) /= 0)  &
        .OR. (SIZE(pnode%gens) /= 0)  &
        .OR. (SIZE(pnode%finals) /= 0) ) THEN
      ALLOCATE(stmt_rnode)
      stmt_rnode%context = istContains
      CALL append_keyword(stmt_rnode, 'CONTAINS', 0)
      CALL append_stmt(seq, stmt_rnode)
      
      DO i = 1, SIZE(pnode%tbps)
        CALL render_type_bound_procedure_stmt(pnode%tbps(i), sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END DO
      
      DO i = 1, SIZE(pnode%gens)
        CALL render_type_bound_generic_stmt(pnode%gens(i), sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END DO
      
      DO i = 1, SIZE(pnode%finals)
        CALL render_final_procedure_stmt(pnode%finals(i), sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END DO
    END IF
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndType
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'TYPE', 0)
    CALL append_name(stmt_rnode, pnode%first%name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_derived_type_def
  
  
  !*****************************************************************************
  !!
  !> Render an /enum-def/.
  
  SUBROUTINE render_enum_def(pnode, rnode, opts)
    
    USE EnumDefs
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumDef), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: ist            ! Child /enumerator-def-stmt/ index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building individual statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptEnumDef
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEnumDef
    CALL append_keyword(stmt_rnode, 'ENUM', 0)
    CALL append_comma(stmt_rnode)
    CALL render_language_binding_spec(pnode%ls, sub_rnode, opts)
    CALL append_stmt(seq, stmt_rnode)
    
    DO ist = 1, SIZE(pnode%children)
      CALL render_enumerator_def_stmt(pnode%children(ist), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndType
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'ENUM', 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_enum_def
  
  
  !*****************************************************************************
  !!
  !> Render an /enumerator-def-stmt/.
  
  SUBROUTINE render_enumerator_def_stmt(pnode, rnode, opts)
    
    USE EnumDefStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EnumeratorDefStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: ienum          ! Enumerator index.
    
    ! Render nodes from child expressions.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istEnumeratorDef
    
    CALL append_keyword(seq, 'ENUMERATOR', 0)
    IF (opts%always_double_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    DO ienum = 1, SIZE(pnode%enumerator_list)
      CALL append_name(seq, pnode%enumerator_list(ienum)%name, 0)
      IF (ALLOCATED(pnode%enumerator_list(ienum)%expr)) THEN
        CALL append_token(seq, ittAssign, '=', '', 0)
        CALL render_expression(  &
            pnode%enumerator_list(ienum)%expr,  &
            sub_rnode,  &
            opts )
      END IF
      IF (ienum /= SIZE(pnode%enumerator_list)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_enumerator_def_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /interface-block/.
  
  SUBROUTINE render_interface_block(pnode, rnode, opts)
    
    USE InterfaceBlocks
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InterfaceBlock), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    INTEGER :: i              ! interface specification index.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the opening and closing statements.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptInterfaceBlock
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istInterface
    
    IF (pnode%abstract) THEN
      CALL append_keyword(stmt_rnode, 'ABSTRACT', 0)
    END IF
    CALL append_keyword(stmt_rnode, 'INTERFACE', 0)
    IF (ALLOCATED(pnode%generic_spec)) THEN
      CALL render_generic_spec(pnode%generic_spec, sub_rnode, opts)
      CALL append_rnode(stmt_rnode, sub_rnode)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    DO i = 1, SIZE(pnode%specs)
      SELECT TYPE (item => pnode%specs(i)%item)
      TYPE IS (ProcedureStmt)
        CALL render_procedure_stmt(item, sub_rnode, opts)
      TYPE IS (FunctionInterfaceBody)
        CALL render_function_interface_body(item, sub_rnode, opts)
      TYPE IS (SubroutineInterfaceBody)
        CALL render_subroutine_interface_body(item, sub_rnode, opts)
      CLASS DEFAULT
        STOP 'Internal error in RenderParseNodes::render_interface_block, &
            &unknown spec.'
      END SELECT
      CALL append_rnode(seq, sub_rnode)
    END DO
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndInterface
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'INTERFACE', 0)
    IF (ALLOCATED(pnode%generic_spec)) THEN
      CALL render_generic_spec(pnode%generic_spec, sub_rnode, opts)
      CALL append_rnode(stmt_rnode, sub_rnode)
    END IF
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_interface_block
  
  
  !*****************************************************************************
  !!
  !> Render an /interface-body/ for a function.
  
  SUBROUTINE render_function_interface_body(pnode, rnode, opts)
    
    USE InterfaceBlocks
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionInterfaceBody), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptInterfaceBody
    
    CALL render_function_stmt(pnode%fn_stmt, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndFunction
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'FUNCTION', 0)
    CALL append_name(stmt_rnode, pnode%fn_stmt%function_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_function_interface_body
  
  
  !*****************************************************************************
  !!
  !> Render an /interface-body/ for a subroutine.
  
  SUBROUTINE render_subroutine_interface_body(pnode, rnode, opts)
    
    USE InterfaceBlocks
    USE Statements
    USE SyntaxParts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineInterfaceBody), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeConstruct), ALLOCATABLE :: seq
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    ! Temporary for building the closing statement.
    TYPE(RenderNodeStatement), ALLOCATABLE :: stmt_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = iptInterfaceBody
    
    CALL render_subroutine_stmt(pnode%sub_stmt, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_specification_part(pnode%spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    ALLOCATE(stmt_rnode)
    stmt_rnode%context = istEndFunction
    CALL append_keyword(stmt_rnode, 'END', 0)
    CALL append_keyword(stmt_rnode, 'SUBROUTINE', 0)
    CALL append_name(stmt_rnode, pnode%sub_stmt%subroutine_name, 0)
    CALL append_stmt(seq, stmt_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_subroutine_interface_body
  
  
  !*****************************************************************************
  !!
  !> Render a /function-stmt/.
  
  SUBROUTINE render_function_stmt(pnode, rnode, opts)
    
    USE FunctionStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! argument index.
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istFunction
    
    IF (ALLOCATED(pnode%ds)) THEN
      CALL render_declaration_type_spec(pnode%ds, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL render_other_prefixes(pnode%prefixes, seq, opts)
    
    CALL append_keyword(seq, 'FUNCTION', 0)
    
    CALL append_name(seq, pnode%function_name, 0)
    
    CALL append_delimiter(seq, '(')
    DO i = 1, SIZE(pnode%args)
      CALL append_name(seq, pnode%args(i), 0)
      IF (i /= SIZE(pnode%args)) CALL append_comma(seq)
    END DO
    CALL append_delimiter(seq, ')')
    
    IF (ALLOCATED(pnode%result_name)) THEN
      CALL append_keyword(seq, 'RESULT', 0)
      CALL append_delimiter(seq, '(')
      CALL append_name(seq, pnode%result_name, 0)
      CALL append_delimiter(seq, ')')
    END IF
    
    IF (ALLOCATED(pnode%ls)) THEN
      CALL render_language_binding_spec(pnode%ls, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_function_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /subroutine-stmt/.
  
  SUBROUTINE render_subroutine_stmt(pnode, rnode, opts)
    
    USE SubroutineStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SubroutineStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! argument index.
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istSubroutine
    
    CALL render_other_prefixes(pnode%prefixes, seq, opts)
    
    CALL append_keyword(seq, 'SUBROUTINE', 0)
    
    CALL append_name(seq, pnode%subroutine_name, 0)
    
    CALL append_delimiter(seq, '(')
    DO i = 1, SIZE(pnode%args)
      IF (pnode%args(i)%is_star) THEN
        CALL append_token(seq, ittOperator, '*', '', 0)
      ELSE
        CALL append_name(seq, pnode%args(i)%name, 0)
      END IF
      IF (i /= SIZE(pnode%args)) CALL append_comma(seq)
    END DO
    CALL append_delimiter(seq, ')')
    
    IF (ALLOCATED(pnode%ls)) THEN
      CALL render_language_binding_spec(pnode%ls, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_subroutine_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /program-stmt/.
  
  SUBROUTINE render_program_stmt(pnode, rnode, opts)
    
    USE ProgramStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProgramStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProgram
    
    CALL append_keyword(seq, 'PROGRAM', 0)
    
    CALL append_name(seq, pnode%name, 0)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_program_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /procedure-stmt/.
  
  SUBROUTINE render_procedure_stmt(pnode, rnode, opts)
    
    USE InterfaceBlocks
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcedureStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! procedure index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProcedure
    
    IF (pnode%has_module) THEN
      CALL append_keyword(seq, 'MODULE', 0)
    END IF
    CALL append_keyword(seq, 'PROCEDURE', 0)
    
    DO i = 1, SIZE(pnode%procedure_names)
      CALL append_name(seq, pnode%procedure_names(i), 0)
      IF (i /= SIZE(pnode%procedure_names)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_procedure_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /final-procedure-stmt/.
  
  SUBROUTINE render_final_procedure_stmt(pnode, rnode, opts)
    
    USE FinalProcedureStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FinalProcedureStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! procedure index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istFinalProcedure
    
    CALL append_keyword(seq, 'FINAL', 0)
    
    DO i = 1, SIZE(pnode%finals)
      CALL append_name(seq, pnode%finals(i), 0)
      IF (i /= SIZE(pnode%finals)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_final_procedure_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /type-bound-generic-stmt/.
  
  SUBROUTINE render_type_bound_generic_stmt(pnode, rnode, opts)
    
    USE TypeBoundGenericStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundGenericStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! declarator index.
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istTypeBoundGeneric
    
    CALL append_keyword(seq, 'GENERIC', 0)
    
    IF (pnode%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%access, seq, opts)
    END IF
    
    CALL append_token(seq, ittDeclare, '::', '', 0)
    
    CALL render_generic_spec(pnode%gs, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL append_token(seq, ittAssociate, '=>', '', 0)
    
    DO i = 1, SIZE(pnode%binding_list)
      CALL append_name(seq, pnode%binding_list(i), 0)
      IF (i /= SIZE(pnode%binding_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_type_bound_generic_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /assignment-stmt/.
  
  SUBROUTINE render_assignment_stmt(pnode, rnode, opts)
    
    USE AssignmentStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AssignmentStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istAssignment
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL render_variable(pnode%var, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL append_token(seq, ittAssign, '=', '', 0)
    
    CALL render_expression(pnode%expr, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_assignment_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /stmt-function-stmt/.
  
  SUBROUTINE render_stmt_function_stmt(pnode, rnode, opts)
    
    USE StmtFunctionStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StmtFunctionStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Dummy argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istStmtFunction
    
    CALL append_name(seq, pnode%function_name, 0)
    
    CALL append_delimiter(seq, '(')
    DO i = 1, SIZE(pnode%args)
      CALL append_name(seq, pnode%args(i), 0)
      IF (i /= SIZE(pnode%args)) CALL append_comma(seq)
    END DO
    CALL append_delimiter(seq, ')')
    
    CALL append_token(seq, ittAssign, '=', '', 0)
    
    CALL render_expression(pnode%expr, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_stmt_function_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /pointer-assignment-stmt/.
  
  SUBROUTINE render_pointer_assignment_stmt(pnode, rnode, opts)
    
    USE PointerAssignmentStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerAssignmentStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istPointerAssignment
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    DO i = 1, SIZE(pnode%pointer)
      CALL render_part_ref(pnode%pointer(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%pointer)) THEN
        CALL append_token(seq, ittPercent, '%', '', 0)
      END IF
    END DO
    
    CALL append_token(seq, ittAssign, '=>', '', 0)
    
    CALL render_expression(pnode%target, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_pointer_assignment_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /type-bound-procedure-stmt/.
  
  SUBROUTINE render_type_bound_procedure_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE TypeBoundProcedureStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeBoundProcedureStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! declarator index.
    LOGICAL :: need_colon     ! Indicates if `::` is required.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istTypeBoundProcedure
    
    CALL append_keyword(seq, 'PROCEDURE', 0)
    
    IF (ALLOCATED(pnode%interface_name)) THEN
      CALL append_delimiter(seq, '(')
      CALL append_name(seq, pnode%interface_name, 0)
      CALL append_delimiter(seq, ')')
    END IF
    
    need_colon = .FALSE.
    
    IF (pnode%attrs%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%attrs%access, seq, opts)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%pass) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'PASS', 0)
      IF (ALLOCATED(pnode%attrs%pass_arg_name)) THEN
        CALL append_delimiter(seq, '(')
        CALL append_name(seq, pnode%attrs%pass_arg_name, 0)
        CALL append_delimiter(seq, ')')
      END IF
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%nopass) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'NOPASS', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%non_overridable) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'NON_OVERRIDABLE', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%deferred) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'DEFERRED', 0)
      need_colon = .TRUE.
    END IF
    
    IF (.NOT. need_colon) THEN
      DO i = 1, SIZE(pnode%decls)
        IF (ALLOCATED(pnode%decls(i)%specific)) THEN
          need_colon = .TRUE.
        END IF
      END DO
    END IF
    
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%decls)
      CALL append_name(seq, pnode%decls(i)%binding, 0)
      IF (ALLOCATED(pnode%decls(i)%specific)) THEN
        CALL append_token(seq, ittAssociate, '=>', '', 0)
        CALL append_name(seq, pnode%decls(i)%specific, 0)
      END IF
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_type_bound_procedure_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /type-param-def-stmt/.
  
  SUBROUTINE render_type_param_def_stmt(pnode, rnode, opts)
    
    USE TypeParamDefStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(TypeParamDefStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! declarator index.
    LOGICAL :: need_colon     ! Indicates if `::` is required.
    
    ! Render nodes from child syntax terms.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istTypeParamDef
    
    CALL append_keyword(seq, 'INTEGER', 0)
    
    IF (ALLOCATED(pnode%kind_expr)) THEN
      CALL append_delimiter(seq, '(')
      CALL render_expression(pnode%kind_expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_delimiter(seq, ')')
    END IF
    
    need_colon = .FALSE.
    
    ! pnode%attr will always be non-zero.
    IF (pnode%attr /= 0) THEN
      CALL append_comma(seq)
      SELECT CASE (pnode%attr)
      CASE (itattrKind) ; CALL append_keyword(seq, 'KIND', 0)
      CASE (itattrLen)  ; CALL append_keyword(seq, 'LEN', 0)
      END SELECT
      need_colon = .TRUE.
    END IF
    
    IF (.NOT. need_colon) THEN
      DO i = 1, SIZE(pnode%params)
        IF (ALLOCATED(pnode%params(i)%default)) THEN
          need_colon = .TRUE.
          EXIT
        END IF
      END DO
    END IF
    
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%params)
      CALL append_name(seq, pnode%params(i)%name, 0)
      IF (ALLOCATED(pnode%params(i)%default)) THEN
        CALL append_token(seq, ittAssign, '=', '', 0)
        CALL render_expression(pnode%params(i)%default, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END IF
      IF (i /= SIZE(pnode%params)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_type_param_def_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /procedure-component-def-stmt/.
  
  SUBROUTINE render_proc_component_def_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE ComponentDefStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcComponentDefStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! procedure declarator index.
    LOGICAL :: need_colon     ! Indicates if `::` is required.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProcComponentDef
    
    need_colon = .FALSE.
    
    CALL append_keyword(seq, 'PROCEDURE', 0)
    CALL append_delimiter(seq, '(')
    IF (ALLOCATED(pnode%attrs%proc_intf)) THEN
      CALL render_proc_interface(pnode%attrs%proc_intf, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    CALL append_delimiter(seq, ')')
    
    IF (pnode%attrs%pointer) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'POINTER', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%pass) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'PASS', 0)
      IF (ALLOCATED(pnode%attrs%pass_arg_name)) THEN
        CALL append_delimiter(seq, '(')
        CALL append_name(seq, pnode%attrs%pass_arg_name, 0)
        CALL append_delimiter(seq, ')')
      END IF
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%nopass) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'NOPASS', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%attrs%access, seq, opts)
      need_colon = .TRUE.
    END IF
    
    ! Double colon is also required if there are any initializers
    !
    ! (We are just going through the motions here, the double colon is 
    ! always required.)
    IF (.NOT. need_colon) THEN
      DO i = 1, SIZE(pnode%decls)
        IF ( ALLOCATED(pnode%decls(i)%null_init)  &
            .OR. ALLOCATED(pnode%decls(i)%init_target) ) THEN
          need_colon = .TRUE.
          EXIT
        END IF
      END DO
    END IF
    
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%decls)
      CALL render_proc_decl(pnode%decls(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_proc_component_def_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /procedure-component-def-stmt/.
  
  SUBROUTINE render_data_component_def_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE ComponentDefStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataComponentDefStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Type parameter index.
    LOGICAL :: need_colon     ! Indicates if `::` is required.
    
    ! Render nodes from child program units.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istDataComponentDef
    
    CALL render_declaration_type_spec(pnode%decl_spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    need_colon = .FALSE.
    
    IF (pnode%attrs%allocatable) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ALLOCATABLE', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%contiguous) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'CONTIGUOUS', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%pointer) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'POINTER', 0)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%attrs%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%attrs%access, seq, opts)
      need_colon = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%attrs%dim_specs)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'DIMENSION', 0)
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%attrs%dim_specs, seq, opts)
      CALL append_delimiter(seq, ')')
      need_colon = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%attrs%codim_specs)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'CODIMENSION', 0)
      CALL append_delimiter(seq, '[')
      CALL render_dim_specs(pnode%attrs%codim_specs, seq, opts)
      CALL append_delimiter(seq, ']')
      need_colon = .TRUE.
    END IF
    
    IF (.NOT. need_colon) THEN
      DO i = 1, SIZE(pnode%names)
        IF ( ALLOCATED(pnode%names(i)%null_init)  &
            .OR. ALLOCATED(pnode%names(i)%expr)  &
            .OR. ALLOCATED(pnode%names(i)%initial_data_target) ) THEN
          need_colon = .TRUE.
          EXIT
        END IF
      END DO
    END IF
    
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%names)
      CALL render_entity_decl(pnode%names(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%names)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_data_component_def_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /derived-type-stmt/.
  
  SUBROUTINE render_derived_type_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE DerivedTypeDefs
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DerivedTypeStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Type parameter index.
    LOGICAL :: need_colon     ! Indicates if `::` is required.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istDerivedType
    
    need_colon = .FALSE.
    
    CALL append_keyword(seq, 'TYPE', 0)
    
    IF (pnode%abstract) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ABSTRACT', 0)
      need_colon = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%extends)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'EXTENDS', 0)
      CALL append_delimiter(seq, '(')
      CALL append_name(seq, pnode%extends, 0)
      CALL append_delimiter(seq, ')')
      need_colon = .TRUE.
    END IF
    
    IF (pnode%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%access, seq, opts)
      need_colon = .TRUE.
    END IF
    
    IF (pnode%lang /= 0) THEN
      CALL append_comma(seq)
      ! This is not a /language-binding-spec/!
      CALL append_keyword(seq, 'BIND', 0)
      CALL append_delimiter(seq, '(')
      CALL append_keyword(seq, 'C', 0)
      CALL append_delimiter(seq, ')')
      need_colon = .TRUE.
    END IF
    
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (SIZE(pnode%params) /= 0) THEN
      CALL append_delimiter(seq, '(')
      DO i = 1, SIZE(pnode%params)
        CALL append_name(seq, pnode%params(i), 0)
        IF (i /= SIZE(pnode%params)) CALL append_comma(seq)
      END DO
      CALL append_delimiter(seq, ')')
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_derived_type_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /allocatable-stmt/.
  
  SUBROUTINE render_allocatable_stmt(pnode, rnode, opts)
    
    USE AllocatableStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AllocatableStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! decl index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProcedure
    
    CALL append_keyword(seq, 'ALLOCATABLE', 0)
    
    DO i = 1, SIZE(pnode%decls)
      CALL append_name(seq, pnode%decls(i)%name, 0)
      IF (ALLOCATED(pnode%decls(i)%array_spec)) THEN
        CALL append_delimiter(seq, '(')
        CALL render_dim_specs(pnode%decls(i)%array_spec, seq, opts)
        CALL append_delimiter(seq, ')')
      END IF
      IF (ALLOCATED(pnode%decls(i)%coarray_spec)) THEN
        CALL append_delimiter(seq, '[')
        CALL render_dim_specs(pnode%decls(i)%coarray_spec, seq, opts)
        CALL append_delimiter(seq, ']')
      END IF
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_allocatable_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /asynchronous-stmt/.
  
  SUBROUTINE render_asynchronous_stmt(pnode, rnode, opts)
    
    USE AsynchronousStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(AsynchronousStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! name index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProcedure
    
    CALL append_keyword(seq, 'ASYNCHRONOUS', 0)
    
    DO i = 1, SIZE(pnode%object_name_list)
      CALL append_name(seq, pnode%object_name_list(i), 0)
      IF (i /= SIZE(pnode%object_name_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_asynchronous_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /bind-stmt/.
  
  SUBROUTINE render_bind_stmt(pnode, rnode, opts)
    
    USE BindStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(BindStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! entity index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istBind
    
    CALL render_language_binding_spec(pnode%ls, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    DO i = 1, SIZE(pnode%entities)
      IF (pnode%entities(i)%common_block) THEN
        CALL append_delimiter(seq, '/')
        CALL append_name(seq, pnode%entities(i)%name, 0)
        CALL append_delimiter(seq, '/')
      ELSE
        CALL append_name(seq, pnode%entities(i)%name, 0)
      END IF
      IF (i /= SIZE(pnode%entities)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_bind_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /codimension-stmt/.
  
  SUBROUTINE render_codimension_stmt(pnode, rnode, opts)
    
    USE CodimensionStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CodimensionStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! decl index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istCodimension
    
    CALL append_keyword(seq, 'CODIMENSION', 0)
    
    DO i = 1, SIZE(pnode%decls)
      CALL append_name(seq, pnode%decls(i)%name, 0)
      CALL append_delimiter(seq, '[')
      CALL render_dim_specs(pnode%decls(i)%array_spec, seq, opts)
      CALL append_delimiter(seq, ']')
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_codimension_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /contiguous-stmt/.
  
  SUBROUTINE render_contiguous_stmt(pnode, rnode, opts)
    
    USE ContiguousStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ContiguousStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! name index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istContiguous
    
    CALL append_keyword(seq, 'CONTIGUOUS', 0)
    
    IF (opts%always_double_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%name_list)
      CALL append_name(seq, pnode%name_list(i), 0)
      IF (i /= SIZE(pnode%name_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_contiguous_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /common-stmt/.
  
  SUBROUTINE render_common_stmt(pnode, rnode, opts)
    
    USE CommonStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: is             ! set index.
    INTEGER :: i              ! Object/value index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istCommon
    
    CALL append_keyword(seq, 'COMMON', 0)
    
    DO is = 1, SIZE(pnode%sets)
      IF (is == 1) THEN
        ! Special processing of the first set.
        IF (ALLOCATED(pnode%sets(is)%name)) THEN
          CALL append_delimiter(seq, '/')
          CALL append_name(seq, pnode%sets(is)%name, 0)
          CALL append_delimiter(seq, '/')
        END IF
      ELSE
        CALL append_delimiter(seq, '/')
        IF (ALLOCATED(pnode%sets(is)%name)) THEN
          CALL append_name(seq, pnode%sets(is)%name, 0)
        END IF
        CALL append_delimiter(seq, '/')
      END IF
      
      DO i = 1, SIZE(pnode%sets(is)%objects)
        CALL render_common_block_object(  &
            pnode%sets(is)%objects(i),  &
            sub_rnode,  &
            opts )
        CALL append_rnode(seq, sub_rnode)
        ! This comma is also permitted at the end of the object list if 
        ! there are following sets.
        IF (i /= SIZE(pnode%sets(is)%objects)) CALL append_comma(seq)
      END DO
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_common_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /data-stmt/.
  
  SUBROUTINE render_data_stmt(pnode, rnode, opts)
    
    USE DataStmts
    USE Statements
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: is             ! set index.
    INTEGER :: i              ! Object/value index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istData
    
    CALL append_keyword(seq, 'DATA', 0)
    
    DO is = 1, SIZE(pnode%data_set_list)
      DO i = 1, SIZE(pnode%data_set_list(is)%objects)
        SELECT TYPE (ds_object => pnode%data_set_list(is)%objects(i)%item)
        TYPE IS (DataStmtObjectVariable)
          CALL render_designator(ds_object%desig, sub_rnode, opts)
          CALL append_rnode(seq, sub_rnode)
        TYPE IS (DataStmtObjectDataImpliedDo)
          CALL render_data_implied_do(ds_object%obj, sub_rnode, opts)
          CALL append_rnode(seq, sub_rnode)
        CLASS DEFAULT
          STOP 'Internal error in RenderParseNodes::render_data_stmt, &
              &unknown object type.'
        END SELECT
        IF (i /= SIZE(pnode%data_set_list(is)%objects)) CALL append_comma(seq)
      END DO
      
      CALL append_delimiter(seq, '/')
      
      DO i = 1, SIZE(pnode%data_set_list(is)%values)
        ASSOCIATE(ds_value => pnode%data_set_list(is)%values(i))
          IF (ALLOCATED(ds_value%repeat)) THEN
            SELECT TYPE (repeat => ds_value%repeat)
            TYPE IS (DataStmtRepeatDesig)
              CALL render_designator(repeat%desig, sub_rnode, opts)
              CALL append_rnode(seq, sub_rnode)
            TYPE IS (DataStmtRepeatLiteral)
              CALL append_token(  &
                  seq,  &
                  QueryType(repeat%tok),  &
                  QueryValue(repeat%tok),  &
                  QueryKind(repeat%tok),  &
                  0 )
            CLASS DEFAULT
              STOP 'Internal error in RenderParseNodes::render_data_stmt, &
                  &unknown repeat type.'
            END SELECT
            CALL append_token(seq, ittOperator, '*', '', 0)
          END IF
          SELECT TYPE (constant => ds_value%constant)
          TYPE IS (DataStmtConstantDesig)
            CALL render_designator(constant%desig, sub_rnode, opts)
            CALL append_rnode(seq, sub_rnode)
          TYPE IS (DataStmtConstantSC)
            CALL render_structure_constructor(constant%sc, sub_rnode, opts)
            CALL append_rnode(seq, sub_rnode)
          TYPE IS (DataStmtConstantSimpleLiteral)
            IF (ALLOCATED(constant%sign_tok)) THEN
              CALL append_token(  &
                  seq,  &
                  QueryType(constant%sign_tok),  &
                  QueryValue(constant%sign_tok),  &
                  QueryKind(constant%sign_tok),  &
                  0 )
            END IF
            CALL append_token(  &
                seq,  &
                QueryType(constant%tok),  &
                QueryValue(constant%tok),  &
                QueryKind(constant%tok),  &
                0 )
          TYPE IS (DataStmtConstantComplexLiteral)
            CALL render_complex_literal_constant(  &
                constant%constant,  &
                sub_rnode,  &
                opts)
            CALL append_rnode(seq, sub_rnode)
          CLASS DEFAULT
            STOP 'Internal error in RenderParseNodes::render_data_stmt, &
                &unknown constant type.'
          END SELECT
        END ASSOCIATE
        IF (i /= SIZE(pnode%data_set_list(is)%values)) CALL append_comma(seq)
      END DO
      CALL append_delimiter(seq, '/')
      ! Note this comma is optional.
      IF (is /= SIZE(pnode%data_set_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_data_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /dimension-stmt/.
  
  SUBROUTINE render_dimension_stmt(pnode, rnode, opts)
    
    USE DimensionStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DimensionStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! procedure index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istCodimension
    
    CALL append_keyword(seq, 'DIMENSION', 0)
    
    DO i = 1, SIZE(pnode%decls)
      CALL append_name(seq, pnode%decls(i)%name, 0)
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%decls(i)%array_spec, seq, opts)
      CALL append_delimiter(seq, ')')
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_dimension_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /equivalence-stmt/.
  
  SUBROUTINE render_equivalence_stmt(pnode, rnode, opts)
    
    USE EquivalenceStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EquivalenceStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: is             ! set index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istEquivalence
    
    CALL append_keyword(seq, 'EQUIVALENCE', 0)
    
    DO is = 1, SIZE(pnode%sets)
      CALL render_equivalence_set(pnode%sets(is), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (is /= SIZE(pnode%sets)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_equivalence_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /pointer-stmt/.
  
  SUBROUTINE render_pointer_stmt(pnode, rnode, opts)
    
    USE PointerStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PointerStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! entity index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istPointer
    
    CALL append_keyword(seq, 'POINTER', 0)
    
    IF (opts%always_double_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%decls)
      CALL append_name(seq, pnode%decls(i)%name, 0)
      IF (ALLOCATED(pnode%decls(i)%array_spec)) THEN
        CALL append_delimiter(seq, '(')
        CALL render_dim_specs(pnode%decls(i)%array_spec, seq, opts)
        CALL append_delimiter(seq, ')')
      END IF
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_pointer_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /protected-stmt/.
  
  SUBROUTINE render_protected_stmt(pnode, rnode, opts)
    
    USE ProtectedStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProtectedStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! name list index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istProtected
    
    CALL append_keyword(seq, 'PROTECTED', 0)
    
    IF (opts%always_double_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%name_list)
      CALL append_name(seq, pnode%name_list(i), 0)
      IF (i /= SIZE(pnode%name_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_protected_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /save-stmt/.
  
  SUBROUTINE render_save_stmt(pnode, rnode, opts)
    
    USE SaveStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(SaveStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! entity index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istSave
    
    CALL append_keyword(seq, 'SAVE', 0)
    
    IF (opts%always_double_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%decls)
      IF (pnode%decls(i)%common) THEN
        CALL append_delimiter(seq, '/')
        CALL append_name(seq, pnode%decls(i)%name, 0)
        CALL append_delimiter(seq, '/')
      ELSE
        CALL append_name(seq, pnode%decls(i)%name, 0)
      END IF
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_save_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a use statement.
  
  SUBROUTINE render_use_stmt(pnode, rnode, opts)
    
    USE Statements
    USE UseStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(UseStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! Rename/only list index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istUse
    
    CALL append_keyword(seq, 'USE', 0)
    IF (pnode%mod_nature /= 0) THEN
      CALL append_comma(seq)
      SELECT CASE (pnode%mod_nature)
      CASE (1) ; CALL append_keyword(seq, 'INTRINSIC', 0)
      CASE (2) ; CALL append_keyword(seq, 'NON_INTRINSIC', 0)
      CASE DEFAULT ; STOP 'Unknown module nature'
      END SELECT
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (pnode%only_flag) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ONLY', 0)
      CALL append_token(seq, ittColon, ':', '', 0)
    ELSE IF (SIZE(pnode%items) /= 0) THEN
      CALL append_comma(seq)
    END IF
    
    DO i = 1, SIZE(pnode%items)
      CALL render_rename(pnode%items(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%items)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_use_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an access statement.
  
  SUBROUTINE render_access_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(AccessStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! generic spec index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istAccess
    
    CALL render_access_spec(pnode%access, seq, opts)
    
    IF (SIZE(pnode%access_id_list) /= 0) THEN
      DO i = 1, SIZE(pnode%access_id_list)
        CALL render_generic_spec(pnode%access_id_list(i), sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
        IF (i /= SIZE(pnode%access_id_list)) THEN
          CALL append_comma(seq)
        END IF
      END DO
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_access_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an import statement.
  
  SUBROUTINE render_type_declaration_stmt(pnode, rnode, opts)
    
    USE Statements
    USE TypeDeclarationStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeDeclarationStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! entity index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istTypeDeclaration
    
    CALL render_declaration_type_spec(pnode%decl_spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL render_attr_specs(pnode%attrs, seq, opts)
    
    ! @todo make this conditional.
    CALL append_token(seq, ittDeclare, '::', '', 0)
    
    DO i = 1, SIZE(pnode%names)
      CALL render_entity_decl(pnode%names(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%names)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_type_declaration_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an import statement.
  
  SUBROUTINE render_import_stmt(pnode, rnode, opts)
    
    USE Statements
    USE ImportStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ImportStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! name index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istImport
    
    CALL append_keyword(seq, 'IMPORT', 0)
    
    DO i = 1, SIZE(pnode%import_names)
      CALL append_name(seq, pnode%import_names(i), 0)
      IF (i /= SIZE(pnode%import_names)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_import_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /implicit-stmt/.
  
  SUBROUTINE render_implicit_stmt(pnode, rnode, opts)
    
    USE ImplicitStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ImplicitStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! spec index.
    INTEGER :: il             ! letter spec index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istImplicit
    
    CALL append_keyword(seq, 'IMPLICIT', 0)
    
    IF (SIZE(pnode%specs) == 0) THEN
      CALL append_keyword(seq, 'NONE', 0)
      CALL MOVE_ALLOC(seq, rnode)
      RETURN
    END IF
    
    DO i = 1, SIZE(pnode%specs)
      CALL render_declaration_type_spec(pnode%specs(i)%dts, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
      CALL append_delimiter(seq, '(')
      DO il = 1, SIZE(pnode%specs(i)%letters)
        ! Is a letter a name or a keyword?
        CALL append_keyword(seq, pnode%specs(i)%letters(il)%start, 0)
        IF (pnode%specs(i)%letters(il)%finish /= '') THEN
          CALL append_token(seq, ittOperator, '-', '', 0)
          CALL append_keyword(  &
              seq,  &
              pnode%specs(i)%letters(il)%finish,  &
              0 )
        END IF
        IF (il /= SIZE(pnode%specs(i)%letters)) THEN
          CALL append_comma(seq)
        END IF
      END DO
      CALL append_delimiter(seq, ')')
      
      IF (i /= SIZE(pnode%specs)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_implicit_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /procedure-declaration-stmt/.
  
  SUBROUTINE render_procedure_declaration_stmt(pnode, rnode, opts)
    
    USE AccessStmts
    USE IntentSpecs
    USE ProcedureDeclarationStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProcedureDeclarationStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_colon     ! Double colon required.
    INTEGER :: i              ! actual argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istExit
    
    CALL append_keyword(seq, 'PROCEDURE', 0)
    
    CALL append_delimiter(seq, '(')
    IF (ALLOCATED(pnode%attrs%proc_intf)) THEN
      IF (ALLOCATED(pnode%attrs%proc_intf%decl)) THEN
        CALL render_declaration_type_spec(  &
            pnode%attrs%proc_intf%decl,  &
            sub_rnode,  &
            opts )
        CALL append_rnode(seq, sub_rnode)
      END IF
      IF (ALLOCATED(pnode%attrs%proc_intf%name)) THEN
        CALL append_name(seq, pnode%attrs%proc_intf%name, 0)
      END IF
    END IF
    CALL append_delimiter(seq, ')')
    
    CALL render_proc_attr_spec_list(pnode%attrs, seq, opts)
    
    need_colon = (pnode%attrs%access /= iaccNone)  &
        .OR. ALLOCATED(pnode%attrs%ls)  &
        .OR. (pnode%attrs%intent /= iintNone)  &
        .OR. pnode%attrs%optional  &
        .OR. pnode%attrs%pointer  &
        .OR. pnode%attrs%save  &
        .OR. pnode%attrs%protected
    IF (need_colon) THEN
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%decls)
      CALL render_proc_decl(pnode%decls(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%decls)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_procedure_declaration_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /format-stmt/.
  
  SUBROUTINE render_format_stmt(pnode, rnode, opts)
    
    USE FormatStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(FormatStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istFormat
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'FORMAT', 0)
    CALL append_token(seq, ittFormatSpecification, pnode%format_spec, '', 0)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_format_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /namelist-stmt/.
  
  SUBROUTINE render_namelist_stmt(pnode, rnode, opts)
    
    USE NamelistStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(NamelistStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: igroup         ! Namelist group index.
    INTEGER :: iname          ! Namelist name index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istNamelist
    
    CALL append_keyword(seq, 'NAMELIST', 0)
    
    DO igroup = 1, SIZE(pnode%groups)
      CALL append_delimiter(seq, '/')
      CALL append_name(seq, pnode%groups(igroup)%name, 0)
      CALL append_delimiter(seq, '/')
      DO iname = 1, SIZE(pnode%groups(igroup)%objects)
        CALL render_namelist_group_object(  &
            pnode%groups(igroup)%objects(iname),  &
            sub_rnode,  &
            opts )
        CALL append_rnode(seq, sub_rnode)
        IF (iname /= SIZE(pnode%groups(igroup)%objects)) THEN
          CALL append_comma(seq)
        END IF
      END DO
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_namelist_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /call-stmt/.
  
  SUBROUTINE render_call_stmt(pnode, rnode, opts)
    
    USE CallStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(CallStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! actual argument index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istCall
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'CALL', 0)
    
    DO i = 1, SIZE(pnode%parts)
      CALL render_part_ref(pnode%parts(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(seq, ittPercent, '%', '', 0)
    END DO
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (opts%always_call_parens .OR. (SIZE(pnode%args) /= 0)) THEN
      CALL append_delimiter(seq, '(')
    END IF
    
    DO i = 1, SIZE(pnode%args)
      CALL render_actual_arg_spec(pnode%args(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%args)) CALL append_comma(seq)
    END DO
    
    IF (opts%always_call_parens .OR. (SIZE(pnode%args) /= 0)) THEN
      CALL append_delimiter(seq, ')')
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_call_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /print-stmt/.
  
  SUBROUTINE render_print_stmt(pnode, rnode, opts)
    
    USE PrintStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(PrintStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: i              ! actual argument index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istPrint
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'PRINT', 0)
    
    CALL render_format(pnode%fmt, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    IF (SIZE(pnode%items) /= 0) THEN
      CALL append_comma(seq)
      DO i = 1, SIZE(pnode%items)
        CALL render_output_item(pnode%items(i)%item, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
        IF (i /= SIZE(pnode%items)) CALL append_comma(seq)
      END DO
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_print_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /open-stmt/.
  
  SUBROUTINE render_open_stmt(pnode, rnode, opts)
    
    USE OpenStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(OpenStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istWrite
    
    CALL append_keyword(seq, 'OPEN', 0)
    
    CALL append_delimiter(seq, '(')
    CALL render_connect_spec_list(  &
        pnode,  &
        seq,  &
        opts )
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_open_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /close-stmt/.
  
  SUBROUTINE render_close_stmt(pnode, rnode, opts)
    
    USE CloseStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(CloseStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istClose
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'CLOSE', 0)
    
    CALL append_delimiter(seq, '(')
    CALL render_close_spec_list(pnode, seq, opts)
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_close_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /flush-stmt/.
  
  SUBROUTINE render_flush_stmt(pnode, rnode, opts)
    
    USE FlushStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(FlushStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istFlush
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'FLUSH', 0)
    
    ! We always render the statement in the form with the parentheses.
    CALL append_delimiter(seq, '(')
    CALL render_flush_spec_list(pnode, seq, opts)
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_flush_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /rewind-stmt/.
  
  SUBROUTINE render_rewind_stmt(pnode, rnode, opts)
    
    USE RewindStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(RewindStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istRewind
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'REWIND', 0)
    
    ! We always render the statement in the form with the parentheses.
    CALL append_delimiter(seq, '(')
    CALL render_position_spec_list(  &
        pnode%position_specs,  &
        seq,  &
        opts )
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_rewind_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /backspace-stmt/.
  
  SUBROUTINE render_backspace_stmt(pnode, rnode, opts)
    
    USE BackspaceStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(BackspaceStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istBackspace
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'BACKSPACE', 0)
    
    ! We always render the statement in the form with the parentheses.
    CALL append_delimiter(seq, '(')
    CALL render_position_spec_list(  &
        pnode%position_specs,  &
        seq,  &
        opts )
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_backspace_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /endfile-stmt/.
  
  SUBROUTINE render_endfile_stmt(pnode, rnode, opts)
    
    USE EndfileStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(EndfileStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istEndfile
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'ENDFILE', 0)
    
    ! We always render the statement in the form with the parentheses.
    CALL append_delimiter(seq, '(')
    CALL render_position_spec_list(  &
        pnode%position_specs,  &
        seq,  &
        opts )
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_endfile_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /inquire-stmt/.
  
  SUBROUTINE render_inquire_stmt(pnode, rnode, opts)
    
    USE InquireStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(InquireStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! actual argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istInquire
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'INQUIRE', 0)
    
    CALL append_delimiter(seq, '(')
    CALL render_inquire_spec_list(pnode, seq, opts)
    CALL append_delimiter(seq, ')')
    
    DO i = 1, SIZE(pnode%output_list)
      CALL render_output_item(pnode%output_list(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%output_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_inquire_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /write-stmt/.
  
  SUBROUTINE render_write_stmt(pnode, rnode, opts)
    
    USE WriteStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(WriteStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! actual argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istWrite
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'WRITE', 0)
    
    CALL append_delimiter(seq, '(')
    CALL render_io_control_spec_list(  &
        pnode%io_control_spec_list,  &
        seq,  &
        opts )
    CALL append_delimiter(seq, ')')
    
    DO i = 1, SIZE(pnode%item_list)
      CALL render_output_item(pnode%item_list(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%item_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_write_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /read-stmt/.
  
  SUBROUTINE render_read_stmt(pnode, rnode, opts)
    
    USE ReadStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ReadStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! actual argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istRead
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'READ', 0)
    
    IF (ALLOCATED(pnode%io_control_spec_list)) THEN
      CALL append_delimiter(seq, '(')
      CALL render_io_control_spec_list(  &
          pnode%io_control_spec_list,  &
          seq,  &
          opts )
      CALL append_delimiter(seq, ')')
    ELSE
      CALL render_format(pnode%fmt, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (SIZE(pnode%item_list) /= 0) CALL append_comma(seq)
    END IF
    
    DO i = 1, SIZE(pnode%item_list)
      CALL render_input_item(pnode%item_list(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%item_list)) CALL append_comma(seq)
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_read_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /return-stmt/.
  
  SUBROUTINE render_return_stmt(pnode, rnode, opts)
    
    USE ReturnStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ReturnStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istReturn
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'RETURN', 0)
    
    IF (ALLOCATED(pnode%expr)) THEN
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_return_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /goto-stmt/.
  
  SUBROUTINE render_goto_stmt(pnode, rnode, opts)
    
    USE GotoStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(GotoStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istGoto
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'GO', 0)
    CALL append_keyword(seq, 'TO', 0)
    CALL append_label(seq, pnode%goto_label)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_goto_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /computed-goto-stmt/.
  
  SUBROUTINE render_computed_goto_stmt(pnode, rnode, opts)
    
    USE GotoStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ComputedGotoStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    INTEGER :: il             ! Label index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istComputedGoto
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'GO', 0)
    CALL append_keyword(seq, 'TO', 0)
    CALL append_delimiter(seq, '(')
    DO il = 1, SIZE(pnode%labels)
      CALL append_label(seq, pnode%labels(il))
      IF (il /= SIZE(pnode%labels)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    CALL append_delimiter(seq, ')')
    
    CALL render_expression(pnode%expr, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_computed_goto_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /continue-stmt/.
  
  SUBROUTINE render_continue_stmt(pnode, rnode, opts)
    
    USE ContinueStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ContinueStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istReturn
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'CONTINUE', 0)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_continue_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /allocate-stmt/.
  
  SUBROUTINE render_allocate_stmt(pnode, rnode, opts)
    
    USE AllocateStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(AllocateStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istAllocate
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'ALLOCATE', 0)
    
    CALL append_delimiter(seq, '(')
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL render_type_spec(pnode%type_spec, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%allocations)
      SELECT TYPE (object => pnode%allocations(i)%object)
      TYPE IS (AllocateObjectVariable)
        CALL append_name(seq, object%name, 0)
      TYPE IS (AllocateObjectStructureComponent)
        CALL render_structure_component(object%parts, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END SELECT
      IF (ALLOCATED(pnode%allocations(i)%shape_spec)) THEN
        CALL append_delimiter(seq, '(')
        CALL render_dim_specs(pnode%allocations(i)%shape_spec, seq, opts)
        CALL append_delimiter(seq, ')')
      END IF
      IF (ALLOCATED(pnode%allocations(i)%coarray_spec)) THEN
        CALL append_delimiter(seq, '[')
        CALL render_dim_specs(pnode%allocations(i)%coarray_spec, seq, opts)
        CALL append_delimiter(seq, ']')
      END IF
      IF (i /= SIZE(pnode%allocations)) CALL append_comma(seq)
    END DO
    
    IF (ALLOCATED(pnode%source)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'SOURCE', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_expression(pnode%source, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%mold)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'MOLD', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_expression(pnode%mold, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%stat)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'STAT', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_variable(pnode%stat, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%errmsg)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ERRMSG', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_variable(pnode%errmsg, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_allocate_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /arithmetic-if-stmt/.
  
  SUBROUTINE render_arithmetic_if_stmt(pnode, rnode, opts)
    
    USE IfStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ArithmeticIfStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istArithmeticIf
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'IF', 0)
    CALL append_delimiter(seq, '(')
    CALL render_expression(pnode%expr, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    CALL append_delimiter(seq, ')')
    
    CALL append_label(seq, pnode%the_labels(1))
    CALL append_comma(seq)
    CALL append_label(seq, pnode%the_labels(2))
    CALL append_comma(seq)
    CALL append_label(seq, pnode%the_labels(3))
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_arithmetic_if_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /allocate-stmt/.
  
  SUBROUTINE render_deallocate_stmt(pnode, rnode, opts)
    
    USE AllocateStmts
    USE DeallocateStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DeallocateStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Allocation index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istDeallocate
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'DEALLOCATE', 0)
    
    CALL append_delimiter(seq, '(')
    
    DO i = 1, SIZE(pnode%objects)
      SELECT TYPE (object => pnode%objects(i)%item)
      TYPE IS (AllocateObjectVariable)
        CALL append_name(seq, object%name, 0)
      TYPE IS (AllocateObjectStructureComponent)
        CALL render_structure_component(object%parts, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END SELECT
      IF (i /= SIZE(pnode%objects)) CALL append_comma(seq)
    END DO
    
    IF (ALLOCATED(pnode%stat)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'STAT', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_variable(pnode%stat, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%errmsg)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ERRMSG', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_variable(pnode%errmsg, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_deallocate_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /cycle-stmt/.
  
  SUBROUTINE render_cycle_stmt(pnode, rnode, opts)
    
    USE DoStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(CycleStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istCycle
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'CYCLE', 0)
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(seq, pnode%construct_name, 0)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_cycle_stmt
  
  
  !*****************************************************************************
  !!
  !> Render an /exit-stmt/.
  
  SUBROUTINE render_exit_stmt(pnode, rnode, opts)
    
    USE ExitStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ExitStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istExit
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'EXIT', 0)
    
    IF (ALLOCATED(pnode%construct_name)) THEN
      CALL append_name(seq, pnode%construct_name, 0)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_exit_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /stop-stmt/.
  
  SUBROUTINE render_stop_stmt(pnode, rnode, opts)
    
    USE StopStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(StopStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istStop
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'STOP', 0)
    
    IF (ALLOCATED(pnode%stop_code)) THEN
      CALL render_expression(pnode%stop_code, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_stop_stmt
  
  
  !*****************************************************************************
  !!
  !> Render a /error-stop-stmt/.
  
  SUBROUTINE render_error_stop_stmt(pnode, rnode, opts)
    
    USE StopStmts
    USE Statements
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ErrorStopStmt), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeStatement), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = istErrorStop
    
    IF (ALLOCATED(pnode%statement_label)) THEN
      CALL append_label(seq, pnode%statement_label)
    END IF
    
    CALL append_keyword(seq, 'ERROR', 0)
    CALL append_keyword(seq, 'STOP', 0)
    
    IF (ALLOCATED(pnode%stop_code)) THEN
      CALL render_expression(pnode%stop_code, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_error_stop_stmt
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Procedures for syntax terms.
  
  
  !*****************************************************************************
  !!
  !> Render an /output-item/.
  
  RECURSIVE SUBROUTINE render_output_item(pnode, rnode, opts)
    
    USE InputOutputItems
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOItem), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrOutputItem
    
    SELECT TYPE (pnode)
    TYPE IS (IOItemExpr)
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    TYPE IS (IOItemImpliedDo)
      CALL render_io_implied_do(pnode, sub_rnode, opts, render_output_item)
      CALL append_rnode(seq, sub_rnode)
    END SELECT
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_output_item
  
  
  !*****************************************************************************
  !!
  !> Render an /input-item/.
  
  RECURSIVE SUBROUTINE render_input_item(pnode, rnode, opts)
    
    USE InputOutputItems
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOItem), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrInputItem
    
    SELECT TYPE (pnode)
    TYPE IS (IOItemExpr)
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    TYPE IS (IOItemImpliedDo)
      CALL render_io_implied_do(pnode, sub_rnode, opts, render_input_item)
      CALL append_rnode(seq, sub_rnode)
    END SELECT
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_input_item
  
  
  !*****************************************************************************
  !!
  !> Render an /io-implied-do/.
  !!
  !! This subroutine doesn't match the typical interface, as we also 
  !! pass along the worker procedure that renders an individual item.
  
  RECURSIVE SUBROUTINE render_io_implied_do(pnode, rnode, opts, proc)
    
    USE InputOutputItems
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOItemImpliedDo), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    PROCEDURE(render_output_item) :: proc
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Child item index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrIoImpliedDo
    
    CALL append_delimiter(seq, '(')
    DO i = 1, SIZE(pnode%items)
      ! This is an /io-implied-do-object/.
      CALL proc(pnode%items(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_comma(seq)
    END DO
    CALL append_name(seq, pnode%control%do_variable, 0)
    CALL append_token(seq, ittAssign, '=', '', 0)
    CALL render_expression(pnode%control%start, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    CALL append_comma(seq)
    CALL render_expression(pnode%control%finish, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    IF (ALLOCATED(pnode%control%step)) THEN
      CALL append_comma(seq)
      CALL render_expression(pnode%control%step, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_io_implied_do
  
  
  !*****************************************************************************
  !!
  !> Render a /proc-decl/.
  
  SUBROUTINE render_proc_decl(pnode, rnode, opts)
    
    USE ProcedureDeclarationStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProcDecl), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrProcDecl
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (ALLOCATED(pnode%null_init) .OR. ALLOCATED(pnode%init_target)) THEN
      CALL append_token(seq, ittAssociate, '=>', '', 0)
      IF (ALLOCATED(pnode%null_init)) THEN
        CALL render_null_init(pnode%null_init, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      ELSE
        CALL append_name(seq, pnode%init_target, 0)
      END IF
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_proc_decl
  
  
  !*****************************************************************************
  !!
  !> Render a /proc-decl/.
  
  SUBROUTINE render_proc_interface(pnode, rnode, opts)
    
    USE ProcedureDeclarationStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(ProcInterface), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrProcInterface
    
    IF (ALLOCATED(pnode%decl)) THEN
      CALL render_declaration_type_spec(pnode%decl, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    ELSE
      CALL append_name(seq, pnode%name, 0)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_proc_interface
  
  
  !*****************************************************************************
  !!
  !> Render a DeclarationTypeSpec.
  
  SUBROUTINE render_declaration_type_spec(pnode, rnode, opts)
    
    USE DeclarationTypeSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DeclTypeSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: use_type       ! Flag to indicate use of TYPE or CLASS.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrDeclarationTypeSpec
    
    ! Set by default.
    use_type = .TRUE.
    IF (pnode%is_polymorphic) THEN
      CALL append_keyword(seq, 'CLASS', 0)
    ELSE
      IF (.NOT. opts%always_use_type) THEN
        ASSOCIATE(type_name => .NameOf. pnode%type_spec%type_name)
          IF ( (type_name == 'INTEGER')  &
              .OR. (type_name == 'REAL')  &
              .OR. (type_name == 'COMPLEX')  &
              .OR. (type_name == 'LOGICAL')  &
              .OR. (type_name == 'CHARACTER') ) THEN
            use_type = .FALSE.
          END IF
        END ASSOCIATE
      END IF
      IF (use_type) THEN
        CALL append_keyword(seq, 'TYPE', 0)
      END IF 
    END IF
    
    IF (use_type) CALL append_delimiter(seq, '(')

    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL render_type_spec(pnode%type_spec, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    ELSE
      CALL append_token(seq, ittOperator, '*', '', 0)
    END IF
    
    IF (use_type) CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_declaration_type_spec
  
  
  !*****************************************************************************
  !!
  !> Render a DeclarationTypeSpec
  
  RECURSIVE SUBROUTINE render_type_spec(pnode, rnode, opts)
    
    USE DeclarationTypeSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(TypeSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    INTEGER :: i              ! type param spec index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrTypeSpec
    
    CALL append_name(seq, pnode%type_name, 0)
    
    IF (SIZE(pnode%params) /= 0) THEN
      CALL append_delimiter(seq, '(')
      DO i= 1, SIZE(pnode%params)
        IF (ALLOCATED(pnode%params(i)%name)) THEN
          CALL append_name(seq, pnode%params(i)%name, 0)
          CALL append_token(seq, ittAssign, '=', '', 0)
        END IF
        IF (pnode%params(i)%is_colon) THEN
          CALL append_token(seq, ittColon, ':', '', 0)
        ELSE IF (pnode%params(i)%is_star) THEN
          CALL append_token(seq, ittOperator, '*', '', 0)
        ELSE
          CALL render_expression(pnode%params(i)%int_expr, sub_rnode, opts)
          CALL append_rnode(seq, sub_rnode)
        END IF
        IF (i /= SIZE(pnode%params)) THEN
          CALL append_comma(seq)
        END IF
      END DO
      CALL append_delimiter(seq, ')')
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_type_spec
  
  
  !*****************************************************************************
  !!
  !> Render a Rename
  
  SUBROUTINE render_rename(pnode, rnode, opts)
    
    USE SyntaxTerms
    USE UseStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Rename), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrRename
    
    CALL render_generic_spec(pnode%GenericSpec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    IF (ALLOCATED(pnode%use)) THEN
      CALL append_token(seq, ittAssociate, '=>', '', 0)
      SELECT CASE (pnode%keyword)
      CASE (0)
        CALL append_name(seq, pnode%use, 0)
      CASE (1)
        CALL append_keyword(seq, 'OPERATOR', 0)
        CALL append_delimiter(seq, '(')
        CALL append_name(seq, pnode%use, 0)
        CALL append_delimiter(seq, ')')
      CASE (2)
        CALL append_keyword(seq, 'ASSIGNMENT', 0)
        CALL append_delimiter(seq, '(')
        CALL append_token(seq, ittAssign, '=', '', 0)
        CALL append_delimiter(seq, ')')
      CASE DEFAULT
        STOP 'Invalid /rename/.'
      END SELECT
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_rename
  
  
  !*****************************************************************************
  !!
  !> Render a GenericSpec
  
  SUBROUTINE render_generic_spec(pnode, rnode, opts)
    
    USE GenericSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(GenericSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrGenericSpec
    
    SELECT CASE (pnode%keyword)
    CASE (0)
      CALL append_name(seq, pnode%name, 0)
    CASE (1)
      CALL append_keyword(seq, 'OPERATOR', 0)
      CALL append_delimiter(seq, '(')
      CALL append_name(seq, pnode%name, 0)
      CALL append_delimiter(seq, ')')
    CASE (2)
      CALL append_keyword(seq, 'ASSIGNMENT', 0)
      CALL append_delimiter(seq, '(')
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL append_delimiter(seq, ')')
    CASE (3)
      CALL append_keyword(seq, 'READ', 0)
      CALL append_delimiter(seq, '(')
      CALL append_keyword(seq, 'FORMATTED', 0)
      CALL append_delimiter(seq, ')')
    CASE (4)
      CALL append_keyword(seq, 'WRITE', 0)
      CALL append_delimiter(seq, '(')
      CALL append_keyword(seq, 'FORMATTED', 0)
      CALL append_delimiter(seq, ')')
    CASE (5)
      CALL append_keyword(seq, 'READ', 0)
      CALL append_delimiter(seq, '(')
      CALL append_keyword(seq, 'UNFORMATTED', 0)
      CALL append_delimiter(seq, ')')
    CASE (6)
      CALL append_keyword(seq, 'WRITE', 0)
      CALL append_delimiter(seq, '(')
      CALL append_keyword(seq, 'UNFORMATTED', 0)
      CALL append_delimiter(seq, ')')
    END SELECT
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_generic_spec
  
  
  !*****************************************************************************
  !!
  !> Render an /attr-spec-list/.
  
  SUBROUTINE render_attr_specs(pnode, seq, opts)
    
    USE AccessStmts
    USE IntentSpecs
    USE TypeDeclarationStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(AttrSpec), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
  
    IF (pnode%allocatable) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ALLOCATABLE', 0)
    END IF
    IF (pnode%asynchronous) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'ASYNCHRONOUS', 0)
    END IF
    IF (pnode%contiguous) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'CONTIGUOUS', 0)
    END IF
    IF (pnode%external) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'EXTERNAL', 0)
    END IF
    IF (pnode%intrinsic) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'INTRINSIC', 0)
    END IF
    IF (pnode%optional) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'OPTIONAL', 0)
    END IF
    IF (pnode%parameter) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'PARAMETER', 0)
    END IF
    IF (pnode%pointer) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'POINTER', 0)
    END IF
    IF (pnode%protected) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'PROTECTED', 0)
    END IF
    IF (pnode%save) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'SAVE', 0)
    END IF
    IF (pnode%target) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'TARGET', 0)
    END IF
    IF (pnode%value) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'VALUE', 0)
    END IF
    IF (pnode%volatile) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'VOLATILE', 0)
    END IF
    
    IF (pnode%access /= 0) THEN
      CALL append_comma(seq)
      CALL render_access_spec(pnode%access, seq, opts)
    END IF
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'DIMENSION', 0)
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%dim_specs, seq, opts)
      CALL append_delimiter(seq, ')')
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'CODIMENSION', 0)
      CALL append_delimiter(seq, '[')
      CALL render_dim_specs(pnode%codim_specs, seq, opts)
      CALL append_delimiter(seq, ']')
    END IF
    
    IF (pnode%intent /= 0) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'INTENT', 0)
      CALL append_delimiter(seq, '(')
      CALL render_intent_spec(pnode%intent, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_delimiter(seq, ')')
    END IF
    
    IF (ALLOCATED(pnode%binding)) THEN
      CALL append_comma(seq)
      CALL render_language_binding_spec(pnode%binding, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
  END SUBROUTINE render_attr_specs
  
  
  !*****************************************************************************
  !!
  !> Render a list of dim specs.
  
  SUBROUTINE render_dim_specs(pnode, seq, opts)
    
    USE SubscriptLists
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: pnode(:)
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    INTEGER :: i              ! spec index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    DO i = 1, SIZE(pnode)
      CALL render_dim_spec(pnode(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    
  END SUBROUTINE render_dim_specs
  
  
  
  !*****************************************************************************
  !!
  !> Render a list of dim specs.
  
  SUBROUTINE render_dim_spec(pnode, rnode, opts)
    
    USE SubscriptLists
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(DimSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    ! seq%context = ??? ! So many options!
    
    IF (ALLOCATED(pnode%start)) THEN
      CALL render_expression(pnode%start, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    IF (pnode%colon) THEN
      CALL append_token(seq, ittColon, ':', '', 0)
    END IF
    IF (ALLOCATED(pnode%finish)) THEN
      CALL render_expression(pnode%finish, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    IF (pnode%star) THEN
      CALL append_token(seq, ittOperator, '*', '', 0)
    END IF
    IF (ALLOCATED(pnode%stride)) THEN
      CALL append_token(seq, ittColon, ':', '', 0)
      CALL render_expression(pnode%stride, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_dim_spec
  
  
  !*****************************************************************************
  !!
  !> Render an /entity-decl/.
  
  SUBROUTINE render_entity_decl(pnode, rnode, opts)
    
    USE SyntaxTerms
    USE TypeDeclarationStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EntityDetails), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrEntityDecl
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (ALLOCATED(pnode%dim_specs)) THEN
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%dim_specs, seq, opts)
      CALL append_delimiter(seq, ')')
    END IF
    
    IF (ALLOCATED(pnode%codim_specs)) THEN
      CALL append_delimiter(seq, '[')
      CALL render_dim_specs(pnode%codim_specs, seq, opts)
      CALL append_delimiter(seq, ']')
    END IF
    
    IF (ALLOCATED(pnode%len)) THEN
      CALL append_token(seq, ittOperator, '*', '', 0)
      CALL render_expression(pnode%len, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%expr)) THEN
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    IF (ALLOCATED(pnode%null_init)) THEN
      CALL append_token(seq, ittAssociate, '=>', '', 0)
      CALL render_null_init(pnode%null_init, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    ELSE IF (ALLOCATED(pnode%initial_data_target)) THEN
      CALL append_token(seq, ittAssociate, '=>', '', 0)
      CALL render_designator(pnode%initial_data_target, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_entity_decl
  
  
  !*****************************************************************************
  !!
  !> Render a designator
  
  RECURSIVE SUBROUTINE render_designator(pnode, rnode, opts)
    
    USE Designators
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Designator), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Part index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrDesignator
    
    IF (ALLOCATED(pnode%literal)) THEN
      IF (ALLOCATED(pnode%literal_kind)) THEN
        CALL append_token(  &
            seq,  &
            ittCharacterLiteral,  &
            pnode%literal,  &
            pnode%literal_kind,  &
            0 )
      ELSE
        CALL append_token(  &
            seq,  &
            ittCharacterLiteral,  &
            pnode%literal,  &
            '',  &
            0 )
      END IF
      
      ! The substring component is always allocated if the literal 
      ! component is allocated.
      CALL append_delimiter(seq, '(')
      IF (ALLOCATED(pnode%substring%start)) THEN
        CALL render_expression(pnode%substring%start, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END IF
      CALL append_token(seq, ittColon, ':', '', 0)
      IF (ALLOCATED(pnode%substring%finish)) THEN
        CALL render_expression(pnode%substring%finish, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END IF
      CALL append_delimiter(seq, ')')
    ELSE
      DO i = 1, SIZE(pnode%part_refs)
        CALL render_part_ref(pnode%part_refs(i), sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
        IF (i /= SIZE(pnode%part_refs)) THEN
          CALL append_token(seq, ittPercent, '%', '', 0)
        END IF
      END DO
      IF (ALLOCATED(pnode%substring)) THEN
        CALL append_delimiter(seq, '(')
        IF (ALLOCATED(pnode%substring%start)) THEN
          CALL render_expression(pnode%substring%start, sub_rnode, opts)
          CALL append_rnode(seq, sub_rnode)
        END IF
        CALL append_token(seq, ittColon, ':', '', 0)
        IF (ALLOCATED(pnode%substring%finish)) THEN
          CALL render_expression(pnode%substring%finish, sub_rnode, opts)
          CALL append_rnode(seq, sub_rnode)
        END IF
        CALL append_delimiter(seq, ')')
      END IF
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_designator
  
  
  !*****************************************************************************
  !!
  !> Render a /part-ref/.
  
  SUBROUTINE render_part_ref(pnode, rnode, opts)
    
    USE Designators
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PartRef), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrPartRef
    
    CALL append_name(seq, pnode%part_name, 0)
    IF (ALLOCATED(pnode%subscripts)) THEN
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%subscripts, seq, opts)
      CALL append_delimiter(seq, ')')
    END IF
    
    IF (ALLOCATED(pnode%image_selector)) THEN
      CALL append_delimiter(seq, '[')
      CALL render_dim_specs(pnode%image_selector, seq, opts)
      CALL append_delimiter(seq, ']')
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_part_ref
  
  
  !*****************************************************************************
  !!
  !> Render an /language-binding-spec/.
  
  SUBROUTINE render_language_binding_spec(pnode, rnode, opts)
    
    USE LanguageBindingSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(LangSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrLanguageBindingSpec
    
    CALL append_keyword(seq, 'BIND', 0)
    CALL append_delimiter(seq, '(')
    SELECT CASE (pnode%lang)
    CASE (1)      ; CALL append_keyword(seq, 'C', 0)
    CASE DEFAULT
      STOP 'Internal error in &
          &RenderParseNodes::render_language_binding_spec, &
          &unknown language spec.'
    END SELECT
    IF (ALLOCATED(pnode%name)) THEN
      CALL append_comma(seq)
      CALL append_keyword(seq, 'NAME', 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_expression(pnode%name, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_language_binding_spec
  
  
  !*****************************************************************************
  !!
  !> Render an /expr/.
  
  RECURSIVE SUBROUTINE render_expression(pnode, rnode, opts)
    
    USE Expressions
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Expression), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrExpr
    
    CALL render_expression_worker(pnode, seq, opts)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_expression
  
  
  !*****************************************************************************
  !!
  !> Render a /variable/.
  
  RECURSIVE SUBROUTINE render_variable(pnode, rnode, opts)
    
    USE SyntaxTerms
    USE Variables
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Variable), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrVariable
    
    CALL render_expression_worker(pnode, seq, opts)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_variable
  
  
  !*****************************************************************************
  !!
  !> Render a /function-reference/.
  
  RECURSIVE SUBROUTINE render_function_reference(pnode, rnode, opts)
    
    USE PrimaryExpressions
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FunctionRefExpr), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Child part or argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrFunctionReference
    
    DO i = 1, SIZE(pnode%parts)
      CALL render_part_ref(pnode%parts(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(seq, ittPercent, '%', '', 0)
    END DO
    
    CALL append_name(seq, pnode%name, 0)
    
    CALL append_delimiter(seq, '(')
    DO i = 1, SIZE(pnode%args)
      CALL render_actual_arg_spec(pnode%args(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%args)) CALL append_comma(seq)
    END DO
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_function_reference
  
  
  !*****************************************************************************
  !!
  !> Render an /array-constructor/.
  
  RECURSIVE SUBROUTINE render_array_constructor(pnode, rnode, opts)
    
    USE ArrayConstructors
    USE PrimaryExpressions
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ArrayConstructorExpr), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Child part or argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrArrayConstructor
    
    CALL append_delimiter(seq, '[')
    
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL render_type_spec(pnode%type_spec, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%values)
      CALL render_ac_value(pnode%values(i)%item, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnode%values)) CALL append_comma(seq)
    END DO
    
    CALL append_delimiter(seq, ']')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_array_constructor
  
  
  !*****************************************************************************
  !!
  !> Render an /ac-value/.
  
  RECURSIVE SUBROUTINE render_ac_value(pnode, rnode, opts)
    
    USE ArrayConstructors
    USE PrimaryExpressions
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ACValue), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Child part or argument index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrACValue
    
    SELECT TYPE (pnode)
    TYPE IS (ACValueExpr)
      CALL render_expression(pnode%ex, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    TYPE IS (ACValueImpliedDo)
      ! /ac-implied-do/.
      CALL append_delimiter(seq, '(')
      DO i = 1, SIZE(pnode%ac_value_list)
        CALL render_ac_value(pnode%ac_value_list(i)%item, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
        CALL append_comma(seq)
      END DO
      CALL render_ac_implied_do_control(pnode%control, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_delimiter(seq, ')')
    CLASS DEFAULT
      STOP 'Internal error in RenderParseNodes::render_ac_value, &
          &unknown type of value.'
    END SELECT
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_ac_value
  
  
  !*****************************************************************************
  !!
  !> Render an /ac-implied-do-control/.
  
  RECURSIVE SUBROUTINE render_ac_implied_do_control(pnode, rnode, opts)
    
    USE ArrayConstructors
    USE PrimaryExpressions
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ACImpliedDoControl), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrACImpliedDoControl
    
    CALL append_name(seq, pnode%do_variable, 0)
    CALL append_token(seq, ittAssign, '=', '', 0)
    CALL render_expression(pnode%start, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    CALL append_comma(seq)
    CALL render_expression(pnode%finish, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    IF (ALLOCATED(pnode%step)) THEN
      CALL append_comma(seq)
      CALL render_expression(pnode%step, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_ac_implied_do_control
  
  
  !*****************************************************************************
  !!
  !> Render an /actual-arg-spec/.
  
  SUBROUTINE render_actual_arg_spec(pnode, rnode, opts)
    
    USE CallStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ActualArgSpec), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrActualArgSpec
    
    IF (ALLOCATED(pnode%keyword)) THEN
      !=========================================================================
      ! ifort 19.0 bug - see ticket #48
      !CALL append_keyword(seq, .NameOf. pnode%keyword, 0)
      !=========================================================================
      CALL append_keyword(seq, name_bug(pnode%keyword), 0)
      !=========================================================================
      CALL append_token(seq, ittAssign, '=', '', 0)
    END IF
    IF (ALLOCATED(pnode%alt_return_label)) THEN
      CALL append_label(seq, pnode%alt_return_label)
    ELSE
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_actual_arg_spec
  
  
  !=============================================================================
  ! ifort 19.0 bug - see ticket #48.
  !=============================================================================
  function name_bug(ident)
    use identifierutils
    
    type(identifier), intent(in) :: ident
    character(:), allocatable :: name_bug
    
    name_bug = .NameOf. ident
  end function name_bug
  !=============================================================================
  
  
  !*****************************************************************************
  !!
  !> Render a /complex-literal-constant/.
  
  SUBROUTINE render_complex_literal_constant(pnode, rnode, opts)
    
    USE ComplexLiteralConstants
    USE SyntaxTerms
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ComplexLiteralConstant), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrComplexLiteralConstant
    
    CALL append_delimiter(seq, '(')
    
    IF (ALLOCATED(pnode%real_sign)) THEN
      CALL append_token(  &
          seq,  &
          QueryType(pnode%real_sign),  &
          QueryValue(pnode%real_sign),  &
          QueryKind(pnode%real_sign),  &
          0 )
    END IF
    
    CALL append_token(  &
        seq,  &
        QueryType(pnode%real_part),  &
        QueryValue(pnode%real_part),  &
        QueryKind(pnode%real_part),  &
        0 )
    
    CALL append_comma(seq)
    
    IF (ALLOCATED(pnode%imag_sign)) THEN
      CALL append_token(  &
          seq,  &
          QueryType(pnode%imag_sign),  &
          QueryValue(pnode%imag_sign),  &
          QueryKind(pnode%imag_sign),  &
          0 )
    END IF
    
    CALL append_token(  &
        seq,  &
        QueryType(pnode%imag_part),  &
        QueryValue(pnode%imag_part),  &
        QueryKind(pnode%imag_part),  &
        0 )
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_complex_literal_constant
  
  
  !*****************************************************************************
  !!
  !> Render a /structure-constructor/.
  
  SUBROUTINE render_structure_constructor(pnode, rnode, opts)
    
    USE StructureConstructors
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(StructureConstructor), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Component spec index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrStructureConstructor
    
    CALL render_type_spec(pnode%type_spec, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    
    CALL append_delimiter(seq, '(')
    
    DO i = 1, SIZE(pnode%components)
      ASSOCIATE(comp => pnode%components(i))
        IF (ALLOCATED(comp%keyword)) THEN
          ! The syntax rules describe this as a component keyword, so 
          ! we call it a keyword.
          CALL append_keyword(seq, .NameOf. comp%keyword, 0)
          CALL append_token(seq, ittAssign, '=', '', 0)
        END IF
        CALL render_expression(comp%data_source, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END ASSOCIATE
      IF (i /= SIZE(pnode%components)) CALL append_comma(seq)
    END DO
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_structure_constructor
  
  
  !*****************************************************************************
  !!
  !> Render a /structure-component/.
  
  SUBROUTINE render_structure_component(pnodes, rnode, opts)
    
    USE Designators
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PartRef), INTENT(IN) :: pnodes(:)
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Part-ref index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrStructureComponent
    
    DO i = 1, SIZE(pnodes)
      CALL render_part_ref(pnodes(i), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (i /= SIZE(pnodes)) THEN
        CALL append_token(seq, ittPercent, '%', '', 0)
      END IF
    END DO
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_structure_component
  
  
  !*****************************************************************************
  !!
  !> Render a /format/.
  
  SUBROUTINE render_io_unit(pnode, rnode, opts)
    
    USE IoControlSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IOUnit), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrIOUnit
    
    IF (pnode%is_star) THEN
      CALL append_token(seq, ittOperator, '*', '', 0)
    ELSE
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_io_unit
  
  
  !*****************************************************************************
  !!
  !> Render a /format/.
  
  SUBROUTINE render_format(pnode, rnode, opts)
    
    USE Formats
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Format), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrFormat
    
    IF (pnode%is_star) THEN
      CALL append_token(seq, ittOperator, '*', '', 0)
    ELSE IF (ALLOCATED(pnode%label_spec)) THEN
      CALL append_label(seq, pnode%label_spec)
    ELSE
      CALL render_expression(pnode%expr, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_format
  
  
  !*****************************************************************************
  !!
  !> Render a /forall-header/.
  
  SUBROUTINE render_forall_header(pnode, rnode, opts)
    
    USE ForallStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ForallHeader), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    INTEGER :: i              ! Triplet index.
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrForallHeader
    
    CALL append_delimiter(seq, '(')
    IF (ALLOCATED(pnode%type_spec)) THEN
      CALL render_type_spec(pnode%type_spec, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(seq, ittDeclare, '::', '', 0)
    END IF
    
    DO i = 1, SIZE(pnode%triplets)
      CALL append_name(seq, pnode%triplets(i)%index, 0)
      CALL append_token(seq, ittAssign, '=', '', 0)
      CALL render_expression(pnode%triplets(i)%start, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(seq, ittColon, ':', '', 0)
      CALL render_expression(pnode%triplets(i)%finish, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (ALLOCATED(pnode%triplets(i)%step)) THEN
        CALL append_token(seq, ittColon, ':', '', 0)
        CALL render_expression(pnode%triplets(i)%step, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      END IF
      IF (i /= SIZE(pnode%triplets)) CALL append_comma(seq)
    END DO
    
    IF (ALLOCATED(pnode%mask)) THEN
      CALL append_comma(seq)
      CALL render_expression(pnode%mask, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_forall_header
  
  
  !*****************************************************************************
  !!
  !> Render an /intent_spec/.
  
  SUBROUTINE render_intent_spec(intent, rnode, opts)
    
    USE IntentSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: intent
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrIntentSpec
    
    SELECT CASE (intent)
    CASE (iintIn)     ; CALL append_keyword(seq, 'IN', 0)
    CASE (iintOut)    ; CALL append_keyword(seq, 'OUT', 0)
    CASE (iintInOut)  ; CALL append_keyword(seq, 'INOUT', 0)
    END SELECT
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_intent_spec
  
  
  !*****************************************************************************
  !!
  !> Render a /data-implied-do/.
  
  ! Recursive because data-implied-do's can contain other data-implied-do's.
  RECURSIVE SUBROUTINE render_data_implied_do(pnode, rnode, opts)
    
    USE DataStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(DataImpliedDo), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    INTEGER :: i              ! object index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrDataImpliedDo
    
    CALL append_delimiter(seq, '(')
    
    DO i = 1, SIZE(pnode%items)
      SELECT TYPE (object => pnode%items(i)%item)
      TYPE IS (DataIDoObjectDesignator)
        CALL render_designator(object%desig, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      TYPE IS (DataIDoObjectDataImpliedDo)
        CALL render_data_implied_do(object%obj, sub_rnode, opts)
        CALL append_rnode(seq, sub_rnode)
      CLASS DEFAULT
        STOP 'Internal error in RenderParseNodes::render_data_implied_do, &
            &unknown type of do object.'
      END SELECT
      CALL append_comma(seq)
    END DO
    
    ! There is no such syntax rule as a /data-implied-do-control/, so we
    ! render the equivalent of that rule in line.
    CALL append_name(seq, pnode%do_control%do_variable, 0)
    CALL append_token(seq, ittAssign, '=', '', 0)
    CALL render_expression(pnode%do_control%start, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    CALL append_comma(seq)
    CALL render_expression(pnode%do_control%finish, sub_rnode, opts)
    CALL append_rnode(seq, sub_rnode)
    IF (ALLOCATED(pnode%do_control%step)) THEN
      CALL append_comma(seq)
      CALL render_expression(pnode%do_control%step, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
    END IF
    
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_data_implied_do
  
  
  !*****************************************************************************
  !!
  !> Render a /common-block-object/.
  
  SUBROUTINE render_common_block_object(pnode, rnode, opts)
    
    USE CommonStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CommonBlockObject), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrCommonBlockObject
    
    CALL append_name(seq, pnode%name, 0)
    
    IF (ALLOCATED(pnode%array_spec)) THEN
      CALL append_delimiter(seq, '(')
      CALL render_dim_specs(pnode%array_spec, seq, opts)
      CALL append_delimiter(seq, ')')
    END IF
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_common_block_object
  
  
  !*****************************************************************************
  !!
  !> Render an /equivalence-set/.
  
  SUBROUTINE render_equivalence_set(pnode, rnode, opts)
    
    USE EquivalenceStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(EquivalenceSet), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    INTEGER :: io             ! Object index.
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrEquivalenceSet
    
    CALL append_delimiter(seq, '(')
    DO io = 1, SIZE(pnode%object_list)
      CALL render_designator(pnode%object_list(io), sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      IF (io /= SIZE(pnode%object_list)) THEN
        CALL append_comma(seq)
      END IF
    END DO
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_equivalence_set
  
  
  !*****************************************************************************
  !!
  !> Render a /namelist-group-object/.
  !!
  !! This is just a name.
  
  SUBROUTINE render_namelist_group_object(pnode, rnode, opts)
    
    USE Identifiers
    USE NamelistStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(Identifier), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrNamelistGroupObject
    
    CALL append_name(seq, pnode, 0)
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_namelist_group_object
  
  
  !*****************************************************************************
  !!
  !> Render a /null-init/.
  
  SUBROUTINE render_null_init(pnode, rnode, opts)
    
    USE ProcedureDeclarationStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(NullInit), INTENT(IN) :: pnode
    CLASS(RenderNode), INTENT(OUT), ALLOCATABLE :: rnode
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! The node we build.
    TYPE(RenderNodeSequence), ALLOCATABLE :: seq
    
    !***************************************************************************
    
    ALLOCATE(seq)
    seq%context = isrNullInit
    
    CALL append_name(seq, pnode%name, 0)
    CALL append_delimiter(seq, '(')
    CALL append_delimiter(seq, ')')
    
    CALL MOVE_ALLOC(seq, rnode)
    
  END SUBROUTINE render_null_init
  
  
  !-----------------------------------------------------------------------------
  !-----------------------------------------------------------------------------
  !
  ! Render helpers.
  
  
  !*****************************************************************************
  !!
  !> Render the other (other than /declaration-type-spec/) prefixes that can 
  !! appear out the front of a /function-stmt/ or /subroutine-stmt/.
  
  SUBROUTINE render_other_prefixes(prefixes, seq, opts)
    
    USE FunctionStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    TYPE(OtherPrefixSpecs), INTENT(IN) :: prefixes
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !***************************************************************************
    
    IF (prefixes%elemental) THEN
      CALL append_keyword(seq, 'ELEMENTAL', 0)
    END IF
    
    SELECT CASE (prefixes%purity)
    CASE (1)  ;   CALL append_keyword(seq, 'IMPURE', 0)
    CASE (2)  ;   CALL append_keyword(seq, 'PURE', 0)
    END SELECT
    
    IF (prefixes%module) THEN
      CALL append_keyword(seq, 'MODULE', 0)
    END IF
    
    IF (prefixes%recursive) THEN
      CALL append_keyword(seq, 'RECURSIVE', 0)
    END IF
    
  END SUBROUTINE render_other_prefixes
  
  
  !*****************************************************************************
  !!
  !> Render an /access-spec/.
  
  SUBROUTINE render_access_spec(access, seq, opts)
    
    USE AccessStmts
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    INTEGER, INTENT(IN) :: access
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !***************************************************************************
    
    SELECT CASE (access)
    CASE (iaccPublic)   ; CALL append_keyword(seq, 'PUBLIC', 0)
    CASE (iaccPrivate)  ; CALL append_keyword(seq, 'PRIVATE', 0)
    END SELECT
    
  END SUBROUTINE render_access_spec
  
  
  !*****************************************************************************
  !!
  !> Render the contents of an Expression object.
  !!
  !! This code is common to expression and variable rendering, due 
  !! to the way we store variables.
  
  RECURSIVE SUBROUTINE render_expression_worker(pnode, seq, opts)
    
    USE Expressions
    USE PrimaryExpressions
    USE Tokens
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(Expression), INTENT(IN) :: pnode
    TYPE(RenderNodeSequence), INTENT(OUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    SELECT TYPE (pnode)
    TYPE IS (UnaryOpExpr)
      CALL append_token(  &
          seq,  &
          ittOperator,  &
          .NameOf. pnode%op,  &
          '',  &
          0 )
      CALL render_expression(pnode%rhs, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (BinaryOpExpr)
      CALL render_expression(pnode%lhs, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_token(  &
          seq,  &
          ittOperator,  &
          .NameOf. pnode%op,  &
          '',  &
          0 )
      CALL render_expression(pnode%rhs, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (ConstantExpr)
      IF (HasKind(pnode%tok)) THEN
        CALL append_token(  &
            seq,  &
            QueryType(pnode%tok),  &
            QueryValue(pnode%tok),  &
            QueryKind(pnode%tok),  &
            0 )
      ELSE
        CALL append_token(  &
            seq,  &
            QueryType(pnode%tok),  &
            QueryValue(pnode%tok),  &
            '',  &
            0 )
      END IF
      
    TYPE IS (ComplexLiteralExpr)
      CALL render_complex_literal_constant(pnode%constant,sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (DesignatorExpr)
      CALL render_designator(pnode%desig, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (ArrayConstructorExpr)
      CALL render_array_constructor(pnode, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (StructureConstructorExpr)
      CALL render_structure_constructor(pnode%sc, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (FunctionRefExpr)
      CALL render_function_reference(pnode, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      
    TYPE IS (ParenthesisedExpr)
      CALL append_delimiter(seq, '(')
      CALL render_expression(pnode%ex, sub_rnode, opts)
      CALL append_rnode(seq, sub_rnode)
      CALL append_delimiter(seq, ')')
      
    CLASS DEFAULT
      PRINT "('The unimplemented ParseNode was for a ',A)", pnode%What()
      STOP 'Rendering not implemented for an Expression.'
    END SELECT
    
  END SUBROUTINE render_expression_worker
  
  
  !*****************************************************************************
  !!
  !> Render a /io-control-spec-list/.
  
  SUBROUTINE render_io_control_spec_list(pnode, seq, opts)
    
    USE IoControlSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(IoControlSpecList), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL render_io_unit(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%fmt_or_nml)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_name(item_rnode, pnode%fmt_or_nml, 0)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%fmt)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'FMT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_format(pnode%fmt, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%nml)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'NML', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_name(item_rnode, pnode%nml, 0)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%advance)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'ADVANCE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%advance, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%asynchronous)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'ASYNCHRONOUS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%asynchronous, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%blank)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'BLANK', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%blank, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%decimal)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'DECIMAL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%decimal, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%delim)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'DELIM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%delim, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%end)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'END', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%end)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%eor)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'EOR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%eor)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%id)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'ID', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%id, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pad)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'PAD', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%pad, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pos)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'POS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%pos, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%rec)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'REC', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%rec, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%round)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'ROUND', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%round, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%sign)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'SIGN', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%sign, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%size)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrIOControlSpec
      CALL append_keyword(item_rnode, 'SIZE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%size, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_io_control_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render a /connect-spec-list/.
  
  SUBROUTINE render_connect_spec_list(pnode, seq, opts)
    
    USE OpenStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(OpenStmt), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'UNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%newunit)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'NEWUNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%newunit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%file)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'FILE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%file, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%access)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ACCESS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%access, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%action)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ACTION', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%action, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%asynchronous)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ASYNCHRONOUS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%asynchronous, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%blank)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'BLANK', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%blank, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%decimal)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'DECIMAL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%decimal, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%delim)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'DELIM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%delim, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%encoding)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ENCODING', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%encoding, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%form)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'FORM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%form, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pad)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'PAD', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%pad, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%position)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'POSITION', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%position, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%recl)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'RECL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%recl, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%round)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'ROUND', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%round, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%sign)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'SIGN', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%sign, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%status)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrConnectSpec
      CALL append_keyword(item_rnode, 'STATUS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%status, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_connect_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render a /close-spec-list/.
  
  SUBROUTINE render_close_spec_list(pnode, seq, opts)
    
    USE CloseStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(CloseStmt), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrCloseSpec
      CALL append_keyword(item_rnode, 'UNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrCloseSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrCloseSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrCloseSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%status)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrCloseSpec
      CALL append_keyword(item_rnode, 'STATUS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%status, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_close_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render a /close-spec-list/.
  
  SUBROUTINE render_flush_spec_list(pnode, seq, opts)
    
    USE FlushStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(FlushStmt), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrFlushSpec
      CALL append_keyword(item_rnode, 'UNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrFlushSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrFlushSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrFlushSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_flush_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render a /position-spec-list/.
  
  SUBROUTINE render_position_spec_list(pnode, seq, opts)
    
    USE PositionSpecs
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(PositionSpecList), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrPositionSpec
      CALL append_keyword(item_rnode, 'UNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrPositionSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrPositionSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrPositionSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_position_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render an /inquire-spec-list/.
  
  SUBROUTINE render_inquire_spec_list(pnode, seq, opts)
    
    USE InquireStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(InquireStmt), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    LOGICAL :: need_comma     ! Comma required before next item.
    
    !***************************************************************************
    
    need_comma = .FALSE.
    
    IF (ALLOCATED(pnode%unit)) THEN
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'UNIT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%unit, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%file)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'FILE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%file, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%access)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ACCESS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%access, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%action)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ACTION', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%action, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%asynchronous)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ASYNCHRONOUS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%asynchronous, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%blank)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'BLANK', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%blank, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%decimal)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'DECIMAL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%decimal, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%delim)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'DELIM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%delim, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%direct)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'DIRECT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%direct, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%encoding)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ENCODING', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%encoding, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%err)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ERR', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL append_label(item_rnode, pnode%err)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%exist)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'EXIST', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%exist, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%form)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'FORM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%form, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%formatted)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'FORMATTED', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%formatted, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%id)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ID', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%id, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iomsg)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'IOMSG', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iomsg, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iostat)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'IOSTAT', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iostat, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%name)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'NAME', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%name, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%named)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'NAMED', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%named, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%nextrec)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'NEXTREC', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%nextrec, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%number)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'NUMBER', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%number, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%opened)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'OPENED', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%opened, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pad)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'PAD', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%pad, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pending)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'PENDING', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%pending, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%pos)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'POS', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%pos, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%position)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'POSITION', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%position, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%read)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'READ', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%read, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%readwrite)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'READWRITE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%readwrite, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%recl)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'RECL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%recl, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%round)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'ROUND', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%round, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%sequential)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'SEQUENTIAL', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%sequential, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%sign)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'SIGN', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_expression(pnode%sign, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%size)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'SIZE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%size, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%stream)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'STREAM', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%stream, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%unformatted)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'UNFORMATTED', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%unformatted, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%write)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'WRITE', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%write, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
    IF (ALLOCATED(pnode%iolength)) THEN
      IF (need_comma) CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrInquireSpec
      CALL append_keyword(item_rnode, 'IOLENGTH', 0)
      CALL append_token(item_rnode, ittAssign, '=', '', 0)
      CALL render_variable(pnode%iolength, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
      need_comma = .TRUE.
    END IF
    
  END SUBROUTINE render_inquire_spec_list
  
  
  !*****************************************************************************
  !!
  !> Render a list of /proc-attr-spec/'s.
  
  SUBROUTINE render_proc_attr_spec_list(pnode, seq, opts)
    
    USE AccessStmts
    USE IntentSpecs
    USE ProcedureDeclarationStmts
    USE SyntaxTerms
    
    !---------------------------------------------------------------------------
    ! Arguments
    
    CLASS(ProcAttrSpec), INTENT(IN) :: pnode
    CLASS(RenderNodeSequence), INTENT(INOUT) :: seq
    CLASS(RenderOptions), INTENT(IN) :: opts
    
    !---------------------------------------------------------------------------
    ! Locals
    
    ! Node for each child spec.
    TYPE(RenderNodeSequence), ALLOCATABLE :: item_rnode
    
    ! For building child nodes.
    CLASS(RenderNode), ALLOCATABLE :: sub_rnode
    
    !***************************************************************************
    
    IF (pnode%access /= iaccNone) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL render_access_spec(pnode%access, item_rnode, opts)
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (ALLOCATED(pnode%ls)) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL render_language_binding_spec(pnode%ls, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (pnode%intent /= iintNone) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL append_keyword(item_rnode, 'INTENT', 0)
      CALL append_delimiter(item_rnode, '(')
      CALL render_intent_spec(pnode%intent, sub_rnode, opts)
      CALL append_rnode(item_rnode, sub_rnode)
      CALL append_delimiter(item_rnode, ')')
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (pnode%optional) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL append_keyword(item_rnode, 'OPTIONAL', 0)
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (pnode%pointer) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL append_keyword(item_rnode, 'POINTER', 0)
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (pnode%save) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL append_keyword(item_rnode, 'SAVE', 0)
      CALL append_seq(seq, item_rnode)
    END IF
    
    IF (pnode%protected) THEN
      CALL append_comma(seq)
      ALLOCATE(item_rnode)
      item_rnode%context = isrProcAttrSpec
      CALL append_keyword(item_rnode, 'PROTECTED', 0)
      CALL append_seq(seq, item_rnode)
    END IF
    
  END SUBROUTINE render_proc_attr_spec_list
  
END MODULE RenderParseNodes
